<?php
/**
 * Main content controller; handles both frontend and administrative utilities.
 *
 * @package Content
 * @since 0.1
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\AccessDeniedException;
use Core\ConfigHandler;
use Core\Controller;
use Core\Forms\Form;
use Core\NotFoundException;
use Core\TemporaryFollowRedirectException;
use Core\ViewControl;


class ContentController extends Controller {
	public function view() {
		// I'm calling check access here because the cached access string is canonical in this case.
		$request      = $this->getRequest();
		$page         = $this->getPageModel();
		$view         = $this->getView();
		$status       = $page->getPublishedStatus();
		$model        = ContentModel::Construct($request->getParameter(0));
		$user         = \Core\user();
		$manageAccess = $user->checkAccess('p:/content/manage_all');
		$editAccess   = $user->checkAccess($model->editpermissions) || $manageAccess;
		$viewAccess   = $user->checkAccess($model->access) || $editAccess;

		if(!$model->checkExists()) {
			throw new NotFoundException();
		}

		if(!$viewAccess) {
			throw new AccessDeniedException();
		}


		if(!$page->isPublished()){
			// If this page is not published yet, provide a notice to editors and prevent users from viewing it.
			if($editAccess){
				\Core\set_message('This page is currently set as ' . $page->getPublishedStatus() . '.', 'warning');
			}
			else{
				throw new NotFoundException();
			}
		}

		if($view->templatename) {
			$view->bodyclasses[] = 'template-' . basename($view->templatename);
		}

		$view->addStylesheet('assets/css/content.css');
		$view->assign('model', $model);
		$view->assign('page', $page);
		$view->addControls($page);

		if($manageAccess) {
			$view->addControl(ViewControl::Factory([
				'title' => 't:STRING_CONTENT_PAGE_ADD_CHILD',
				'link' => '/content/create?page_template=' . $page->get('base_template') . '&parenturl=' . $page->get('baseurl'),
				'icon' => 'add',
				'class' => 'add',
				'group' => ViewControl::GROUP_MANAGE,
			]));
		}
	}

	/**
	 * Edit an existing page
	 * 
	 * This has dynamic permissions, so access must be checked manually
	 * 
	 * @throws NotFoundException
	 * @throws AccessDeniedException
	 */
	public function edit(){
		$view    = $this->getView();
		$request = $this->getRequest();
		$model   = new ContentModel($request->getParameter(0));
		$page    = $model->getLink('Page');
		

		if(!$model->checkExists()){
			throw new NotFoundException();
		}

		$editor  = (\Core\user()->checkAccess($model->get('editpermissions')) || \Core\user()->checkAccess('p:/content/manage_all'));
		$manager = \Core\user()->checkAccess('p:/content/manage_all');

		if(!($editor || $manager)){
			throw new AccessDeniedException();
		}
		
		$form = new Form();
		$form->set('callsmethod', 'ContentController::_SaveHandler');

		$form->addModel($page, 'page');
		$form->addModel($model, 'model');

		// Tack on a submit button
		$form->addElement('submit', ['value' => 't:STRING_UPDATE', 'name' => 'save']);

		// Editors have certain permissions here, namely limited.
		if(!$manager){
			$form->removeElement('model[nickname]');
			$form->removeElement('model[editpermissions]');
			$form->removeElement('page[rewriteurl]');
			$form->removeElement('page[parenturl]');
		}
		
		$pageTitle = $page->get('title') ?: t('STRING_NO_PAGE_TITLE');
		
		$view->mastertemplate = 'admin';
		$view->addBreadcrumb($pageTitle, $page->get('baseurl'));
		$view->title        = 'Edit ' . $pageTitle;
		$view->assignVariable('model', $model);
		$view->assignVariable('form', $form);
	}

	/**
	 * Create a new content page on the site
	 * 
	 * @core-access-string p:/content/manage_all
	 */
	public function create() {

		$view = $this->getView();
		$request = $this->getRequest();

		$model = new ContentModel();
		$page = $model->getLink('Page');
		
		// Allow the user to specify a parent URL to default to.
		// This is used with the "add page 'here'" option.
		if($request->getParameter('parenturl')) {
			$page->set('parenturl', $request->getParameter('parenturl'));
		}
		if($request->getParameter('page_template')) {
			$page->set('page_template', $request->getParameter('page_template'));
		}
		if($request->getParameter('rewriteurl')) {
			$page->set('rewriteurl', $request->getParameter('rewriteurl'));
		}
		if($request->getParameter('title')) {
			$page->set('title', $request->getParameter('title'));
		}

		$form = new Form();
		$form->set('callsmethod', 'ContentController::_SaveHandler');

		$form->addModel($page, 'page');
		$form->addModel($model, 'model');
		
		// Tack on a submit button
		$form->addElement('submit', array('value' => 't:STRING_CREATE'));


		$view->mastertemplate = 'admin';
		$view->templatename = '/pages/content/create.tpl';
		$view->title        = 't:STRING_NEW_CONTENT_PAGE';
		$view->assignVariable('model', $model);
		$view->assignVariable('form', $form);
	}

	public static function _SaveHandler(Form $form) {

		/** @var $model ContentModel */
		$model = $form->getModel('model');
		/** @var $page PageModel Page object for this model, already linked up! */
		$page = $form->getModel('page');

		$isnew = $model->checkIsNew();

		// The content nickname is derived from the page title.
		$model->set('nickname', $page->get('title'));
		if($page->checkExists() && $page->checkIsChanged('site')){
			$model->set('site', $page->get('site'));
		}
		$model->save();

		$page->set('editurl', '/content/edit/' . $model->get('id'));
		$page->set('deleteurl', '/content/delete/' . $model->get('id'));
		$page->set('component', 'content');
		// Force this page to be updated at the time of updating.
		$page->set('updated', \Core\Date\DateTime::NowGMT());
		$page->save();

		// Clear the page cache
		$page->purgePageCache();

		// w00t
		$msg = ($isnew ? 'Added' : 'Updated');
		\Core\set_message($msg . ' page successfully!', 'success');

		$link = \Core\resolve_link($page->get('baseurl'));

		// Redirect to the new page immediately for user convenience.
		return $link;
	}

	/**
	 * @core-request-method POST
	 * @core-accessstring p:/content/manage_all
	 */
	public function delete() {
		$view    = $this->getView();
		$request = $this->getRequest();

		$m    = new ContentModel($request->getParameter(0));
		$link = \Core\resolve_link($m->get('baseurl'));

		if (!$m->exists()){
			throw new NotFoundException();
		}
		$m->delete();

		\Core\set_message('Removed ' . $m->get('nickname') . ' successfully!', 'success');

		$hist = $request->getReferrer();
		if($hist == $link){
			throw new TemporaryFollowRedirectException('/admin/pages');
		}
		else{
			throw new TemporaryFollowRedirectException('BACK');
		}
	}

	public static function GetConfigurationLinks() {
		$ret = [];
		$user = \Core\user();
		$canManage = $user->checkAccess('p:/content/manage_all');

		try {
			$page = PageModel::GetByURI('/');
			if($page->checkExists()) {
				$hp = $page;
			}
			else {
				$hp = null;
			}
		}
		catch(Exception $ex) {
			$hp = null;
		}

		if($canManage) {
			if($hp) {
				$ret[] = [
					'title' => 'Edit Home Page',
					'link' => '/content/edit/' . $hp->id,
					'description' => 'A "home page" was created, this page is any page that has "/" as the rewrite URL.',
				];
			}
			else {
				$ret[] = [
					'title' => 'Create Home Page',
					'link' => '/content/create?rewriteurl=/&title=Home Page',
					'description' => 'Create a "home page", this page is usually a content page, but can be any page that has "/" as the rewrite URL.',
				];
			}
		}

		return $ret;
	}
}

