<?php
/**
 * Class file for the controller CronController
 *
 * @package Cron
 */

use Core\ConfigHandler;
use Core\Controller;
use Core\Date\DateTime;
use Core\Date\Timezone;
use Core\FilterForm;
use Core\Forms\Form;
use Core\HookHandler;
use Core\ListingTable\Table;
use Core\View;


class CronController extends Controller {

	/**
	 * The new auto cron handler, this is safe to call minutely.
	 */
	public function index(){
		$request = $this->getRequest();
		$view    = $this->getView();
		$view->mode = View::MODE_NOOUTPUT;
		// Do not record this page view in the user activity log; it's a system utility.
		$view->record = false;

		$cronsEnabled = ConfigHandler::Get('/cron/enabled');
		if(!$cronsEnabled) {
			echo "Crons are disabled on this server.";
			return;
		}
		
		$crons = [
			'1-minute'  => SECONDS_ONE_MINUTE,
			'5-minute'  => SECONDS_ONE_MINUTE*5,
			'15-minute' => SECONDS_ONE_MINUTE*15,
			'30-minute' => SECONDS_ONE_MINUTE*30,
			'hourly'    => SECONDS_ONE_HOUR,
			'2-hour'    => SECONDS_ONE_HOUR*2,
			'3-hour'    => SECONDS_ONE_HOUR*3,
			'6-hour'    => SECONDS_ONE_HOUR*6,
			'12-hour'   => SECONDS_ONE_HOUR*12,
			'daily'     => SECONDS_ONE_DAY,
			'weekly'    => SECONDS_ONE_WEEK,
			'monthly'   => SECONDS_ONE_MONTH,
		];
		
		// Iterate over each and ensure that it was recently enough.
		// Only process as many as can be ran within 30 seconds.
		// This will ensure that each cron, (with the exception of the 1-minute), is staggered.
		// eg: if the script is called * * * * * like it should be, it'll be executed once-per minute.
		// The first minute, (0:00), 1-minute and 5-minute will run.
		// The next, (0:01), 1-minute and 15-minute will run... etc.
		
		$maxTime = time() + 30;
		$currentTime = DateTime::NowGMT();

		foreach($crons as $cron => $iteration){
			if($maxTime <= time()) {
				// Only process crons if the time allowed is low enough.
				continue;
			}

			$timeVarianceAllowance = 5;

			// daily, weekly, and monthly have additional checks to allow the user to set expected times.
			if($cron === 'daily') {
				// Daily has even more, as due to day light savings time, the number of seconds between the last run may not be 86400.
				$runTime = ConfigHandler::Get('/cron/run/daily');
				if($runTime !== -1) {
					// Daily allows an hour to be specified.
					$timeVarianceAllowance = 3600;
					if($runTime !== (int)DateTime::Now('G')) {
						continue;
					}
				}
			}
			elseif($cron === 'weekly') {
				$runTime = ConfigHandler::Get('/cron/run/weekly');
				if ($runTime !== -1 && $runTime !== (int)DateTime::Now('N')) {
					// Weekly allows a day of the week to be specified.
					continue;
				}
			}
			elseif($cron === 'monthly') {
				$runTime = ConfigHandler::Get('/cron/run/monthly');
				if ($runTime !== -1 && $runTime !== (int)DateTime::Now('j')) {
					// Monthly allows a day of the month to be specified.
					continue;
				}
			}

			$lastRan = ConfigHandler::Get('/cron/last/' . $cron);
			if($lastRan + $iteration - $timeVarianceAllowance > $currentTime) {
				// Last cron of this instance ran recently enough.
				continue;
			}

			$this->_performCron($cron);
			ConfigHandler::Set('/cron/last/' . $cron, $currentTime);
		}
	}
	
	/**
	 * Run a persistent job that's registered on this server.
	 * 
	 * This is meant to be executed on the CLI and will abort otherwise!
	 *
	 * @core-exec-mode CLI
	 */
	public function persistent() {
		$request = $this->getRequest();
		$view = $this->getView();
		
		// Pull the script to run based on the inbound data.
		$jid = $request->getParameter(0);
		$job = CronPersistentJobModel::Construct($jid);
		if(!$job->checkExists()) {
			\Core\log()
				->type('application')
				->code('/cron/persistent')
				->level(LOG_LEVEL_INFO)
				->message('Requested persistent job does not exist - ' . $jid . ', it may have been deleted already')
				->send();
			\Core\CLI\CLI::PrintError('Requested job not found');
			exit(2);
		}

		if($job->server !== SERVER_ID){
			\Core\CLI\CLI::PrintError('Requested job managed by another server');
			exit(2);
		}

		if($job->state !== 'starting') {
			\Core\CLI\CLI::PrintError('Requested job is not marked as starting');
			exit(2);
		}

		// Check this job before blindly running.
		$status = $job->getProcessInformation();
		if($status['state'] === 'running') {
			\Core\CLI\CLI::PrintError('Requested job already running');
			exit(2);
		}


		$job->pid = getmypid();
		$job->state = 'running';
		$job->save();
		
		try{
			// Try to call the requested script!
			if($job->parameters) {
				call_user_func_array($job->call, $job->parameters);
			}
			else {
				call_user_func($job->call);
			}
		}
		catch (Exception $ex) {
			$job = new CronPersistentJobModel($jid);
			$job->state = 'stopped';
			$job->pid = 0;
			$job->save();

			\Core\log_exception($ex);
			exit(3);
		}
		
		// After it's done, mark it as complete!
		// This will only be hit if the called function returns.
		$job = new CronPersistentJobModel($jid);
		$job->state = 'stopped';
		$job->pid = 0;
		$job->save();
		exit(0);
	}

	/**
	 * @core-access-string g:admin
	 *
	 * @todo FINISH
	 */
	public function admin() {
		$view = $this->getView();
		$request = $this->getRequest();


		$jobs = [];


		$crons = [
			'1-minute',
			'5-minute',
			'15-minute',
			'30-minute',
			'hourly',
			'2-hour',
			'3-hour',
			'6-hour',
			'12-hour',
			'daily',
			'weekly',
			'monthly',
		];
		foreach($crons as $cron) {
			$hook = HookHandler::GetHook('/cron/' . $cron);
			$bindings = $hook->getBindings();

			foreach($bindings as $binding) {
				$jobs[] = [
					'title' => $binding['call'],
					'hook' => $hook->description,
					'code' => '/cron/' . $cron,
				];
			}
		}

		$persistent = CronPersistentJobModel::Find();
		foreach($persistent as $cron) {
			$jobs[] = [
				'title' => $cron->call,
				'hook' => 'Persistent Background Process',
				'code' => '/cron/persistent',
			];
		}

		var_dump($jobs); die();
	}

	/**
	 * @core-access-string g:admin
	 */
	public function configure() {
		$view = $this->getView();
		$request = $this->getRequest();

		$keys = [
			'/cron/enabled',
			'/cron/run/daily',
			'/cron/run/weekly',
			'/cron/run/monthly',
		];

		$form = new Form();
		$form->set('callsmethod', 'AdminController::_ConfigSubmit');

		foreach($keys as $k){
			$c = ConfigHandler::GetConfig($k);
			$f = $c->asFormElement();
			// Don't need them grouped
			$f->set('group', '');
			$form->addElement($f);
		}
		$form->addElement('submit', ['value' => t('STRING_SAVE')]);

		// Modify the default form elements to be more friendly.
		$form->switchElementType('config[/cron/run/daily]', 'select');
		$form->switchElementType('config[/cron/run/weekly]', 'select');
		$form->switchElementType('config[/cron/run/monthly]', 'select');

		$daily = $form->getElement('config[/cron/run/daily]');
		$options = ['-1' => 'Any time of day'];
		for($i = 0; $i <= 23; $i++) {
			$options[$i] = 'During hour ' . $i;
		}
		$daily->set('options', $options);

		$weekly = $form->getElement('config[/cron/run/weekly]');
		$options = [
			'-1' => 'Any day of the week',
			'1' => 'Monday',
			'2' => 'Tuesday',
			'3' => 'Wednesday',
			'4' => 'Thursday',
			'5' => 'Friday',
			'6' => 'Saturday',
			'7' => 'Sunday',
		];
		$weekly->set('options', $options);

		$monthly = $form->getElement('config[/cron/run/monthly]');
		$options = ['-1' => 'Any day of the month'];
		for($i = 1; $i <= 28; $i++) {
			$options[$i] = 'On day ' . $i;
		}
		$monthly->set('options', $options);


		$view->title = 't:STRING_CRON_CONFIGURE';
		$view->assign('url', ROOT_URL_SSL);
		$view->assign('sitename', SITENAME);
		$view->assign('form', $form);
	}


	/**
	 * Execute the actual cron for the requested type.
	 *
	 * @param string $cron Cron type to execute.
	 *
	 * @return void
	 * @throws Exception
	 */
	private function _performCron(string $cron) {
		$start = microtime(true);
		$sep = '==========================================' . "\n";
		$contents = "Starting cron execution for $cron\n$sep";

		// This uses the hook system, but will be slightly different than most things.
		$hook = HookHandler::GetHook('/cron/' . $cron);
		$hookcount = 0;
		$hooksuccesses = 0;

		if($hook) {
			if($hook->getBindingCount()) {
				$hookcount = $hook->getBindingCount();
				$bindings = $hook->getBindings();
				foreach($bindings as $b) {
					$contents .= sprintf(
						"\nExecuting Binding %s...\n",
						$b['call']
					);
					// Since these systems will just be writing to STDOUT, I'll need to capture that.
					try{
						ob_start();
						$execution = $hook->callBinding($b, array());
						$executiondata = ob_get_clean();
					}
					catch(Exception $ex){
						$execution     = false;
						$executiondata = 'EXCEPTION: ' . $ex->getMessage() . ob_get_clean();
					}
					catch(Error $er) {
						$execution     = false;
						$executiondata = 'EXCEPTION: ' . $er->getMessage() . ob_get_clean();
					}


					if($executiondata == '' && $execution){
						$contents .= "Cron executed successfully with no output\n";
						++$hooksuccesses;
					}
					elseif($execution){
						$contents .= $executiondata . "\n";
						++$hooksuccesses;
					}
					else{
						$contents .= $executiondata . "\n!!FAILED\n";
					}
				}
			}
			else{
				$contents = 'No bindings located for requested cron';
			}
		}
		else{
			$contents = 'Invalid hook requested: ' . $cron;
		}


		// Just in case the contents are returning html... (they should be plain text).
		// Replace the most common line endings with things that make sense for plain text.
		// This is to ensure that all the available scenarios are met and saved/displayed without extra whitespace.
		//
		// Since some systems will provide plain text (easy!), windows/os9 line endings,
		// HTML (br and br/), and formatted HTML (br + \n). 
		$contents = str_ireplace(
			[
				"\r\n<br>", "\r\n<br/>", "\r\n<br />",
				"\n<br>", "\n<br/>", "\n<br />",
				"<br>", "<br/>", "<br />",
			    "\r\n", "\r"
			],
			"\n",
			$contents
		);

		// Make a copy of this in the system log too if applicable.
		// This time is listed in ms
		$time = \Core\time_duration_format(microtime(true) - $start, 0);

		$msg = 'Cron ' . $cron . ' completed in ' . $time . '.  ' . $hooksuccesses . ' out of ' . $hookcount . ' hooks called successfully.';
		SystemLogModel::LogInfoEvent('/cron/' . $cron, $msg, $contents);
	}
}
