/**
 * Inline style editor for Core Plus,
 * allows live editing of CSS variables in real time
 */

// a = window.getMatchedCSSRules(document.querySelector('.page-header'))
// Object.values(a[0].style).forEach(e => {console.log(e, a[0].style[e]); })

(() => {

	let styleEditorUI;

	/**
	 * Get all the fonts available on the system
	 */
	function getFonts() {
		let fonts = document.fonts.entries(),
			font = fonts.next(),
			fontsAvailable = {},
			fallbackFonts = [
				'sans-serif',
				'Arial, sans-serif',
				'Arial Black, sans-serif',
				'Verdana, sans-serif',
				'Tahoma, sans-serif',
				'Trebuchet MS, sans-serif',
				'Impact, sans-serif',
				'serif',
				'Times New Roman, serif',
				'Didot, serif',
				'Georgia, serif',
				'American Typewriter, serif',
				'monospace',
				'Lucida Console, monospace',
				'Monaco, monospace',
				'cursive',
				'Bradley Hand, cursive',
				'Brush Script MT, cursive',
				'fantasy',
				'Luminari, fantasy'
			];

		while(!font.done) {
			for(let i = 0; i < font.value.length; i++) {
				let fontName = font.value[i].family,
					fontWeight = font.value[i].weight;

				// Remove quotes around the font names.
				fontName = fontName.replaceAll('"', '');

				if(fontName === 'Font Awesome 5 Free' || fontName === 'Font Awesome 5 Brands') {
					// Skip some fonts that don't need to display
					continue;
				}

				fontsAvailable[ fontName ] = [];
				fontsAvailable[ fontName ].push(fontWeight);
			}
			font = fonts.next();
		}

		Object.keys(fontsAvailable).forEach(fontName => {
			fontsAvailable[ fontName ] = fontName + ' (' + fontsAvailable[ fontName ].join(', ') + ')';
		})

		fontsAvailable[''] = 'No Font Selected';

		return [fontsAvailable, fallbackFonts];
	}

	/**
	 * Parse a value for its size and suffix
	 *
	 * Value is key 0
	 * Suffix is key 1
	 *
	 * @param {string }value
	 * @returns {(string)[]}
	 */
	function parseSize(value) {
		if(value === 0 || value === '0' || value === '') {
			return [
				'0',
				'px'
			];
		}
		else if(value.match(/px$/) !== null) {
			return [
				value.replace(/px$/, ''),
				'px'
			];
		}
		else if(value.match(/pt$/) !== null) {
			return [
				value.replace(/pt$/, ''),
				'pt'
			];
		}
		else if(value.match(/rem$/) !== null) {
			return [
				value.replace(/rem$/, ''),
				'rem'
			];
		}
		else if(value.match(/em$/) !== null) {
			return [
				value.replace(/em$/, ''),
				'em'
			];
		}
		else {
			return [
				value,
				''
			];
		}
	}

	/**
	 * Generate the processed value for a given value row
	 * 
	 * @param {Element} row
	 *
	 * @return {string}
	 */
	function generateValue(row) {
		let valueInput = row.querySelector('.value'),
			values = row.querySelector('.values'),
			valueSuffix = row.querySelector('.value-suffix');

		if(values) {
			// This row is a multi value
			let newValues = [];
			values.querySelectorAll('.multi-value').forEach(el => {
				let newValue = el.value;

				if(values.querySelector('[data-suffix-key="' + el.dataset.valueKey + '"]')) {
					let suffix = values.querySelector('[data-suffix-key="' + el.dataset.valueKey + '"]');
					if(suffix.dataset.suffixJoin) {
						newValue += suffix.dataset.suffixJoin;
					}

					newValue += suffix.value;
				}

				newValues.push(newValue);
			});

			console.log(newValues);
			return newValues.join(' ');
		}
		else {
			// Single value
			let newValue = valueInput.value;

			if(valueSuffix && valueSuffix.value) {

				if(valueSuffix.dataset.suffixJoin) {
					newValue += valueSuffix.dataset.suffixJoin;
				}

				newValue += valueSuffix.value;
			}

			console.log(newValue);
			return newValue;
		}
	}

	/**
	 * @param {Element} target
	 * @param {string[]} options
	 */
	function createStyleEditorUI(target, options) {
		styleEditorUI = document.createElement('div');
		let html = '',
			title,
			renderedKeys = [],
			fonts = getFonts(),
			fontWeights = {
				'inherit': 'Inherit From Parent',
				'100': 'Extra Light (100)',
				'200': 'Light (200)',
				'300': 'Book (300)',
				'400': 'Normal (400)',
				'500': 'Medium (500)',
				'600': 'Semibold (600)',
				'700': 'Bold (700)',
				'800': 'Extra Bold (800)',
				'900': 'Fat (900)'
			},
			fontWeightAliases = {
				'normal': '400',
				'bold': '700',
				'': 'inherit'
			},
			borderStyles = [
				'initial',
				'none',
				'hidden',
				'dotted',
				'dashed',
				'solid',
				'double',
				'groove',
				'ridge',
				'inset',
				'outset',
				'inherit'
			],
			textAligns = {
				'initial': 'Default',
				'left': 'Left Align',
				'right': 'Right Align',
				'center': 'Center Align',
				'justify': 'Justify Text',
				'inherit': 'Inherit From Parent'
			},
			sizeSuffixes = {
				'px': 'PX',
				'pt': 'PT',
				'em': 'EM',
				'REM': 'REM'
			};

		title = target.nodeName;
		if(target.className) {
			title += '.' + target.className.replaceAll(' ', '.');
		}

		html += '<h2>' + title + ' Styles</h2>';


		for(let i = options.length - 1; i >= 0; i--) {
			let varName = options[i],
				varNameProcessed = varName,
				varNameSuffix = null,
				varValue = window.getComputedStyle(target).getPropertyValue('--' + options[i]),
				varValues = null,
				varType = 'text',
				varStep = '1',
				varKey = null,
				varSuffix;

			// Take out redundant modifier fields
			if(varNameProcessed.endsWith('-active')) {
				varNameSuffix = '(Active)';
				varNameProcessed = varNameProcessed.replace(/-active$/, '');
			}
			else if(varNameProcessed.endsWith('-hover')) {
				varNameSuffix = '(Hover)';
				varNameProcessed = varNameProcessed.replace(/-hover$/, '');
			}
			else if(varNameProcessed.endsWith('-disabled')) {
				varNameSuffix = '(Disabled)';
				varNameProcessed = varNameProcessed.replace(/-disabled$/, '');
			}
			else if(varNameProcessed.endsWith('-selected')) {
				varNameSuffix = '(Selected)';
				varNameProcessed = varNameProcessed.replace(/-selected$/, '');
			}

			// Retrieve some information about the variable based on its name
			if(varNameProcessed.endsWith('-border-color')) {
				varType = 'color';
				varNameProcessed = 'Border Color';
			}
			else if(varNameProcessed.endsWith('-background-color')) {
				varType = 'color';
				varNameProcessed = 'Background Color';
			}
			else if(varNameProcessed.endsWith('-color')) {
				varType = 'color';
				varNameProcessed = 'Color';
			}
			else if(varNameProcessed.endsWith('-font-size')) {
				varType = 'size';
				varNameProcessed = 'Font Size';
			}
			else if(varNameProcessed.endsWith('-border-width')) {
				varType = 'size';
				varNameProcessed = 'Border Width';
			}
			else if(varNameProcessed.endsWith('-line-height')) {
				varType = 'size';
				varNameProcessed = 'Line Height';
			}
			else if(varNameProcessed.endsWith('-font-weight')) {
				varType = 'font-weight';
				varNameProcessed = 'Font Weight';
			}
			else if(varNameProcessed.endsWith('-font-family')) {
				varType = 'font-family';
				varNameProcessed = 'Font Family';
			}
			else if(varNameProcessed.endsWith('-border-style')) {
				varType = 'border-style';
				varNameProcessed = 'Border Style';
			}
			else if(varNameProcessed.endsWith('-text-align')) {
				varType = 'text-align';
				varNameProcessed = 'Text Align';
			}
			else if(varNameProcessed.endsWith('-padding')) {
				varType = 'sizes';
				varNameProcessed = 'Padding';
			}


			// Color expects a full 6 digit code
			if(varType === 'color' && varValue.match(/^#[0-9a-fA-F][0-9a-fA-F][0-9a-fA-F]$/) !== null) {
				varValue = varValue.replace(/^#([0-9a-fA-F])([0-9a-fA-F])([0-9a-fA-F])$/, '#$1$1$2$2$3$3');
			}

			// Font weight can be aliased.
			if(varType === 'font-weight') {
				if(typeof(fontWeightAliases[ varValue ]) !== 'undefined') {
					varValue = fontWeightAliases[ varValue ];
				}
			}

			// Sizes support a unit suffix
			if(varType === 'size') {
				let dat = parseSize(varValue);
				if(dat !== null) {
					varValue = dat[0];
					varSuffix = dat[1];
				}
				else {
					varType = 'text';
				}
			}

			// Sizes are complicated; they support 0 value, 2 values, 3 values, or 4 values.
			if(varType === 'sizes') {
				let s = varValue.trim().split(' ');

				if(s.length === 1) {
					// top/left/bottom/right
					varValues = [
						parseSize(s[0]),
						parseSize(s[0]),
						parseSize(s[0]),
						parseSize(s[0])
					];
				}
				else if(s.length === 2) {
					// top/bottom, left/right
					varValues = [
						parseSize(s[0]),
						parseSize(s[1]),
						parseSize(s[0]),
						parseSize(s[1])
					];
				}
				else if(s.length === 3) {
					// top, left/right, bottom
					varValues = [
						parseSize(s[0]),
						parseSize(s[1]),
						parseSize(s[2]),
						parseSize(s[1])
					];
				}
				else if(s.length === 4) {
					// top, right, bottom, left
					varValues = [
						parseSize(s[0]),
						parseSize(s[1]),
						parseSize(s[2]),
						parseSize(s[3])
					];
				}
				else {
					varType = 'text';
				}
			}

			// Font families should have fallback fonts.
			if(varType === 'font-family') {
				if(fonts[1].indexOf(varValue) !== -1) {
					// The full font is selected as a fallback only with no custom font.
					varValue = '';
					varSuffix = varValue;
				}
				else if(varValue.indexOf(',') !== -1) {
					// Contains a fallback.
					varSuffix = varValue.substring(varValue.indexOf(',') + 1).trim();
					varValue = varValue.substr(0, varValue.indexOf(',')).trim();
				}
				else {
					varSuffix = '';
				}
			}

			html += '<label data-var="' + varName + '" title="' + varName + '">' +
				'<span>' + varNameProcessed +
				(varNameSuffix !== null ? ' ' + varNameSuffix : '') +
				'</span> ';

			if(varType === 'text') {
				html += '<input class="value" type="text" value="' + varValue + '"/>';
			}
			else if(varType === 'color') {
				html += '<input class="value" type="color" value="' + varValue + '"/>';
			}
			else if(varType === 'size') {
				html += '<input class="value" type="number" step="0.05" value="' + varValue + '"/>';
				if(varSuffix) {
					html += '<select class="value-suffix" data-suffix-join="">';
					Object.keys(sizeSuffixes).forEach(k => {
						html += '<option value="' + k + '"' + (varValue === k ? ' selected="selected"' : '') + '>' + sizeSuffixes[k] + '</option>';
					});
					html += '</select>';
				}
			}
			else if(varType === 'sizes') {
				let l = [
					[ 'icon-chevron-up', 'Top' ],
					[ 'icon-chevron-right', 'Right' ],
					[ 'icon-chevron-down', 'Bottom' ],
					[ 'icon-chevron-left', 'Left' ]
				];

				html += '<span class="values">';

				for(let i = 0; i < varValues.length; i++) {
					html += '<i class="icon ' + l[i][0] + '" title="' + l[i][1] + '"></i>';
					html += '<input class="multi-value" type="number" data-value-key="' + i + '" step="0.05" value="' + varValues[i][0] + '"/>';
					if(varValues[i][1]) {
						html += '<select class="value-suffix" data-suffix-key="' + i + '" data-suffix-join="">';
						Object.keys(sizeSuffixes).forEach(k => {
							html += '<option value="' + k + '"' + (varValues[i][1] === k ? ' selected="selected"' : '') + '>' + sizeSuffixes[k] + '</option>';
						});
						html += '</select>';
					}
				}

				html += '</span>';
			}
			else if(varType === 'font-weight') {
				html += '<select class="value">';
				Object.keys(fontWeights).forEach(k => {
					html += '<option value="' + k + '"' + (varValue === k ? ' selected="selected"' : '') + '>' + fontWeights[k] + '</option>';
				});
				html += '</select>';
			}
			else if(varType === 'font-family') {
				html += '<select class="value">';
				Object.keys(fonts[0]).forEach(k => {
					html += '<option value="' + k + '"' + (varValue === k ? ' selected="selected"' : '') + '>' + fonts[0][k] + '</option>';
				});
				if(typeof(fonts[0][varValue]) === 'undefined') {
					// Font not registered, at least provide an option for it!
					html += '<option value="' + varValue + ' selected="selected">' + varValue + ' (Font Not Found!)</option>';
				}
				html += '</select>';

				html += '<select class="value-suffix" data-suffix-join=", ">';
				for(let i = 0; i < fonts[1].length; i++) {
					html += '<option value="' + fonts[1][i] + '"' + (varSuffix === fonts[1][i] ? ' selected="selected"' : '') + '>' + fonts[1][i] + '</option>';
				}
				html += '</select>';
			}
			else if(varType === 'border-style') {
				html += '<select class="value">';
				for(let i = 0; i < borderStyles.length; i++) {
					html += '<option value="' + borderStyles[i] + '"' + (varValue === borderStyles[i] ? ' selected="selected"' : '') + '>' + borderStyles[i] + '</option>';
				}
				html += '</select>';
			}
			else if(varType === 'text-align') {
				html += '<select class="value">';
				Object.keys(textAligns).forEach(k => {
					html += '<option value="' + k + '"' + (varValue === k ? ' selected="selected"' : '') + '>' + textAligns[k] + '</option>';
				});
				html += '</select>';
			}
			else {
				html += '[INVALID TYPE - ' + varType + ']';
			}


			html += '</label>'
		}
		
		html += '<label><a href="#" class="save button">Save Settings</a></label>';

		styleEditorUI.classList.add('mediamanager-styleeditor-ui');
		styleEditorUI.innerHTML = html;
		document.body.appendChild(styleEditorUI);

		// Update the position.
		styleEditorUI.style.marginTop = window.pageYOffset + 'px';

		styleEditorUI.querySelectorAll('input,select').forEach(inp => {
			inp.addEventListener('change', evt => {
				let label = evt.target.closest('label'),
					newValue = generateValue(label);

				document.documentElement.style.setProperty('--' + label.dataset.var, newValue);
			});
		});

		styleEditorUI.querySelector('.save').addEventListener('click', evt => {
			evt.preventDefault();

			let values = {};
			styleEditorUI.querySelectorAll('label').forEach(l => {
				if(l.dataset.var) {
					// This label has a variable attached to it!
					values[ l.dataset.var ] = generateValue(l);
				}
			});

			Core.Ajax(
				Core.ROOT_URL + 'mediamanagerstyles/save',
				{
					method: 'POST',
					data: values,
					dataType: 'json',
					callback: ret => {
						if(ret.status === 200) {
							$.notify('Updated styles', 'success');
						}
						else {
							$.notify(ret.message, 'error');
						}
					}
				}
			);

			document.body.removeChild(styleEditorUI);
			styleEditorUI = null;
		});
	}

	/**
	 * Select an element to inspect and open the dialog
	 *
	 * @param {Element} inspectElement
	 */
	function selectElement(inspectElement) {
		let targets = [],
			variables = [],
			manualStyles = [
				'border-color',
				'border-width',
				'border-style',
				'padding'
			];

		// Build a list of parent nodes; we'll do this so we can reverse through them to compile an accurate cascade.
		while(inspectElement !== document.body.parentElement) {
			targets.push(inspectElement);
			inspectElement = inspectElement.parentElement;
		}

		for(let t = 0; t < targets.length; t++) {
			//console.log('Processing CSS rules for', targets[t]);

			let rules = getMatchedCSSRules(targets[t]);

			for(let r = rules.length - 1; r >= 0; r--) {
				// Some styles are difficult and refuse to expose themselves.
				for(let s = 0; s < manualStyles.length; s++) {
					let styleValue = rules[r].style[ manualStyles[s] ], styleVariable;
					//console.log(s, styleValue);

					if(styleValue.indexOf('var(--') !== -1) {
						styleVariable = styleValue.replace(/var\(--([a-zA-Z\-0-9]*).*\)/, '$1').trim();

						variables.push(styleVariable);
					}
				}

				Object.values(rules[r].style).forEach(s => {
					let styleValue = rules[r].style[s], styleVariable;
					//console.log(s, styleValue);

					if(styleValue.indexOf('var(--') !== -1) {
						styleVariable = styleValue.replace(/var\(--([a-zA-Z\-0-9]*).*\)/, '$1').trim();

						variables.push(styleVariable);
					}
				});
			}

			if(variables.length > 0) {
				// Once there are variables, break out of the loop.
				// This way, we can sift up through the DOM stack until we find something that's editable.
				inspectElement = targets[t];
				break;
			}
		}

		if(!variables.length) {
			$.notify('Selected element contains no customizable styles', 'error');
			return;
		}

		// Remove duplicates
		variables = [...new Set(variables)];

		createStyleEditorUI(inspectElement, variables);
	}

	document.addEventListener('click', evt => {
		if(evt.target.closest('.mediamanager-styleeditor-ui')) {
			// Click target was on the editor UI itself
			return;
		}

		if(evt.target.closest('.mediamanager-styleeditor-toggle')) {
			// Click target was the fixed controls
			return;
		}

		if(!document.body.classList.contains('mediamanager-styleeditor-enabled')) {
			// Body needs the class to show it's enabled.

			if(styleEditorUI) {
				// Close the currently open window.
				document.body.removeChild(styleEditorUI);
				styleEditorUI = null;
			}

			return;
		}

		document.body.classList.remove('mediamanager-styleeditor-enabled');
		evt.preventDefault();
		selectElement(evt.target);
	});


	document.addEventListener('contextmenu', evt => {

		if(!document.body.classList.contains('mediamanager-styleeditor-enabled')) {
			// Body needs the class to show it's enabled.
			return;
		}

		document.body.classList.remove('mediamanager-styleeditor-enabled');
		evt.preventDefault();
		selectElement(evt.target);
	});



	let closeUI = document.createElement('div');
	closeUI.classList.add('mediamanager-styleeditor-toggle');
	closeUI.innerHTML = '<a href="#" class="button toggle"><i class="icon icon-eye-dropper"></i> <span>Inspect Element</span></a>' +
		'<a href="' + Core.ROOT_URL + 'mediamanagerstyles/toggle' + '" class="button"><i class="icon icon-check"></i> <span>Finish Editing</span></a>';
	document.body.appendChild(closeUI);

	closeUI.querySelector('.toggle').addEventListener('click', evt => {
		document.body.classList.toggle('mediamanager-styleeditor-enabled');
		evt.preventDefault();
	});

	//alert('Inline style editing is enabled!' + "\n\n" + 'To change how something looks, simply click on it and you will be presented with options.');
})();