<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace MediaManager;

use Core\AccessDeniedException;
use Core\Filestore\Contents\ContentImage;
use Core\Filestore\Contents\ContentVideo;
use Core\Filestore\FileMetaHelper;
use Core\Forms\Form;
use Core\Forms\FormElement;
use Core\NotFoundException;
use Core\View;
use Exception;
use Core\BadRequestException;
use Core\Core;
use Core\ConfigHandler;
use Core\Date\DateTime;
use Core\Filestore\Directory;
use Core\Filestore\File;
use Core\Filestore\Factory as FileFactory;
use Core\Theme;
use MediaManager\Navigators\PageNavigator;
use CodeMirror;

/**
 * Description of Navigator2
 *
 * @author charlie
 */
class Navigator2 {
	
	protected $_isAdmin   = false;
	
	/**
	 * @var bool Permission for access, update, and create
	 */
	protected $_canUpload = false;
	
	/**
	 * @var bool Permission for access, and update, (EXISTING files only)
	 */
	protected $_canUpdate = false;
	
	/**
	 * @var bool Permission for access
	 */
	protected $_canAccess = false;
	protected $_basedir;
	protected $_cwd       = '';
	protected $_ignores   = [];
	protected $_type;
	
	protected $_listings;

	
	/**
	 * Get the relative base directory for media lookups.
	 * 
	 * This is usually "assets/", "private/", "public/", or some other Core-supported prefix.
	 * 
	 * @return string
	 */
	public function getBaseDir() : string {
		return $this->_basedir;
	}
	
	/**
	 * Get slightly more inclusive information about a subset of files+directories for this navigator.
	 * 
	 * This is used to reduce the number of operations required on the initial listing,
	 * because if the user never browses to a subdirectory, there's no need to load the image information.
	 * 
	 * @return array
	 */
	public function getInfo(array $paths) : array {
		if(!$this->_canAccess){
			// No access, no files.
			return [];
		}
		
		$results = [];
		$dir = FileFactory::Directory($this->getBaseDir());
		$basedir = $dir->getPath();
		$dirlen = strlen($dir->getPath());
		
		foreach($paths as $rel){
			$path = $basedir . $this->_sanitizeFilename($rel);
			
			if(!file_exists($path)){
				// Simply skip invalid/unlocated files.
				continue;
			}
			elseif(is_dir($path)){
				$entry = FileFactory::Directory($path);
				$rel = substr($entry->getPath(), $dirlen);
				
				$results[] = $this->_getDirectoryInfo($entry, $rel);
			}
			elseif(is_file($path)){
				$entry = FileFactory::File($path);
				$rel = substr($entry->getFilename(), $dirlen);
				
				$results[] = $this->_getFileInfo($entry, $rel);
			}
		}
		
		return $results;
	}

	public function getCWD() : string {
		return $this->_cwd;
	}
	
	/**
	 * Get an array of file and directory listings for this lookup
	 * 
	 * @return array
	 */
	public function getListing() : array {
		if(!$this->_canAccess){
			// No access, no files.
			return [];
		}
		
		$this->_listings = [];
		
		$dir     = FileFactory::Directory($this->getBaseDir());
		$dirlen  = strlen($dir->getPath());
		
		
		foreach($dir->ls(null, true) as $entry){
			if($entry instanceof Directory){
				$path = $entry->getPath();
				
				if(array_search($path, $this->_ignores) !== false){
					// Skip certain directories here that are unimportant.
					continue;
				}
				
				$rel = substr($path, $dirlen);
				$parent = dirname($rel) . '/';
				
				if($parent === './'){
					$parent = '';
				}
				
				$this->_addDirectoryToListings($rel, $parent, filemtime($path));
			}
			elseif($entry instanceof File){
				
				foreach($this->_ignores as $skip){
					if($entry->inDirectory($skip)){
						// Skip files in skipped directories.
						continue 2;
					}
				}
				
				$path = $entry->getFilename();
				$rel  = substr($path, $dirlen);
				$parent = dirname($rel) . '/';
				
				if($parent === './'){
					$parent = '';
				}
				
				$this->_addFileToListings($entry, $rel, $parent);
			}
		}
		
		// Sorting would be nice, just because.
		ksort($this->_listings, SORT_STRING | SORT_FLAG_CASE);
		
		return array_values($this->_listings);
	}

	public function getPermissions(string $path = null) : array {
		$ret = [
			'up' => $this->_cwd !== '',
			'mkdir' => $this->checkCanMkdir(),
			'upload' => $this->checkCanUpload(),
			'touch' => $this->checkCanTouch(),
			'rename' => false,
			'rm' => false,
		];

		if($path) {
			$ret['rename'] = $this->checkCanRename($path);
			$ret['rm'] = $this->checkCanDelete($path);
		}

		return $ret;
	}
	
	/**
	 * Get the entry of a given path as either a File or Directory.
	 * 
	 * NULL is returned if the requested path does not match anything.
	 * 
	 * @param string $path Requested path to lookup
	 * @return File|Directory|null
	 */
	public function getEntry(string $path) {
		if(!$this->_canAccess){
			// No access, no files.
			return null;
		}
		
		$results = [];
		$dir = FileFactory::Directory($this->getBaseDir());
		$basedir = $dir->getPath();
		$dirlen = strlen($dir->getPath());
		
		$path = $basedir . $this->_sanitizeFilename($path);

		if(!file_exists($path)){
			// Simply skip invalid/unlocated files.
			return null;
		}
		elseif(is_dir($path)){
			return FileFactory::Directory($path);
		}
		elseif(is_file($path)){
			return FileFactory::File($path);
		}
		else{
			// umm
			return null;
		}
	}

	/**
	 * Get the details for an entry, useful for rendering in the details view template
	 *
	 * @param string $path
	 * @return array|null
	 */
	public function getEntryDetails(string $path) : ?array {
		$entry = $this->getEntry($path);
		if(!$entry) {
			return null;
		}

		$details = [
			'preview' => null,
			'name' => $entry->getBasename(),
			'type' => null,
			'attributes' => [],
			'data' => [],
		];

		if($entry instanceof Directory) {
			$details['preview'] = $entry->getMimetypeIconURL('200x200');
			$details['type'] = 'Directory';
			$details['attributes']['Size'] = t('STRING_MEDIAMANAGER_DIR_N_ITEMS', $this->getEntryChildrenCount($path));
			$details['attributes']['Modified'] = DateTime::FormatString($entry->getMTime(), 'SDT');
		}
		else {

			$details['attributes']['URL'] = $entry->getURL();

			$contents = $entry->getContentsObject();

			$details['preview'] = $entry->isImage() ? $entry->getPreviewURL('200x200') : $entry->getMimetypeIconURL('200x200');

			try {
				$metas = new FileMetaHelper($entry);
				foreach($metas->getMetas() as $meta) {
					$details['attributes'][ $meta->meta_key ] = $meta->meta_value_title;
				}
			}
			catch(Exception $ex) {
				// Continue.
			}

			$details['type'] = $entry->getMimetype();
			$details['attributes']['Size'] = $entry->getFilesize(true);
			$details['attributes']['Modified'] = DateTime::FormatString($entry->getMTime(), 'SDT');

			if($entry->isImageRaster() && $contents instanceof ContentImage) {
				$details['attributes']['Dimensions'] = $contents->getDimensions();
				list($width, $height) = explode('x', $contents->getDimensions());
				$details['data']['width'] = (int)$width;
				$details['data']['height'] = (int)$height;

				$exif = $contents->getExif();
				if(count($exif->fields)) {
					$details['attributes']['Image EXIF Metadata'] = '';
					foreach($exif->fields as $field) {
						$details['attributes'][ t($field->title) ] = $field->formatted;
					}
				}
			}
			elseif($contents instanceof ContentVideo) {
				if($contents->getBitrate()) {
					$details['attributes']['Bitrate'] = $contents->getBitrate();
				}

				if($contents->getDuration()) {
					$parts = explode(':', $contents->getDuration());
					$length = (int)$parts[2];

					if((int)$parts[1] > 0) {
						$length += ((int)$parts[1]) * 60;
					}

					if((int)$parts[0] > 0) {
						$length += ((int)$parts[0]) * 3600;
					}

					$details['attributes']['Length'] = \Core\Formatter\GeneralFormatter::TimeDuration($length);
				}

				$streams = $contents->getStreams();
				foreach($streams as $stream) {
					if($stream['codec']) {
						$details['attributes']['Stream ' . $stream['id']] = ucfirst($stream['type']) . ' - ' . $stream['codec'];
					}
					else {
						$details['attributes']['Stream ' . $stream['id']] = ucfirst($stream['type']);
					}

					if($stream['dimensions']) {
						$details['attributes'][$stream['id'] . ' Size'] = $stream['dimensions'];

						// This is a hack to get the video width/height, as videos can technically
						// contain multiple video resolutions in the same container.
						// This however will probably be a fringe case for web, as it will greatly increase
						// the filesizes of videos, which is not something commonly desired.
						list($width, $height) = explode('x', $stream['dimensions']);
						$details['data']['width'] = (int)$width;
						$details['data']['height'] = (int)$height;
					}
					if($stream['bitrate']) {
						$details['attributes'][$stream['id'] . ' Bitrate'] = $stream['bitrate'];
					}
					if($stream['frequency']) {
						$details['attributes'][$stream['id'] . ' Frequency'] = $stream['frequency'];
					}
					if($stream['channels']) {
						$details['attributes'][$stream['id'] . ' Channels'] = $stream['channels'];
					}
					if(isset($stream['meta']['handler_name']) && $stream['meta']['handler_name']) {
						$details['attributes'][$stream['id'] . ' Handler'] = $stream['meta']['handler_name'];
					}
					if(isset($stream['meta']['encoder']) && $stream['meta']['encoder']) {
						$details['attributes'][$stream['id'] . ' Encoder'] = $stream['meta']['encoder'];
					}
				}
			}
		}

		return $details;
	}

	public function getEntryChildrenCount(string $path) : int {
		$entry = $this->getEntry($path);

		return ($entry instanceof Directory) ? count($entry->ls()) : 0;
	}
	
	/**
	 * Get a human-friendly and translated title for this Navigator based on the type
	 * 
	 * @return string
	 */
	public function getTitle() : string {
		return t('STRING_MEDIA_FILE_BROWSER');
	}

	/**
	 * Get the URL of the respective navigator
	 */
	public function getURL() : string {
		return '/mediamanagernavigator';
	}
	
	/**
	 * Get the key-version of the type of this Navigator
	 * 
	 * @return string
	 */
	public function getType() : string {
		return $this->_type;
	}
	
	/**
	 * Check if the current user has permission to *view* files
	 * 
	 * @return bool
	 */
	public function checkHasAccess() : bool {
		return $this->_canAccess;
	}
	
	/**
	 * Check if the current user has permission to upload files in the CWD
	 *
	 * Directory-level permission
	 * 
	 * @return bool
	 */
	public function checkCanUpload() : bool {
		return $this->_canUpload;
	}

	/**
	 * Check if the user can perform mkdir in the CWD
	 *
	 * Directory-level permission
	 *
	 * @return bool
	 */
	public function checkCanMkdir() : bool {
		return $this->_canUpload;
	}

	/**
	 * Check if the user can create a new file in the CWD
	 *
	 * Directory-level permission
	 *
	 * @return bool
	 */
	public function checkCanTouch() : bool {
		return $this->_canUpload;
	}

	/**
	 * Check if the user can rename the provided file
	 *
	 * Entry-level permission
	 *
	 * @return bool
	 */
	public function checkCanRename(string $path) : bool {
		return $this->_canUpdate;
	}

	/**
	 * Check if the user can delete the provided file
	 *
	 * Entry-level permission
	 *
	 * @return bool
	 */
	public function checkCanDelete(string $path) : bool {
		return $this->_canUpdate;
	}
	
	/**
	 * Check if the current user has permission to edit *existing* files
	 * 
	 * @return bool
	 */
	public function checkCanUpdate() : bool {
		return $this->_canUpdate;
	}
	
	public function setCWD(string $directory) {
		// No working directory should ever have '..' in the filename.
		$this->_cwd = str_replace('..', '', $directory);
		
		if($this->_cwd !== '' && !\Core\str_ends_with($this->_cwd, '/')){
			$this->_cwd .= '/';
		}
	}
	
	/**
	 * Make a directory in the current working directory.
	 *
	 * @param $directory
	 *
	 * @throws \Exception
	 * 
	 * @return string relative path of the newly created directory, useful for browsing immediately to it.
	 */
	public function mkdir(string $directory) : string {
		if(!$this->checkCanUpload()){
			throw new AccessDeniedException();
		}
		
		// Sanitize the filename
		if(
			strpos($directory, '..') !== false ||
			strpos($directory, '/') !== false
		){
			throw new Exception('Invalid directory name');
		}

		$base = $this->getBaseDir();
		$directory = trim(\Core\str_to_url($directory));

		$dir = FileFactory::Directory($base . $this->_cwd . $directory);
		
		if($dir->exists()){
			throw new Exception('Directory ' . $directory . ' already exists!');
		}

		if(!$dir->mkdir()){
			throw new Exception('Unable to create directory ' . $directory);
		}
		
		return $this->_cwd . $directory . '/';
	}

	public function touch(string $newFilename) : string {
		if(!$this->checkCanTouch()) {
			throw new AccessDeniedException();
		}

		// Sanitize the filename
		if(
			strpos($newFilename, '..') !== false ||
			strpos($newFilename, '/') !== false
		){
			throw new Exception('Invalid file name');
		}

		$base = $this->getBaseDir();
		$newFilename = trim(\Core\str_to_url($newFilename, true));

		$file = FileFactory::File($base . $this->_cwd . $newFilename);

		if($file->exists()) {
			touch($file->getFilename());
		}
		else {
			$file->putContents('');
		}

		return $this->_cwd . $newFilename;
	}

	public function loadEditor(View $view, string $path) {
		// In your own Navigator, this MUST be overridden and set with your requested functionality!
		throw new Exception('The edit functionality for this media type has not been defined');

		/*
		elseif($editor === 'image') {
			// And now, $file is a valid File object!
			$meta = new FileMetaHelper($entry);
			$meta->addElementsToForm($form, 'metas');
		}
*/
	}

	/**
	 * Save an uploaded file to its final destination
	 *
	 * @param string $filename
	 * @param string $tmp
	 * @throws Exception
	 */
	public function setUploadedFile(string $filename, string $tmp) {
		if(!$this->checkCanUpload()){
			throw new AccessDeniedException();
		}
		
		$base = $this->getBaseDir();
		
		$file = FileFactory::File($base . $this->_cwd . \Core\str_to_url($filename, true));

		// Ensure the parent directory exists first!
		// This usually isn't required, but may be needed for the first upload in media.
		$parentdir = FileFactory::Directory($file->getDirectoryName());
		$parentdir->mkdir();

		if(!move_uploaded_file($tmp, $file->getFilename())) {
			throw new Exception('Unable to upload file, please try again later.');
		}
	}
	
	public static function Factory(string $type) : Navigator2 {
		switch($type){
			case 'media':
				return new Navigators\MediaNavigator();
			case 'assets':
				return new Navigators\AssetsNavigator();
			case 'public':
				return new Navigators\PublicNavigator();
			case 'page':
				return new PageNavigator();
			default:
				throw new BadRequestException();
		}
	}
	
	/**
	 * Get all available navigators for the given user
	 * 
	 * @return Navigator2[]
	 */
	public static function GetAccessableNavigators() : array {
		
		$ret = [];
		
		$admin = \Core\user()->checkAccess('g:admin');
		$media = (
			\Core\user()->checkAccess('p:/mediamanager/browse') || 
			\Core\user()->checkAccess('p:/mediamanager/upload')
		);
		
		if($media){
			$ret[] = new Navigators\MediaNavigator();
		}
		if($admin){
			$ret[] = new Navigators\PublicNavigator();
			$ret[] = new Navigators\AssetsNavigator();
			$ret[] = new PageNavigator();
		}
		
		return $ret;
	}

	/*
	public static function FileSaveHandler(Form $form) {

		$helper   = new FileMetaHelper($entry);

		// Run through each element and save its metadata to the table.
		foreach($form->getElementsByName('metas\[.*\]') as $el){
			$name = $el->get('name');

			// The name will contain metas[ ... ], trim that part out.
			$name = substr($name, 6, -1);

			$helper->setMeta($name, $el->get('value'));
		}
	}
	*/

	protected function _addDirectoryToListings(string $path, string $parent, int $mtime) {
		$this->_listings[ $path ] = [
			'type'    => 'directory',
			'name'    => basename($path),
			'path'    => $path,
			'parent'  => $parent,
			'mtime'   => $mtime,
		];
	}
	
	protected function _addFileToListings(File $file, string $path, string $parent) {
		$this->_listings[ $path ] = [
			'type'   => $file->getMimetype(),
			'name'   => $file->getBasename(),
			'path'   => $path,
			'parent' => $parent,
			'mtime'  => $file->getMTime(),
			'url'    => $file->getURL(),
		];
	}
	
	/**
	 * Sanitize an incoming filename of a requested file.
	 * 
	 * This simply just attempts to keep the request inside the contained directory by removing "/..".
	 * 
	 * @param string $filename
	 * @return string
	 */
	protected function _sanitizeFilename(string $filename) : string {
		$f = str_replace('/..', '', $filename);
		
		return $f;
	}
	
	protected function _getDirectoryInfo(Directory $dir, string $path) : array {
		$mtime = filemtime($dir->getPath());
		$count = $this->getEntryChildrenCount($path);
		
		return [
			'path'  => $path,
			'mtimr' => $mtime,
			'mtimf' => DateTime::FormatString($mtime, DateTime::SHORTDATE),
			'sizer' => $count,
			'sizef' => t('STRING_MEDIAMANAGER_DIR_N_ITEMS', $count),
			'icon'  => $dir->getMimetypeIconURL(),
			'prvw'  => $dir->getMimetypeIconURL(),
			'large' => $dir->getMimetypeIconURL(),
			'url'   => null,
		];
	}

	protected function _getFileInfo(File $file, string $path) : array {
		return [
			'path'  => $path,
			'mtimr' => $file->getMTime(),
			'mtimf' => DateTime::FormatString($file->getMTime(), DateTime::SHORTDATE),
			'sizer' => $file->getFilesize(),
			'sizef' => $file->getFilesize(true),
			'icon'  => $file->getMimetypeIconURL(),
			'prvw'  => $file->getPreviewURL('96x96'),
			'large' => $file->getPreviewURL('200x200'),
			'url'   => $file->getURL(),
		];
	}
}
