<?php


namespace PHPMailerCore;


use Core\GPG\GPG;
use Exception;
use PHPMailer\PHPMailer\PHPMailer;

class CorePHPMailer extends PHPMailer {
	public $encryption = null;
	public $gpgKeyID = null;

	public function setGPGEncryptionKey(string $keyid) {
		$this->encryption = 'pgp-encrypted';
		$this->gpgKeyID = $keyid;
	}

	public function send() {
		if($this->encryption === 'pgp-encrypted') {
			if(!$this->gpgKeyID) {
				throw new Exception('Unable to use GPG encryption if no key is set!');
			}

			return $this->_sendGPGEncrypted();
		}
		elseif($this->encryption === null) {
			return parent::send();
		}
		else {
			throw new Exception('Unsupported encryption type: ' . $this->encryption);
		}
	}

	private function _sendGPGEncrypted() {

		// Retrieve the message before encryption, because this is what we'll be encrypting!
		$this->preSend();
		$messagePlain     = $this->MIMEHeader . $this->MIMEBody;
		$gpg              = new GPG();
		$messageEncrypted = $gpg->encryptData($messagePlain, $this->gpgKeyID);

		// Remove all the content that just got encrypted, we'll recreate it.
		$this->clearAttachments();
		$this->Body = null;
		$this->AltBody = null;
		$this->AllowEmpty = true;

		$this->addStringEmbeddedImage(
			'Version: 1',
			'openpgp-version',
			'',
			'7bit',
			'application/pgp-encrypted',
			'inline'
		);
		$this->addStringEmbeddedImage(
			$messageEncrypted,
			'openpgp-message',
			'encrypted.asc',
			'base64',
			'application/octet-stream',
			'inline'
		);

		// Recreate the mime data for the encrypted message now.
		$this->message_type = 'pgp-encrypted';

		$this->MIMEBody = $this->createBody();
		$this->MIMEHeader = $this->createHeader();

		return $this->postSend();
	}

	/**
	 * Get the message MIME type headers.
	 *
	 * @return string
	 */
	public function getMailMIME() {
		$result = '';
		$ismultipart = true;
		switch ($this->message_type) {
			case 'inline':
				$result .= $this->headerLine('Content-Type', static::CONTENT_TYPE_MULTIPART_RELATED . ';');
				$result .= $this->textLine("\tboundary=\"" . $this->boundary[1] . '"');
				break;
			case 'attach':
			case 'inline_attach':
			case 'alt_attach':
			case 'alt_inline_attach':
				$result .= $this->headerLine('Content-Type', static::CONTENT_TYPE_MULTIPART_MIXED . ';');
				$result .= $this->textLine("\tboundary=\"" . $this->boundary[1] . '"');
				break;
			case 'alt':
			case 'alt_inline':
				$result .= $this->headerLine('Content-Type', static::CONTENT_TYPE_MULTIPART_ALTERNATIVE . ';');
				$result .= $this->textLine("\tboundary=\"" . $this->boundary[1] . '"');
				break;
			case 'pgp-encrypted':
				$result .= $this->headerLine('Content-Type', 'multipart/encrypted; protocol="application/pgp-encrypted";');
				$result .= $this->textLine("\tboundary=\"" . $this->boundary[1] . '"');
				break;
			default:
				// Catches case 'plain': and case '':
				$result .= $this->textLine('Content-Type: ' . $this->ContentType . '; charset=' . $this->CharSet);
				$ismultipart = false;
				break;
		}
		// RFC1341 part 5 says 7bit is assumed if not specified
		if (static::ENCODING_7BIT != $this->Encoding) {
			// RFC 2045 section 6.4 says multipart MIME parts may only use 7bit, 8bit or binary CTE
			if ($ismultipart) {
				if (static::ENCODING_8BIT == $this->Encoding) {
					$result .= $this->headerLine('Content-Transfer-Encoding', static::ENCODING_8BIT);
				}
				// The only remaining alternatives are quoted-printable and base64, which are both 7bit compatible
			} else {
				$result .= $this->headerLine('Content-Transfer-Encoding', $this->Encoding);
			}
		}

		if ('mail' != $this->Mailer) {
			$result .= static::$LE;
		}

		return $result;
	}

	/**
	 * Assemble the message body.
	 * Returns an empty string on failure.
	 *
	 * @throws \PHPMailer\PHPMailer\Exception
	 *
	 * @return string The assembled message body
	 */
	public function createBody() {
		$body = '';
		//Create unique IDs and preset boundaries
		$this->uniqueid = $this->generateId();
		$this->boundary[1] = 'b1_' . $this->uniqueid;
		$this->boundary[2] = 'b2_' . $this->uniqueid;
		$this->boundary[3] = 'b3_' . $this->uniqueid;

		if ($this->sign_key_file) {
			$body .= $this->getMailMIME() . static::$LE;
		}

		$this->setWordWrap();

		$bodyEncoding = $this->Encoding;
		$bodyCharSet = $this->CharSet;
		//Can we do a 7-bit downgrade?
		if (static::ENCODING_8BIT == $bodyEncoding and !$this->has8bitChars($this->Body)) {
			$bodyEncoding = static::ENCODING_7BIT;
			//All ISO 8859, Windows codepage and UTF-8 charsets are ascii compatible up to 7-bit
			$bodyCharSet = 'us-ascii';
		}
		//If lines are too long, and we're not already using an encoding that will shorten them,
		//change to quoted-printable transfer encoding for the body part only
		if (static::ENCODING_BASE64 != $this->Encoding and static::hasLineLongerThanMax($this->Body)) {
			$bodyEncoding = static::ENCODING_QUOTED_PRINTABLE;
		}

		$altBodyEncoding = $this->Encoding;
		$altBodyCharSet = $this->CharSet;
		//Can we do a 7-bit downgrade?
		if (static::ENCODING_8BIT == $altBodyEncoding and !$this->has8bitChars($this->AltBody)) {
			$altBodyEncoding = static::ENCODING_7BIT;
			//All ISO 8859, Windows codepage and UTF-8 charsets are ascii compatible up to 7-bit
			$altBodyCharSet = 'us-ascii';
		}
		//If lines are too long, and we're not already using an encoding that will shorten them,
		//change to quoted-printable transfer encoding for the alt body part only
		if (static::ENCODING_BASE64 != $altBodyEncoding and static::hasLineLongerThanMax($this->AltBody)) {
			$altBodyEncoding = static::ENCODING_QUOTED_PRINTABLE;
		}
		//Use this as a preamble in all multipart message types
		$mimepre = 'This is a multi-part message in MIME format.' . static::$LE;
		switch ($this->message_type) {
			case 'pgp-encrypted':
				$body .= 'This is an OpenPGP/MIME encrypted message (RFC 4880 and 3156), please use a compatible client to view the message.' . static::$LE;
				$body .= $this->attachAll('inline', $this->boundary[1]);
				break;
			case 'inline':
				$body .= $mimepre;
				$body .= $this->getBoundary($this->boundary[1], $bodyCharSet, '', $bodyEncoding);
				$body .= $this->encodeString($this->Body, $bodyEncoding);
				$body .= static::$LE;
				$body .= $this->attachAll('inline', $this->boundary[1]);
				break;
			case 'attach':
				$body .= $mimepre;
				$body .= $this->getBoundary($this->boundary[1], $bodyCharSet, '', $bodyEncoding);
				$body .= $this->encodeString($this->Body, $bodyEncoding);
				$body .= static::$LE;
				$body .= $this->attachAll('attachment', $this->boundary[1]);
				break;
			case 'inline_attach':
				$body .= $mimepre;
				$body .= $this->textLine('--' . $this->boundary[1]);
				$body .= $this->headerLine('Content-Type', static::CONTENT_TYPE_MULTIPART_RELATED . ';');
				$body .= $this->textLine("\tboundary=\"" . $this->boundary[2] . '"');
				$body .= static::$LE;
				$body .= $this->getBoundary($this->boundary[2], $bodyCharSet, '', $bodyEncoding);
				$body .= $this->encodeString($this->Body, $bodyEncoding);
				$body .= static::$LE;
				$body .= $this->attachAll('inline', $this->boundary[2]);
				$body .= static::$LE;
				$body .= $this->attachAll('attachment', $this->boundary[1]);
				break;
			case 'alt':
				$body .= $mimepre;
				$body .= $this->getBoundary($this->boundary[1], $altBodyCharSet, static::CONTENT_TYPE_PLAINTEXT, $altBodyEncoding);
				$body .= $this->encodeString($this->AltBody, $altBodyEncoding);
				$body .= static::$LE;
				$body .= $this->getBoundary($this->boundary[1], $bodyCharSet, static::CONTENT_TYPE_TEXT_HTML, $bodyEncoding);
				$body .= $this->encodeString($this->Body, $bodyEncoding);
				$body .= static::$LE;
				if (!empty($this->Ical)) {
					$body .= $this->getBoundary($this->boundary[1], '', static::CONTENT_TYPE_TEXT_CALENDAR . '; method=REQUEST', '');
					$body .= $this->encodeString($this->Ical, $this->Encoding);
					$body .= static::$LE;
				}
				$body .= $this->endBoundary($this->boundary[1]);
				break;
			case 'alt_inline':
				$body .= $mimepre;
				$body .= $this->getBoundary($this->boundary[1], $altBodyCharSet, static::CONTENT_TYPE_PLAINTEXT, $altBodyEncoding);
				$body .= $this->encodeString($this->AltBody, $altBodyEncoding);
				$body .= static::$LE;
				$body .= $this->textLine('--' . $this->boundary[1]);
				$body .= $this->headerLine('Content-Type', static::CONTENT_TYPE_MULTIPART_RELATED . ';');
				$body .= $this->textLine("\tboundary=\"" . $this->boundary[2] . '"');
				$body .= static::$LE;
				$body .= $this->getBoundary($this->boundary[2], $bodyCharSet, static::CONTENT_TYPE_TEXT_HTML, $bodyEncoding);
				$body .= $this->encodeString($this->Body, $bodyEncoding);
				$body .= static::$LE;
				$body .= $this->attachAll('inline', $this->boundary[2]);
				$body .= static::$LE;
				$body .= $this->endBoundary($this->boundary[1]);
				break;
			case 'alt_attach':
				$body .= $mimepre;
				$body .= $this->textLine('--' . $this->boundary[1]);
				$body .= $this->headerLine('Content-Type', static::CONTENT_TYPE_MULTIPART_ALTERNATIVE . ';');
				$body .= $this->textLine("\tboundary=\"" . $this->boundary[2] . '"');
				$body .= static::$LE;
				$body .= $this->getBoundary($this->boundary[2], $altBodyCharSet, static::CONTENT_TYPE_PLAINTEXT, $altBodyEncoding);
				$body .= $this->encodeString($this->AltBody, $altBodyEncoding);
				$body .= static::$LE;
				$body .= $this->getBoundary($this->boundary[2], $bodyCharSet, static::CONTENT_TYPE_TEXT_HTML, $bodyEncoding);
				$body .= $this->encodeString($this->Body, $bodyEncoding);
				$body .= static::$LE;
				if (!empty($this->Ical)) {
					$body .= $this->getBoundary($this->boundary[2], '', static::CONTENT_TYPE_TEXT_CALENDAR . '; method=REQUEST', '');
					$body .= $this->encodeString($this->Ical, $this->Encoding);
				}
				$body .= $this->endBoundary($this->boundary[2]);
				$body .= static::$LE;
				$body .= $this->attachAll('attachment', $this->boundary[1]);
				break;
			case 'alt_inline_attach':
				$body .= $mimepre;
				$body .= $this->textLine('--' . $this->boundary[1]);
				$body .= $this->headerLine('Content-Type', static::CONTENT_TYPE_MULTIPART_ALTERNATIVE . ';');
				$body .= $this->textLine("\tboundary=\"" . $this->boundary[2] . '"');
				$body .= static::$LE;
				$body .= $this->getBoundary($this->boundary[2], $altBodyCharSet, static::CONTENT_TYPE_PLAINTEXT, $altBodyEncoding);
				$body .= $this->encodeString($this->AltBody, $altBodyEncoding);
				$body .= static::$LE;
				$body .= $this->textLine('--' . $this->boundary[2]);
				$body .= $this->headerLine('Content-Type', static::CONTENT_TYPE_MULTIPART_RELATED . ';');
				$body .= $this->textLine("\tboundary=\"" . $this->boundary[3] . '"');
				$body .= static::$LE;
				$body .= $this->getBoundary($this->boundary[3], $bodyCharSet, static::CONTENT_TYPE_TEXT_HTML, $bodyEncoding);
				$body .= $this->encodeString($this->Body, $bodyEncoding);
				$body .= static::$LE;
				$body .= $this->attachAll('inline', $this->boundary[3]);
				$body .= static::$LE;
				$body .= $this->endBoundary($this->boundary[2]);
				$body .= static::$LE;
				$body .= $this->attachAll('attachment', $this->boundary[1]);
				break;
			default:
				// Catch case 'plain' and case '', applies to simple `text/plain` and `text/html` body content types
				//Reset the `Encoding` property in case we changed it for line length reasons
				$this->Encoding = $bodyEncoding;
				$body .= $this->encodeString($this->Body, $this->Encoding);
				break;
		}

		if ($this->isError()) {
			$body = '';
			if ($this->exceptions) {
				throw new Exception($this->lang('empty_message'), self::STOP_CRITICAL);
			}
		} elseif ($this->sign_key_file) {
			try {
				if (!defined('PKCS7_TEXT')) {
					throw new Exception($this->lang('extension_missing') . 'openssl');
				}
				// @TODO would be nice to use php://temp streams here
				$file = tempnam(sys_get_temp_dir(), 'mail');
				if (false === file_put_contents($file, $body)) {
					throw new Exception($this->lang('signing') . ' Could not write temp file');
				}
				$signed = tempnam(sys_get_temp_dir(), 'signed');
				//Workaround for PHP bug https://bugs.php.net/bug.php?id=69197
				if (empty($this->sign_extracerts_file)) {
					$sign = @openssl_pkcs7_sign(
						$file,
						$signed,
						'file://' . realpath($this->sign_cert_file),
						['file://' . realpath($this->sign_key_file), $this->sign_key_pass],
						[]
					);
				} else {
					$sign = @openssl_pkcs7_sign(
						$file,
						$signed,
						'file://' . realpath($this->sign_cert_file),
						['file://' . realpath($this->sign_key_file), $this->sign_key_pass],
						[],
						PKCS7_DETACHED,
						$this->sign_extracerts_file
					);
				}
				@unlink($file);
				if ($sign) {
					$body = file_get_contents($signed);
					@unlink($signed);
					//The message returned by openssl contains both headers and body, so need to split them up
					$parts = explode("\n\n", $body, 2);
					$this->MIMEHeader .= $parts[0] . static::$LE . static::$LE;
					$body = $parts[1];
				} else {
					@unlink($signed);
					throw new Exception($this->lang('signing') . openssl_error_string());
				}
			} catch (Exception $exc) {
				$body = '';
				if ($this->exceptions) {
					throw $exc;
				}
			}
		}

		return $body;
	}
}