<?php
/**
 * Email wrapper around the system mail utility, (phpMailer in this case).
 *
 * @package PHPMailer
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace PHPMailerCore;

use Core\Filestore\File;
use Core\GPG\GPG;
use Exception;
use PHPMailer;
use Core\EmailAPI\EmailInterface;
use Core\EmailAPI\Email;
use Core\EmailAPI\EmailAddress;
use Core\EmailAPI\EmailAttachment;
use ReflectionClass;


abstract class EmailBackend extends Email implements EmailInterface {
	/**
	 * @var CorePHPMailer The mailer object to send this email with.
	 */
	protected $_mailer;

	/**
	 * Construct a new email backend to send an email.
	 */
	public function __construct() {
		//$this->_mailer = new PHPMailer(true);
		$this->_mailer = new CorePHPMailer(true);
		
		parent::__construct();
	}
	
	/**
	 * Get all "BCC" addresses for this email.
	 * 
	 * @return EmailAddress[]
	 */
	public function getBCC(): array {
		return $this->_convertAddresses($this->_mailer->getBccAddresses());
	}

	/**
	 * Get all "CC" addresses for this email.
	 * 
	 * @return EmailAddress[]
	 */
	public function getCC(): array {
		return $this->_convertAddresses($this->_mailer->getCcAddresses());
	}

	/**
	 * Get all "to" addresses for this email.
	 * 
	 * @return EmailAddress[]
	 */
	public function getTo(): array {
		return $this->_convertAddresses($this->_mailer->getToAddresses());
	}
	
	/**
	 * Get the "from" address for this email.
	 * 
	 * @return EmailAddress
	 */
	public function getFrom(): EmailAddress {
		return $this->_convertAddress($this->_mailer->From, $this->_mailer->FromName);
	}
	
	/**
	 * Get the "Reply-To" address for this email.
	 * 
	 * @return EmailAddress
	 */
	public function getReplyTo(): EmailAddress {
		return $this->_convertAddress(
			array_values($this->_mailer->getReplyToAddresses())
		);
	}
	
	/**
	 * Get the subject for this email, as set by setSubject usually.
	 * 
	 * @return string
	 */
	public function getSubject(): string {
		return $this->_mailer->Subject;
	}
	
	/**
	 * Get all custom headers for this email
	 * 
	 * @return array
	 */
	public function getHeaders(): array {
		$hdr = $this->_mailer->getCustomHeaders();
		$ret = [];
		foreach($hdr as $dat){
			// The header key is the first parameter.
			$k = $dat[0];
			// phpMailer for some reason prepends the value with a space...
			$v = trim($dat[1]);
			
			if(isset($ret[$k]) && !is_array($ret[$k])){
				// Key is already set, but is not an array.
				$ret[$k] = [
					$ret[$k],
					$v
				];
			}
			elseif(isset($ret[$k]) && is_array($ret[$k])){
				// Key already set and IS an array.
				$ret[$k][] = $v;
			}
			else{
				// Just plain set.
				$ret[$k] = $v;
			}
		}
		
		return $ret;
	}
	
	/**
	 * Get the raw and complete EML for this email, with headers, body, and all attachments.
	 * 
	 * @return string
	 */
	public function getFullEML(): string {
		// The EML contains everything, so I need to render the email as it will be sent.
		// To do that, call phpMailer's preSend method to prep the data.
		$this->preSend();
		$this->_mailer->PreSend();
		$header = $this->_mailer->CreateHeader();
		$body   = $this->_mailer->CreateBody();
		
		// Now, the EML is simply those two elements combined.
		return $header . $body;
	}
	
	/**
	 * Get any attachments on this email as an array
	 * 
	 * @return EmailAttachment[]
	 */
	public function getAttachments() : array {
		$ret = [];
		
		foreach($this->_mailer->getAttachments() as $row) {
			$a = new EmailAttachment();
			$a->filename    = $row[1];
			$a->encoding    = $row[3];
			$a->mimetype    = $row[4];
			$a->disposition = $row[6];
			
			if($row[5]) {
				// Streamed attachment
				$a->contents = $row[0];
			}
			else {
				// Standard file include from filesystem
				$a->filepath = $row[0];
			}
			
			$ret[] = $a;
		}
		
		return $ret;
	}
	
	/**
	 * Get the main body, probably as HTML
	 * 
	 * @return string
	 */
	public function getMainBody() : string {
		return $this->_mailer->Body;
	}
	
	/**
	 * Set the "Sender" address for this email
	 * 
	 * Return the parent object to allow chaining.
	 * 
	 * @param string $address
	 * @param string $name
	 * 
	 * @return EmailInterface
	 */
	public function setSender(string $address, string $name = ''): EmailInterface {
		/*if($address && $name) {
			$this->_mailer->Sender = $name . ' <' . $address . '>';
		}
		elseif($address) {
			$this->_mailer->Sender = $address;
		}*/
		
		if($address) {
			$this->_mailer->Sender = $address;
		}
		
		return $this;
	}
	
	/**
	 * Set the "to" address for this email.
	 *
	 * Will clear out any other to address.
	 * 
	 * Return the parent object to allow chaining.
	 *
	 * @param string $address
	 * @param string $name
	 * 
	 * @return EmailInterface
	 */
	public function setTo(string $address, string $name = ''): EmailInterface{
		// Reset any "to" address already on the mailer.
		$this->_mailer->ClearAddresses();
		$this->_mailer->AddAddress($address, $name);
		
		return $this;
	}

	/**
	 * Set the "from" address for this email
	 * 
	 * Return the parent object to allow chaining.
	 * 
	 * @param string $address
	 * @param string $name
	 * 
	 * @return EmailInterface
	 */
	public function setFrom(string $address, string $name = ''): EmailInterface{
		// Will only affect the from, not the sender!
		$this->_mailer->From = $address;
		if($name){
			$this->_mailer->FromName = $name;
		}
		
		return $this;
	}
	
	/**
	 * Set the Reply To address for this email.
	 *
	 * @param $address
	 * @param string $name
	 * 
	 * @return EmailInterface
	 */
	public function setReplyTo(string $address, string $name = ''): EmailInterface {
		$this->_mailer->clearReplyTos();
		$this->_mailer->AddReplyTo($address, $name);
		
		return $this;
	}
	
	/**
	 * Set the main body on this email, optionally setting the mimetype.
	 * 
	 * This usually sets the rich/HTML version, but can be different.
	 * 
	 * @param string $body
	 * @param string $mimetype
	 * 
	 * @return EmailInterface
	 */
	public function setMainBody(string $body, string $mimetype = 'text/html') : EmailInterface {
		$this->_mailer->Body = $body;
		$this->_mailer->ContentType = $mimetype;
		
		return $this;
	}
	
	/**
	 * Set the alternate plantext-only body on this email.
	 * 
	 * @param string $body
	 * 
	 * @return EmailInterface
	 */
	public function setAltBody(string $body) : EmailInterface {
		$this->_mailer->AltBody = $body;
		
		return $this;
	}
	
	/**
	 * Set the subject of this email.
	 *
	 * @param string $subject
	 * 
	 * @return EmailInterface
	 */
	public function setSubject(string $subject): EmailInterface {
		$this->_mailer->Subject = $subject;
		
		return $this;
	}
	
	/**
	 * Set a custom header on this Email object.
	 * 
	 * @param string $key
	 * @param string $value
	 * 
	 * @return EmailInterface
	 */
	public function setHeader(string $key, string $value): EmailInterface {
		switch($key){
			case 'X-Mailer':
				$this->_mailer->XMailer = $value;
				break;
			default:
				$this->_mailer->addCustomHeader($key . ': ' . $value);
				break;
		}
		
		return $this;
	}
	
	/**
	 * Add a blind carbon copy email address.
	 * 
	 * Return the parent object to allow chaining.
	 * 
	 * @param string $address
	 * @param string $name
	 * 
	 * @return EmailInterface
	 */
	public function addBCC(string $address, string $name = ''): EmailInterface {
		$this->_mailer->AddBCC($address, $name);
		
		return $this;
	}
	
	/**
	 * Add a carbon copy email address.
	 * 
	 * Return the parent object to allow chaining.
	 * 
	 * @param string $address
	 * @param string $name
	 * 
	 * @return EmailInterface
	 */
	public function addCC(string $address, string $name = ''): EmailInterface {
		$this->_mailer->AddCC($address, $name);
		
		return $this;
	}

	/**
	 * Add a "to" address for this email.
	 *
	 * Can be called multiple times for sending to multiple people, will
	 * add addresses to the "to" recipient on each call.
	 *
	 * @param string $address
	 * @param string $name
	 * 
	 * @return EmailInterface
	 */
	public function addTo(string $address, string $name = ''): EmailInterface {
		$this->_mailer->AddAddress($address, $name);
		
		return $this;
	}
	
	/**
	 * Add a Reply To address for this email.
	 *
	 * @param $address
	 * @param string $name
	 * 
	 * @return EmailInterface
	 */
	public function addReplyTo(string $address, string $name = ''): EmailInterface {
		$this->_mailer->AddReplyTo($address, $name);
		
		return $this;
	}
	
	/**
	 * Add a file as an attachment
	 *
	 * @param File $file
	 * @param bool                 $inline Set to true to request this attachment to be viewed inline.
	 *
	 * @throws \Exception
	 * 
	 * @return EmailInterface
	 */
	public function addAttachmentFromFile(File $file, $inline = false): EmailInterface {
		$this->_mailer->addAttachment(
			$file->getFilename(),               // Full Path
			$file->getBasename(),               // Base Filename (to be exposed in client)
			'base64',                           // Yup, just do this
			$file->getMimetype(),               // Mimetype, try to use correct hinting for client
			($inline ? 'inline' : 'attachment') // Set the disposition based on inline or not.
		);
		
		return $this;
	}
	
	/**
	 * Add a file stream as an attachment
	 *
	 * @param string $data     The data to attach to this email
	 * @param string $filename Client Rendering Hint for file name
	 * @param string $mimetype Client Rendering Hint for file mimetype
	 * @param bool   $inline   Set to true to request this attachment to be viewed inline.
	 *
	 * @throws \Exception
	 * 
	 * @return EmailInterface
	 */
	public function addAttachmentFromStream(string $data, string $filename, string $mimetype, $inline = false): EmailInterface {
		$this->_mailer->addStringAttachment(
			$data,                              // Full data contents of the attachment
			$filename,                          // Base filename to render to the client
			'base64',                           // Base64 Encode the data
			$mimetype,                          // Mimetype, to assist with the client for hinting.
			($inline ? 'inline' : 'attachment') // Render this attachment inline if possible, more client hinting.
		);
		
		return $this;
	}
	
	/**
	 * Send this email!
	 *
	 * @throws \Exception
	 */
	public function send() {
		if(!$this->preSend()) {
			// If presend indicated the message not to be sent, then do not send.
			return;
		}

		if($this->encrypted && \Core\Core::IsComponentAvailable('gpg-key-manager')) {
			// This message was requested to be encrypted, and the GPG Key Manager is available.

			$recipients  = $this->getTo();

			// This is required for multiple recipients!
			// This requires a little more overhead, as I need to lookup each recipient's user account
			// to retrieve their GPG key.
			foreach($recipients as $recipient){
				$clone = clone $this->_mailer;
				$clone->clearAddresses();
				$keyID = $recipient->getGPGKeyID();

				$clone->addAddress($recipient->address, $recipient->name);
				if($keyID) {
					$clone->setGPGEncryptionKey($keyID);
				}

				$clone->send();
			}
		}
		else {
			// Not requested to be encrypted, just send a normal message.
			$this->_mailer->send();
		}
		
		$this->postSend();
	}
	
	
	
	
/*
	public function setOnBehalfOf($address, $name = ''){
		if($this->_mailer->From){
			// Move these over to the Sender header.
			$sender = $this->_mailer->FromName;
			if($sender){
				$sender .= ' <' . $this->_mailer->From . '>';
			}
			else{
				$sender = $this->_mailer->From;
			}
			$this->_mailer->AddCustomHeader('Sender', $sender);
		}
		
		$this->_mailer->From = $address;
		if($name){
			$this->_mailer->FromName = $name;
		}
	}*/
	
	/**
	 * Clear all recipients, (to, cc, and bcc), for this email.
	 *
	 * @throws \Exception
	 * 
	 * @return EmailInterface
	 */
	public function clearRecipients(): EmailInterface {
		$this->_mailer->clearAllRecipients();
		
		return $this;
	}
	
	/**
	 * Clear all attachments for this email
	 * 
	 * @return EmailInterface
	 */
	public function clearAttachments() : EmailInterface {
		$this->_mailer->clearAttachments();
		
		return $this;
	}
	
	/**
	 * Convert the internal indexed array to EmailAddress classes
	 * 
	 * @param array $arr
	 * 
	 * @return EmailAddress[]
	 */
	protected function _convertAddresses($arr) : array {
		$ret = [];
		foreach($arr as $row) {
			$a = new EmailAddress();
			$a->address = $row[0];
			$a->name = $row[1] ?: '';
			$ret[] = $a;
		}
		
		return $ret;
	}
	
	/**
	 * Convert a single address/name to the standardized version
	 * 
	 * @param string $email
	 * @param string $name
	 * 
	 * @return EmailAddress
	 */
	protected function _convertAddress(string $email, string $name) : EmailAddress {
		$a = new EmailAddress();
		$a->address = $email;
		$a->name = $name;
		return $a;
	}
}
