<?php


use Core\ConfigHandler;
use Core\Controller;
use Core\Core;
use Core\RedirectException;
use Core\Theme;
use Core\View;
use Core\ViewControl;
use Core\ViewStatusExceptionInterface;
use Favicon\FaviconHelper;
use Minishlink\WebPush\VAPID;

class PwaController extends Controller {
	/**
	 * Generate the manifest file for this application
	 *
	 * @throws Exception
	 */
	public function manifest() {
		$view = $this->getView();

		$view->contenttype = View::CTYPE_JSON;
		$view->record = false;

		$sizes = FaviconHelper::GetFaviconSizes();

		$background = defined('THEME_STYLE_SITE_BACKGROUND_COLOR') ? THEME_STYLE_SITE_BACKGROUND_COLOR : '#3E4EB8';
		$theme = defined('THEME_STYLE_TOOLBAR_BACKGROUND_COLOR') ? THEME_STYLE_TOOLBAR_BACKGROUND_COLOR : '#2F3BA2';

		$view->jsondata = [
			'name' => SITENAME,
			'short_name' => SITENAME,
			'start_url' => ROOT_URL,
			'display' => 'standalone',
			'background_color' => $background,
			'theme_color' => $theme,
			'icons' => [],
		];

		foreach(['512', '256', '192', '152', '144', '128'] as $s) {
			if(isset($sizes[ $s ])) {
				$file = $sizes[ $s ];
				$sxs = $s . 'x' . $s;
				$view->jsondata['icons'][] = [
					'src' => $file->getPreviewURL($sxs . '!'),
					'sizes' => $sxs,
					'type' => $file->getMimetype(),
				];
			}
		}
	}

	/**
	 * Generate the service worker that controls applications.
	 *
	 * Pulls addon scripts from [component]/resources/service-workers/{common,install,activate,fetch}.js
	 */
	public function serviceworker() {
		$view = $this->getView();

		$view->contenttype = 'application/javascript';
		$view->mastertemplate = false;
		$view->templatename = 'pages/pwa/serviceworker';

		// Allow components to hook into the configuration step to update behaviour of their code.
		$config   = \Core\HookHandler::DispatchHook('/pwa/serviceworker/config');
		$common   = [];
		$install  = [];
		$activate = [];
		$fetch    = [];
		$push     = [];

		$sources = Core::GetComponents();
		// And add on the current theme.
		$sources[] = Theme::GetTheme();

		foreach($sources as $s) {
			/** @var \Core\Component $s */
			$dir = $s->getBaseDir() . 'resources/service-workers/';
			if(is_dir($dir)) {

				if(DEVELOPMENT_MODE) {
					$type = $s instanceof Theme ? 'Theme' : 'Component';
					$name = $s->getName();
					$header = <<<EOD

/////////////////////////////
// Loaded from $type $name

EOD;
					$footer = <<<EOD
// End $type $name
/////////////////////////////


EOD;

				}
				else {
					$header = null;
					$footer = null;
				}


				if(file_exists($dir . 'common.js')) {
					if($header) {
						$common[] = $header;
					}
					$common[] = file_get_contents($dir . 'common.js');
					if($footer) {
						$common[] = $footer;
					}
				}

				if(file_exists($dir . 'install.js')) {
					if($header) {
						$install[] = $header;
					}
					$install[] = file_get_contents($dir . 'install.js');
					if($footer) {
						$install[] = $footer;
					}
				}

				if(file_exists($dir . 'activate.js')) {
					if($header) {
						$activate[] = $header;
					}
					$activate[] = file_get_contents($dir . 'activate.js');
					if($footer) {
						$activate[] = $footer;
					}
				}

				if(file_exists($dir . 'fetch.js')) {
					if($header) {
						$fetch[] = $header;
					}
					$fetch[] = file_get_contents($dir . 'fetch.js');
					if($footer) {
						$fetch[] = $footer;
					}
				}

				if(file_exists($dir . 'push.js')) {
					if($header) {
						$push[] = $header;
					}
					$push[] = file_get_contents($dir . 'push.js');
					if($footer) {
						$push[] = $footer;
					}
				}
			}
		}

		$view->assign('config', $config);
		$view->assign('common', $common);
		$view->assign('install', $install);
		$view->assign('activate', $activate);
		$view->assign('fetch', $fetch);
		$view->assign('push', $push);
		$view->assign('devmode', DEVELOPMENT_MODE);
	}

	/**
	 * Provide an interface for users to manage their push notification settings, ie: opt-in / opt-out.
	 *
	 * @core-access-string g:authenticated
	 *
	 * @throws ViewStatusExceptionInterface
	 */
	public function notifications_manage() {
		$view = $this->getView();

		$pub = ConfigHandler::Get('/pwa/serverkey/public');
		$pri = ConfigHandler::Get('/pwa/serverkey/private');

		if(!$pub || !$pri) {
			try {
				// No public or private keys are set yet!
				$keys = VAPID::createVapidKeys();

				$pub = $keys['publicKey'];
				$pri = $keys['privateKey'];

				ConfigHandler::Set('/pwa/serverkey/public', $pub);
				ConfigHandler::Set('/pwa/serverkey/private', $pri);
			}
			catch(Exception $ex) {
				\Core\log_exception($ex);
				\Core\set_message('Notification support is offline at the moment, sorry', 'error');
				throw new RedirectException('BACK');
			}
		}

		$view->addBreadcrumb('t:STRING_MY_PROFILE', '/user/me');
		$view->title = 't:STRING_PWA_MANAGE_PUSH_NOTIFICATIONS';
		$view->assign('public_key', $pub);
	}

	/**
	 * Manage user subscriptions, for creating a new one, deleting, and updated.
	 *
	 * @throws ViewStatusExceptionInterface
	 */
	public function notifications_subscribe() {
		$view = $this->getView();
		$request = $this->getRequest();

		$view->contenttype = View::CTYPE_JSON;

		$subscription = json_decode(file_get_contents('php://input'), true);

		if(!$subscription) {
			throw new \Core\BadRequestException('Invalid subscription requested');
		}

		if($request->isPost()) {
			$model = new PwaUserSubscriptionModel();
			$model->setFromArray([
				'user'       => \Core\user()->id,
				'useragent'  => $request->useragent,
				'endpoint'   => $subscription['endpoint'],
				'encoding'   => $subscription['contentEncoding'],
				'auth'       => $subscription['authToken'],
				'public_key' => $subscription['publicKey'],
			]);
			$model->save();

			$view->setJSONSuccessMessage('Created subscription successfully');
		}
		elseif($request->isPut()) {
			$model = PwaUserSubscriptionModel::Find(['user = ' . \Core\user()->id, 'auth' => $subscription['authToken']], 1);
			if(!$model) {
				// Auto create a new registration.
				$model = new PwaUserSubscriptionModel();
			}
			$model->setFromArray([
				'user'       => \Core\user()->id,
				'useragent'  => $request->useragent,
				'endpoint'   => $subscription['endpoint'],
				'encoding'   => $subscription['contentEncoding'],
				'auth'       => $subscription['authToken'],
				'public_key' => $subscription['publicKey'],
			]);
			$model->save();

			$view->setJSONSuccessMessage('Updated subscription successfully');
		}
		elseif($request->isDelete()) {
			$model = PwaUserSubscriptionModel::Find(['user = ' . \Core\user()->id, 'auth' => $subscription['authToken']], 1);
			if($model) {
				$model->delete();
				$view->setJSONSuccessMessage('Removed subscription successfully');
			}
			else {
				throw new \Core\NotFoundException();
			}
		}
		else {
			throw new \Core\BadRequestException();
		}
	}

	/**
	 * Allow the user to test out the notification system.
	 * This will only send a message to the one device they're using at the moment.
	 *
	 * @core-access-string g:authenticated
	 * @core-request-method PUT
	 *
	 * @throws ViewStatusExceptionInterface
	 */
	public function notifications_test(string $auth) {
		$view = $this->getView();
		$view->contenttype = View::CTYPE_JSON;

		$model = PwaUserSubscriptionModel::Find(['user = ' . \Core\user()->id, 'auth = ' . $auth], 1);
		if(!$model) {
			throw new \Core\NotFoundException('Authentication token not registered, please re-subscribe your notification subscription.');
		}
		
		$notification = new \PWA\PushNotification();

		$logo = ConfigHandler::Get('/core/site_logo');
		if(!$logo) {
			$logo = 'assets/images/logo.png';
		}

		$notification->title('Test Message From ' . SITENAME)
			->body('This is a test notification!')
			->image($logo)
			->icon($logo)
			->action('This is a Button', 'thing');

		try {
			$notification->sendToSubscription($model);
			$view->setJSONSuccessMessage('Successfully queued notification!');
		}
		catch(Exception $ex) {
			\Core\log_exception($ex);
			$view->setJSONErrorMessage($ex->getMessage(), $ex->getCode());
		}
	}

	/**
	 * Hook to add in the necessary favicon ViewMeta attribute into the View.
	 *
	 * The ViewMeta_favicon handles the rest of the magic.
	 *
	 * @param View $view
	 *
	 * @return bool
	 */
	public static function PageHook(View $view) {

		if($view->mode !== View::MODE_PAGE) {
			return true;
		}

		$view->addMetaName('manifest', null);

		$url = \Core\resolve_link('/pwa/serviceworker');
		$path = ROOT_WDIR;
		$script = <<<EOD
<script>
if ('serviceWorker' in navigator) {
  window.addEventListener('load', function() {
    navigator.serviceWorker.register('$url', {scope: '$path'}).then(function(registration) {
      // Registration was successful
      console.log('ServiceWorker registration successful with scope: ', registration.scope);
    }, function(err) {
      // registration failed :(
      console.log('ServiceWorker registration failed: ', err);
    });
  });
  
  window.addEventListener('beforeunload', evt => {
  	document.body.classList.add('pwa-loading-frame-open');
  });
}
</script>

EOD;

		$view->addScript($script);

		$view->addStylesheet('assets/css/pwa.css');


		// Generate content to use between page loads.
		// This is important because apps do not otherwise provide a UX that anything is happening.
		$loadingTemplate = \Core\Templates\Template::Factory('includes/pwa/loading');
		$view->appendBodyContent($loadingTemplate->fetch());

		return true;
	}

	/**
	 * Hook to add necessary configuration values for the service workers.
	 *
	 * @return array
	 */
	public static function ConfigHook() : array {

		return [
			'ROOT_URL' => ROOT_URL,
			'ROOT_WDIR' => ROOT_WDIR,
		];
	}

	/**
	 * @param UserModel $user
	 * @return array
	 * @throws Exception
	 */
	public static function UserModelControlHook(UserModel $user) : array {
		$ret = [];

		$currentUser = \Core\user();
		$isSelf      = $user->id === $currentUser->id;
		$isManager   = $currentUser->checkAccess('p:/user/users/manage');

		if($user->active === '1' && ($isSelf || $isManager)) {
			$ret[] = [
				'title'  => 't:STRING_PWA_MANAGE_PUSH_NOTIFICATIONS',
				'link'   => '/pwa/notifications/manage',
				'icon'   => 'archive',
				'weight' => ViewControl::WEIGHT_UTILITIES,
				'group'  => ViewControl::GROUP_MANAGE,
			];
		}

		return $ret;
	}
}