<?php


namespace PWA;


use Core\Filestore\Factory;

class PushNotification extends PushObject {

	protected $_title;
	protected $_body;
	protected $_icon;
	protected $_image;
	protected $_badge;
	protected $_sound;
	protected $_tag;
	protected $_actions = [];

	/**
	 *
	 * ## Platform Support
	 *
	 * Browser | Gnome   | KDE     | Windows | MacOS   | Android
	 * --------|---------|---------|---------|---------|---------
	 * Firefox | Y       | Y       | *       | *       | *
	 * Chrome  | Y       | Y       | Y       | Y       | Y
	 * Opera   | *       | *       | *       | *       | *
	 *
	 * "*" Indicates Unknown or Untested Support
	 *
	 * @param string $title
	 * @return PushNotification
	 */
	public function title(string $title) : self {
		$this->_title = $title;

		return $this;
	}

	/**
	 * ## Platform Support
	 *
	 * Browser | Gnome   | KDE     | Windows | MacOS   | Android
	 * --------|---------|---------|---------|---------|---------
	 * Firefox | Y       | Y       | *       | *       | *
	 * Chrome  | Y       | Y       | Y       | Y       | Y
	 * Opera   | *       | *       | *       | *       | *
	 *
	 * "*" Indicates Unknown or Untested Support
	 *
	 * @param string $body
	 * @return PushNotification
	 */
	public function body(string $body) : self {
		$this->_body = $body;

		return $this;
	}


	/**
	 * The icon option is essentially a small image you can show next to the title and body text.
	 *
	 * Recommended image size is 192px x 192px
	 *
	 * ## Platform Support
	 *
	 * Browser | Gnome   | KDE     | Windows | MacOS   | Android
	 * --------|---------|---------|---------|---------|---------
	 * Firefox | Y       | Y       | *       | *       | *
	 * Chrome  | Y       | Y       | Y       | *       | *
	 * Opera   | *       | *       | *       | *       | *
	 *
	 * "*" Indicates Unknown or Untested Support
	 * @param string $icon
	 * @return PushNotification
	 */
	public function icon(string $icon) : self {
		try {
			$file = Factory::File($icon);
			$icon = $file->getURL();
		}
		catch(\Exception $ex) {
			\Core\log_exception($ex);
			$icon = null;
		}

		$this->_icon = $icon;

		return $this;
	}


	/**
	 * The image option can be used to display a larger image to the user.
	 * This is particularly useful to display a preview image to the user.
	 *
	 * Recommended image width is 1350px
	 *
	 * ## Platform Support
	 *
	 * Browser | Gnome   | KDE     | Windows | MacOS   | Android
	 * --------|---------|---------|---------|---------|---------
	 * Firefox | N       | N       | *       | *       | *
	 * Chrome  | *       | *       | Y       | *       | Y
	 * Opera   | *       | *       | *       | *       | *
	 *
	 * "*" Indicates Unknown or Untested Support
	 *
	 * @param string $image
	 * @return PushNotification
	 */
	public function image(string $image) : self {
		try {
			$file = Factory::File($image);
			$image = $file->getURL();
		}
		catch(\Exception $ex) {
			\Core\log_exception($ex);
			$image = null;
		}

		$this->_image = $image;

		return $this;
	}


	/**
	 * The badge is a small monochrome icon that is used to portray a little more information
	 * to the user about where the notification is from.
	 * 
	 * Recommended image size is 72px x 72px
	 *
	 * ## Platform Support
	 *
	 * Browser | Gnome   | KDE     | Windows | MacOS   | Android
	 * --------|---------|---------|---------|---------|---------
	 * Firefox | N       | N       | N       | N       | N
	 * Chrome  | N       | N       | N       | N       | Y
	 * Opera   | N       | N       | N       | N       | N
	 *
	 * "*" Indicates Unknown or Untested Support
	 *
	 * @param string $badge
	 * @return PushNotification
	 */
	public function badge(string $badge) : self {
		$this->_badge = $badge;

		return $this;
	}


	/**
	 * ## Platform Support
	 *
	 * Browser | Gnome   | KDE     | Windows | MacOS   | Android
	 * --------|---------|---------|---------|---------|---------
	 * Firefox | N       | N       | N       | N       | N
	 * Chrome  | N       | N       | N       | N       | N
	 * Opera   | N       | N       | N       | N       | N
	 *
	 * "*" Indicates Unknown or Untested Support
	 *
	 * @param string $sound
	 * @return PushNotification
	 */
	public function sound(string $sound) : self {
		$this->_sound = $sound;

		return $this;
	}

	/**
	 * ## Platform Support
	 *
	 * Browser | Gnome   | KDE     | Windows | MacOS   | Android
	 * --------|---------|---------|---------|---------|---------
	 * Firefox | *       | *       | *       | *       | *
	 * Chrome  | *       | *       | *       | *       | *
	 * Opera   | *       | *       | *       | *       | *
	 *
	 * "*" Indicates Unknown or Untested Support
	 *
	 * @param string $tag
	 * @return PushNotification
	 */
	public function tag(string $tag) : self {
		$this->_tag = $tag;

		return $this;
	}

	/**
	 * You can define actions to display buttons with a notification.
	 *
	 * ## Platform Support
	 *
	 * Browser | Gnome   | KDE     | Windows | MacOS   | Android
	 * --------|---------|---------|---------|---------|---------
	 * Firefox | N       | N       | N       | N       | N
	 * Chrome  | N       | N       | N       | N       | Y
	 * Opera   | N       | N       | N       | N       | Y
	 *
	 * "*" Indicates Unknown or Untested Support
	 *
	 * @param string $title
	 * @param string $action
	 * @param string|null $icon
	 * @return PushNotification
	 */
	public function action(string $title, string $action, string $icon = null) : self {

		if($icon) {
			try {
				$file = Factory::File($icon);
				$icon = $file->getURL();
			}
			catch(\Exception $ex) {
				\Core\log_exception($ex);
				$icon = null;
			}
		}

		$this->_actions[] = [
			'action' => $action,
			'title' => $title,
			'icon' => $icon,
		];

		return $this;
	}

	/**
	 * Get this push object as a transmittable payload
	 *
	 * @return string
	 */
	public function getAsPayload() : string {
		$payload = [
			'type' => 'notification',
		];

		if($this->_title) {
			$payload['title'] = $this->_title;
		}

		if($this->_body) {
			$payload['body'] = $this->_body;
		}

		if($this->_icon) {
			$payload['iconUrl'] = $this->_icon;
			$payload['icon'] = $this->_icon;
		}

		if($this->_image) {
			$payload['image'] = $this->_image;
		}

		if($this->_sound) {
			$payload['sound'] = $this->_sound;
		}

		if($this->_tag) {
			$payload['tag'] = $this->_tag;
		}

		if(count($this->_actions)) {
			$payload['actions'] = $this->_actions;
		}

		return json_encode($payload);
	}
}