<?php


namespace PWA;


use Core\ConfigHandler;
use Core\RedirectException;
use Exception;
use Jose\Component\KeyManagement\Analyzer\Message;
use Minishlink\WebPush\MessageSentReport;
use Minishlink\WebPush\VAPID;

class PushObject {
	/**
	 * Get this push object as a transmittable payload
	 *
	 * @return string
	 */
	public function getAsPayload() : string {
		trigger_error('This method must be extended!');
		return '';
	}

	/**
	 * Send a push object to a single subscription device
	 *
	 * @param \PwaUserSubscriptionModel $model
	 *
	 * @throws \ErrorException
	 * @throws \Exception
	 *
	 * @return MessageSentReport
	 */
	public function sendToSubscription(\PwaUserSubscriptionModel $model) : MessageSentReport {
		$webpush = new \Minishlink\WebPush\WebPush(
			[
				'VAPID' => self::_GetVAPID(),
			]
		);

		/** @var \Generator|MessageSentReport[] $reports */
		$reports = $webpush->sendNotification($model->getAsSubscription(), $this->getAsPayload(), true);

		foreach($reports as $report) {
			if(!$report->isSuccess()) {
				throw new Exception($report->getReason(), $report->getResponse()->getStatusCode());
			}
			return $report;
		}
	}

	/**
	 * Send a push object to any device for a single user
	 *
	 * For tracking purposes, the number of *successful* sends are returned.
	 *
	 * @param string $userid
	 *
	 * @throws \ErrorException
	 * @throws \Exception
	 *
	 * @return int
	 */
	public function sentToUser(string $userid) : int {
		$webpush = new \Minishlink\WebPush\WebPush(
			[
				'VAPID' => self::_GetVAPID(),
			]
		);

		$payload = $this->getAsPayload();

		$subscriptions = \PwaUserSubscriptionModel::Find(['user = ' . $userid]);
		foreach($subscriptions as $subscription) {
			/** @var \PwaUserSubscriptionModel $subscription */
			$webpush->sendNotification($subscription->getAsSubscription(), $payload);
		}

		$counter = 0;
		foreach($webpush->flush() as $report) {
			$report->isSuccess();
		}
	}

	/**
	 * Get the server public key, will auto-generate one if needed.
	 *
	 * If the key cannot be created, an Exception is thrown.
	 *
	 * @return string
	 *
	 * @throws Exception
	 */
	public static function GetPublicKey() : string {
		return self::_GetVAPID()['privateKey'];
	}

	/**
	 * Get the VAPID object that can be used in push events
	 *
	 * @return array
	 *
	 * @throws \Exception
	 */
	protected static function _GetVAPID() : array {
		$pub = ConfigHandler::Get('/pwa/serverkey/public');
		$pri = ConfigHandler::Get('/pwa/serverkey/private');
		$sub = SITENAME . ' (' . ROOT_URL . ')';

		if(!$pub || !$pri) {
			// No public or private keys are set yet!
			// Auto generate them.

			$keys = VAPID::createVapidKeys();

			$pub = $keys['publicKey'];
			$pri = $keys['privateKey'];

			ConfigHandler::Set('/pwa/serverkey/public', $pub);
			ConfigHandler::Set('/pwa/serverkey/private', $pri);
		}

		return [
			'subject' => $sub,
			'publicKey' => $pub,
			'privateKey' => $pri,
		];
	}


}