<?php
/**
 * Admin controller, handles all /Admin requests
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\BaconIpsumGenerator;
use Core\AccessDeniedException;
use Core\BadRequestException;
use Core\CLI\CLI;
use Core\Cache;
use Core\Component;
use Core\ConfigHandler;
use Core\Controller;
use Core\Core;
use Core\Datamodel\Dataset;
use Core\Datamodel\DMI;
use Core\Datamodel\DMIException;
use Core\Date\DateTime;
use Core\EmailAPI\Email;
use Core\Filestore\Factory;
use Core\Filestore\Factory as FileFactory;
use Core\Filestore\File;
use Core\FilterForm;
use Core\Forms\Form;
use Core\Forms\FormGroup;
use Core\HealthCheckResult;
use Core\HookHandler;
use Core\Installer\InstallHelper;
use Core\Installer\Checks;
use Core\I18N\I18NLoader;
use Core\ListingTable\Table;
use Core\ModelFactory;
use Core\Session;
use Core\TemporaryFollowRedirectException;
use Core\Templates\Backends\Smarty;
use Core\Theme;
use Core\Updater\ArchiveExporter;
use Core\Updater\ArchiveImporter;
use Core\Utilities\Profiler\Profiler;
use Core\Utilities\Profiler\ProfilerEvent;
use Core\View;
use Core\Widget;
use Core\Page\PageTable;
use Core\ViewControls;
use Core\ViewControl;
use Ifsnop\Mysqldump\Mysqldump;


class AdminController extends Controller {

	/**
	 * Display the admin dashboard.
	 *
	 * This page is primarily made up of widgets added by other systems.
	 *
	 * @return int
	 */
	public function index() {
		$view     = $this->getView();
		$pages    = PageModel::Find(array('admin' => '1'));
		$viewable = array();

		foreach ($pages as $p) {
			if (!\Core\user()->checkAccess($p->get('access'))) continue;

			$viewable[] = $p;
		}

		// If there are no viewable pages... don't display any admin dashboard.
		if(!sizeof($viewable)){
			throw new AccessDeniedException();
		}

		$view->title = 't:STRING_ADMIN';
		$view->assign('links', $viewable);

		// Dispatch the hook that other systems can hook into and perform checks or operations on the admin dashboard page.
		HookHandler::DispatchHook('/core/admin/view');
	}
	
	/**
	 * Autofix underlying Core issues in the filesystem, (if any)
	 * 
	 * @core-access-string g:admin
	 */
	public function autofix() {
		$checks = [];
		$allPassed = true;
		
		$autos = InstallHelper::GetChecks();
		
		// Add checks for the web server itself, (since we're running from the CLI, that may not be a given)
		$autos[] = new Checks\ApacheServerCheck();
		
		// Append the dynamic checks based off of configuration variables.
		$autos[] = new Checks\AssetDirectoryCheck();
		$autos[] = new Checks\PublicDirectoryCheck();
		$autos[] = new Checks\PrivateDirectoryCheck();
		
		
		// Perform all the standard installer checks and provide the fixes for any that fail.
		foreach($autos as $check){
			if(!$check->test()){
				$allPassed = false;
				$checks[] = [
					'title' => $check->getTitle(),
					'desc'  => $check->getDetails(),
					'fix'   => $check->getFix(),
				];
			}
		}
		
		$this->view->title = 'Auto Fix Core';
		$this->view->assign('checks', $autos);
	}
	
	/**
	 * Disable a given component by its keyname.
	 * 
	 * This component will require to be present on the server and ready to be installed already!
	 * 
	 * @param string $component
	 * 
	 * @core-access-string g:admin
	 * @core-request-method POST
	 */
	public function component_disable(string $component){
		$comp = Core::GetComponent($component);
		
		if(!$comp->isEnabled()){
			\Core\set_message('t:MESSAGE_ERROR_COMPONENT_S_ALREADY_DISABLED', $comp->getName());
			throw new TemporaryFollowRedirectException('BACK');
		}
		
		$comp->disable();
		
		\Core\set_message('t:MESSAGE_SUCCESS_COMPONENT_S_DISABLED', $comp->getName());
		
		// Flush the system cache, just in case
		Cache::Flush();
		Smarty::FlushCache();
		
		throw new TemporaryFollowRedirectException('BACK');
	}
	
	/**
	 * Enable a given component by its keyname.
	 * 
	 * This component will require to be present on the server and ready to be installed already!
	 * 
	 * @param string $component
	 * 
	 * @core-access-string g:admin
	 * @core-request-method POST
	 */
	public function component_enable(string $component){
		$comp = Core::GetComponent($component);
		
		if($comp->isEnabled()){
			\Core\set_message('t:MESSAGE_ERROR_COMPONENT_S_ALREADY_ENABLED', $comp->getName());
			throw new TemporaryFollowRedirectException('BACK');
		}
		
		$comp->enable();
		
		\Core\set_message('t:MESSAGE_SUCCESS_COMPONENT_S_ENABLED', $comp->getName());
		
		// Flush the system cache, just in case
		Cache::Flush();
		Smarty::FlushCache();
		
		throw new TemporaryFollowRedirectException('BACK');
	}
	
	/**
	 * Install a given component by its keyname.
	 * 
	 * This component will require to be present on the server and ready to be installed already!
	 * 
	 * @param string $component
	 * 
	 * @core-access-string g:admin
	 * @core-request-method POST
	 */
	public function component_install(string $component){
		$comp = Core::GetComponent($component);
		
		if($comp->isInstalled()){
			\Core\set_message('t:MESSAGE_ERROR_COMPONENT_S_ALREADY_INSTALLED', $comp->getName());
			throw new TemporaryFollowRedirectException('BACK');
		}
		
		$results = $comp->install();
		
		if(count($results['errors'])){
			\Core\set_message('t:MESSAGE_WARNING_COMPONENT_S_INSTALLED_WITH_ERRORS', $comp->getName());
			\Core\set_message(implode('<br/>', array_merge($results['changes'], $results['errors'])), 'log');
		}
		else{
			\Core\set_message('t:MESSAGE_SUCCESS_COMPONENT_S_INSTALLED', $comp->getName());
			\Core\set_message(implode('<br/>', $results['changes']), 'log');
		}
		
		// Flush the system cache, just in case
		Cache::Flush();
		
		throw new TemporaryFollowRedirectException('BACK');
	}
	
	/**
	 * Install ALL available components
	 *
	 * @core-access-string g:admin
	 * @core-request-method POST
	 */
	public function component_installall(){
		$components = Core::GetDisabledComponents();
		
		$resultsAllChanges = [];
		$resultsAllErrors  = [];
		
		foreach($components as $comp) {
			if(!$comp->isInstalled()){
				$results = $comp->install();
				
				$resultsAllChanges = array_merge($resultsAllChanges, $results['changes']);
				$resultsAllErrors = array_merge($resultsAllErrors, $results['errors']);
			}
		}
		
		if(count($resultsAllErrors)){
			\Core\set_message('t:MESSAGE_WARNING_COMPONENTS_INSTALLED_WITH_ERRORS');
			\Core\set_message(implode('<br/>', array_merge($resultsAllChanges, $resultsAllErrors)), 'log');
		}
		else{
			\Core\set_message('t:MESSAGE_SUCCESS_COMPONENTS_INSTALLED');
			\Core\set_message(implode('<br/>', $resultsAllChanges), 'log');
		}
		
		// Flush the system cache, just in case
		Cache::Flush();
		
		throw new TemporaryFollowRedirectException('BACK');
	}
	
	/**
	 * Uninstall a given component by its keyname.
	 * 
	 * This component will require to be present on the server and ready to be installed already!
	 * 
	 * @param string $component
	 * 
	 * @core-access-string g:admin
	 * @core-request-method POST
	 */
	public function component_uninstall(string $component){
		$comp = Core::GetComponent($component);
		
		if(!$comp->isInstalled()){
			\Core\set_message('t:MESSAGE_ERROR_COMPONENT_S_ALREADY_UNINSTALLED', $comp->getName());
			throw new TemporaryFollowRedirectException('BACK');
		}
		
		$results = $comp->uninstall();
		
		if(count($results['errors'])){
			\Core\set_message('t:MESSAGE_WARNING_COMPONENT_S_UNINSTALLED_WITH_ERRORS', $comp->getName());
			\Core\set_message(implode('<br/>', array_merge($results['changes'], $results['errors'])), 'log');
		}
		else{
			\Core\set_message('t:MESSAGE_SUCCESS_COMPONENT_S_UNINSTALLED', $comp->getName());
			\Core\set_message(implode('<br/>', $results['changes']), 'log');
		}
		
		// Flush the system cache, just in case
		Cache::Flush();
		Smarty::FlushCache();
		
		throw new TemporaryFollowRedirectException('BACK');
	}

	/**
	 * Full page to display the health checks of the site.
	 * 
	 * @core-access-string g:admin
	 */
	public function health(){
		$view    = $this->getView();
		$request = $this->getRequest();
		
		$checks = HookHandler::DispatchHook('/core/healthcheck');
		
		$view->title = 't:STRING_SYSTEM_HEALTH';
		$view->assign('checks', $checks);
	}
	
	/**
	 * Display the server ID of this site
	 * 
	 * @core-access-string g:admin
	 */
	public function serverid(){
		$view    = $this->getView();
		$request = $this->getRequest();
		
		$serverid = defined('SERVER_ID') ? SERVER_ID : null;
		
		if($serverid === null || $serverid == ''){
			\Core\set_message('t:MESSAGE_ERROR_SERVER_ID_NOT_SET_ADD_TO_CONFIGURATION');
			$newkey = \Core\random_hex(32);
		}
		elseif(strlen($serverid) < 32){
			\Core\set_message('t:MESSAGE_WARNING_SERVER_ID_LEGACY_UPDATE_NOW');
			$newkey = \Core\random_hex(32);
		}
		else{
			// Format the server ID to be human-readable (ish).
			$serverid = wordwrap($serverid, 4, '-', true);
			$newkey = null;
		}
		
		
		$view->title = 't:STRING_SERVER_ID';
		$view->assign('server_id', $serverid);
		$view->assign('new_key', $newkey);
	}

	/**
	 * Simple page to clear system cache
	 *
	 * @core-exec-mode BOTH
	 * @core-access-string g:admin
	 */
	public function clearcache() {
		$view    = $this->getView();
		$request = $this->getRequest();

		if(EXEC_MODE === 'CLI') {
			CLI::PrintWarning('CLI and Web caches are usually in two different locations, so you may also need to call /admin/clearcache from the web interface.');

			Cache::Flush();
			Smarty::FlushCache();
		}
		elseif($request->isPost()) {
			Cache::Flush();
			Smarty::FlushCache();

			Core::SetMessage('Cleared system cache!', 'success');
		}

		$view->mastertemplate = 'admin';
		$view->title = 't:STRING_CLEAR_SYSTEM_CACHE';
	}

	/**
	 * Run through and reinstall all components and themes.
	 *
	 * @core-exec-mode BOTH
	 * @core-access-string g:admin
	 * @return int
	 */
	public function reinstall() {
		// Just run through every component currently installed and reinstall it.
		// This will just ensure that the component is up to date and correct as per the component.xml metafile.
		$view    = $this->getView();
		$request = $this->getRequest();


		if($request->isPost() || EXEC_MODE === 'CLI'){
			$view->mode = View::MODE_STREAM;
			$view->contenttype = View::CTYPE_HTML;
			$view->record = false;
			$view->templatename = null;
			$view->render();
			
			// Calculate the amount of actions required to be performed.
			// This will be used to render an accurate progress bar.
			$actions    = 2;
			$components = [];
			$theme      = null;
			$report     = [];
			
			foreach(Core::GetComponents() as $c){
				/** @var $c Component */
				
				if($c->isInstalled() && $c->isEnabled()){
					$components[] = $c;
					$actions += 5; // Components have 5 actions to perform.
				}
			}
			
			// There's always one and only one theme enabled.
			try {
				$theme = Theme::GetTheme();
				$actions += 2; // Themes only have 2 actions to perform.

				// Render how much progress each action consists of.
				$progressEach = '+' . (100 / $actions);
			}
			catch(Exception $ex) {
				\Core\log_exception($ex);
				$report['Theme'] = [
					'changes' => [],
					'errors' => [ $ex->getMessage() ],
				];
				$theme = null;
			}


			
			
			// Now we can perform the actual install!
			foreach($components as $component){
				/** @var $component Component */
				
				$report[$component->getName()] = [
					'changes' => [],
					'errors'  => []
				];
				
				CLI::PrintProgressBar($progressEach, 'Reinstalling DB Schema for ' . $component->getName());
				$report[$component->getName()] = array_merge_recursive(
					$report[$component->getName()],
					$component->installDBSchema()
				);
				usleep(100000);
				
				CLI::PrintProgressBar($progressEach, 'Reinstalling Configs for ' . $component->getName());
				$report[$component->getName()] = array_merge_recursive(
					$report[$component->getName()],
					$component->installConfigs()
				);
				usleep(100000);
				
				CLI::PrintProgressBar($progressEach, 'Reinstalling Pages for ' . $component->getName());
				$report[$component->getName()] = array_merge_recursive(
					$report[$component->getName()],
					$component->installPages()
				);
				usleep(100000);
				
				CLI::PrintProgressBar($progressEach, 'Reinstalling Widgets for ' . $component->getName());
				$report[$component->getName()] = array_merge_recursive(
					$report[$component->getName()],
					$component->installWidgets()
				);
				usleep(100000);
				
				CLI::PrintProgressBar($progressEach, 'Reinstalling Assets for ' . $component->getName());
				$report[$component->getName()] = array_merge_recursive(
					$report[$component->getName()],
					$component->installAssets()
				);
				usleep(100000);
			}
			

			if($theme) {
				$report[$theme->getName()] = [
					'changes' => [],
					'errors'  => []
				];

				CLI::PrintProgressBar($progressEach, 'Reinstalling Assets for Theme ' . $theme->getName());
				$report[$theme->getName()] = array_merge_recursive(
					$report[$theme->getName()],
					$theme->installAssets()
				);
				usleep(100000);

				CLI::PrintProgressBar($progressEach, 'Reinstalling Configs for Theme ' . $theme->getName());
				$report[$theme->getName()] = array_merge_recursive(
					$report[$theme->getName()],
					$theme->installConfigs()
				);
				usleep(100000);
			}
			
			
			// Flush the system cache, just in case
			CLI::PrintProgressBar($progressEach, 'Flushing System Cache');
			Cache::Flush();
			Smarty::FlushCache();
			usleep(100000);

			// Increment the version counter.
			CLI::PrintProgressBar($progressEach, 'Incrementing Asset Version Counter');
			$version = ConfigHandler::Get('/core/filestore/assetversion');
			ConfigHandler::Set('/core/filestore/assetversion', ++$version);
			usleep(100000);
			
			
			$errorsDetected = false;
			$changesDetected = false;
			CLI::Clear();
			foreach($report as $title => $results){
				if(count($results['changes']) || count($results['errors'])){
					CLI::PrintHeader('Reinstallation results for ' . $title);
				}
				if(count($results['changes'])){
					CLI::PrintLine($results['changes']);
					$changesDetected = true;
				}
				if(count($results['errors'])){
					CLI::PrintError($results['errors']);
					$errorsDetected = true;
				}
			}
			
			if($errorsDetected){
				CLI::PrintError('Reinstallation completed with errors');
			}
			elseif($changesDetected){
				CLI::PrintSuccess('Reinstallation completed successfully');
			}
			else{
				CLI::PrintLine('Reinstallation completed with no changes');
			}
		} // End if is post.

		$view->addBreadcrumb('t:STRING_ADMIN', '/admin');
		$view->title = 't:STRING_RESYNC_AND_REDEPLOY';
	}

	/**
	 * Display ALL the system configuration options.
	 *
	 * @core-access-string g:admin
	 * 
	 * @return int
	 */
	public function config() {
		$view = $this->getView();

		$where = array();
		// If the enterprise component is installed and multisite is enabled, configurations have another layer of complexity.
		if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::GetCurrentSiteID()){
			$where['overrideable'] = '1';
		}

		$configs = ConfigModel::Find($where, null, 'key');

		$groups  = array();
		foreach ($configs as $c) {
			/** @var ConfigModel $c */
			// Export out the group for this config option.
			$el = $c->getAsFormElement();
			$gname = $el->get('group');

			if (!isset($groups[$gname])){
				$groups[$gname] = new FormGroup(
					[
						'title' => $gname,
						'name' => $gname,
						//'class' => 'collapsible collapsed'
						'class' => 'system-config-group'
					]
				);
			}

			$groups[$gname]->addElement($el);
		}


		$form = new Form();
		$form->set('callsmethod', 'AdminController::_ConfigSubmit');
		// This form gives me more trouble with its persistence....
		// @todo Implement a better option than this.
		// This hack is designed to prevent this form from using stale values from a previous page load instead of
		// pulling them from the database.
		$form->set('uniqueid', 'admin-config-' . \Core\random_hex(6));
		foreach ($groups as $g) {
			$form->addElement($g);
		}

		$form->addElement('submit', array('value' => t('STRING_SAVE')));

		$this->setTemplate('/pages/admin/config.tpl');
		$view->assign('form', $form);
		$view->assign('config_count', sizeof($configs));
	}

	/**
	 * Sync the search index fields of every model on the system.
	 *
	 * @core-access-string g:admin
	 */
	public function syncsearchindex() {
		// Just run through every component currently installed and reinstall it.
		// This will just ensure that the component is up to date and correct as per the component.xml metafile.
		$view       = $this->getView();
		$request    = $this->getRequest();
		$models     = [];
		$components = \Core\Core::GetComponents();


		foreach($components as $component) {
			$classes = $component->getClassList();

			foreach(array_keys($classes) as $class) {
				if(strrpos($class, 'Model') === strlen($class) - 5 && strpos($class, '\\') === false) {
					try {
						$ref = new ReflectionClass($class);
					}
					catch(ReflectionException $ex) {
						Core::SetMessage($ex->getMessage(), 'warning');
						continue;
					}

					if($ref->getProperty('HasSearch')->getValue(null)) {
						$models[] = $class;
					}
				}
			}
		}

		sort($models);


		if($request->isPost() && $request->getParameter('model')) {
			$view->mode = View::MODE_AJAX;
			$view->contenttype = View::CTYPE_JSON;
			$class = $request->getParameter('model');
			$fac = new ModelFactory($class);

			$profiler = new ProfilerEvent('Syncing ' . $class . ' search indexes');
			$profiler->start();
			foreach($fac->generator() as $model) {
				$profiler->addEventString('Indexing ' . $model->getLabel());
				$model->set('search_index_pri', '!');
				$model->save();
			}

			$profiler->stop();
			$view->setJSONSuccessMessage('Updated indexes');
			return;
		}


		$view->mastertemplate = 'admin';
		$view->title = 'Sync Searchable Index';
		$view->assign('models', $models);
	}

	/**
	 * Display a list of system logs that have been recorded.
	 *
	 * @core-access-string p:/core/systemlog/view
	 * 
	 * @return int
	 */
	public function log(){
		$view = $this->getView();
		$request = $this->getRequest();

		$codes = ['' => '-- All --'];
		$ds = Dataset::Init()
			->table('system_log')
			->select('code')
			->unique(true)
			->order('code')
			->execute();

		foreach($ds as $row){
			$codes[$row['code']] = $row['code'];
		}

		$types = ['' => '-- All --'];
		$ds = Dataset::Init()
			->table('system_log')
			->select('type')
			->unique(true)
			->order('type')
			->execute();

		foreach($ds as $row){
			$types[$row['type']] = ucfirst($row['type']);
		}

		$listings = new Table();
		$listings->setModelName('SystemLogModel');
		$listings->setDefaultSort('datetime');

		$listings->addFilter(
			'select',
			array(
				'title' => 'Type',
				'name' => 'type',
				'options' => $types,
				'link' => FilterForm::LINK_TYPE_STANDARD,
			)
		);

		$listings->addFilter(
			'select',
			[
				'title' => 'Code',
				'name' => 'code',
				'options' => $codes,
				'link' => FilterForm::LINK_TYPE_STANDARD,
			]
		);

		$listings->addFilter(
			'date',
			[
				'title' => 'On or After',
				'name' => 'datetime_onafter',
				'linkname' => 'datetime',
				'link' => FilterForm::LINK_TYPE_GE,
			]
		);
		$listings->addFilter(
			'date',
			[
				'title' => 'On or Before',
				'name' => 'datetime_onbefore',
				'linkname' => 'datetime',
				'link' => FilterForm::LINK_TYPE_LE,
				'linkvaluesuffix' => ' 23:59:59'
			]
		);

		$listings->addFilter(
			'hidden',
			array(
				'title' => 'Session',
				'name' => 'session_id',
				'link' => FilterForm::LINK_TYPE_STANDARD,
			)
		);
		$listings->addFilter(
			'user',
			array(
				'title' => 'User',
				'name' => 'affected_user_id',
				'linkname' => [
					'affected_user_id',
					'user_id',
				],
				'link' => FilterForm::LINK_TYPE_STANDARD,
			)
		);
		$listings->addFilter(
			'text',
			[
				'title' => 'IP Address',
				'name' => 'ip_addr',
				'link' => FilterForm::LINK_TYPE_STARTSWITH,
			]
		);
		
		$listings->addFilter(
			'text',
			[
				'title' => 'Message Contains',
				'name' => 'message',
				'link' => FilterForm::LINK_TYPE_CONTAINS,
			]
		);

		$listings->addColumn('message');
		$listings->addColumn('type');
		$listings->addColumn('level')->hidden();
		$listings->addColumn('datetime');
		$listings->addColumn('ip_addr');
		$listings->addColumn('url')->hidden();
		$listings->addColumn('referrer')->hidden();
		$listings->addColumn('useragent')->hidden();
		$listings->addColumn('user_id')->hidden();
		$listings->addColumn('affected_user_id')->hidden();

		$listings->loadFiltersFromRequest($request);

		$view->mastertemplate = 'admin';
		$view->title = 't:STRING_SYSTEM_LOG';
		$view->assign('listings', $listings);
	}

	/**
	 * Page to display full details of a system log, usually opened in an ajax dialog.
	 * 
	 * @core-access-string p:/core/systemlog/view
	 * 
	 * @return int
	 * @throws DMI_Exception
	 */
	public function log_details(){
		$view = $this->getView();
		$request = $this->getRequest();

		$view->mode = View::MODE_PAGEORAJAX;

		$log = SystemLogModel::Construct($request->getParameter(0));
		if(!$log->exists()){
			return View::STATUS_NOTFOUND;
		}

		$view->mastertemplate = $request->isAjax() ? false : 'admin';
		$view->addBreadcrumb('t:STRING_SYSTEM_LOG', '/admin/log');
		$view->title = 't:STRING_SYSTEM_LOG_DETAILS';
		$view->assign('entry', $log);
	}

	/**
	 * Display log management for Core.
	 * 
	 * @core-access-string p:/core/systemlog/view
	 * 
	 * @return int
	 */
	public function log_config(){
		$view = $this->getView();
		$request = $this->getRequest();

		if($request->getParameter('download')){
			$name = $request->getParameter('download');
			$dir = Factory::Directory(ROOT_PDIR . 'logs/');
			$file = Factory::File(ROOT_PDIR . 'logs/' . $name);
			if(!$file->inDirectory($dir->getPath())){
				return View::STATUS_BADREQUEST;
			}
			if(!$file->exists()){
				return View::STATUS_NOTFOUND;
			}
			$file->sendToUserAgent(true);
			return;
		}

		$keys = [
			'/core/logs/rotate/frequency',
			'/core/logs/rotate/compress',
			'/core/logs/rotate/keep',
			'/core/logs/db/keep',
			'/core/logs/level',
			'/core/logs/db/level',
			'/core/logs/syslog/level',
		];

		$form = new Form();
		$form->set('callsmethod', 'AdminController::_ConfigSubmit');

		foreach($keys as $k){
			$c = ConfigHandler::GetConfig($k);
			$f = $c->asFormElement();
			// Don't need them grouped
			$f->set('group', '');
			$form->addElement($f);
		}
		$form->addElement('submit', ['value' => t('STRING_SAVE')]);
		
		// Give me some information about the logs currently on the system.
		$dir = Factory::Directory(ROOT_PDIR . 'logs/');
		$logs = [];
		$archived = [];
		foreach($dir->ls() as $file){
			/** @var File $file */
			$b = $file->getBaseFilename(true);
			if($b == 'info'){
				$logs[] = [
					'type' => 'info',
					'file' => $file,
				];
			}
			elseif($b == 'security'){
				$logs[] = [
					'type' => 'security',
					'file' => $file,
				];
			}
			elseif($b == 'error'){
				$logs[] = [
					'type' => 'error',
					'file' => $file,
				];
			}
			elseif(strpos($b, 'info.log') === 0){
				$archived[] = [
					'type' => 'info',
					'file' => $file,
				];
			}
			elseif(strpos($b, 'security.log') === 0){
				$archived[] = [
					'type' => 'security',
					'file' => $file,
				];
			}
			elseif(strpos($b, 'error.log') === 0){
				$archived[] = [
					'type' => 'error',
					'file' => $file,
				];
			}
		}
		
		// Sort the archived ones.
		usort($archived, function($a, $b){
			return $a['file']->getBasename() < $b['file']->getBasename();
		});

		$view->title = 't:STRING_LOG_CONFIG';
		$view->assign('form', $form);
		$view->assign('logs', $logs);
		$view->assign('archived', $archived);
	}

	/**
	 * Display a listing of all pages registered in the system.
	 * 
	 * @core-access-string p:/core/pages/view
	 */
	public function pages() {
		$view = $this->getView();
		$request = $this->getRequest();

		$table = new PageTable();

		$table->loadFiltersFromRequest();
		
		$links = [];
		foreach(Core::GetComponents() as $c){
			/** @var Core\Component $c */
			
			$pageCreates = $c->getPageCreatesDefined();
			foreach($pageCreates as $dat){
				// Support i18n.
				if(strpos($dat['title'], 't:') === 0){
					$dat['title'] = t(substr($dat['title'], 2));
				}
				if(strpos($dat['description'], 't:') === 0){
					$dat['description'] = t(substr($dat['description'], 2));
				}
				$links[] = $dat;
			}
		}
		
		if(count($links)) {
			$view->addControl([
				'title' => 't:STRING_CREATE_NEW_PAGE',
				'href' => '#',
				'icon' => 'add',
				'class' => 'toggle-create-links'
			]);
		}

		$view->title = 't:STRING_ALL_PAGES';
		$view->assign('links', $links);
		$view->assign('listing', $table);
	}

	/**
	 * Shortcut for publishing a page.
	 * 
	 * @core-access-string p:/core/pages/manage
	 * @core-request-method POST
	 */
	public function page_publish() {
		$view    = $this->getView();
		$request = $this->getRequest();

		$baseurl = $request->getParameter('baseurl');

		$page = new PageModel($baseurl);
		if(!$page->exists()){
			return View::STATUS_NOTFOUND;
		}

		// Is this page already published?
		if($page->get('published_status') == 'published'){
			\Core\set_message('t:MESSAGE_ERROR_PAGE_ALREADY_PUBLISHED');
			\Core\go_back();
		}

		$page->set('published_status', 'published');
		$page->save();

		\Core\set_message('t:MESSAGE_SUCCESS_PAGE_PUBLISHED');
		\Core\go_back();
	}

	/**
	 * Shortcut for unpublishing a page.
	 * 
	 * @core-access-string p:/core/pages/manage
	 * @core-request-method POST
	 */
	public function page_unpublish() {
		$view    = $this->getView();
		$request = $this->getRequest();

		$baseurl = $request->getParameter('baseurl');

		$page = new PageModel($baseurl);
		if(!$page->exists()){
			return View::STATUS_NOTFOUND;
		}

		// Is this page already un-published?
		if($page->get('published_status') == 'draft'){
			\Core\set_message('t:MESSAGE_ERROR_PAGE_ALREADY_UNPUBLISHED');
			\Core\go_back();
		}

		$page->set('published_status', 'draft');
		$page->save();

		\Core\set_message('t:MESSAGE_SUCCESS_PAGE_UNPUBLISHED');
		\Core\go_back();
	}

	/**
	 * Display a listing of all pages registered in the system.
	 * 
	 * @deprecated since version 6.0.0
	 */
	public function widgets(){
		$view = $this->getView();
		$request = $this->getRequest();

		$viewer = \Core\user()->checkAccess('p:/core/widgets/manage');
		$manager = \Core\user()->checkAccess('p:/core/widgets/manage');
		if(!($viewer || $manager)){
			return View::STATUS_ACCESSDENIED;
		}

		\Core\redirect('/widget/admin');
	}

	/**
	 * Create a simple widget with the standard settings configurations.
	 * 
	 * @deprecated since version 6.0.0
	 */
	public function widget_create(){
		$view = $this->getView();
		$request = $this->getRequest();

		if(!\Core\user()->checkAccess('p:/core/widgets/manage')){
			return View::STATUS_ACCESSDENIED;
		}
		\Core\redirect('/widget/create');
	}

	/**
	 * Create a simple widget with the standard settings configurations.
	 * 
	 * @deprecated since version 6.0.0
	 */
	public function widget_update(){
		$view = $this->getView();
		$request = $this->getRequest();

		$baseurl = $request->getParameter('baseurl');
		\Core\redirect('/widget/update?baseurl=' . $baseurl);
	}

	/**
	 * Delete a simple widget.
	 * 
	 * @deprecated since version 6.0.0
	 */
	public function widget_delete(){
		$view = $this->getView();
		$request = $this->getRequest();

		if(!\Core\user()->checkAccess('p:/core/widgets/manage')){
			return View::STATUS_ACCESSDENIED;
		}

		if(!$request->isPost()){
			return View::STATUS_BADREQUEST;
		}

		$baseurl = $request->getParameter('baseurl');
		if($baseurl[0] == '/'){
			// Trim off the beginning '/' from the URL.
			$class = substr($baseurl, 1, strpos($baseurl, '/', 1)-1) . 'widget';
		}
		else{
			$class = substr($baseurl, 0, strpos($baseurl, '/')) . 'widget';
		}


		if(!class_exists($class)){
			\Core\set_message('t:MESSAGE_ERROR_CLASS_S_NOT_AVAILABLE', $class);
			\Core\go_back();
		}

		/** @var Widget $obj */
		$obj = new $class();

		if(!($obj instanceof \Core\Widget)){
			\Core\set_message('t:MESSAGE_ERROR_CLASS_S_NOT_VALID_WIDGET', $class);
			\Core\go_back();
		}

		if(!$obj->is_simple){
			\Core\set_message('t:MESSAGE_ERROR_CLASS_S_NOT_SIMPLE_WIDGET', $class);
			\Core\go_back();
		}

		$model = new WidgetModel($baseurl);

		$model->delete();
		\Core\set_message('t:MESSAGE_SUCCESS_DELETED_WIDGET_S', $baseurl);
		\Core\go_back();
	}

	public function widgetinstances_save(){
		$view = $this->getView();
		$request = $this->getRequest();

		if(!\Core\user()->checkAccess('p:/core/widgets/manage')){
			return View::STATUS_ACCESSDENIED;
		}

		if(!$request->isPost()){
			return View::STATUS_BADREQUEST;
		}

		$counter = 0;
		$changes = ['created' => 0, 'updated' => 0, 'deleted' => 0];

		$selected = $_POST['selected'];
		// For the incoming options, I want an explicit NULL if it's empty.
		$theme    = $_POST['theme'] == '' ? null : $_POST['theme'];
		$skin     = $_POST['skin'] == '' ? null : $_POST['skin']; $_POST['skin'];
		$template = $_POST['template'] == '' ? null : $_POST['template'];
		$baseurl  = $_POST['page_baseurl'] == '' ? null : $_POST['page_baseurl'];

		foreach($_POST['widgetarea'] as $id => $dat){

			// Merge in the global information for this request
			//$dat['theme']         = $theme;
			//$dat['skin']          = $skin;
			$dat['template']      = $template;
			$dat['page_baseurl']  = $baseurl;

			$dat['weight'] = ++$counter;
			$dat['access'] = $dat['widgetaccess'];

			$w = WidgetModel::Construct($dat['baseurl']);
			$dat['site'] = $w->get('site');

			if(strpos($id, 'new') !== false){
				$w = new WidgetInstanceModel();
				$w->setFromArray($dat);
				$w->save();
				$changes['created']++;
			}
			elseif(strpos($id, 'del-') !== false){
				$w = new WidgetInstanceModel(substr($id, 4));
				$w->delete();
				// Reset the counter back down one notch since this was a deletion request.
				--$counter;
				$changes['deleted']++;
			}
			else{
				$w = new WidgetInstanceModel($id);
				$w->setFromArray($dat);
				if($w->save()) $changes['updated']++;
			}
		} // foreach($_POST['widgetarea'] as $id => $dat)

		// Display some human friendly status message.
		if($changes['created'] || $changes['updated'] || $changes['deleted']){
			$changetext = [];

			if($changes['created'] == 1) $changetext[] = 'One widget added';
			elseif($changes['created'] > 1) $changetext[] = $changes['created'] . ' widgets added';

			if($changes['updated'] == 1) $changetext[] = 'One widget updated';
			elseif($changes['updated'] > 1) $changetext[] = $changes['updated'] . ' widgets updated';

			if($changes['deleted'] == 1) $changetext[] = 'One widget deleted';
			elseif($changes['deleted'] > 1) $changetext[] = $changes['deleted'] . ' widgets deleted';

			\Core\set_message(implode('<br/>', $changetext), 'success');
		}
		else{
			\Core\set_message('t:MESSAGE_INFO_NO_CHANGES_PERFORMED');
		}

		if($baseurl){
			\Core\redirect($baseurl);
		}
		else{
			\Core\redirect('/admin/widgets?selected=' . $selected);
		}

	}

	/**
	 * Page to test the UI of various Core elements
	 * 
	 * @core-access-string g:admin
	 */
	public function testui(){
		$view = $this->getView();
		$request = $this->getRequest();

		$lorem = new BaconIpsumGenerator();

		$skins = \Core\Templates\Template::GetSkins();

		if($request->getParameter('skin')) {
			$skin = $request->getParameter('skin');
			if(!isset($skins[ $skin ])) {
				// Only allow registered / located skins.
				$skin = 'admin';
			}
		}
		else {
			$skin = 'admin';
		}

		
		// Build some sample controls to render.
		$controls = new ViewControls();
		$buttons = [];
		foreach(
			[
				ViewControl::ICON_ADD, 
				ViewControl::ICON_DELETE, 
				ViewControl::ICON_DOWNLOAD, 
				ViewControl::ICON_VIEW, 
				ViewControl::ICON_EDIT,
				'coffee', 'utensils', 'gem', 'comments'
			]
			as $icon
		) {
			$c = new ViewControl();
			$c->icon    = $icon;
			$c->title   = $lorem->getWord(1);
			$c->confirm = $lorem->getWord(5) . '?';
			$c->link    = '#';
			$controls->addLink($c);

			$buttons[] = [
				'icon' => $icon,
				'text' => $lorem->getWord(1),
			];
		}

		// Retrieve a list of all icons available from FontAwesome.
		// we'll dynamically create the list because there are SOOOO many icons.
		$contents = explode("\n", file_get_contents(ROOT_PDIR . 'core/assets/scss/fontawesome/_icons.scss'));
		$icons = [];
		$brands = [
			'500px',
			'accessible-icon',
			'accusoft',
			'acquisitions-incorporated',
			'adn',
			'adobe',
			'adversal',
			'affiliatetheme',
			'airbnb',
			'algolia',
			'alipay',
			'amazon',
			'amazon-pay',
			'amilia',
			'android',
			'angellist',
			'angrycreative',
			'angular',
			'app-store',
			'app-store-ios',
			'apper',
			'apple',
			'apple-pay',
			'artstation',
			'asymmetrik',
			'atlassian',
			'audible',
			'autoprefixer',
			'avianex',
			'aviato',
			'aws',
			'bandcamp',
			'battle-net',
			'behance',
			'behance-square',
			'bimobject',
			'bitbucket',
			'bitcoin',
			'bity',
			'black-tie',
			'blackberry',
			'blogger',
			'blogger-b',
			'bluetooth',
			'bluetooth-b',
			'bootstrap',
			'btc',
			'buffer',
			'buysellads',
			'buromobelexperte',
			'canadian-maple-leaf',
			'cc-amazon-pay',
			'cc-amex',
			'cc-apple-pay',
			'cc-diners-club',
			'cc-paypal',
			'cc-discover',
			'cc-jcb',
			'cc-mastercard',
			'cc-stripe',
			'cc-visa',
			'centercode',
			'centos',
			'chrome',
			'chromecast',
			'cloudscale',
			'cloudsmith',
			'cloudversify',
			'codepen',
			'codiepie',
			'confluence',
			'connectdevelop',
			'contao',
			'cpanel',
			'creative-commons',
			'creative-commons-by',
			'creative-commons-nc',
			'creative-commons-nc-eu',
			'creative-commons-nc-jp',
			'creative-commons-nd',
			'creative-commons-pd',
			'creative-commons-pd-alt',
			'creative-commons-remix',
			'creative-commons-sa',
			'creative-commons-sampling',
			'creative-commons-sampling-plus',
			'creative-commons-share',
			'creative-commons-zero',
			'critical-role',
			'css3',
			'css3-alt',
			'cuttlefish',
			'd-and-d',
			'd-and-d-beyond',
			'dashcube',
			'delicious',
			'deploydog',
			'deskpro',
			'dev',
			'deviantart',
			'dhl',
			'diaspora',
			'digg',
			'digital-ocean',
			'discord',
			'discourse',
			'dochub',
			'docker',
			'draft2digital',
			'dribbble',
			'dribbble-square',
			'dropbox',
			'drupal',
			'dyalog',
			'earlybirds',
			'ebay',
			'edge',
			'elementor',
			'ello',
			'ember',
			'empire',
			'envira',
			'erlang',
			'ethereum',
			'etsy',
			'evernote',
			'expeditedssl',
			'facebook',
			'facebook-f',
			'facebook-messenger',
			'facebook-square',
			'fantasy-flight-games',
			'fedex',
			'fedora',
			'figma',
			'firefox',
			'firstdraft',
			'flickr',
			'font-awesome',
			'font-awesome-alt',
			'font-awesome-logo-full',
			'first-order',
			'first-order-alt',
			'flipboard',
			'fly',
			'font-awesome-flag',
			'fonticons',
			'fonticons-fi',
			'fort-awesome',
			'fort-awesome-alt',
			'forumbee',
			'foursquare',
			'free-code-camp',
			'freebsd',
			'fulcrum',
			'galactic-republic',
			'galactic-senate',
			'get-pocket',
			'gg',
			'gg-circle',
			'git',
			'git-alt',
			'git-square',
			'github',
			'github-alt',
			'github-square',
			'gitkraken',
			'gitlab',
			'gitter',
			'glide',
			'glide-g',
			'gofore',
			'goodreads',
			'goodreads-g',
			'google',
			'google-drive',
			'google-play',
			'google-plus',
			'google-plus-g',
			'google-plus-square',
			'google-wallet',
			'gratipay',
			'grav',
			'gripfire',
			'grunt',
			'gulp',
			'hacker-news',
			'hacker-news-square',
			'hackerrank',
			'hips',
			'hire-a-helper',
			'hooli',
			'hornbill',
			'hotjar',
			'houzz',
			'html5',
			'hubspot',
			'imdb',
			'instagram',
			'intercom',
			'internet-explorer',
			'invision',
			'ioxhost',
			'itch-io',
			'itunes',
			'itunes-note',
			'java',
			'jedi-order',
			'jenkins',
			'jira',
			'joget',
			'joomla',
			'js',
			'js-square',
			'jsfiddle',
			'kaggle',
			'keybase',
			'keycdn',
			'kickstarter',
			'kickstarter-k',
			'korvue',
			'laravel',
			'lastfm',
			'lastfm-square',
			'leanpub',
			'less',
			'line',
			'linkedin',
			'linkedin-in',
			'linode',
			'linux',
			'lyft',
			'magento',
			'mailchimp',
			'mandalorian',
			'markdown',
			'mastodon',
			'maxcdn',
			'medapps',
			'medium',
			'medium-m',
			'medrt',
			'meetup',
			'megaport',
			'mendeley',
			'microsoft',
			'mix',
			'mixcloud',
			'mizuni',
			'modx',
			'monero',
			'napster',
			'neos',
			'nimblr',
			'node',
			'node-js',
			'npm',
			'ns8',
			'nutritionix',
			'odnoklassniki',
			'odnoklassniki-square',
			'old-republic',
			'opencart',
			'openid',
			'opera',
			'optin-monster',
			'osi',
			'page4',
			'pagelines',
			'palfed',
			'patreon',
			'paypal',
			'penny-arcade',
			'periscope',
			'phabricator',
			'phoenix-framework',
			'phoenix-squadron',
			'php',
			'pied-piper',
			'pied-piper-alt',
			'pied-piper-hat',
			'pied-piper-pp',
			'pinterest',
			'pinterest-p',
			'pinterest-square',
			'playstation',
			'product-hunt',
			'pushed',
			'python',
			'qq',
			'quinscape',
			'quora',
			'r-project',
			'raspberry-pi',
			'ravelry',
			'react',
			'reacteurope',
			'readme',
			'rebel',
			'red-river',
			'reddit',
			'reddit-alien',
			'reddit-square',
			'redhat',
			'renren',
			'replyd',
			'researchgate',
			'resolving',
			'rev',
			'rocketchat',
			'rockrms',
			'safari',
			'salesforce',
			'sass',
			'schlix',
			'scribd',
			'searchengin',
			'sellcast',
			'sellsy',
			'servicestack',
			'shirtsinbulk',
			'shopware',
			'simplybuilt',
			'sistrix',
			'sith',
			'sketch',
			'skyatlas',
			'skype',
			'slack',
			'slack-hash',
			'slideshare',
			'snapchat',
			'snapchat-ghost',
			'snapchat-square',
			'soundcloud',
			'sourcetree',
			'speakap',
			'speaker-deck',
			'spotify',
			'squarespace',
			'stack-exchange',
			'stack-overflow',
			'stackpath',
			'staylinked',
			'steam',
			'steam-square',
			'steam-symbol',
			'sticker-mule',
			'strava',
			'stripe-s',
			'studiovinari',
			'stumbleupon',
			'stumbleupon-circle',
			'superpowers',
			'supple',
			'suse',
			'symfony',
			'teamspeak',
			'telegram',
			'telegram-plane',
			'tencent-weibo',
			'the-red-yeti',
			'themeco',
			'themeisle',
			'think-peaks',
			'trade-federation',
			'trello',
			'tripadvisor',
			'tumblr',
			'tumblr-square',
			'twitch',
			'twitter',
			'twitter-square',
			'typo3',
			'uber',
			'ubuntu',
			'uikit',
			'uniregistry',
			'untappd',
			'ups',
			'usb',
			'usps',
			'ussunnah',
			'vaadin',
			'viacoin',
			'viadeo',
			'viadeo-square',
			'viber',
			'vimeo',
			'vimeo-square',
			'vimeo-v',
			'vine',
			'vk',
			'vnv',
			'vuejs',
			'waze',
			'weebly',
			'weibo',
			'weixin',
			'whatsapp',
			'whatsapp-square',
			'whmcs',
			'wikipedia-w',
			'windows',
			'wix',
			'wizards-of-the-coast',
			'wolf-pack-battalion',
			'wordpress',
			'wordpress-simple',
			'wpbeginner',
			'stripe',
			'wpexplorer',
			'wpforms',
			'wpressr',
			'xbox',
			'xing',
			'xing-square',
			'y-combinator',
			'yahoo',
			'yammer',
			'yandex',
			'yandex-international',
			'yarn',
			'yelp',
			'yoast',
			'youtube',
			'youtube-square',
			'zhihu',
		];
		foreach($contents as $line) {
			if(strpos($line, '.#{$fa-css-prefix}-') === 0) {
				$icon = substr($line, 19);
				$icon = substr($icon, 0, strpos($icon, ':before'));

				if(!in_array($icon, $brands)) {
					$icons[] = $icon;
				}
			}
		}
		
		
		if($request->isPost()){
			// Test output for CLI modes.
			$view->mode = View::MODE_NOOUTPUT;
			$view->render();
			CLI::PrintHeader('Doing Something Important');
			for($i = 0; $i < 4; $i++){
				sleep(1);
				CLI::PrintProgressBar('+10');
			}
			sleep(1);
			CLI::PrintWarning('Something unexpected happened!');
			sleep(1);
			CLI::PrintProgressBar('+10'); // 50% now
			sleep(1);
			CLI::PrintError('Something bad happened!');
			for($i = 0; $i < 5; $i++){
				sleep(1);
				CLI::PrintProgressBar('+10');
			}
			
			return;
		}

		$view->mastertemplate = $skin;
		$view->title = 'Test General UI/UX';
		$view->assign('lorem_p', $lorem->getParagraphsAsMarkup(3));
		$view->assign(
			'lis', [
				$lorem->getWord(3),
				$lorem->getWord(3),
				$lorem->getWord(3),
				$lorem->getWord(3),
				$lorem->getWord(3),
			]
		);
		$view->assign('skins', $skins);
		$view->assign('skin', $skin);
		$view->assign('controls', $controls);
		$view->assign('buttons', $buttons);
		$view->assign('icons', $icons);
		$view->assign('icon_brands', $brands);
	}

	/**
	 * Page to view and test the i18n settings and strings of this site.
	 *
	 * Also useful for viewing what strings are currently installed and where they came from!
	 *
	 * @core-access-string g:admin
	 * 
	 * @return int
	 */
	public function i18n(){
		$view = $this->getView();
		$request = $this->getRequest();
		
		// I need to use GetLocalesAvailable because the higher level functions will only return what's currently enabled,
		// which is the entire point of this page!
		$locales = I18NLoader::GetLocalesAvailable();
		
		// Make this full list a more flat list suitable for populating directly into the form element.
		$all = [];
		foreach($locales as $key => $dat){
			$all[$key] = t($dat['lang']) . ' (' . t($dat['dialect']) . ')';
		}
		
		$enabled = ConfigHandler::Get('/core/language/languages_enabled');
		// This is expected to be a pipe-seperated list of languages/locales enabled.
		$enabled = array_map('trim', explode('|', $enabled));

		// Did the user request a specific language?
		$requested = $request->getParameter('lang');
		if($requested){
			$showStrings = false;
			$showForm = true;
		}
		else{
			$showStrings = true;
			$showForm = false;
			$requested = I18NLoader::GetUsersLanguage();
		}

		$strings = I18NLoader::GetAllStrings($requested);

		$form = new Form();
		$form->set('callsmethod', 'AdminController::_i18nSaveHandler');
		
		$form->addElement(
			'checkboxes', 
			[
				'name' => 'languages[]',
				'title' => t('STRING_CONFIG_CORE_LANGUAGE_LANGUAGES_ENABLED'),
				'description' => t('MESSAGE_CONFIG_CORE_LANGUAGE_LANGUAGES_ENABLED'),
			    'options' => $all,
			    'value' => $enabled,
			]
		);

		$form->addElement('submit', ['value' => t('STRING_SAVE')]);

		$view->addBreadcrumb('t:STRING_ADMIN', '/admin');
		$view->title = 't:STRING_I18N_LANGUAGES';
		//$view->assign('languages', $languages);
		$view->assign('form', $form);
		$view->assign('strings', $strings);
		$view->assign('show_strings', $showStrings);
		$view->assign('show_form', $showForm);
		$view->assign('requested', $requested);
	}

	/**
	 * Configure several of the SEO-based options on Core.
	 * 
	 * @core-access-string g:admin
	 */
	public function seo_config(){
		$view = $this->getView();

		$keys = [
			'/core/page/title_remove_stop_words',
			'/core/page/url_remove_stop_words',
			'/core/page/title_template',
			'/core/page/indexable',
		];

		$form = new Form();
		$form->set('callsmethod', 'AdminController::_ConfigSubmit');

		foreach($keys as $k){
			$c = ConfigHandler::GetConfig($k);
			$f = $c->asFormElement();
			// Don't need them grouped
			$f->set('group', '');
			$form->addElement($f);
		}
		$form->addElement('submit', ['value' => 'Save Options']);

		$view->title = 'SEO Options';
		$view->assign('form', $form);
	}

	/**
	 * Configure several of the performance-based options on Core.
	 * 
	 * @core-access-string g:admin
	 */
	public function performance_config(){
		$view = $this->getView();

		$keys = [
			'/core/javascript/minified',
			'/core/markup/minified',
			//'/core/filestore/assetversion',
			'/core/assetversion/proxyfriendly',
			'/core/performance/anonymous_user_page_cache',
		];

		$form = new Form();
		$form->set('callsmethod', 'AdminController::_ConfigSubmit');

		foreach($keys as $k){
			$c = ConfigHandler::GetConfig($k);
			$f = $c->asFormElement();
			// Don't need them grouped
			$f->set('group', '');
			$form->addElement($f);
		}

		$form->addElement('submit', ['value' => 'Save Options']);

		$view->title = 'Performance Options';
		$view->assign('form', $form);
	}

	/**
	 * Landing page to export or export the site, database, and necessary files
	 *
	 * @core-access-string g:admin
	 */
	public function importexport() {
		$view = $this->getView();

		$view->mastertemplate = 'admin';
		$view->title = 't:STRING_IMPORT_EXPORT_MANAGER';
	}

	/**
	 * Export the site content and save the file for transferring to another server
	 *
	 * @core-access-string g:admin
	 */
	public function export() {
		$view        = $this->getView();
		$models      = [];
		$modelGroups = [];

		foreach(\Core\Core::GetComponents() as $component) {
			$models = array_merge($models, $component->getModelList());
		}

		foreach(array_keys($models) as $modelName) {
			$ref = new ReflectionClass($modelName);
			$group = $ref->getStaticPropertyValue('ExportGroup', null);
			$name  = $ref->getMethod('GetTableName')->invoke(null);

			if($group) {
				if(!isset($modelGroups[ $group ])) {
					$modelGroups[ $group ] = [];
				}
				$modelGroups[ $group ][] = $name;
			}
		}

		$form = new Form();
		$form->set('action', '/admin/export/submit');

		$fileGroup = new FormGroup([
			'title' => 't:STRING_EXPORT_FILES',
		]);

		$dbGroup = new FormGroup([
			'title' => 't:STRING_EXPORT_DATABASE',
		]);

		$fileGroup->addElement(
			'checkbox',
			[
				'name' => 'export[files][public]',
				'title' => 'File Uploads',
				'description' => 'All "public" files that have been uploaded for content, users, etc.',
			]
		);

		foreach($modelGroups as $group => $models) {

			$list = implode(', ', $models);


			$dbGroup->addElement(
				'checkbox',
				[
					'name' => 'export[database][' . $group . ']',
					'title' => $group,
					'description' => 'Export database content, includes tables ' . $list,
				]
			);
		}

		$form->addElement($fileGroup);
		$form->addElement($dbGroup);

		$form->addElement('submit', ['value' => 't:STRING_EXPORT']);

		$view->mastertemplate = 'admin';
		$view->addBreadcrumb('t:STRING_IMPORT_EXPORT_MANAGER', '/admin/importexport');
		$view->title = 't:STRING_EXPORT';
		$view->assign('form', $form);
	}

	/**
	 * Actual handler for the export, required because we do NOT want to use form tokens here.
	 *
	 * @core-access-string g:admin
	 * @core-request-method POST
	 */
	public function export_submit() {
		$view = $this->getView();
		$request = $this->getRequest();
		$models  = [];
		$exporter = new ArchiveExporter();

		foreach(\Core\Core::GetComponents() as $component) {
			$models = array_merge($models, $component->getModelList());
		}

		foreach(array_keys($models) as $modelName) {
			$ref = new ReflectionClass($modelName);
			$group = $ref->getStaticPropertyValue('ExportGroup', null);
			$name  = $ref->getMethod('GetTableName')->invoke(null);

			if($group) {
				if($request->getPost('export[database][' . $group  . ']')) {
					$exporter->exportTable($name);
				}
			}
		}


		if($request->getPost('export[files][public]')) {
			$exporter->exportPublicFiles();
		}

		$archive = $exporter->process();

		$view->mode = View::MODE_NOOUTPUT;

		// Wrap up and send the data to the user agent.
		$archive->sendToUserAgent(true);
	}

	/**
	 * Import the site content from a zip archive
	 *
	 * @core-access-string g:admin
	 */
	public function import() {
		$view        = $this->getView();


		$form = new Form();
		$form->set('callsmethod', 'AdminController::ImportSave');

		$form->addElement(
			'file',
			[
				'name'     => 'importfile',
				'required' => true,
				'title'    => 't:STRING_ARCHIVE_FILE',
				'basedir'  => 'tmp/import',
			]
		);

		$form->addElement('submit', ['value' => 't:STRING_IMPORT']);


		$view->mastertemplate = 'admin';
		$view->addBreadcrumb('t:STRING_IMPORT_EXPORT_MANAGER', '/admin/importexport');
		$view->title = 't:STRING_IMPORT';
		$view->assign('form', $form);
	}

	/**
	 * @core-access-string g:admin
	 * 
	 * @return type
	 */
	public function email_config(){
		$view = $this->getView();

		$keys = [
			'/core/email/enable_sending',
			'/core/email/from',
			'/core/email/from_name',
			'/core/email/dynamic_sender',
			'/core/email/sandbox_to',
		];

		$form = new Form();
		$form->set('callsmethod', 'AdminController::_ConfigSubmit');

		foreach($keys as $k){
			$c = ConfigHandler::GetConfig($k);
			$f = $c->asFormElement();
			// Don't need them grouped
			$f->set('group', '');
			$form->addElement($f);
		}
		$form->addElement('submit', ['value' => t('STRING_SAVE')]);
		
		$backends = Email::GetBackends();
		$backend = ConfigHandler::Get('/core/email/mailer');
		if(isset($backends[$backend])){
			\Core\set_message('t:MESSAGE_INFO_EMAIL_CURRENT_BACKEND_IS_S', $backends[$backend]);
			$good = true;
		}
		elseif(!count($backends)){
			\Core\set_message('t:MESSAGE_ERROR_EMAIL_NO_BACKENDS_INSTALLED');
			$good = false;
		}
		else{
			\Core\set_message('t:MESSAGE_ERROR_EMAIL_NO_BACKEND_CONFIGURED');
			$good = false;
		}

		$view->title = 'Email Options &amp; Diagnostics';
		$view->assign('form', $form);
		$view->assign('email_enabled', ConfigHandler::Get('/core/email/enable_sending'));
		$view->assign('good', $good);
	}

	/**
	 * Test the email delivery ability
	 * 
	 * @core-access-string g:admin
	 * @core-request-method POST
	 *
	 * @throws BadRequestException
	 */
	public function email_test(){
		$request = $this->getRequest();
		$view = $this->getView();

		if(!$request->getPost('email')){
			throw new BadRequestException('Email is required');
		}

		$view->mode = View::MODE_NOOUTPUT;
		$view->contenttype = View::CTYPE_HTML;
		//$view->render();

		$dest         = $request->getPost('email');
		$method       = ConfigHandler::Get('/core/email/mailer');
		/*
		$smtpHost     = ConfigHandler::Get('/core/email/smtp_host');
		$smtpUser     = ConfigHandler::Get('/core/email/smtp_user');
		$smtpPass     = ConfigHandler::Get('/core/email/smtp_password');
		$smtpPort     = ConfigHandler::Get('/core/email/smtp_port');
		$smtpSec      = ConfigHandler::Get('/core/email/smtp_security');
		$sendmailPath = ConfigHandler::Get('/core/email/sendmail_path');*/
		$emailDebug   = [];

		//$emailDebug[] = 'Sending Method: ' . $method;

		/*switch($method){
			case 'smtp':
				$emailDebug[] = 'SMTP Host: ' . $smtpHost . ($smtpPort ? ':' . $smtpPort : '');
				$emailDebug[] = 'SMTP User/Pass: ' . ($smtpUser ? $smtpUser . '//' . ($smtpPass ? '*** saved ***' : 'NO PASS') : 'Anonymous');
				$emailDebug[] = 'SMTP Security: ' . $smtpSec;
				break;
			case 'sendmail':
				$emailDebug[] = 'Sendmail Path: ' . $sendmailPath;
				break;
		}*/

		CLI::PrintHeader('Sending test email to ' . $dest);

		CLI::PrintActionStart('Initializing Email System');
		try{
			$email = Email::Factory('/emails/admin/test_email.tpl');
			$email->setTo($dest);
			$email->setSubject('Test Email');
			$email->enableDebug();
			//$email->assign('debugs', $emailDebug);

			CLI::PrintActionStatus(true);
		}
		catch(Exception $e){
			\Core\log_exception($e);
			CLI::PrintActionStatus(false);
			CLI::PrintError($e->getMessage());
			CLI::PrintLine($e->getTrace());

			return;
		}

		CLI::PrintActionStart('Sending Email via ' . $method);
		try{
			$email->send();

			CLI::PrintActionStatus(true);
		}
		catch(Exception $e){
			\Core\log_exception($e);
			CLI::PrintActionStatus(false);
			CLI::PrintError($e->getMessage());
			CLI::PrintLine(explode("\n", $e->getTraceAsString()));
		}

		CLI::PrintHeader('Sent Data:');
		CLI::PrintLine(explode("\n", $email->getFullEML()));
	}

	public static function _WidgetCreateUpdateHandler(Form $form){
		$baseurl = $form->getElement('baseurl')->get('value');

		$model = new WidgetModel($baseurl);
		$model->set('editurl', '/admin/widget/update?baseurl=' . $baseurl);
		$model->set('deleteurl', '/admin/widget/delete?baseurl=' . $baseurl);
		$model->set('title', $form->getElement('title')->get('value'));
		if($form->getElement('template')){
			$model->set('template', $form->getElementValue('template'));
		}

		$elements = $form->getElements();
		foreach($elements as $el){
			/** @var FormElement $el */
			if(strpos($el->get('name'), 'setting[') === 0){
				$name = substr($el->get('name'), 8, -1);
				$model->setSetting($name, $el->get('value'));
			}
		}
		$model->save();

		return 'back';
	}

	public static function _ConfigSubmit(Form $form) {
		$elements = $form->getElements();

		$updatedcount = 0;

		foreach ($elements as $e) {
			/** @var FormElement $e */
			
			// I'm only interested in config options.
			if (strpos($e->get('name'), 'config[') === false) continue;

			// Make the name usable a little.
			$n = $e->get('name');
			if (($pos = strpos($n, '[]')) !== false) $n = substr($n, 0, $pos);
			$n = substr($n, 7, -1);

			// And get the config object
			$c = new ConfigModel($n);
			$val = null;

			switch ($c->get('type')) {
				case 'string':
				case 'text':
				case 'enum':
				case 'boolean':
				case 'int':
					$val = $e->get('value');
					break;
				case 'set':
					$val = implode('|', $e->get('value'));
					break;
				default:
					throw new Exception('Unsupported configuration type for ' . $c->get('key') . ', [' . $c->get('type') . ']');
					break;
			}

			// This is required because enterprise multisite mode has a different location for site configs.
			if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::GetCurrentSiteID()){
				$siteconfig = MultiSiteConfigModel::Construct($c->get('key'), MultiSiteHelper::GetCurrentSiteID());
				$siteconfig->setValue($val);
				if($siteconfig->save()) ++$updatedcount;
			}
			else{
				$c->setValue($val);
				if ($c->save()) ++$updatedcount;
			}

		}

		\Core\set_message('t:MESSAGE_SUCCESS_UPDATED_N_CONFIGURATION', $updatedcount);

		return true;
	}

	/**
	 * The save handler for /admin/pages quick edit.
	 *
	 * @param Form $form
	 *
	 * @return bool
	 */
	public static function PagesSave(Form $form) {
		$models = [];

		foreach($form->getElements() as $el){
			/** @var FormElement $el */
			$n = $el->get('name');

			// i only want model
			if(strpos($n, 'model[') !== 0){
				continue;
			}

			$baseurl = substr($n, 6, strpos($n, ']')-6);
			$n = substr($n, strpos($n, ']')+1);

			// Is this a meta attribute?
			if(strpos($n, '[meta][') === 0){
				$ismeta = true;
				$n = substr($n, 7, -1);
			}
			else{
				$ismeta = false;
				$n = substr($n, 1, -1);
			}

			// Make sure the model is available.
			if(!isset($models[$baseurl])){
				$models[$baseurl] = PageModel::Construct($baseurl);
			}
			/** @var PageModel $p */
			$p = $models[$baseurl];

			if($ismeta){
				$p->setMeta($n, $el->get('value'));
			}
			else{
				$p->set($n, $el->get('value'));
			}
		}

		foreach($models as $p){
			/** @var PageModel $p */
			$p->save();
		}

		return true;
	}

	public static function _i18nSaveHandler(Form $form) {
		
		// NEW IDEA!
		// Instead of setting the override for keys, (possibly useful, just somewhere else)...
		// Set the enabled languages for this site.
		// This allows site administrators to NOT have every language under the sun appear if they're running SuSE.
		$selected = $form->getElement('languages[]')->get('value');
		
		// Implode them into a single string.
		$enabled = implode('|', $selected);
		// Strip out any invalid character.
		$enabled = preg_replace('/[^a-zA-Z_|]/', '', $enabled);
		
		// And save!
		ConfigHandler::Set('/core/language/languages_enabled', $enabled);
		return true;
		
		// Create a custom ini for just these options.
		// This will allow the site admin to change a string without worrying about it getting overridden from an update.

		$lang = $form->getElementValue('lang');
		$ini = "[$lang]\n; Custom locale strings set by the site manager!\n\n";

		foreach($form->getElements() as $el){
			/** @var FormElement $el */

			$name = $el->get('name');
			$val  = $el->get('value');

			if(strpos($name, 'MESSAGE') === 0 || strpos($name, 'FORMAT') === 0 || strpos($name, 'STRING') === 0){
				$ini .= $name . ' = "' . str_replace('"', '\\"', $val) . '";' . "\n";
			}
		}

		// Save this ini out to a custom i18n file.
		$fileout = Factory::File(ROOT_PDIR . 'themes/custom/i18n/' . $lang . '.ini');
		$fileout->putContents($ini);

		\Core\set_message('t:MESSAGE_SUCCESS_UPDATED_TRANSLATION_STRINGS');
		return true;
	}

	/**
	 * Call to check some of the core requirements on Core, such as file permissions and the like.
	 *
	 * @return array
	 */
	public static function _HealthCheckHook(){
		
		$autos = InstallHelper::GetChecks();
		
		// Add checks for the web server itself, (since we're running from the CLI, that may not be a given)
		//$autos[] = new Checks\ApacheServerCheck();
		
		// Append the dynamic checks based off of configuration variables.
		$autos[] = new Checks\AssetDirectoryCheck();
		$autos[] = new Checks\PublicDirectoryCheck();
		$autos[] = new Checks\PrivateDirectoryCheck();


		$checks      = [];
		
		foreach($autos as $check){
			if($check->test()){
				$checks[] = HealthCheckResult::ConstructGood($check->getTitle(), $check->getDetails());
			}
			else{
				$checks[] = HealthCheckResult::ConstructWarn($check->getTitle(), $check->getDetails(), '/admin/autofix');
			}
		}

		if(defined('SERVER_ID') && strlen(SERVER_ID) == 32){
			$checks[] = HealthCheckResult::ConstructGood(
				t('STRING_CHECK_SERVER_ID_VALID'),
				t('MESSAGE_SUCCESS_CHECK_SERVER_ID_IS_S', wordwrap(SERVER_ID, 4, '-', true))
			);
		}
		else{
			$checks[] = HealthCheckResult::ConstructWarn(
				t('STRING_CHECK_SERVER_ID_NOT_VALID'),
				t('MESSAGE_ERROR_CHECK_SERVER_ID_NOT_VALID'),
				'/admin/serverid'
			);
		}
		
		foreach(Core::GetComponents() as $c) {
			/** @var Component $c */

			$rc = $c->getRequires();

			foreach($rc as $result) {
				try{
					$result->check();
					
					$checks[] = HealthCheckResult::ConstructGood($c->getName() . ': ' . $result->getLabel(), '');
				}
				catch (Exception $ex) {
					$checks[] = HealthCheckResult::ConstructWarn($c->getName() . ': ' . $result->getLabel(), $ex->getMessage(), '');
				}
			}

			// Check this component's license data as well by performing an actual query against the licensing server.
			try{
				$c->queryLicenser();
			}
			catch (Exception $ex) {
				$checks[] = HealthCheckResult::ConstructError($ex->getMessage(), null, null);
			}
			
			
			$licenseCheck = $c->getLicenseData();
			if(sizeof($licenseCheck)){
				$check = new HealthCheckResult();
				if($licenseCheck['status']){
					$check->result = HealthCheckResult::RESULT_GOOD;
					$check->title = $c->getName() . ' has a valid license until ' . DateTime::FormatString($licenseCheck['expires'], DateTime::SHORTDATE);
					$check->description = '';
					foreach($licenseCheck['features'] as $k => $v){
						$check->description .= $k . ': ' . $v . '<br/>';
					}
				}
				else{
					$check->result = HealthCheckResult::RESULT_ERROR;
					$check->title = $c->getName() . ' ' . $licenseCheck['message'];
				}

				$checks[] = $check;
			}
		}

		return $checks;
	}

	/**
	 * The weekly health report to email to admins.
	 * 
	 * Will only send an email if there was an issue found with the site.
	 * Otherwise, this will be used by the upstream maintainers to know what versions clients are connecting with.
	 */
	public static function _HealthCheckReport(){
		$checks = HookHandler::DispatchHook('/core/healthcheck');
		$toReport = [];
		
		foreach($checks as $check){
			/** @var HealthCheckResult $check */
			if($check->result != HealthCheckResult::RESULT_GOOD){
				$toReport[] = $check;
			}
		}
		
		if(sizeof($toReport) == 0){
			// YAY!
			return true;
		}
		
		if(!defined('SERVER_ADMIN_EMAIL') || SERVER_ADMIN_EMAIL == ''){
			echo 'Health issues found but unable to send an email, please set SERVER_ADMIN_EMAIL in your configuration.xml!';
			return false;
		}
		
		$email = Email::Factory('/emails/admin/health_report.tpl');
		if(strpos(SERVER_ADMIN_EMAIL, ',') !== false){
			$emails = explode(',', SERVER_ADMIN_EMAIL);
			foreach($emails as $e){
				$e = trim($e);
				if($e){
					$email->addTo($e);
				}
			}
		}
		else{
			$email->addTo(SERVER_ADMIN_EMAIL);	
		}
		$email->setSubject('Site Health Report');
		$email->assign('checks', $toReport);
		
		try{
			if($email->send()){
				echo 'Sent health issues to the server admin successfully.';
				return true;
			}
			else{
				echo 'Unable to send health issues to server admin, please check your email settings.';
				return false;
			}
		}
		catch(Exception $e){
			echo 'Unable to send health issues to server admin, please check your email settings.';
			return false;
		}
	}

	public static function ImportSave(Form $form) {
		/** @var \Core\Forms\FileInput $el */
		$el      = $form->getElement('importfile');
		$file    = $el->getFile();

		$importer = new ArchiveImporter();
		$importer->process($file);

		\Core\set_message('Imported archive successfully.  Please note, you may need to reinstall in order to see all changes.', 'success');
		return true;
	}
}


