<?php
/**
 * File for class DatasetAuthController definition in the coreplus project
 *
 * @package Core\User
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20140320.1636
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\AccessDeniedException;
use Core\ConfigHandler;
use Core\Controller;
use Core\Date\DateTime;
use Core\Date\Timezone;
use Core\EmailAPI\Email;
use Core\Forms\Form;
use Core\Forms\FormElement;
use Core\GeneralValidationException;
use Core\HookHandler;
use Core\ModelValidationException;
use Core\NotFoundException;
use Core\RedirectException;
use Core\Session;
use Core\User;
use Core\User\Helper;
use Core\View;
use Core\User\AuthDrivers\datastore;
use Core\Logger\LogEntry;
use Core\Utilities\PasswordGenerator;
use Core\ViewControl;


/**
 * A short teaser of what DatasetAuthController does.
 *
 * More lengthy description of what DatasetAuthController does and why it's fantastic.
 *
 * <h3>Usage Examples</h3>
 *
 *
 * @todo Write documentation for DatasetAuthController
 * <h4>Example 1</h4>
 * <p>Description 1</p>
 * <code>
 * // Some code for example 1
 * $a = $b;
 * </code>
 *
 *
 * <h4>Example 2</h4>
 * <p>Description 2</p>
 * <code>
 * // Some code for example 2
 * $b = $a;
 * </code>
 *
 * @package Core\User
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
class DatastoreAuthController extends Controller {

	/**
	 * View to set the user's password, both administratively and from the user's profile.
	 *
	 */
	public function password(){

		$view        = $this->getView();
		$req         = $this->getRequest();
		$userid      = $req->getParameter(0);
		$currentUser = \Core\user();
		$isManager   = $currentUser->checkAccess('p:/user/users/manage');
		$allowPublic = ConfigHandler::Get('/user/register/allowpublic');


		// Default to current user.
		if($userid === null){
			$userid = $currentUser->id;
		}


		// Only allow this if the user is either the same user or has the user manage permission.
		if($userid !== $currentUser->id && !$isManager) {
			throw new AccessDeniedException();
		}


		$user = UserModel::Construct($userid);
		if(!$user->checkExists()) {
			throw new NotFoundException();
		}


		try {
			$auth = $user->getAuthDriver('datastore');
		}
		catch(Exception $ex) {
			if(!$allowPublic && !$isManager) {
				// Password logins are not enabled and public register is disabled,
				// do not allow the user to switch auth backends in this case!
				throw new AccessDeniedException('Password logins are not enabled on your account, please contact a site administrator for more details.');
			}
			// If this user doesn't have password logins enabled, but public registration is enabled, just allow it to continue.
			$auth = new \Core\User\AuthDrivers\datastore($user);
		}


		$form = new Form();
		$form->set('callsmethod', 'DatastoreAuthController::PasswordSave');
		$form->addElement('system', ['name' => 'user', 'value' => $user->id]);

		if($isManager){
			$form->addElement(
				'checkbox', [
					'name'        => 'pwgen',
					'value'       => '1',
					'checked'     => false,
					'title'       => t('STRING_GENERATE_SECURE_PASSWORD'),
					'description' => t('MESSAGE_GENERATE_SECURE_PASSWORD'),
				]
			);
		}
		
		$form->addElement(
			'password', 
			[
				'name' => 'p1',
				'title' => t('STRING_PASSWORD'),
				'required' => (!$isManager)
			]
		);
		$form->addElement(
			'password', 
			[
				'name' => 'p2',
				'title' => t('STRING_CONFIRM_PASSWORD'),
				'required' => (!$isManager)
			]
		);

		$form->addElement('submit', array('value' => t('STRING_SAVE')));

		$complexity = $auth->getPasswordComplexityAsHTML();
		if($complexity){
			\Core\set_message($complexity, 'tutorial');
		}

		UserController::SetBreadcrumbsUserPage($view, $user);
		$view->title = 't:STRING_CHANGE_PASSWORD';
		$view->assign('form', $form);
		$view->assign('is_manager', $isManager);
	}

	/**
	 * Reset a user's password, generally when it's forgotten.
	 *
	 * @throws Exception
	 */
	public function password_reset() {
		$view = $this->getView();
		$request = $this->getRequest();


		// Create a simple form to render.  This is better than doing it in the template.
		$form = new Form();
		$form->set('callsmethod', 'DatastoreAuthController::PasswordResetRequest');
		$form->addElement('text', ['name' => 'email', 'title' => 't:STRING_MODEL_USERMODEL_EMAIL', 'required' => true]);
		$form->addElement('submit', ['name' => 'submit', 'value' => 't:STRING_USER_SEND_PASSWORD_RESET_INSTRUCTIONS']);

		$view->title = 't:STRING_USER_RESET_PASSWORD';
		$view->addMetaName('robots', 'noindex');
		$view->assign('form', $form);
	}

	/**
	 * Provides confirmation for resetting a password and does all necessary checks for nonce validation
	 *
	 * @throws RedirectException
	 */
	public function password_confirm() {
		$view    = $this->getView();
		$request = $this->getRequest();
		$n       = $request->getParameter(0);
		$nonce   = NonceModel::Construct($n);

		if(!$nonce->isValid()){
			\Core\log()
				->type('security')
				->code('/user/password/reset/failed')
				->message('Password reset confirmation had an invalid nonce')
				->send();

			\Core\set_message('Invalid key provided, please double check the link and try again.', 'error');
			throw new RedirectException('/');
		}

		$nonce->decryptData();
		$data = $nonce->get('data');

		if($data['type'] !== 'password-reset') {
			\Core\log()
				->type('security')
				->code('/user/password/reset/failed')
				->message('Password reset confirmation had an invalid nonce')
				->send();

			throw new NotFoundException('Invalid key provided, please double check the link and try again.');
		}

		$user = UserModel::Construct($data['user']);
		if(!$user->checkExists()) {
			\Core\log()
				->type('security')
				->code('/user/password/reset/failed')
				->message('Password reset confirmation had an invalid user')
				->send();

			throw new NotFoundException('Invalid user account, please contact the site administrator.');
		}


		$genericauth = new \Core\User\AuthDrivers\datastore();


		$form = new Form();
		$form->set('callsmethod', 'DatastoreAuthController::PasswordSave');
		$form->addElement('system', ['name' => 'user', 'value' => $user->id]);
		$form->addElement('system', ['name' => 'nonce', 'value' => $nonce->key]);
		$form->addElement('password', ['name' => 'p1', 'title' => 't:STRING_PASSWORD', 'required' => true]);
		$form->addElement('password', ['name' => 'p2', 'title' => 't:STRING_CONFIRM_PASSWORD', 'required' => true]);
		$form->addElement('submit', ['name' => 'submit', 'value' => 't:STRING_SAVE']);


		$view->title = 't:STRING_USER_RESET_PASSWORD';
		$view->addMetaName('robots', 'noindex');
		$view->assign('form', $form);
		$view->assign('requirements', $genericauth->getPasswordComplexityAsHTML());
	}

	/**
	 * Front-end view to allow users to reset their password.
	 */
	public function forgotPassword(){

		// This logic is no longer used as of Core 7.0
		throw new RedirectException('/datastoreauth/password/reset');

		$request = $this->getRequest();

		// If e and k are set as parameters... it's on step 2.
		if($request->getParameter(0)){
			return $this->_forgotPassword2();
		}
		// Else, just step 1.
		else{
			return $this->_forgotPassword1();
		}
	}

	/**
	 * POST-only view to disable a user's password login ability.
	 * @return int
	 */
	public function disable(){
		$view    = $this->getView();
		$request = $this->getRequest();
		$userid  = $request->getParameter(0);
		$isadmin = \Core\user()->checkAccess('p:/user/users/manage'); // Current user an admin?
		$isself  = (\Core\user()->get('id') == $userid);

		if(!($isadmin || $isself)){
			return View::STATUS_ACCESSDENIED;
		}

		if(!$request->isPost()){
			return View::STATUS_BADREQUEST;
		}

		/** @var UserModel $user */
		$user = UserModel::Construct($userid);

		if(!$user->exists()){
			return View::STATUS_NOTFOUND;
		}

		if(sizeof($user->getEnabledAuthDrivers()) == 1){
			return View::STATUS_OTHER;
		}

		$user->disableAuthDriver('datastore');
		$user->save();

		\Core\set_message('t:MESSAGE_SUCCESS_DISABLED_PASSWORD_AUTH');
		\Core\go_back();
	}

	private function _forgotPassword1(){
		$view = $this->getView();
		$request = $this->getRequest();


		// Create a simple form to render.  This is better than doing it in the template.
		$form = new Form();
		$form->set('method', 'POST');

		if(\Core\user()->exists()){
			// This may happen with the enable-password feature for facebook accounts.
			// They shouldn't have the option to change the email, but it should display where the information will go to.
			$form->addElement('system', ['name' => 'email', 'value' => \Core\user()->get('email')]);
			$current = \Core\user()->get('email');
		}
		else{
			$form->addElement('text', ['name' => 'email', 'title' => 'Email', 'required' => true]);
			$current = false;
		}

		$form->addElement('submit', ['name' => 'submit', 'value' => 'Send Reset Instructions']);

		$view->title = 'Forgot Password';
		// This is step 1
		$view->assign('step', 1);
		$view->assign('form', $form);
		$view->assign('current', $current);
		$view->assign('can_change_email', ConfigHandler::Get('/user/email/allowchanging'));
		// Google has no business indexing user-action pages.
		$view->addMetaName('robots', 'noindex');

		// There's really nothing to do here except for check the email and send it.

		if($request->isPost()){

			if(\Core\user()->exists()){
				/** @var UserModel $u */
				$u = \Core\user();
			}
			else{
				/** @var UserModel $u */
				$u = UserModel::Find(array('email' => $_POST['email']), 1);
			}

			if(!$u){
				\Core\set_message('t:MESSAGE_ERROR_USER_LOGIN_EMAIL_NOT_FOUND');
				SystemLogModel::LogSecurityEvent('/user/forgotpassword/send', 'Failed Forgot Password. Invalid email requested for reset: [' . $_POST['email'] . ']');
				return;
			}

			// Use the Nonce system to generate a one-time key with this user's data.
			$nonce = NonceModel::Generate(
				'20 minutes',
				null,
				[
					'type' => 'password-reset',
					'user' => $u->get('id'),
				]
			);

			$link = '/datastoreauth/forgotpassword/' . $nonce;

			$e = Email::Factory('/emails/user/datastoreauth_forgotpassword.tpl');
			$e->setSubject('Forgot Password Request');
			$e->setTo($u->get('email'));
			$e->assign('link', \Core\resolve_link($link));
			$e->assign('ip', REMOTE_IP);
			$e->encrypted = true;
			try{
				$e->send();
				SystemLogModel::LogSecurityEvent('/user/forgotpassword/send', 'Forgot password request sent successfully', null, $u->get('id'));
			}
			catch(Exception $e){
				\Core\set_message('Error sending the email, ' . $e->getMessage(), 'error');
				SystemLogModel::LogErrorEvent('/user/forgotpassword/send', $e->getMessage());
				return;
			}

			// Otherwise, it must have sent, (hopefully)...
			\Core\set_message('t:MESSAGE_SUCCESS_PLEASE_CHECK_EMAIL_FOR_PASSWORD_RESET_INSTRUCTIONS');
			\core\redirect('/');
		}
	}

	private function _forgotPassword2(){
		$view = $this->getView();
		$request = $this->getRequest();

		$genericauth = new \Core\User\AuthDrivers\datastore();

		// Create a simple form to render.  This is better than doing it in the template.
		$form = new Form();
		$form->set('method', 'POST');
		$form->addElement('password', ['name' => 'p1', 'title' => 'Password', 'required' => true]);
		$form->addElement('password', ['name' => 'p2', 'title' => 'Confirm', 'required' => true]);
		$form->addElement('submit', ['name' => 'submit', 'value' => 'Set New Password']);

		$view->title = 'Forgot Password';
		$view->assign('step', 2);
		$view->assign('form', $form);
		$view->assign('requirements', $genericauth->getPasswordComplexityAsHTML());

		$n = $request->getParameter(0);

		/** @var $nonce NonceModel */
		$nonce = NonceModel::Construct($n);

		if(!$nonce->isValid()){
			SystemLogModel::LogSecurityEvent('/user/forgotpassword/confirm', 'Failed Forgot Password. Invalid nonce requested: [' . $n . ']');
			\Core\set_message('t:MESSAGE_ERROR_USER_LOGIN_EMAIL_NOT_FOUND');
			\core\redirect('/');
			return;
		}

		$nonce->decryptData();
		$data = $nonce->get('data');

		/** @var UserModel $u */
		$u = UserModel::Construct($data['user']);
		if(!$u){
			SystemLogModel::LogSecurityEvent('/user/forgotpassword/confirm', 'Failed Forgot Password. Invalid user account requested: [' . $data['user'] . ']');
			\Core\set_message('t:MESSAGE_ERROR_USER_LOGIN_EMAIL_NOT_FOUND');
			\core\redirect('/');
			return;
		}


		if($request->isPost()){
			// Validate the password.
			if($_POST['p1'] != $_POST['p2']){
				\Core\set_message('t:MESSAGE_ERROR_USER_REGISTER_PASSWORD_MISMATCH');
				return;
			}

			// Else, try to set it... the user model will complain if it's invalid.
			try{
				$u->enableAuthDriver('datastore');
				/** @var \Core\User\AuthDrivers\datastore $auth */
				$auth = $u->getAuthDriver('datastore');

				$auth->setPassword($_POST['p1']);
				$u->save();
				// NOW I can invalidate that nonce!
				$nonce->markUsed();
				SystemLogModel::LogSecurityEvent('/user/forgotpassword/confirm', 'Reset password successfully!', null, $u->get('id'));
				\Core\set_message('Reset password successfully', 'success');
				if($u->get('active')){
					Session::SetUser($u);
				}
				\core\redirect('/');
			}
			catch(ModelValidationException $e){
				SystemLogModel::LogSecurityEvent('/user/forgotpassword/confirm', 'Failed Forgot Password. ' . $e->getMessage(), null, $u->get('id'));
				\Core\set_message($e->getMessage(), 'error');
				return;
			}
			catch(Exception $e){
				SystemLogModel::LogSecurityEvent('/user/forgotpassword/confirm', 'Failed Forgot Password. ' . $e->getMessage(), null, $u->get('id'));
				\Core\set_message((DEVELOPMENT_MODE ? $e->getMessage() : 'An unknown error occured'), 'error');
				return;
			}
		}
	}

	/**
	 * This is the handler the standard registration login form.
	 *
	 * Since datastore is a basic auth source, supplemental fields and everything are included in one form.
	 *
	 * @param Form $form
	 *
	 * @return bool|string
	 */
	public static function RegisterHandler(Form $form) {
		$p1 = $form->getElement('pass');
		$p2 = $form->getElement('pass2');
				
		$user = new UserModel();
		
		// Set the user's requested email; will throw a ModelValidationException if it fails.
		$user->set('email', $form->getElementValue('email'));
		
		// This is the datastore controller, so enable that auth driver.
		$user->enableAuthDriver('datastore');
		/** @var datastore $auth */
		$auth = $user->getAuthDriver('datastore');
		
		if($form->getElement('pwgen') && $form->getElementValue('pwgen')){
			// Admins have the ability to generate a random password for the user.
			$password = PasswordGenerator::QuickGen();
			$auth->setPassword($password);
			
			// This is required to inform the user what their new dynamic password is via email.
			$user->set('password_raw', $password);
		}
		else{
			// Users are required to set their password initially.
			if($p1->get('value') !== $p2->get('value')){
				throw new ModelValidationException('t:MESSAGE_ERROR_USER_REGISTER_PASSWORD_MISMATCH');
			}
			
			$auth->setPassword($p1->get('value'));
		}


		// Load up all the additional fields.
		foreach($form->getElementsByName('user\[.*\]') as $el) {
			/** @var \Core\Forms\FormElement $el */
			$key = substr(substr($el->get('name'), 5, -1));
			$user->set($key, $el->getValue());
		}

		if(! \Core\user()->checkAccess('g:admin') ) {
			$user->setDefaultActiveStatuses();
			$user->setDefaultGroups();
		}

		$user->setDefaultMetaFields();
		$user->generateNewApiKey();
		$user->save();

		\Core\log()
			->type('security')
			->code('/user/register')
			->message('User registration successful')
			->user($user)
			->send();

		if(!\Core\user()->checkExists() && $user->active === '1') {
			$user->login();
		}

		if($user->active === '1') {
			// Send a thank you for registering email to the user.
			try{
				$user->sendWelcomeEmail();
			}
			catch(Exception $e){
				\Core\ErrorManagement\exception_handler($e);
				\Core\set_message('t:MESSAGE_ERROR_CANNOT_SEND_WELCOME_EMAIL');
			}
		}

		if($form->getElementValue('redirect')) {
			// The preferred default redirect method.
			$url = $form->getElementValue('redirect');
		}
		elseif(($overrideurl = HookHandler::DispatchHook('/user/postlogin/getredirecturl'))) {
			// Allow an external script to override the redirecting URL.
			$url = $overrideurl;
		}
		elseif(strpos(REL_REQUEST_PATH, '/user/register') === 0) {
			// If the user came from the registration page, get the page before that.
			$url = '/';
		}
		else{
			// else the registration link is now on the same page as the 403 handler.
			$url = REL_REQUEST_PATH;
		}

		return $url;
	}

	/**
	 * Form Handler for logging in.
	 *
	 * @static
	 *
	 * @param Form $form
	 *
	 * @return bool|null|string
	 */
	public static function LoginHandler(Form $form) {
		/** @var FormElement $e */
		$e = $form->getElement('email');
		/** @var FormElement $p */
		$p = $form->getElement('pass');


		/** @var UserModel $u */
		$u = UserModel::Find(array('email' => $e->get('value')), 1);

		if(!$u) {
			LogEntry::Init()
				->type('security')
				->level(LOG_LEVEL_HIGH)
				->code('/user/login')
				->tag('email', $e->get('value'))
				->message('Login via password failed; email not registered')
				->send();
			$e->setError('t:MESSAGE_ERROR_USER_LOGIN_FAILED');
			HookHandler::DispatchHook('/user/login/failed', null);
			return false;
		}

		if($u->get('active') === '-1') {
			// Account was deactivated
			LogEntry::Init()
				->type('security')
				->level(LOG_LEVEL_HIGH)
				->code('/user/login')
				->user($u)
				->message('Login via password failed; account deactivated')
				->send();
			$e->setError('t:MESSAGE_ERROR_USER_LOGIN_FAILED');
			HookHandler::DispatchHook('/user/login/failed', $u);
			return false;
		}

		try{
			/** @var \Core\User\AuthDrivers\datastore $auth */
			$auth = $u->getAuthDriver('datastore');
		}
		catch(Exception $err){
			// Datastore auth backend not enabled for this user
			LogEntry::Init()
				->type('security')
				->level(LOG_LEVEL_HIGH)
				->code('/user/login')
				->user($u)
				->message('Login via password failed; password authentication not enabled')
				->send();
			$e->setError('t:MESSAGE_ERROR_USER_LOGIN_FAILED');
			HookHandler::DispatchHook('/user/login/failed', $u);
			return false;
		}


		// This is a special case if the password isn't set yet.
		// It can happen with imported users or if a password is invalidated.
		if($u->get('password') === '' && $u->isActive()) {
			// Use the Nonce system to generate a one-time key with this user's data.
			$nonce = NonceModel::Generate(
				'20 minutes',
				['type' => 'password-reset', 'user' => $u->get('id')]
			);

			$link = '/datastoreauth/forgotpassword?e=' . urlencode($u->get('email')) . '&n=' . $nonce;

			$email = Email::Factory('/emails/user/initialpassword.tpl');
			$email->setSubject('Initial Password Request');
			$email->setTo($u->get('email'));
			$email->assign('link', \Core\resolve_link($link));
			$email->assign('ip', REMOTE_IP);
			try{
				$email->send();
				LogEntry::Init()
					->type('security')
					->level(LOG_LEVEL_INFO)
					->code('/user/login')
					->user($u)
					->message('Login via password attempt; no password set, sent email instructions')
					->send();

				\Core\set_message('t:MESSAGE_INFO_USER_LOGIN_MUST_SET_NEW_PASSWORD_INSTRUCTIONS_HAVE_BEEN_EMAILED');
				return true;
			}
			catch(Exception $e){
				LogEntry::Init()->exception($e)->icon('envelope')->send();
				\Core\set_message('t:MESSAGE_ERROR_USER_LOGIN_MUST_SET_NEW_PASSWORD_UNABLE_TO_SEND_EMAIL');
				return false;
			}
		}


		if(!$auth->checkPassword($p->get('value'))) {
			// Log this as a login attempt!
			LogEntry::Init()
				->type('security')
				->level(LOG_LEVEL_HIGH)
				->code('/user/login')
				->user($u)
				->tag('password', $p->get('value'))
				->icon('user-times')
				->message('Login via password failed; bad password')
				->send();
			$e->setError('t:MESSAGE_ERROR_USER_LOGIN_FAILED');
			HookHandler::DispatchHook('/user/login/failed', $u);
			return false;
		}

		if($u->get('active') === '0') {
			// User account not active yet
			LogEntry::Init()
				->type('security')
				->level(LOG_LEVEL_INFO)
				->code('/user/login')
				->user($u)
				->icon('user-times')
				->message('Login via password attempt; account not activated')
				->send();
			$e->setError('t:MESSAGE_ERROR_USER_ACCOUNT_NOT_ACTIVATED');
			return false;
		}
		
		if($form->getElementValue('redirect')) {
			// The page was set via client-side javascript on the login page.
			// This is the most reliable option.
			$url = $form->getElementValue('redirect');
		}
		elseif(strpos($form->originalurl, '/user/login') !== false) {
			// If the user came from the registration page, get the page before that.
			$url = $form->referrer ? $form->referrer : '/';
		}
		else {
			// else the registration link is now on the same page as the 403 handler.
			$url = $form->originalurl;
		}
		
		// Well, record this too!
		LogEntry::Init()
			->code('/user/login')
			->type('security')
			->level(LOG_LEVEL_INFO)
			->message('Login via password successful')
			->icon('user-lock')
			->user($u)
			->send();

		// yay...
		$u->set('last_login', DateTime::Now('U', Timezone::TIMEZONE_GMT));
		$u->save();
		Session::SetUser($u);

		// This backend supports remembering logins!
		if($form->getElement('remember')->get('checked')) {
			UserDeviceModel::RememberDevice($u);
		}

		// Allow an external script to override the redirecting URL.
		$overrideurl = HookHandler::DispatchHook('/user/postlogin/getredirecturl');
		if($overrideurl){
			$url = $overrideurl;
		}

		return $url;
	}

	public static function PasswordResetRequest(Form $form) {
		$email = $form->getElementValue('email');


		$user = UserModel::Find(['email = ' . $email ], 1);
		if(!$user) {
			// User email was not registered, but for security reasons, don't reveal that info.

			\Core\log()
				->type('security')
				->code('/user/password/reset/failed')
				->message('Password reset attempt on invalid email')
				->tag('email', $email)
				->send();

			// Sleep between 1.5 and 3.0 seconds, to hinder timing attacks.
			$timer = rand(15, 30);
			usleep($timer * 100000);

			\Core\set_message('t:MESSAGE_INFO_PLEASE_CHECK_EMAIL_FOR_PASSWORD_RESET_INSTRUCTIONS');
			return '/';
		}


		try {
			$user->getAuthDriver('datastore');
		}
		catch(Exception $ex) {
			if(!ConfigHandler::Get('/user/register/allowpublic')) {
				// Password logins are not enabled and public register is disabled,
				// do not allow the user to switch auth backends in this case!
				\Core\log()
					->type('security')
					->code('/user/password/reset/failed')
					->message('Password reset attempt on account without password logins enabled')
					->user($user)
					->send();

				$e = Email::Factory('/emails/datastoreauth/password_reset_notenabled');
				$e->setSubject(t('STRING_USER_RESET_PASSWORD'));
				$e->setToUser($user);
				$e->assign('ip', REMOTE_IP);

				// The Geolookup package may provide geographic locations for IPs automatically.
				$e->assign('geo_city', defined('REMOTE_CITY') ? REMOTE_CITY : null);
				$e->assign('geo_province', defined('REMOTE_PROVINCE') ? REMOTE_PROVINCE : null);
				$e->assign('geo_country', defined('REMOTE_COUNTRY') ? REMOTE_COUNTRY : null);

				$e->encrypted = true;
				try{
					$e->send();
					\Core\set_message('t:MESSAGE_INFO_PLEASE_CHECK_EMAIL_FOR_PASSWORD_RESET_INSTRUCTIONS');
				}
				catch(Exception $e){
					\Core\set_message('Error sending the email, ' . $e->getMessage(), 'error');
					\Core\log_exception($e);
				}

				return '/';
			}

			// If this user doesn't have password logins enabled, but public registration is enabled, just allow it to continue.
		}


		if(!$user->isActive()) {
			// Inactive user accounts don't get the option to do a password reset either.
			\Core\log()
				->type('security')
				->code('/user/password/reset/failed')
				->message('Password reset attempt on an inactive account')
				->user($user)
				->send();

			$e = Email::Factory('/emails/datastoreauth/password_reset_inactive');
			$e->setSubject(t('STRING_USER_RESET_PASSWORD'));
			$e->setToUser($user);
			$e->assign('ip', REMOTE_IP);

			// The Geolookup package may provide geographic locations for IPs automatically.
			$e->assign('geo_city', defined('REMOTE_CITY') ? REMOTE_CITY : null);
			$e->assign('geo_province', defined('REMOTE_PROVINCE') ? REMOTE_PROVINCE : null);
			$e->assign('geo_country', defined('REMOTE_COUNTRY') ? REMOTE_COUNTRY : null);

			$e->encrypted = true;
			try{
				$e->send();
				\Core\set_message('t:MESSAGE_INFO_PLEASE_CHECK_EMAIL_FOR_PASSWORD_RESET_INSTRUCTIONS');
			}
			catch(Exception $e){
				\Core\set_message('Error sending the email, ' . $e->getMessage(), 'error');
				\Core\log_exception($e);
			}

			return '/';
		}


		// Use the Nonce system to generate a one-time key with this user's data.
		$nonce = NonceModel::Generate(
			'30 minutes',
			null,
			[
				'type' => 'password-reset',
				'user' => $user->id,
			]
		);

		$link = '/datastoreauth/password/confirm/' . $nonce;

		$e = Email::Factory('/emails/datastoreauth/password_reset');
		$e->setSubject(t('STRING_USER_RESET_PASSWORD'));
		$e->setToUser($user);
		$e->assign('link', \Core\resolve_link($link));
		$e->assign('ip', REMOTE_IP);

		// The Geolookup package may provide geographic locations for IPs automatically.
		$e->assign('geo_city', defined('REMOTE_CITY') ? REMOTE_CITY : null);
		$e->assign('geo_province', defined('REMOTE_PROVINCE') ? REMOTE_PROVINCE : null);
		$e->assign('geo_country', defined('REMOTE_COUNTRY') ? REMOTE_COUNTRY : null);

		$e->encrypted = true;
		try{
			$e->send();
			\Core\set_message('t:MESSAGE_INFO_PLEASE_CHECK_EMAIL_FOR_PASSWORD_RESET_INSTRUCTIONS');

			\Core\log()
				->type('security')
				->code('/user/password/reset/request')
				->message('Password reset requested')
				->user($user)
				->send();
		}
		catch(Exception $e){
			\Core\set_message('Error sending the email, ' . $e->getMessage(), 'error');
			\Core\log_exception($e);
		}

		return '/';
	}

	/**
	 * Hook receiver for /core/controllinks/user/view
	 *
	 * @param UserModel $user
	 *
	 * @return array
	 */
	public static function GetUserControlLinks(UserModel $user) : array {

		$globalEnabled = Helper::GetEnabledAuthDrivers();
		if(!isset($globalEnabled['datastore'])){
			// Datastore isn't enabled at all, disable any control links from the system.
			return [];
		}

		if(!$user->checkExists()) {
			// Invalid user.
			return [];
		}

		$currentUser = \Core\user();
		$isSelf      = $user->id === $currentUser->id;
		$isManager   = $currentUser->checkAccess('p:/user/users/manage');
		$allowPublic = ConfigHandler::Get('/user/register/allowpublic');

		if(!($isSelf || $isManager)){
			// Current user does not have access to manage the provided user's data.
			return [];
		}

		if($user->active !== '1') {
			// User is not active, nothing to provide.
			return [];
		}

		$ret = [];
		$enabled = $user->isAuthDriverEnabled('datastore');
		
		if(!$enabled){
			if($allowPublic || $isManager) {
				// Enabling password logins is only available for admins or users when the allowpublic config is set.
				$ret[] = [
					'link' => '/datastoreauth/password/' . $user->get('id'),
					'title' => t('STRING_ENABLE_PASSWORD_LOGIN'),
					'icon' => 'key',
					'group' => $allowPublic ? ViewControl::GROUP_MANAGE : ViewControl::GROUP_ADMIN,
				];
			}
		}
		else{
			$ret[] = [
				'link' => '/datastoreauth/password/' . $user->get('id'),
				'title' => t('STRING_CHANGE_PASSWORD'),
				'icon' => 'key',
				'group' => ViewControl::GROUP_MANAGE,
			];

			if(sizeof($user->getEnabledAuthDrivers()) > 1){
				// Only give the option to disable the password logins if there is another type available.
				$ret[] = [
					'link' => '/datastoreauth/disable/' . $user->get('id'),
					'title' => t('STRING_DISABLE_PASSWORD_LOGIN'),
					'icon' => 'ban',
					'confirm' => 'Are you sure you want to disable password-based logins?  (They can be re-enabled if requested.)',
					'group' => ViewControl::GROUP_MANAGE,
				];
			}
		}
		
		return $ret;

	}

	public static function PasswordSave(Form $form) {

		$p1          = $form->getElementValue('p1');
		$p2          = $form->getElementValue('p2');
		$nkey        = $form->getElementValue('nonce');
		$uid         = $form->getElementValue('user');
		$pwgen       = $form->getElementValue('pwgen');
		$user        = UserModel::Construct($uid);
		$currentUser = \Core\user();
		$isManager   = $currentUser->checkAccess('p:/user/users/manage');

		if(!$user->checkExists()) {
			throw new NotFoundException();
		}

		$user->enableAuthDriver('datastore');
		/** @var \Core\User\AuthDrivers\datastore $auth */
		$auth = $user->getAuthDriver('datastore');

		// Unblock the user account at this stage too.
		$user->set('active', 1);

		if($isManager && $pwgen) {
			// Allow managers to use the password gen utility
			$password = PasswordGenerator::QuickGen();
			$auth->setPassword($password);

			$user->save();

			\Core\log()
				->type('security')
				->code('/user/password/success')
				->message('Password changed (administratively)')
				->user($user)
				->send();

			// Send an email to this user notifying them of the new password.
			$email = Email::Factory('/emails/datastoreauth/password_admin');
			$email->setToUser($user);
			$email->setSubject('New Password Set');
			$email->assign('user', $user);
			$email->assign('new_password', $password);
			$email->assign('sitename', SITENAME);
			$email->encrypted = true;
			try{
				$email->send();
				\Core\set_message('Updated user password!', 'success');
			}
			catch(Exception $e){
				\Core\log_exception($e);
				\Core\set_message('Set password as "' . $password . '" but could not send the notification email.', 'warning');
			}
		}
		else{
			if($p1 !== $p2) {
				throw new GeneralValidationException('t:MESSAGE_ERROR_USER_REGISTER_PASSWORD_MISMATCH');
			}

			$auth->setPassword($p1);

			if(!$currentUser->checkExists()) {
				// User is not currently logged in, this is caused from the password reset page.
				$user->set('last_login', DateTime::Now('U', Timezone::TIMEZONE_GMT));
				$user->save();
				Session::SetUser($user);
			}
			else {
				$user->save();
			}


			\Core\log()
				->type('security')
				->code('/user/password/success')
				->message('Password changed')
				->user($user)
				->send();

			\Core\set_message('t:MESSAGE_SUCCESS_UPDATED_PASSWORD');
		}


		if($nkey) {
			// If a nonce key was used to access the previous page, mark that key as used now that the operation is complete.
			$nonce = NonceModel::Construct($nkey);
			$nonce->markUsed();
		}


		if($isManager) {
			return '/user/admin';
		}
		else {
			return '/user/view/' . $user->id;
		}
	}
}

