<?php
/**
 * Provides some controllers for any of the system form elements that need it.
 *
 * These generally provide some ajax functionality or similar to the UA.
 *
 * @package
 * @since 2.4.1
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\Controller;
use Core\Datamodel\Dataset;
use Core\Forms\Form;
use Core\Forms\FormElement;
use Core\GeneralValidationException;
use Core\ModelValidationException;
use Core\NotFoundException;
use Core\Search\ModelResult;
use Core\Search\SearchResults;
use Core\Session;
use Core\Templates\Template;
use Core\View;
use Core\Logger\LogEntry;
use Core\RedirectException;
use Core\UserPermittedExceptionInterface;
use Core\BadRequestException;
use Core\AccessDeniedException;



class FormController extends Controller {
	
	/**
	 * Generic form submission destination when the callsmethod parameter is set on forms.
	 * 
	 * This replaced previous functionality that was located in Form::CheckSavedSessionData
	 * 
	 * @param Form $form
	 */
	public function submit(Form $form) {
		
		$this->view->assign('form', $form);
		
		// Try to load the form from that form.  That will call all of the model's validation logic
		// and will throw exceptions if it doesn't.
		try{

			if($form->hasError()) {
				// Check against basic form processing errors.
				// These technically do not perform full validation, but at least will check against required fields.
				throw new ModelValidationException(implode('<br/>', $form->getErrors()));
			}

			LogEntry::Init()
				->caller(__FILE__ . ':' . __LINE__)
				->code('/form/submit')
				->level(LOG_LEVEL_TRACE)
				->message('Calling form auto-method ' . $form->get('callsmethod'))
				->send();
			
			$result = call_user_func($form->get('callsmethod'), $form);
			
			LogEntry::Init()
				->caller(__FILE__ . ':' . __LINE__)
				->code('/form/submit')
				->level(LOG_LEVEL_TRACE)
				->message($form->get('callsmethod') . ' successfully ran and returned ' . gettype($result))
				->send();
		}
		catch(ModelValidationException $e){
			\Core\log_exception($e);
			\Core\set_message($e->getMessage(), 'error');
			
			$form->saveToSession();
			throw new RedirectException('BACK', 302);
		}
		catch(GeneralValidationException $e){
			\Core\log_exception($e);
			\Core\set_message($e->getMessage(), 'error');
			
			$form->saveToSession();
			throw new RedirectException('BACK', 302);
		}
		catch( UserPermittedExceptionInterface $e ) {
			\Core\log_exception($e);
			\Core\set_message($e->getMessage(), 'error');
			
			$form->saveToSession();
			throw new RedirectException('BACK', 302);
		}
		catch(Exception $e){
			if(DEVELOPMENT_MODE){
				// Developers get the full message
				\Core\log_exception($e);
				\Core\set_message($e->getMessage(), 'error');
			}
			else{
				// While users of production-enabled sites get a friendlier message.
				\Core\log_exception($e);
				\Core\set_message('t:MESSAGE_ERROR_FORM_SUBMISSION_UNHANDLED_EXCEPTION');
			}
			
			$form->saveToSession();
			throw new RedirectException('BACK', 302);
		}
		
		if( $result === false || $result === null ) {
			// Both instances of this indicate a failure of some sort and the form should be re-displayed.
			$form->saveToSession();
			throw new RedirectException('BACK', 302);
		}
		elseif ($result === 'die' || $result === 'DIE') {
			// If it's set to die, simply exit the script without outputting anything.
			exit;
		}
		elseif($result === 'back' || $result === 'BACK') {
			
			throw new RedirectException(
				($form->referrer && $form->referrer !== REL_REQUEST_PATH) ? $form->referrer : 'BACK', 
				302
			);
		}
		elseif ($result === true) {
			// If the return code is boolean true, it's a reload.
			\Core\redirect($form->originalurl);
		}
		elseif($result === REL_REQUEST_PATH || $result === CUR_CALL) {
			// If the page returned the same page as the current url, force a reload, (as redirect will ignore it)
			\Core\reload();
		}
		else {
			// Anything else gets sent to the redirect system.
			\Core\redirect($result);
		}
	}
	
	/**
	 * Resets the cache for a given form, useful in development when the form won't change
	 * 
	 * @core-request-method GET
	 * @param string $form
	 */
	public function resetcache(string $form) {
		$formid = preg_replace('/[^a-f0-9]*/', '', $form);
		
		if(strlen($formid) !== 32) {
			\Core\set_message('Invalid form requested', 'error');
			throw new RedirectException('BACK');
		}
		
		// Allow "onreset" to do something!
		// This is meant to provide an event handler to call custom logic if the user clicks 'Reset'.
		$formSerialized = Session::Get('FormData/' . $formid);
		if($formSerialized) {
			$form = unserialize($formSerialized);
			if($form && $form instanceof Form && $form->get('onreset')) {
				try{
					call_user_func($form->get('onreset'), $form);
				}
				catch( Exception $ex) {
					// meh, we don't care.
				}
			}
		}
		
		Session::UnsetKey('FormData/' . $formid);
		\Core\set_message('Reset form successfully!', 'success');
		throw new RedirectException('BACK');
	}

	/**
	 * A page load to save a given form in session only.
	 * This applies all user-supplied values from the UA, but does not call the overall save method.
	 */
	public function save(Form $form) {
		$request = $this->getRequest();
		$view = $this->getView();
		$view->record = false;

		if($request->isAjax()) {
			$view->contenttype = View::CTYPE_JSON;
		}

		// Rebuild any Model on the Form, just in case something changed internally.
		// This usually doesn't do anything, but is provided in case a Model changes its behaviour
		// based on user-supplied values, such as the case with page templates,
		// (it loads the insertables and insertable groups from the template)
		foreach($form->getModels() as $prefix => $model) {
			$form->replaceModel($model, $prefix);
		}

		// Save this form back to the Session for continued use on the next page load.
		$form->saveToSession();

		if($request->isAjax()) {
			// AJAX requests should return JSON data as they're a background process.
			// If we throw a redirect exception, the browser will try to request the new URL in the background.
			$view->jsondata = [
				'status' => 200,
				'message' => 'Saved form successfully',
			];
		}
		else {
			throw new RedirectException('BACK', 302);
		}
	}

	/**
	 * Update and rebuild a Form based on submitted data
	 */
	public function pageinsertables_update() {
		$request = $this->getRequest();
		$view = $this->getView();
		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		$view->record = false;

		// This is an ajax-only request.
		if(!$request->isAjax()){
			throw new BadRequestException();
		}

		$formid = $request->getPost('___formid');
		if(!$formid){
			throw new NotFoundException();
		}

		// Lookup that form!
		if(Session::Get('FormData/' . $formid) === null){
			throw new NotFoundException();
		}

		/** @var $form Form */
		$form = unserialize(Session::Get('FormData/' . $formid));

		if(!$form){
			throw new NotFoundException();
		}

		// Run though each element submitted and try to validate it.
		if (strtoupper($form->get('method')) == 'POST') $src =& $_POST;
		else $src =& $_GET;

		$form->loadFrom($src, true);

		// Now that the form has been loaded with the data, reinitialize the page's insertable elements.
		foreach($form->getModels() as $prefix => $model) {
			$form->replaceModel($model, $prefix);
		}

		$form->saveToSession();

		$view->jsondata = array(
			'status' => '1',
			'message' => 'Switched templatename successfully',
			'formid' => $form->generateUniqueHash(),
		);
	}

	public function pagemetas_autocompleteuser(){
		$request = $this->getRequest();
		$view = $this->getView();
		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		$view->record = false;

		// This is an ajax-only request.
		if(!$request->isAjax()){
			throw new BadRequestException();
		}

		// Does the user have access to search for users?
		if(!\Core\user()->checkAccess('p:/user/search/autocomplete')){
			throw new AccessDeniedException();
		}

		$includeInactive = ($request->getParameter('inactive') == '1');
		$term = $request->getParameter('term');
		$results = UserModel::Search($term);
		
		$filteredresults = array();
		foreach($results as $model){
			/** @var UserModel $model */
			// This model will only be added to the form if it's active.
			if(!($model->get('active') || $includeInactive)){
				continue;
			}

			$filteredresults[] = array(
				'id'    => $model->get('id'),
				'label' => $model->getDisplayName(),
				'value' => $model->get('id'),
				'__rel' => $model->get('__search_relevancy'),
			);
		}

		// The json data will contain the following keys for each element:
		// id, label, value

		$view->jsondata = $filteredresults;
	}

	public function pagemetas_autocompletekeyword(){
		$request = $this->getRequest();
		$view = $this->getView();
		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		$term = $request->getParameter('term');
		$view->record = false;

		// This is an ajax-only request.
		if(!$request->isAjax()){
			return View::STATUS_BADREQUEST;
		}

		$ds = new Dataset();
		$ds->table('page_meta');
		$ds->uniquerecords = true;
		$ds->select('meta_value', 'meta_value_title');
		$ds->where('meta_key = keyword');
		$ds->where('meta_value_title LIKE ' . $term . '%');

		// Just in case there are a huge number of records...
		$stream = new DatasetStream($ds);
		$view->jsondata = array();
		while(($record = $stream->getRecord())){
			$view->jsondata[] = array(
				'id' => $record['meta_value'],
				'label' => $record['meta_value_title'],
				'value' => $record['meta_value'],
			);
		}

		// Does the user have access to search for users?
		// if so include that search here to!  This is for the subject matter tag, or "This x is about person y!"
		if(\Core\user()->checkAccess('p:/user/search/autocomplete')){
			$results = UserModel::Search($term);
			foreach($results as $r){
				/** @var $r ModelResult */

				/** @var UserModel $user */
				$user = $r->_model;
				$view->jsondata[] = array(
					'id' => 'u:' . $user->get('id'),
					'label' => $user->getDisplayName(),
					'value' => 'u:' . $user->get('id'),
				);
			}
		}
	}

	/**
	 * Page to test the UI of form elements.
	 *
	 * This will generate a form with every registered form element.
	 */
	public function testui(){
		$view = $this->getView();
		$request = $this->getRequest();

		if(!\Core\user()->checkAccess('g:admin')){
			// This test page is an admin-only utility.
			return View::STATUS_ACCESSDENIED;
		}

		$form = new Form();

		// What type of orientation do you want to see?
		$orientation = $request->getParameter('orientation');
		if(!$orientation){
			$orientation = 'horizontal';
		}
		$required = ($request->getParameter('required'));
		$error    = ($request->getParameter('error'));


		$form->set('orientation', $orientation);

		$mappings = Form::$Mappings;
		// Make them alphabetical.
		ksort($mappings);

		foreach($mappings as $k => $v){
			try{
				$atts = [
					'name' => $k,
					'title' => $v,
					'description' => 'This form element is a ' . $v . ', registered to the key ' . $k . '.',
				];

				if($required) $atts['required'] = true;

				// Some form elements have particular requirements.
				switch($v){
					case '\\Core\\Forms\\FileInput':
					case 'MultiFileInput':
						$atts['basedir'] = 'public/form-testui';
						break;
					case '\\Core\\Forms\\PagePageSelectInput':
						$atts['templatename'] = 'foo';
						break;
					case '\\Core\\Forms\\PageInsertables':
						$atts['baseurl'] = '/';
						break;
					case '\\Core\\Forms\\PageMeta':
						$atts['name'] = 'test';
						break;
					case '\\Core\\Forms\\ButtonInput':
						$atts['value'] = 'Button Text';
						break;
					case '\\Core\\Forms\\CheckboxesInput':
					case '\\Core\\Forms\\RadioInput':
						$atts['options'] = ['key1' => 'Key 1', 'key2' => 'Key 2'];
						break;
				}
				$el = FormElement::Factory($k, $atts);

				if($error && $el instanceof \Core\Forms\FormElement){
					$el->setError('Something bad happened', false);
				}
				$form->addElement( $el );
			}
			catch(Exception $e){
				\Core\set_message('Form element ' . $v . ' failed to load due to ' . $e->getMessage(), 'error');
			}
		}

		$view->title = 'Test Form Element UI/UX';
		$view->assign('form', $form);
		$view->assign('orientation', $orientation);
		$view->assign('required', $required);
		$view->assign('error', $error);
	}

	public function test() {
		$form = new Form();
		$form->set('callsmethod', 'FormController::TestSubmit');
		$form->addElement('text', ['name' => 'thing']);
		$form->addElement('submit', ['value' => 'Test This']);

		$this->getView()->assign('form', $form);
	}

	public static function TestSubmit(Form $form) {
		throw new ModelValidationException('Blerp failed');
	}
}



