<?php
/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

use Core\Controller;
use Core\Core;
use Core\Cache;
use Core\AccessDeniedException;
use Core\Installer\Configuration;
use Core\Session;
use Core\Installer\Checks;
use Core\Installer\InstallHelper;
use Core\Datamodel\DMI;
use Core\Datamodel\DMI_Authentication_Exception;
use Core\Theme;
use Core\CLI\CLI;
use Core\Dict;

/**
 * Description of InstallController
 *
 * @author charlie
 */
class InstallController extends Controller {

	/**
	 * Main loader for the baseline installation system
	 *
	 * @core-exec-mode BOTH
	 *
	 * @throws AccessDeniedException
	 */
	public function index(){
		if(Core::IsInstalled()){
			throw new AccessDeniedException('Core is already installed!');
		}

		$this->view->disableCache();
		$this->view->stylesheets[] = '<style>' . file_get_contents(ROOT_PDIR . 'core/resources/installer.css') . '</style>';
		$this->view->stylesheets[] = '<style>pre.code-shell code .cli-line::before { content: "' . get_current_user() . '@' . gethostname() . ': $ "; } pre.code-shell code .cli-root-line::before { content: "[root@' . gethostname() . ' ~]# "; }</style>';

		// If there's a configuration.xml set, skip straight to the component installation.
		// This is a security precaution so as to not reveal configured parameters to the web
		// and to not overwrite those configured parameters.
		if(file_exists(ROOT_PDIR . 'config/configuration.xml')){
			$this->_installCore();
			return;
		}

		if(EXEC_MODE === 'CLI') {
			$this->_installCLI();
			return;
		}

		if($this->request->isGet()) {
			// GET requests always get routed to the initial license step.
			$this->_license();
			return;
		}

		$submit = filter_input(INPUT_POST, '_submit', FILTER_SANITIZE_STRING);
		$retry  = filter_input(INPUT_POST, '_retry', FILTER_SANITIZE_STRING);

		if($submit === 'license') {
			$this->_setupWelcome();
			return;
		}

		if($submit === 'welcome' || $retry === 'checks') {
			$this->_runChecks();
			return;
		}

		if(filter_has_var(INPUT_POST, '_config')){
			// Allow any step to submit additional config parameters if necessary.
			$this->_storeConfiguration();
		}

		if($submit === 'checks') {
			$this->_setupConfiguration();
			return;
		}

		if($submit === 'config' || $retry === 'database') {
			$this->_setupDatabase();
			return;
		}

		if($submit === 'database' || $retry === 'install'){
			$this->_installConfiguration();
		}

		if($submit === 'install'){
			$this->_installCore();
		}
	}

	private function _installCLI() {

		$config = new Configuration();
		$values = $config->getValues();
		$elements = $config->getFormElements();
		$user = InstallHelper::GetUser();

		foreach($elements as $idx => $el){
			if(!$el['advanced']){
				CLI::Clear();
				CLI::PrintHeader('Initial Configuration');

				$values[ $el['name'] ] = $elements[ $idx ]['value'] = CLI::PromptFormElement($el['type'], $el);
			}
		}
		// Save the values from the initial config step
		$config->setValues($values);


		// The one parameter that is required is the site URL; this is used to overwrite the rewritebase when that's installed.
		CLI::Clear();
		CLI::PrintHeader('Initial Configuration');
		$url = CLI::PromptUser('Please enter the full URL for the site, including any directories', 'text-required');
		$rootdir = parse_url($url, PHP_URL_PATH);


		CLI::Clear();
		CLI::PrintHeader('Advanced Configuration');
		if(CLI::PromptUser('Would you like to configure additional advanced parameters?', 'bool', false)){
			$parameters = new Dict();
			foreach($elements as $idx => $el){
				$parameters[ $idx ] = $el['title'];
			}
			$parameters[ '_NEXT_' ] = '_NEXT_';

			do{
				CLI::Clear();
				CLI::PrintHeader('Advanced Configuration');
				$idx = CLI::PromptUser('What would you like to configure?', $parameters);

				if($idx !== '_NEXT_'){
					CLI::Clear();
					CLI::PrintHeader('Advanced Configuration');
					$values[ $elements[ $idx ]['name'] ] = $elements[ $idx ]['value'] = CLI::PromptFormElement($elements[ $idx ]['type'], $elements[ $idx ]);
				}
			} while ($idx !== '_NEXT_');

			// Save the values from the advanced config step
			$config->setValues($values);
		}

		CLI::PrintSuccess('Proceeding...');
		sleep(2);


		CLI::Clear();
		CLI::PrintHeader('Requirement Checks');
		$checks = InstallHelper::GetChecks();

		// Add checks for the web server itself, (since we're running from the CLI, that may not be a given)
		if(CLI::PromptUser('Check and configure Apache web server?', 'bool', false)) {
			$checks[] = new Checks\ApacheServerCheck();
		}
		$checks[] = new Checks\MysqlServerCheck($values['database_server']);

		// Append the dynamic checks based off of configuration variables.
		$checks[] = new Checks\WriteableDirectoryCheck(ROOT_PDIR . $values['CDN_LOCAL_ASSETDIR']);
		$checks[] = new Checks\WriteableDirectoryCheck(ROOT_PDIR . $values['CDN_LOCAL_PUBLICDIR']);
		$checks[] = new Checks\WriteableDirectoryCheck(ROOT_PDIR . $values['CDN_LOCAL_PRIVATEDIR']);
		$checks[] = new Checks\DirectoryProtectionCheck(ROOT_PDIR . $values['CDN_LOCAL_ASSETDIR'], 'assets');
		$checks[] = new Checks\DirectoryProtectionCheck(ROOT_PDIR . $values['CDN_LOCAL_PUBLICDIR'], 'public');
		$checks[] = new Checks\DirectoryProtectionCheck(ROOT_PDIR . $values['CDN_LOCAL_PRIVATEDIR'], 'private');


		$passed = true;
		foreach($checks as $c){
			CLI::PrintActionStart($c->getTitle());
			usleep(100000);

			if($c instanceof Checks\DirectoryProtectionCheck){
				// This is required because the app doesn't know where the ROOT_WDIR actually is.
				$c->setTemplateValue('@{rewritebase}@', $rootdir);
			}

			if($c->test()){
				CLI::PrintActionStatus(true);
			}
			else{
				CLI::PrintActionStatus(false);
				CLI::PrintLine($c->getDetails());
				$passed = false;
			}
		}

		if(!$passed){
			return;
		}
		else{
			CLI::PrintSuccess('All checks passed!');
			sleep(2);
		}


		CLI::Clear();
		CLI::PrintHeader('Database Setup');
		$server = $values['database_server'];
		$port   = $values['database_port'];
		$type   = $values['database_type'];
		$user   = $values['database_user'];
		$pass   = $values['database_pass'];
		$name   = $values['database_name'];
		$quickfix = "CREATE USER '$user' IDENTIFIED BY '$pass'; CREATE DATABASE IF NOT EXISTS $name; GRANT ALL ON $name.* TO '$user'; FLUSH PRIVILEGES;";


		CLI::PrintLine('Attempting connection to ' . $server . ' as root with no password...');
		usleep(250000);
		$ret = null;
		passthru('mysql -uroot -h' . $server . ' -e "' . $quickfix . '"', $ret);
		if($ret !== 0){
			CLI::PrintLine('Attempting connection to ' . $server . ' as root WITH a password...');
			CLI::PrintDebug('Please note, you will need to enter the database root password when asked.');
			usleep(250000);
			$counter = 0;
			while($counter < 3 && $ret !== 0) {
				++$counter;
				passthru('mysql -uroot -h' . $server . ' -p -e "' . $quickfix . '"', $ret);

				if($ret !== 0) {
					CLI::PrintError('MySQL root password incorrect!');
					sleep(1);
				}
			}
		}

		if($ret !== 0 && $user !== 'root') {
			if(CLI::PromptUser('Connection to database failed, try as sudo?', 'bool', false)) {
				$counter = 0;
				while($counter < 3 && $ret !== 0) {
					++$counter;
					passthru('sudo mysql -uroot -h' . $server . ' -p -e "' . $quickfix . '"', $ret);

					if($ret !== 0) {
						CLI::PrintError('MySQL root password incorrect!');
						sleep(1);
					}
				}
			}
		}

		if($ret !== 0) {
			CLI::PrintError('Connection to database failed, unable to install Core.');
			return;
		}

		try{
			new DMI($type, $server, $user, $pass, $name);
			CLI::PrintSuccess('Connection to ' . $server . ' with ' . $user . ' succeeded!');
			sleep(2);
		}
		catch (Exception $ex) {
			CLI::PrintError($ex->getMessage());
			return;
		}


		CLI::Clear();
		CLI::PrintHeader('Installing App');

		CLI::PrintActionStart('Saving Configuration');
		usleep(100000);
		if(!is_dir(ROOT_PDIR . 'config/')){
			mkdir(ROOT_PDIR . 'config/');
		}
		file_put_contents(ROOT_PDIR . 'config/configuration.xml', $config->getContents());
		CLI::PrintActionStatus(true);

		CLI::PrintActionStart('Securing Configuration');
		usleep(100000);
		chmod(ROOT_PDIR . 'config/', 0500);
		chown(ROOT_PDIR . 'config/', InstallHelper::GetWebServerUser());
		chgrp(ROOT_PDIR . 'config/', InstallHelper::GetWebServerGroup());

		chmod(ROOT_PDIR . 'config/configuration.xml', 0400);
		chown(ROOT_PDIR . 'config/configuration.xml', InstallHelper::GetWebServerUser());
		chgrp(ROOT_PDIR . 'config/configuration.xml', InstallHelper::GetWebServerGroup());
		CLI::PrintActionStatus(true);


		CLI::PrintSuccess('Application deployed and ready for first-run!');
		CLI::PrintLine('Please browse to this application in a web browser to get started.');
	}

	private function _installConfiguration() {
		$values = Session::Get('config');
		$config = new Configuration();

		// If the user already has config parameters set, load them into the configuration system first.
		if(Session::IsKeySet('config')){
			$config->setValues(Session::Get('config'));
		}

		if(
			is_dir(ROOT_PDIR . 'config/') &&
			is_writable(ROOT_PDIR . 'config/') &&
			!file_exists(ROOT_PDIR . 'config/configuration.xml')
		){
			// Simple enough!  Write the config to the filesystem.
			file_put_contents(ROOT_PDIR . 'config/configuration.xml', $config->getContents());
			chmod(ROOT_PDIR . 'config/configuration.xml', 0600);
			$passed = true;
			$contents = '';
		}
		elseif(file_exists(ROOT_PDIR . 'config/configuration.xml')){
			$passed = true;
			$contents = '';
		}
		else{
			$contents = $config->getContents();
			$passed = false;
		}

		$submit = filter_input(INPUT_POST, '_submit', FILTER_SANITIZE_STRING);
		$retry  = filter_input(INPUT_POST, '_retry', FILTER_SANITIZE_STRING);


		$this->view->title = 'Core Plus Installation (Deploy Configuration)';
		$this->view->templatename = 'pages/install/installconfiguration.tpl';
		$this->view->assign('contents', $contents);
		$this->view->assign('passed', $passed);
		$this->view->assign('path', ROOT_PDIR . 'config/configuration.xml');
		$this->view->assign('first_try', $retry !== 'install');
		$this->view->assign('is_cpanel', InstallHelper::CheckIsCPanel());
	}

	private function _installCore(){
		$core = Core::GetComponent('core');
		$core->install();

		// Install the default theme for this application
		Theme::GetTheme('AUTO')->install();

		Cache::Flush();

		$this->view->title = 'Core Plus Installation (Complete)';
		$this->view->templatename = 'pages/install/installcomplete.tpl';
		$this->view->assign('url', ROOT_WDIR);
	}

	private function _license() {
		$this->view->title = 'Core Plus Installation (License)';
		$this->view->templatename = 'pages/install/license.tpl';
	}

	private function _runChecks() {
		$allPassed = true;
		$checks = InstallHelper::GetChecks();

		$cmds = '';
		foreach($checks as $c){
			if(!$c->test()){
				$cmds .= '<span class="cli-line">' . htmlentities($c->getFix()) . '</span>' . NL;
				$allPassed = false;
			}
		}

		$this->view->title = 'Core Plus Installation (Requirement Checks)';
		$this->view->templatename = 'pages/install/checks.tpl';
		$this->view->assign('checks', $checks);
		$this->view->assign('quickfix', $cmds);
		$this->view->assign('passed', $allPassed);
	}

	/**
	 * Display the setup configuration step to the user
	 */
	private function _setupConfiguration(){
		$config = new Configuration();

		// If the user already has config parameters set, load them into the configuration system first.
		if(Session::IsKeySet('config')){
			$config->setValues(Session::Get('config'));
		}

		$this->view->title = 'Core Plus Installation (Configuration)';
		$this->view->templatename = 'pages/install/setupconfiguration.tpl';
		$this->view->assign('elements', $config->getFormElements());
	}

	/**
	 * "Welcome" page of the installer and display some basic options for installation.
	 */
	private function _setupWelcome() {
		// Check if the root directory is writable by the current user.
		// If it is, we can skip the notices and get straight to the welcome.
		$isWritable = is_writable(ROOT_PDIR);
		$scriptOwner = get_current_user();
		$currentUser = posix_getpwuid(posix_geteuid())['name'];
		$isCpanel = InstallHelper::CheckIsCPanel();
		$cpanelVersion = $isCpanel ? file_get_contents('/usr/local/cpanel/version') : null;

		$this->view->title = 'Core Plus Installation (Welcome!)';
		$this->view->templatename = 'pages/install/welcome.tpl';
		$this->view->assign('is_cpanel', $isCpanel);
		$this->view->assign('is_writable', $isWritable);
		$this->view->assign('script_owner', $scriptOwner);
		$this->view->assign('current_user', $currentUser);
		$this->view->assign('root_pdir', ROOT_PDIR);
		$this->view->assign('hostname', gethostname());
		$this->view->assign('cpanel_version', $cpanelVersion);
	}

	/**
	 * Display the setup database step to the user
	 */
	private function _setupDatabase(){
		$values = Session::Get('config');
		$config = new Configuration();
		$config->setValues($values);

		// The keys used here are:
		$server = $values['database_server'];
		$port   = $values['database_port'];
		$type   = $values['database_type'];
		$user   = $values['database_user'];
		$pass   = $values['database_pass'];
		$name   = $values['database_name'];

		$submit = filter_input(INPUT_POST, '_submit', FILTER_SANITIZE_STRING);
		$retry  = filter_input(INPUT_POST, '_retry', FILTER_SANITIZE_STRING);

		if(InstallHelper::CheckIsCPanel()) {
			$values = [
				'name' => substr($name, strpos($name, '_') + 1),
				'user' => substr($user, strpos($user, '_') + 1),
				'pass' => $pass,
			];
		}
		else {
			$values = [
				'name' => $name,
				'user' => $user,
				'pass' => $pass,
			];
		}

		if($retry === 'database') {
			// Only perform the test to connect if retry was requested.
			// This provides the user the opportunity to test before proceeding.
			try{
				new DMI($type, $server, $user, $pass, $name);
				$passed = true;
				$error = '';
			}
			catch (\Core\Datamodel\DMIAuthenticationFailedException $ex){
				$passed = false;
				$error = $ex->getMessage();
			}
			catch (Exception $ex) {
				$passed = false;
				$error = $ex->getMessage();
			}
		}
		else {
			$passed = false;
			$error = null;
		}
		



		// If there was an issue, allow the user to change the database parameters.
		if(!$passed){
			$formElements = [];
			foreach($config->getFormElements() as $element){
				if(strpos($element['name'], 'database_') === 0){
					$formElements[] = $element;
				}
			}
		}
		else{
			$formElements = [];
		}

		$this->view->title = 'Core Plus Installation (Database)';
		$this->view->templatename = 'pages/install/database.tpl';
		$this->view->assign('passed', $passed);
		$this->view->assign('error', $error);
		$this->view->assign('elements', $formElements);
		$this->view->assign('is_cpanel', InstallHelper::CheckIsCPanel());
		$this->view->assign('values', $values);
		$this->view->assign('first_try', $retry !== 'database');
	}

	/**
	 * Store the variables submitted in the configuration step to the session for saving once installation is ready.
	 */
	private function _storeConfiguration(){
		$config = new Configuration();

		// If the user already has config parameters set, load them into the configuration system first.
		if(Session::IsKeySet('config')){
			$config->setValues(Session::Get('config'));
			$values = Session::Get('config');
		}
		else{
			$values = $config->getValues();
		}

		// Load all the configurable keys so I know what to pull from the POST.
		$keys = array_keys($values);
		foreach($keys as $key){
			if(filter_has_var(INPUT_POST, $key)){
				$values[ $key ] = filter_input(INPUT_POST, $key, FILTER_SANITIZE_STRING);
			}
		}

		// Save these and continue!
		Session::Set('config', $values);
	}
}
