<?php
/**
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\Component;
use Core\ComponentConfig;
use Core\ConfigHandler;
use Core\Controller;
use Core\Core;
use Core\CoreStyles;
use Core\Filestore\Backends\FileLocal;
use Core\Filestore\Directory;
use Core\Filestore\Factory;
use Core\Filestore\File;
use Core\Forms\Form;
use Core\Forms\FormElement;
use Core\Templates\Template;
use Core\Theme;
use Core\View;
use Core\RedirectException;


/**
 * The main theme administration controller.
 *
 * This is responsible for all administrative tasks associated with the theme.
 */
class ThemeController extends Controller{

	public function __construct() {
		$this->accessstring = 'g:admin';
	}
	
	/**
	 * Disable or Enable inline widget editing and hinting
	 */
	public function widgets_toggle() {
		$new = ! \Core\user()->getPreference('core-widgets-inline-controls', false);
		\Core\user()->setPreference('core-widgets-inline-controls', $new);
		
		\Core\set_message(($new ? 'Enabled' : 'Disabled') . ' inline widget controls', 'success');
		throw new RedirectException('BACK');
	}

	/**
	 * Page to display the currently installed themes and shortcuts to various operations therein.
	 */
	public function index(){
		$view = $this->getView();
		$selected = ConfigHandler::Get('/theme/selected');
		
		$themes = Theme::GetAllThemes();
		$current = Theme::GetTheme($selected);
		// Set to true if multisite is enabled AND the page is currently on a child site.
		$multisite = Core::IsLibraryAvailable('multisite') && MultiSiteHelper::IsEnabled() && MultiSiteHelper::GetCurrentSiteID() != 0;
		$configDefault  = ConfigHandler::GetConfig('/theme/default_template');
		$configSelected = ConfigHandler::GetConfig('/theme/selected');
		$configEmailDefault  = ConfigHandler::GetConfig('/theme/default_email_template');
		// Only allow changing the theme if it's on the root site OR both config options are overrideable.
		$themeSelectionEnabled = (!$multisite || ($configDefault->get('overrideable') && $configSelected->get('overrideable')));
		$emailSelectionEnabled = (!$multisite || $configEmailDefault->get('overrideable'));

		$configOptions = $current->getConfigs();
		if(!sizeof($configOptions)){
			$optionForm = null;
		}
		else{
			$optionForm = new Form();
			$optionForm->set('callsmethod', 'AdminController::_ConfigSubmit');
			foreach($configOptions as $c){
				/** @var $c ComponentConfig */
				
				$configModel = ConfigHandler::GetConfig($c->key);
				
				if($multisite){
					// Only pull the config options that are enabled for this specific site.
					if($configModel->get('overrideable')){
						$optionForm->addElement($configModel->getAsFormElement());
					}
				}
				else{
					// Sites that either
					// do NOT have multisite installed
					// nor have multisite enabled
					// or on the root site, get all options.
					$optionForm->addElement($configModel->getAsFormElement());
				}
			}
			
			if(sizeof($optionForm->getElements()) > 0){
				// There is at least one element in the option forms!
				$optionForm->addElement('submit', ['value' => 'Save Configurable Options']);
			}
			else{
				// Reset the form back to null so that the section doesn't display.
				$optionForm = null;
			}
		}


		// Get the templates throughout the site.  These can include pages, emails, form elements, etc.
		$components = Core::GetComponents();
		$templates = array();

		foreach($components as $c) {
			/** @var $c Component */
			$dir = $c->getBaseDir() . 'templates/';
			if(!is_dir($dir)){
				continue;
			}

			$dirlen = strlen($dir);
			$component = $c->getName();

			$dh = Factory::Directory($dir);
			//$pagetplfiles = $dh->ls('tpl', true);
			$pagetplfiles = $dh->ls(null, true);

			// not sure why getFilename(path) isn't working as expected, but this works too.
			foreach($pagetplfiles as $obj){

				// I don't want directories.
				if($obj instanceof \Core\Filestore\Directory) continue;

				/** @var $obj File */
				$file = substr($obj->getFilename(), $dirlen);

				try {
					// Since this is a template, it may actually be in a different location than where the package maintainer put it.
					// ie: user template user/templates/pages/user/view.tpl may be installed to themes/myawesometheme/pages/user/view.tpl instead.
					$tpl = Template::Factory($file);
					$resolved = Template::ResolveFile($file);

					$newobj = Factory::File($resolved);

					$templates[$file] = array(
						'file' => $file,
						'resolved' => $resolved,
						'obj' => $newobj,
						'haswidgets' => $tpl->hasWidgetAreas(),
						'component' => $component,
						'has_stylesheets' => $tpl->hasOptionalStylesheets(),
						'type' => 'template',
					);
				}
				catch(Exception $ex) {
					continue;
				}
			}
		}

		// Now that the template files have been loaded into a flat array, I need to convert that to the properly nested version.
		ksort($templates);
		$nestedtemplates = array();
		foreach($templates as $k => $obj){
			$parts = explode('/', $k);
			$lastkey = sizeof($parts) - 1;
			$thistarget =& $nestedtemplates;
			foreach($parts as $i => $bit){
				if($i == $lastkey){
					$thistarget[$bit] = $obj;
				}
				else{
					if(!isset($thistarget[$bit])){
						$thistarget[$bit] = [];
					}
					$thistarget =& $thistarget[$bit];
				}
			}
		}

		// Setup the skin selection based on the various siteskin keys added to the defaults
		/** @var ConfigModel[] $siteSkinApps */
		$siteSkinApps = array_merge(
			[
				ConfigHandler::GetConfig('/theme/default_template'),
				ConfigHandler::GetConfig('/theme/default_admin_template'),
			],
			ConfigHandler::FindConfigs('/theme/siteskin/')
		);

		
		$siteskinform = new Form();
		$siteskinform->set('callsmethod', 'ThemeController::SaveSiteSkins');
		$opts = ['' => '-- Site Default --'];
		foreach($current->getSkins() as $skin){
			// Skip system skins
			if($skin['is_system']) {
				continue;
			}

			$opts[$skin['file']] = $skin['title'];
		}
		foreach($siteSkinApps as $config) {
			// We use the built-in subsystem to retrieve some necessary fields
			$configAtts = $config->getFormAttributes();

			$siteskinform->addElement(
				'select',
				[
					'name' => 'config[' . $config->key . ']',
					'title' => $configAtts['title'],
					'description' => $configAtts['description'],
					'value' => $config->getValue(),
					'options' => $opts,
				]
			);
		}

		$siteskinform->addElement('submit', ['value' => t('STRING_SAVE')]);
		
		$view->assign('themes', $themes);
		$view->assign('current', $current);
		$view->assign('options_form', $optionForm);
		$view->assign('templates', $nestedtemplates);
		$view->assign('url_themeeditor', \Core\resolve_link('/theme/editor'));
		$view->assign('url_themewidgets', \Core\resolve_link('/theme/widgets'));
		$view->assign('url_themestylesheets', \Core\resolve_link('/theme/selectstylesheets'));
		$view->assign('site_skins_form', $siteskinform);
		$view->assign('multisite', $multisite);
		$view->assign('theme_selection_enabled', $themeSelectionEnabled);
		$view->assign('email_selection_enabled', $emailSelectionEnabled);
	}

	/**
	 * Adjust theme styles
	 *
	 * Will use CoreStyles to retrieve all the customizable styles and provide an interface for managing them
	 */
	public function styles() {
		$view = $this->getView();
		$request = $this->getRequest();

		$themeStyles = CoreStyles::GetThemeStyles();
		$customStyles = CoreStyles::GetCustomStyles();
		$processedStyles = [];

		foreach($themeStyles as $k => $v) {
			// Do an additional step to check if it's a colour
			$val = CoreStyles::GetStyle($k);

			if(preg_match('/^#[a-f0-9]{6}$/', $val)) {
				$type = 'color';
			}
			elseif(preg_match('/^[0-9]*px$/', $val)) {
				$type = 'size';
			}
			elseif(preg_match('/^[0-9\.]*rem$/', $val)) {
				$type = 'size';
			}
			elseif(preg_match('/^[0-9\.]*em$/', $val)) {
				$type = 'size';
			}
			else {
				$type = null;
			}

			$processedStyles[ $k ] = [
				'theme' => $val,
				'custom' => null,
				'type' => $type,
			];
		}

		foreach($customStyles as $k => $v) {
			if(isset($processedStyles[ $k ])) {
				$processedStyles[ $k ]['custom'] = $v;
			}
			else {
				$processedStyles[ $k ] = [
					'theme' => null,
					'custom' => $v,
					'type' => null,
				];
			}
		}

		// @todo finish this.
		var_dump($processedStyles); die();


		$groups = [];
		$configs = ConfigHandler::GetAll('/core/style/*');

		if($request->isPost()) {
			$posts = $request->getPost('colors');
			foreach(array_keys($configs) as $key) {
				if(isset($posts[ $key ])) {
					if(!preg_match('/#[0-9a-fA-F]{6}/', $posts[$key])) {
						$value = '';
					}
					else {
						$value = $posts[$key];
					}

					ConfigHandler::Set($key, $value);
				}
			}

			throw new \Core\TemporaryFollowRedirectException('/theme/colors');
		}

		foreach($configs as $key => $val) {
			if(strpos($key, 'color') === false) {
				// Only process colours
				continue;
			}

			if(strpos($key, '/core/style/base/') === 0) {
				$group = '0base';
			}
			elseif(strpos($key, '/core/style/interactive/') === 0) {
				$group = '1interactive';
			}
			else {
				$group = substr($key, 12);
				$group = substr($group, 0, strpos($group, '/'));
			}

			if(!isset($groups[ $group ])) {
				$groups[ $group ] = [];
			}

			$groups[ $group ][$key] = [
				'value' => $val,
				'title' => 'STRING_CONFIG_' . \Core\I18N\I18NLoader::KeyifyString($key),
			];
		}

		ksort($groups);

		$view->title = 'Site Colors';
		$view->assign('groups', $groups);
	}

	/**
	 * View to display a list of currently installed themes, their templates, and be able to manage
	 * their templates and set them as default.
	 *
	 * @todo Implement an Add/Upload Theme link on this page.
	 */
	public function index2(){
		$view = $this->getView();
		$default = ConfigHandler::Get('/theme/selected');

		$themes = array();
		$dir = ROOT_PDIR . 'themes';
		$dh = opendir($dir);
		if($dh){
			while(($file = readdir($dh)) !== false){
				if($file[0] == '.') continue;
				if(!is_dir($dir . '/' . $file)) continue;

				// Load up the templates for this theme.
				$templates = Theme::GetTheme($file)->getTemplates();

				$themes[] = array(
					'name' => $file,
					'default' => ($default == $file),
					'templates' => $templates,
				);
			}
			closedir($dh);
		}

		// Get the page templates that have widgetareas defined within.
		$components = Core::GetComponents();
		$pagetemplates = array();

		foreach($components as $c) {
			/** @var $c Component */
			$dir = $c->getViewSearchDir();
			if(!$dir) continue;

			$dirlen = strlen($dir);
			$component = $c->getName();

			$dh = Factory::Directory($dir);
			$pagetplfiles = $dh->ls('tpl', true);

			// not sure why getFilename(path) isn't working as expected, but this works too.
			foreach($pagetplfiles as $obj){
				/** @var $obj File */
				$file = substr($obj->getFilename(), $dirlen);

				// Since this is a template, it may actually be in a different location than where the package maintainer put it.
				// ie: user template user/templates/pages/user/view.tpl may be installed to themes/myawesometheme/pages/user/view.tpl instead.
				$resolved = Template::ResolveFile($file);
				$newobj = Factory::File($resolved);

				// Check the contents of the file and see if there is a {widgetarea...} here.
				$contents = $newobj->getContents();
				if(strpos($contents, '{widgetarea') !== false){
					$haswidgets = true;
				}
				else{
					$haswidgets = false;
				}


				$pagetemplates[] = array(
					'file' => $file,
					'resolved' => $resolved,
					'obj' => $newobj,
					'haswidgets' => $haswidgets,
					'component' => $component,
				);
			}
		}

		// The CSS files for the current theme and every component.
		// This is keyed by the filename so that duplicate entries don't show up more than once,
		// ie: if an asset exists in both the component and the theme.
		$cssfiles = array();

		// Give me the current theme!
		$dir = ROOT_PDIR . 'themes/' . ConfigHandler::Get('/theme/selected') . '/assets';
		$dirlen = strlen($dir);
		$component = 'Theme/' . ConfigHandler::Get('/theme/selected');
		$dh = Factory::Directory($dir);
		$cssls = $dh->ls('css', true);
		foreach($cssls as $obj){
			/** @var $obj FileLocal */
			$file = 'assets' . substr($obj->getFilename(), $dirlen);

			// Since this is a template, it may actually be in a different location than where the package maintainer put it.
			// ie: user template user/templates/pages/user/view.tpl may be installed to themes/myawesometheme/pages/user/view.tpl instead.
			$newobj = Factory::File($file);

			$cssfiles[$file] = array(
				'file' => $file,
				'obj' => $newobj,
				'component' => $component,
			);
		}

		// And the rest of the components... I suppose
		foreach($components as $c) {
			// Now, give me all this component's CSS files!
			$dir = $c->getAssetDir();
			if(!$dir) continue;

			$dirlen = strlen($dir);
			$component = $c->getName();

			$dh = Factory::Directory($dir);
			$cssls = $dh->ls('css', true);

			// not sure why getFilename(path) isn't working as expected, but this works too.
			foreach($cssls as $obj){
				/** @var $obj FileLocal */
				$file = 'assets' . substr($obj->getFilename(), $dirlen);

				// Since this is a template, it may actually be in a different location than where the package maintainer put it.
				// ie: user template user/templates/pages/user/view.tpl may be installed to themes/myawesometheme/pages/user/view.tpl instead.
				$newobj = Factory::File($file);

				$cssfiles[$file] = array(
					'file' => $file,
					'obj' => $newobj,
					'component' => $component,
				);
			}
		}


		$view->assign('pages', $pagetemplates);
		$view->assign('css', $cssfiles);
		$view->assign('themes', $themes);
		$view->title = 'Theme Manager';
	}

	/**
	 * Set a requested theme and template as default for the site.
	 */
	public function setdefault(){
		$request = $this->getRequest();
		$view = $this->getView();

		$themename = $this->getRequest()->getParameter(0);
		$template = $this->getRequest()->getParameter('template');

		// If the browser prefers JSON data, send that.
		if($request->prefersContentType(View::CTYPE_JSON)){
			$view->contenttype = View::CTYPE_JSON;
		}

		// Validate the theme name
		if(!Theme::ValidateThemeName($themename)){
			\Core\set_message('Invalid theme requested', 'error');
			\Core\go_back();
		}

		$theme = Theme::GetTheme($themename);


		if($template){
			// The template itself can be ignored.
			if(!Theme::ValidateTemplateName($themename, $template)){
				\Core\set_message('Invalid template requested', 'error');
				\Core\go_back();
			}
		}
		else{
			// and the default one is used otherwise.
			$allskins = $theme->getSkins();
			$template = $allskins[0]['file'];
		}


		if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::GetCurrentSiteID()){
			$config_default  = ConfigHandler::GetConfig('/theme/default_template');
			$config_selected = ConfigHandler::GetConfig('/theme/selected');

			if($config_default->get('overrideable') == 0){
				// It's a child site and the admin never gave them permission to change default themes!
				\Core\set_message('Unable to set the default template on a child site, please ensure that the "/theme/default_template" config is set to be overrideable!', 'error');
				\Core\go_back();
			}
			if($config_selected->get('overrideable') == 0){
				// It's a child site and the admin never gave them permission to change default themes!
				\Core\set_message('Unable to set the selected theme on a child site, please ensure that the "/theme/selected" config is set to be overrideable!', 'error');
				\Core\go_back();
			}
		}

		if($request->isPost()){

			if($themename != ConfigHandler::Get('/theme/selected')){
				// The theme changed, change the admin skin too!
				ConfigHandler::Set('/theme/default_admin_template', $template);
				// And the email skin.
				ConfigHandler::Set('/theme/default_email_template', '');
			}
			ConfigHandler::Set('/theme/default_template', $template);
			ConfigHandler::Set('/theme/selected', $themename);

			// reinstall theme and zee assets
			$t = Theme::GetTheme();

			$change = $t->reinstall(0);

			\Core\set_message('Updated default theme', 'success');
			\Core\redirect('/theme');
		}

		$view->assign('theme', $themename);
		$view->assign('template', $template);
	}

	/**
	 * Set a given skin for default use on admin pages.
	 *
	 * Will NOT affect the theme selected.
	 */
	public function setadmindefault(){
		$request = $this->getRequest();
		$view = $this->getView();

		$theme = $this->getRequest()->getParameter(0);
		$template = $this->getRequest()->getParameter('template');

		// If the browser prefers JSON data, send that.
		if($request->prefersContentType(View::CTYPE_JSON)){
			$view->contenttype = View::CTYPE_JSON;
		}

		// Validate
		if(!Theme::ValidateThemeName($theme)){
			\Core\set_message('Invalid theme requested', 'error');
			\Core\go_back();
		}

		if(!Theme::ValidateTemplateName($theme, $template)){
			\Core\set_message('Invalid template requested', 'error');
			\Core\go_back();
		}

		if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::GetCurrentSiteID()){
			$config_default  = ConfigHandler::GetConfig('/theme/default_admin_template');

			if($config_default->get('overrideable') == 0){
				// It's a child site and the admin never gave them permission to change default themes!
				\Core\set_message('Unable to set the default template on a child site, please ensure that the "/theme/default_template" config is set to be overrideable!', 'error');
				\Core\go_back();
			}
		}

		if($request->isPost()){

			if($theme != ConfigHandler::Get('/theme/selected')){
				\Core\set_message('The admin skin must be on the same theme as the site!', 'error');
				\Core\go_back();
			}
			ConfigHandler::Set('/theme/default_admin_template', $template);

			\Core\set_message('Updated admin skin', 'success');
			\Core\go_back();
		}
		else{
			return View::STATUS_BADREQUEST;
		}
	}


	/**
	 * Set a given skin for default use on email communications.
	 *
	 * Will NOT affect the theme selected.
	 */
	public function setemaildefault(){
		$request = $this->getRequest();
		$view = $this->getView();

		$theme = $this->getRequest()->getParameter(0);
		$template = $this->getRequest()->getParameter('template');

		// If the browser prefers JSON data, send that.
		if($request->prefersContentType(View::CTYPE_JSON)){
			$view->contenttype = View::CTYPE_JSON;
		}

		// Validate
		if(!Theme::ValidateThemeName($theme)){
			\Core\set_message('Invalid theme requested', 'error');
			\Core\go_back();
		}

		if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::GetCurrentSiteID()){
			$config_default  = ConfigHandler::GetConfig('/theme/default_email_template');

			if($config_default->get('overrideable') == 0){
				// It's a child site and the admin never gave them permission to change default themes!
				\Core\set_message('Unable to set the default template on a child site, please ensure that the "/theme/default_email_template" config is set to be overrideable!', 'error');
				\Core\go_back();
			}
		}

		if($request->isPost()){

			if($theme != ConfigHandler::Get('/theme/selected')){
				\Core\set_message('The admin skin must be on the same theme as the site!', 'error');
				\Core\go_back();
			}
			ConfigHandler::Set('/theme/default_email_template', $template);

			\Core\set_message('Updated email skin', 'success');
			\Core\go_back();
		}
		else{
			return View::STATUS_BADREQUEST;
		}
	}

	/**
	 * Page to display a user interface to select the optional stylesheets.
	 */
	public function selectstylesheets(){
		$request = $this->getRequest();
		$view = $this->getView();

		$file = $request->getParameter('template');
		$tpl = Template::Factory($file);
		$stylesheets = $tpl->getOptionalStylesheets();

		$form = new Form();
		foreach($stylesheets as $style){
			$model = TemplateCssModel::Construct($file, $style['src']);
			if(!$model->exists() && isset($style['default']) && $style['default']){
				$model->set('enabled', 1);
			}

			$form->addElement(
				'checkbox',
				[
					'title' => $style['title'],
					'name' => 'stylesheets[]',
					'value' => $style['src'],
					'checked' => $model->get('enabled'),
				]
			);
		}
		$form->addElement('submit', ['name' => 'submit', 'value' => 'Update Stylesheets']);


		// If it was a POST... then save that and go back.
		if($request->isPost()){
			if(!isset($_POST['stylesheets'])) $_POST['stylesheets'] = array();
			// Run through the stylesheets retrieved and save their setting.
			foreach($stylesheets as $style){
				$model = TemplateCssModel::Construct($file, $style['src']);
				$model->set('enabled', (in_array($style['src'], $_POST['stylesheets']) ? 1 : 0));
				$model->save();
			}

			\Core\set_message('Updated optional stylesheets successfully', 'success');
			\Core\go_back(1);
		}


		//$view->addBreadcrumb('Theme Manager', '/theme');
		$view->title = 'Select Optional Stylesheets';
		$view->assign('file', $file);
		$view->assign('form', $form);
	}

	/**
	 * Helper function for the setdefault method.
	 * @param $message
	 */
	private function _sendError($message){
		$request = $this->getRequest();
		$view = $this->getView();

		if($request->prefersContentType(View::CTYPE_JSON)){
			$view->jsondata = array('message' => $message, 'status' => 0);
		}
		else{
			\Core\set_message($message, 'error');
			\core\redirect('/theme');
		}
	}

	public static function SaveSiteSkins(Form $form){
		$theme = Theme::GetTheme();
		$siteDefault = null;
		$adminDefault = null;

		foreach($theme->getSkins() as $skin) {
			if($skin['default']) {
				$siteDefault = $skin['file'];
			}
			elseif($skin['admindefault']) {
				$adminDefault = $skin['file'];
			}
		}

		foreach($form->getElements() as $el){
			/** @var FormElement $el */

			$n = $el->get('name');
			if(strpos($n, 'config[') === 0) {
				$k = substr($n, 7, -1);
				$v = $el->getValue();

				// If the value is empty, remap this to the default for this theme.
				if($v === '') {
					switch($k) {
						case '/theme/default_admin_template':
							$v = $adminDefault;
							break;
						case '/theme/default_email_template':
							$v = '';
							break;
						default:
							$v = $siteDefault;
					}
				}

				ConfigHandler::Set($k, $v);
			}
		}

		return true;
	}
}






