<?php
/**
 * File for the class UserGroupAdminController
 *
 * @package Core\User
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\Controller;
use Core\Datamodel\DatasetWhereClause;
use Core\Forms\CheckboxesInput;
use Core\Forms\Form;
use Core\Forms\FormGroup;
use Core\Forms\TabsGroup;
use Core\I18N\I18NLoader;
use Core\Core;
use Core\Dict;
use Core\ModelFactory;
use Core\User;
use Core\View;
use Core\NotFoundException;
use Core\ListingTable\Table;
use Core\ViewControl;


/**
 * Group admin controller for the User system.
 *
 * @core-access-string p:/user/groups/manage
 * 
 * @package Core\User
 */
class UserGroupAdminController extends Controller{
	public function index(){

		$view = $this->getView();

		$canManagePerms = \Core\user()->checkAccess('p:/user/permissions/manage');
		
		$listing = new Table();
		$listing->setModelName('UserGroupModel');
		
		if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::IsEnabled()){
			$listing->addColumnKey('site');
		}
		$listing->addColumnKey('name');
		$listing->addColumnKey('description');
		if($canManagePerms) {
			$listing->addColumnKey('permissions');
		}
		
		$listing->loadFromRequest();
		
		/*
		
		$factory = new ModelFactory('UserGroupModel');

		if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::IsEnabled()){
			if(MultiSiteHelper::GetCurrentSiteID()){
				// Child site, only display global and site-specific sites.
				$w = new DatasetWhereClause();
				$w->setSeparator('or');
				$w->addWhere('site = ' . MultiSiteHelper::GetCurrentSiteID());
				$w->addWhere('site = -1');
				$factory->where($w);

				$displayglobal = true;
				$multisite = false;
			}
			else {
				// Root site, display all groups across all sites.
				$factory->where('site != -2');
				$displayglobal = false;
				$multisite = true;
			}
			$site = MultiSiteHelper::GetCurrentSiteID();
		}
		else{
			$displayglobal = false;
			$multisite = false;
			$site = null;
		}

		$factory->order('name');
		$groups = $factory->get();*/
		
		$view->addControlTLI('t:STRING_CREATE_USER_GROUP', '/usergroupadmin/create', ViewControl::ICON_ADD);
		
		$view->title = 't:STRING_USER_GROUP_ADMIN';
		$view->assign('listing', $listing);
	}

	public function create(){
		$view           = $this->getView();
		$req            = $this->getRequest();
		$model          = new UserGroupModel();
		$canManagePerms = \Core\user()->checkAccess('p:/user/permissions/manage');

		
		$form  = Form::BuildFromModel($model);
		
		if($canManagePerms) {
			$this->_addPermissionsToForm($form, $model);
		}
		
		$form->set('callsmethod', 'UserGroupAdminController::_UpdateFormHandler');

		$form->addElement('submit', ['value' => 't:STRING_SAVE']);

		$view->mastertemplate = 'admin';
		$view->templatename = 'pages/usergroupadmin/create_update.tpl';
		$view->addBreadcrumb('t:STRING_USER_GROUP_ADMIN', '/usergroupadmin');
		$view->title = 't:STRING_CREATE_USER_GROUP';
		$view->assign('form', $form);
	}

	public function update(){
		$view           = $this->getView();
		$req            = $this->getRequest();
		$model          = new UserGroupModel($req->getParameter(0));
		$canManagePerms = \Core\user()->checkAccess('p:/user/permissions/manage');
		
		if(!$model->checkExists()) {
			throw new NotFoundException();
		}

		$form  = Form::BuildFromModel($model);
		
		if($canManagePerms) {
			$this->_addPermissionsToForm($form, $model);
		}
		
		$form->set('callsmethod', 'UserGroupAdminController::_UpdateFormHandler');

		$form->addElement('submit', ['value' => 't:STRING_SAVE']);

		$view->mastertemplate = 'admin';
		$view->templatename = 'pages/usergroupadmin/create_update.tpl';
		$view->addBreadcrumb('t:STRING_USER_GROUP_ADMIN', '/usergroupadmin');
		$view->title = $model->name;
		$view->assign('form', $form);
	}

	/**
	 * @core-request-method POST
	 */
	public function delete(){
		$view  = $this->getView();
		$req   = $this->getRequest();
		$id    = $req->getParameter(0);
		$model = new UserGroupModel($id);

		if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::IsEnabled()){
			$where['site'] = MultiSiteHelper::GetCurrentSiteID();
		}

		$model->delete();
		\Core\set_message('Removed group successfully', 'success');
		\core\redirect('/usergroupadmin');
	}
	
	/**
	 * Admin page to render a report view of the current groups and all users attached to those groups.
	 */
	public function report(){
		$view = $this->getView();

		$permissionmanager = \Core\user()->checkAccess('p:/user/permissions/manage');
		
		$factory = new ModelFactory('UserGroupModel');

		if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::IsEnabled()){
			if(MultiSiteHelper::GetCurrentSiteID()){
				// Child site, only display global and site-specific sites.
				$w = new DatasetWhereClause();
				$w->setSeparator('or');
				$w->addWhere('site = ' . MultiSiteHelper::GetCurrentSiteID());
				$w->addWhere('site = -1');
				$factory->where($w);

				$displayglobal = true;
				$multisite = false;
			}
			else {
				// Root site, display all groups across all sites.
				$factory->where('site != -2');
				$displayglobal = false;
				$multisite = true;
			}
			$site = MultiSiteHelper::GetCurrentSiteID();
		}
		else{
			$displayglobal = false;
			$multisite = false;
			$site = null;
		}

		$factory->order('name');
		$groups = $factory->get();
		$data = [];
		
		// For each group, give me all the users attached to that group.
		foreach($groups as $g){
			/** @var $g UserGroupModel */
			
			$uugs = $g->getLink('UserUserGroup');
			$users = [];
			foreach($uugs as $uug){
				$users[] = $uug->getLink('User');
			}
			
			$data[] = [
				'group' => $g,
				'users' => $users
			];
		}
		
		$view->title = 't:STRING_USER_GROUP_REPORT';
		$view->assign('data', $data);
	}

	/**
	 * Set the site permissions to a given Form object.
	 *
	 * Used by the create and update pages.
	 *
	 * @param Form $form
	 * @param UserGroupModel $model
	 */
	private function _addPermissionsToForm(Form $form, UserGroupModel $model){
		$everything = new TabsGroup([
			'title' => 't:STRING_PERMISSIONS',
		]);
		
		// Flatten the permissions to a simple array
		$allPermissions = Core::GetPermissions();
		$setPermissions = $model->getPermissions();
		$sections = [];
		$groups = [];
		// Sort'em for pretty output.
		sort($allPermissions);
		
		foreach($allPermissions as $perm){
			$exploded = explode('/', substr($perm, 1));
			$title = 't:STRING_PERMISSION_' . strtoupper(implode('_', $exploded));
			$section = array_shift($exploded);
			$group = array_shift($exploded);

			if(!isset($sections[$section])) {
				$sections[$section] = [
					'title' => 't:STRING_PERMISSION_' . strtoupper($section),
					'groups' => []
				];
			}
			
			if(!isset($sections[$section]['groups'][ $group ])){
				$sections[$section]['groups'][ $group ] = [
					'title' => 't:STRING_PERMISSION_' . strtoupper($section) . '_' . strtoupper($group),
					'options' => new Dict(),
					'value' => [],
				];
				$sections[$section]['groups'][ $group ]['options']->autoI18N = true;
			}

			$sections[$section]['groups'][ $group ]['options']->addItem($perm, $title);
			
			if(in_array($perm, $setPermissions)){
				$sections[$section]['groups'][ $group ]['value'][] = $perm;
			}
		}


		$everything->addElement('message', ['name' => 'permissions-howto', 'message_type' => 'tutorial', 'value' => 't:MESSAGE_USER_GROUP_PERMISSIONS_HOWTO']);


		
		// Now, `$groups` contains all the sorted permissions and the associated values for each group.
		foreach($sections as $sectionName => $section) {
			$sectionGroup = new FormGroup(['title' => $section['title']]);
			$everything->addElement($sectionGroup);

			foreach($section['groups'] as $name => $dat){
				$sectionGroup->addElement(
					new CheckboxesInput([
						'title'   => $dat['title'],
						'name'    => 'permissions[' . \Core\str_to_url($sectionName . '_' . $name) . '][]',
						'options' => $dat['options'],
						'value'   => $dat['value'],
						'parent'  => $model,
					])
				);
			}
		}

		
		$form->addElement($everything);
	}

	public static function _UpdateFormHandler(Form $form){
		/** @var UserGroupModel $model */
		$model = $form->getModel();

		//var_dump($form->getElements(), $model->asArray()); die();

		if(\Core\user()->checkAccess('p:/user/permissions/manage')){
			// Merge all the various permission groups together to set back on the Model.
			$perms = [];
			$elements = $form->getElementsByName('permissions\[.*\]');
			foreach($elements as $el){
				if( is_array($el->get('value')) ) {
					$perms = array_merge($perms, $el->get('value'));
				}
			}
			$model->setPermissions($perms);
		}

		if($model->get('context') != ''){
			// Non-global context groups can never be default!
			$model->set('default', 0);
		}

		$model->save();

		return '/usergroupadmin';
	}
}

