<?php

/**
 * Widget controller, management interface for widgets and the like.
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\AccessDeniedException;
use Core\ConfigHandler;
use Core\Core;
use Core\Controller;
use Core\Datamodel\DatasetWhereClause;
use Core\Filestore\Directory;
use Core\Filestore\Factory;
use Core\Filestore\File;
use Core\Forms\Form;
use Core\ListingTable\Table;
use Core\RedirectException;
use Core\Templates\Template;
use Core\Theme;
use Core\View;
use Core\Widget;
use Core\ModelFactory;
use Core\NotFoundException;


class WidgetController extends Controller {

	/**
	 * Display a listing of all widgets registered in the system.
	 * 
	 * @core-accessstring p:/core/widgets/manage
	 */
	public function admin(){
		$view = $this->getView();
		$request = $this->getRequest();

		// Build a list of create pages for all registered components.
		$components    = Core::GetComponents();
		$pages         = [];
		$skins         = [];
		$selected      = null;
		$selectedtype  = null;
		$baseurl       = null;
		$selectoptions = ['' => '-- Select Template --'];
		$links         = [];
		$theme         = Theme::GetTheme();
		$formtheme     = null;
		$formskin      = null;
		$formtemplate  = null;
		$pageSources   = $components;

		// Add the theme to the page sources.
		$pageSources[] = $theme;


		foreach($pageSources as $c){
			if($c instanceof Theme) {
				$viewdir = $c->getBaseDir() . 'pages/';
				$dirlen = strlen($c->getBaseDir());
				$cname = 'Theme/' . $c->getName();
			}
			elseif($c instanceof \Core\Component) {
				$viewdir = $c->getBaseDir() . 'templates/';
				$dirlen = strlen($viewdir);
				$cname = $c->getName();
			}

			$dh = Factory::Directory($viewdir);
			$pagetplfiles = $dh->ls(array_keys(Template::$Renderers), true);

			// not sure why getFilename(path) isn't working as expected, but this works too.
			foreach($pagetplfiles as $obj){

				/** @var $obj File */
				$file = substr($obj->getFilename(), $dirlen);

				// Since this is a template, it may actually be in a different location than where the package maintainer put it.
				// ie: user template user/templates/pages/user/view.tpl may be installed to themes/myawesometheme/pages/user/view.tpl instead.
				$tpl = Template::Factory($file);

				if($tpl->hasWidgetAreas()){
					$pagetitle = $file;

					if(strpos($pagetitle, 'pages/') === 0){
						$pagetitle = substr($pagetitle, 6);
					}

					// And trim off the extension
					$pagetitle = substr($pagetitle, 0, strrpos($pagetitle, '.'));

					$pagefile = $pagetitle;

					// Replace directory slashes with a space
					$pagetitle = str_replace(['/', '-'], ' ', $pagetitle);

					// Capitalize them
					$pagetitle = ucwords($pagetitle);

					// Provide the component that sourced this file and some other meta info.
					$pagetitle = $pagefile . ' - ' . $pagetitle . ' (provided in ' . $cname . ')';

					$pages[$file] = $pagetitle;
				}
			}
			
			$links = array_merge($links, $c->getWidgetCreatesDefined());
		}

		// Build the array of skins for the current theme
		$themeskins  = $theme->getSkins();
		$defaultskin = null;
		foreach($themeskins as $dat){

			if($dat['is_system']) {
				continue;
			}

			$skins[ 'skins/' . $dat['file'] ] = $dat['title'];

			if($dat['default']){
				$defaultskin = 'skins/' . $dat['file'];
			}
		}

		// Now that the various templates have been loaded into a flat array, I need to sort them.
		asort($pages);
		asort($skins);

		foreach($skins as $k => $v){
			$selectoptions[ $k ] = 'Skin: ' . $v;
		}
		foreach($pages as $k => $v){
			$selectoptions[ $k ] = 'Page: ' . $v;
		}
		
		
		$factory = new ModelFactory('WidgetInstanceModel');
		$factory->order('weight');
		if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::IsEnabled()){
			$factory->whereGroup('or', ['site = -1', 'site = ' . MultiSiteHelper::GetCurrentSiteID()]);
		}
		

		if($request->getParameter('baseurl')){
			// It's a URL-specific request, lookup which template that page used last.
			$baseurl  = $request->getParameter('baseurl');
			$page     = PageModel::Construct($baseurl);
			
			if(!$page->checkExists()) {
				throw new NotFoundException();
			}
			
			//$selected = $page->last_template;

			$selected = $page->theme_template;
			if(!$selected) {
				$selected = ConfigHandler::Get('/theme/default_template');
			}
			$selected = '/skins/' . $selected;

			$requested_param = 'baseurl';
			$requested_value = $baseurl;
			$requested_label = $page->title . ' (' . $page->rewriteurl . ')';
			$factory->where('page_baseurl = ' . $baseurl);
			
			$view->addBreadcrumb($page->title, $page->baseurl);
			$view->title = 't:STRING_PAGE_WIDGETS';
			
			$selectedtype = 'url';
			$formtemplate = $selected;
		}
		elseif($request->getParameter('template')){
			$selected = $request->getParameter('template');
			
			// Ensure that it does not start with a '/'.
			// This is because the query is without one.
			if($selected[0] == '/'){
				$selected = substr($selected, 1);
			}
			
			$factory->where('template = ' . $selected);

			if(isset($pages[ $selected ])){
				$selectedtype = 'page';
				$formtemplate = $selected;
			}
			else{
				$selectedtype = 'skin';
				$formtheme = $theme->getKeyName();
				$formskin  = $selected;
			}
			
			$requested_param = 'template';
			$requested_value = $selected;
			$requested_label = $selectedtype . ' ' . $selected;
		}
		else{
			// Just use the default theme skin.
			$selected = '';
			$selectedtype = 'skin';
			$formtheme = $theme->getKeyName();
			$formskin  = $selected;
			$requested_param = 'template';
			$requested_value = $defaultskin;
			$requested_label = 'Default Site Skin';
		}


		$installables = [0 => ''];


		if($selected) {
			$template     = Template::Factory($selected);
			$areas        = $template->getWidgetAreas();
			foreach($areas as $k => $dat){
				// Ensure that each area has a widgets array, (even if it's empty)
				$areas[$k]['widgets'] = [];
				$installables[] = $dat['installable'];
			}

			foreach($factory->get() as $wi){
				/** @var $wi WidgetInstanceModel */

				$a = $wi->get('widgetarea');
				$areas[$a]['widgets'][] = $wi;
			}
		}

		$installables = array_unique($installables);
		$available = WidgetModel::Find(['installable IN ' . implode(', ', $installables)]);


		$view->mastertemplate = 'admin';
		//$view->title = 'All Widgets';
		//$view->assign('table', $table);
		$view->assign('available_widgets', $available);
		$view->assign('links', $links);
		$view->assign('manager', true);
		$view->assign('theme', $formtheme);
		$view->assign('skin', $formskin);
		$view->assign('template', $selected);
		$view->assign('page_template', $formtemplate);
		$view->assign('page_baseurl', $baseurl);
		$view->assign('options', $selectoptions);
		$view->assign('selected', $selected);
		$view->assign('areas', $areas);
		$view->assign('requested_param', $requested_param);
		$view->assign('requested_value', $requested_value);
		$view->assign('requested_label', $requested_label);
		//$view->assign('multisite', $ms);
	}

	/**
	 * Create a simple widget with the standard settings configurations.
	 */
	public function create(){
		$view = $this->getView();
		$request = $this->getRequest();

		// Only widget managers can create widgets.
		if(!\Core\user()->checkAccess('p:/core/widgets/manage')){
			return View::STATUS_ACCESSDENIED;
		}

		$class = $request->getParameter('class');

		// If it doesn't end in "widget", it should!
		if(stripos($class, 'widget') != strlen($class) - 6){
			$class .= 'Widget';
		}

		if(!class_exists($class)){
			\Core\set_message('t:MESSAGE_ERROR_CORE_WIDGET_S_NOT_FOUND_ON_SYSTEM', $class);
			\Core\go_back();
		}

		/** @var Widget $obj */
		$obj = new $class();

		if(!($obj instanceof \Core\Widget)){
			\Core\set_message('t:MESSAGE_ERROR_CORE_WIDGET_S_WRONG_PARENT_CLASS', $class);
			\Core\go_back();
		}

		if(!$obj->is_simple){
			\Core\set_message('t:MESSAGE_ERROR_CORE_WIDGET_S_NOT_SIMPLE_WIDGET', $class);
			\Core\go_back();
		}

		$form = new Form();
		$form->set('callsmethod', 'WidgetController::_CreateUpdateHandler');

		// Make the widget's "baseurl", which for simple widgets will be the widget class followed by a UUID.
		$baseurl = '/' . strtolower(substr($class, 0, -6)) . '/execute/';
		$baseurl .= \Core\generate_uuid();

		$form->addElement('system', ['name' => 'baseurl', 'value' => $baseurl]);
		
		$form->addElement(
			'text',
			[
				'name' => 'title',
				'required' => true,
				'title' => t('STRING_CORE_WIDGET_ADMIN_TITLE'),
				'description' => t('MESSAGE_CORE_WIDGET_ADMIN_TITLE')
			]
		);

		$defaults = $obj->settings;
		$formdata = $obj->getFormSettings();

		foreach($formdata as $dat){
			$type = $dat['type'];
			$name = $dat['name'];

			$dat['name'] = 'setting[' . $name . ']';

			$form->addElement($type, $dat);
		}
		
		// Managers can manage the edit permissions for this particular widget.
		// This allows managers to delegate edit permissions for particular areas on the site.
		$form->addElement(
			'access',
			[
				'name' => 'editpermissions',
				'title' => t('STRING_CORE_WIDGET_EDIT_PERMISSIONS'),
				'description' => t('MESSAGE_CORE_WIDGET_EDIT_PERMISSIONS'),
				'value' => '!*',
			]
		);


		$form->addElement('submit', ['value' => t('STRING_CORE_WIDGET_CREATE')]);

		$view->mastertemplate = 'admin';
		$view->addBreadcrumb('All Widgets', '/widget/admin');
		$view->title = t('STRING_CORE_WIDGET_CREATE');
		$view->assign('form', $form);
	}

	/**
	 * Create a simple widget with the standard settings configurations.
	 */
	public function update(){
		$view = $this->getView();
		$request = $this->getRequest();
		
		$manager = \Core\user()->checkAccess('p:/core/widgets/manage');

		if(!$manager){
			throw new AccessDeniedException();
		}

		$baseurl = $request->getParameter('baseurl');
		$controller = substr($baseurl, 1, strpos($baseurl, '/', 1)-1);
		$widgets = Core::GetWidgets();
		if(!isset($widgets[ $controller ])){
			// Widget not found in any enabled components.
			throw new NotFoundException();
		}
		$class = $widgets[ $controller ];


		/** @var Widget $obj */
		$obj = new $class();

		if(!($obj instanceof \Core\Widget)){
			\Core\set_message('t:MESSAGE_ERROR_CORE_WIDGET_S_WRONG_PARENT_CLASS', $class);
			\Core\go_back();
		}

		if(!$obj->is_simple){
			\Core\set_message('t:MESSAGE_ERROR_CORE_WIDGET_S_NOT_SIMPLE_WIDGET', $class);
			\Core\go_back();
		}

		$model = new WidgetModel($baseurl);
		
		if(!$model->exists()){
			throw new NotFoundException();
		}
		
		// Check that this user has permission to edit if they are not a manager.
		if(!$manager && !\Core\user()->checkAccess($model->get('editpermissions'))){
			throw new AccessDeniedException();
		}

		$form = new Form();
		$form->set('callsmethod', 'WidgetController::_CreateUpdateHandler');

		$form->addElement('system', ['name' => 'baseurl', 'value' => $baseurl]);

		if($manager) {
			$form->addElement(
			'text',
				[
					'name' => 'title',
					'required' => true,
					'value' => $model->title,
					'title' => t('STRING_CORE_WIDGET_ADMIN_TITLE'),
					'description' => t('MESSAGE_CORE_WIDGET_ADMIN_TITLE')
				]
			);

			$form->addElement(
				'select',
				[
					'name' => 'template',
					'required' => false,
					'value' => $model->template,
					'options' => $model->getAlternativeTemplateOptions(),
					'title' => t('STRING_CORE_WIDGET_TEMPLATE'),
					'description' => t('MESSAGE_CORE_WIDGET_TEMPLATE')
				]
			);
		}
		

		$defaults = $obj->settings;
		$formdata = $obj->getFormSettings();

		foreach($formdata as $dat){
			$type = $dat['type'];
			$name = $dat['name'];

			$dat['value'] = $model->getSetting($name) !== null ? $model->getSetting($name) : $defaults[$name];
			$dat['name'] = 'setting[' . $name . ']';

			$form->addElement($type, $dat);
		}
		
		if($manager){
			// Managers can manage the edit permissions for this particular widget.
			// This allows managers to delegate edit permissions for particular areas on the site.
			$form->addElement(
				'access',
				[
					'name' => 'editpermissions',
					'title' => t('STRING_CORE_WIDGET_EDIT_PERMISSIONS'),
					'description' => t('MESSAGE_CORE_WIDGET_EDIT_PERMISSIONS'),
					'value' => $model->get('editpermissions'),
				]
			);
		}


		$form->addElement('submit', ['value' => t('STRING_CORE_WIDGET_UPDATE')]);

		$view->mastertemplate = 'admin';
		$view->addBreadcrumb('All Widgets', '/widget/admin');
		$view->title = t('STRING_CORE_WIDGET_UPDATE');
		$view->assign('form', $form);
	}

	/**
	 * Delete a simple widget.
	 */
	public function delete(){
		$view = $this->getView();
		$request = $this->getRequest();

		if(!\Core\user()->checkAccess('p:/core/widgets/manage')){
			return View::STATUS_ACCESSDENIED;
		}

		if(!$request->isPost()){
			return View::STATUS_BADREQUEST;
		}

		$baseurl = $request->getParameter('baseurl');
		$baseurl = $request->getParameter('baseurl');
		if($baseurl[0] == '/'){
			// Trim off the beginning '/' from the URL.
			$class = substr($baseurl, 1, strpos($baseurl, '/', 1)-1) . 'widget';
		}
		else{
			$class = substr($baseurl, 0, strpos($baseurl, '/')) . 'widget';
		}
		
		if(!class_exists($class)){
			\Core\set_message('t:MESSAGE_ERROR_CORE_WIDGET_S_NOT_FOUND_ON_SYSTEM', $class);
			\Core\go_back();
		}

		/** @var Widget $obj */
		$obj = new $class();

		if(!($obj instanceof \Core\Widget)){
			\Core\set_message('Wrong parent class for [' . $class . '], it does not appear to be a Widget instance, invalid widget!', 'error');
			\Core\go_back();
		}

		if(!$obj->is_simple){
			\Core\set_message('Widget [' . $class . '] does not appear to be a simple widget.  Only simple widgets can be created via this page.', 'error');
			\Core\go_back();
		}

		$model = new WidgetModel($baseurl);

		$model->delete();
		\Core\set_message('Deleted widget ' . $model->get('title') . ' successfully!', 'success');
		\Core\go_back();
	}

	public function instances_save(){
		$view = $this->getView();
		$request = $this->getRequest();

		if(!\Core\user()->checkAccess('p:/core/widgets/manage')){
			return View::STATUS_ACCESSDENIED;
		}

		if(!$request->isPost()){
			return View::STATUS_BADREQUEST;
		}

		$counter = 0;
		$changes = ['created' => 0, 'updated' => 0, 'deleted' => 0];

		$selected = $_POST['selected'];
		// For the incoming options, I want an explicit NULL if it's empty.
		$theme    = $_POST['theme'] == '' ? null : $_POST['theme'];
		$skin     = $_POST['skin'] == '' ? null : $_POST['skin']; $_POST['skin'];
		$template = $_POST['template'] == '' ? null : $_POST['template'];
		$baseurl  = $_POST['page_baseurl'] == '' ? null : $_POST['page_baseurl'];

		foreach($_POST['widgetarea'] as $id => $dat){

			// Merge in the global information for this request
			//$dat['theme']         = $theme;
			//$dat['skin']          = $skin;
			$dat['template']      = $template;
			$dat['page_baseurl']  = $baseurl;

			$dat['weight'] = ++$counter;
			$dat['access'] = $dat['widgetaccess'];

			$w = WidgetModel::Construct($dat['baseurl']);
			$dat['site'] = $w->get('site');

			if(strpos($id, 'new') !== false){
				$w = new WidgetInstanceModel();
				$w->setFromArray($dat);
				$w->save();
				$changes['created']++;
			}
			elseif(strpos($id, 'del-') !== false){
				$w = new WidgetInstanceModel(substr($id, 4));
				$w->delete();
				// Reset the counter back down one notch since this was a deletion request.
				--$counter;
				$changes['deleted']++;
			}
			else{
				$w = new WidgetInstanceModel($id);
				$w->setFromArray($dat);
				if($w->save()) $changes['updated']++;
			}
		} // foreach($_POST['widgetarea'] as $id => $dat)

		// Display some human friendly status message.
		if($changes['created'] || $changes['updated'] || $changes['deleted']){
			$changetext = [];

			if($changes['created'] == 1) $changetext[] = 'One widget added';
			elseif($changes['created'] > 1) $changetext[] = $changes['created'] . ' widgets added';

			if($changes['updated'] == 1) $changetext[] = 'One widget updated';
			elseif($changes['updated'] > 1) $changetext[] = $changes['updated'] . ' widgets updated';

			if($changes['deleted'] == 1) $changetext[] = 'One widget deleted';
			elseif($changes['deleted'] > 1) $changetext[] = $changes['deleted'] . ' widgets deleted';

			\Core\set_message(implode('<br/>', $changetext), 'success');
		}
		else{
			\Core\set_message('t:MESSAGE_INFO_NO_CHANGES_PERFORMED');
		}

		if($baseurl){
			\Core\redirect($baseurl);
		}
		else{
			\Core\redirect('/widget/admin?selected=' . $selected);
		}
	}

	/**
	 * Controller view to install 1 widget into one selected area, be that area a skin, or page template.
	 *
	 * @return int
	 */
	public function instance_install() {
		$view = $this->getView();
		$request = $this->getRequest();

		if($request->isAjax()) {
			// This page supports AJAX calls
			$view->mode = View::MODE_AJAX;
			$view->contenttype = View::CTYPE_JSON;
		}

		if(!\Core\user()->checkAccess('p:/core/widgets/manage')) {
			throw new AccessDeniedException();
		}

		if(!$request->isPost()) {
			throw new \Core\BadRequestException();
		}

		// For the incoming options, I want an explicit NULL if it's empty.
		$template       = $request->getPost('template');
		$page_baseurl   = $request->getPost('baseurl');
		$widget_baseurl = $request->getPost('widget_baseurl');
		$widgetarea     = $request->getPost('area');

		$wm = WidgetModel::Construct($widget_baseurl);
		$dat = [];
		$dat['site']         = $wm->get('site');
		$dat['baseurl']      = $widget_baseurl;
		$dat['template']     = $template;
		$dat['page_baseurl'] = $page_baseurl;
		$dat['widgetarea']   = $widgetarea;


		if($template) {
			$counter = WidgetInstanceModel::Count(['template = ' . $template, 'widgetarea = ' . $widgetarea]);
			$w = new WidgetInstanceModel();
			$w->setFromArray($dat);
			$w->set('weight', $counter + 1);
			$w->save();

			\Core\set_message('Installed widget into requested template', 'success');
		}
		elseif($page_baseurl) {
			$counter = WidgetInstanceModel::Count(['page_baseurl = ' . $page_baseurl, 'widgetarea = ' . $widgetarea]);
			$w = new WidgetInstanceModel();
			$w->setFromArray($dat);
			$w->set('weight', $counter + 1);
			$w->save();

			\Core\set_message('Installed widget into requested page', 'success');
		}
		else {
			throw new Exception('Unknown request!');
		}

		if($request->isAjax()) {
			$view->jsondata = [
				'status' => 200,
				'location' => \Core\resolve_link('/widget/instance/update/' . $w->id),
			];
		}
		else {
			// Immediately drop to the display configuration for this widget;
			// this is usually the first thing requested when installing a widget.
			throw new RedirectException('/widget/instance/update/' . $w->id);
		}
	}

	/**
	 * Controller view to update any instance-specific options for a given template.
	 *
	 * Usually consists of just access permissions and display template, but more options could come in the future.
	 */
	public function instance_update(){
		$view = $this->getView();
		$request = $this->getRequest();

		if(!\Core\user()->checkAccess('p:/core/widgets/manage')){
			return View::STATUS_ACCESSDENIED;
		}

		$instance = WidgetInstanceModel::Construct($request->getParameter(0));
		if(!$instance->exists()){
			return View::STATUS_NOTFOUND;
		}
		
		$widget = $instance->getWidget();

		$form = new Form();
		$form->set('callsmethod', 'WidgetController::_InstanceHandler');

		$form->addModel($instance);
		
		// Add any of the widget display settings that may be set in the Widget.
		foreach($widget->displaySettings as $idx => $setting){
			// Pull the type of the form element from the array, or text as the default.
			$type = $setting['type'] ?: 'text';
			$name = $setting['name'] ?: $idx;
			
			// Remap the name to a name that the form handler will know about.
			$setting['name'] = 'display_setting[' . $name . ']';
			
			if($type == 'checkbox'){
				// This one requires some additional work;
				// The value is true or false, but the attribute to set is 'checked'.
				if($setting['value'] === 1 || $setting['value'] === '1' || $setting['value'] === true){
					$setting['checked'] = true;
				}
				else{
					$setting['checked'] = false;
				}
				
				$setting['value'] = '1';
			}
			
			$form->addElement($type, $setting);
		}
		
		$form->addElement('submit', ['value' => 'Save Options']);

		$view->mastertemplate = 'admin';
		$view->addBreadcrumb('All Widgets', '/widget/admin');
		$view->title = 'Update Installed Options';
		$view->assign('form', $form);
	}

	/**
	 * Controller view to update any instance-specific options for a given template.
	 *
	 * Usually consists of just access permissions and display template, but more options could come in the future.
	 */
	public function instance_remove(){
		$view = $this->getView();
		$request = $this->getRequest();

		if($request->isAjax()) {
			// This page supports AJAX calls
			$view->mode = View::MODE_AJAX;
			$view->contenttype = View::CTYPE_JSON;
		}

		if(!\Core\user()->checkAccess('p:/core/widgets/manage')) {
			throw new AccessDeniedException();
		}

		$instance = WidgetInstanceModel::Construct($request->getParameter(0));
		if(!$instance->checkExists()) {
			throw new NotFoundException();
		}

		if(!$request->isPost()) {
			throw new \Core\BadRequestException();
		}

		$otherCriteria = [
			'site = ' . $instance->get('site'),
			'template = ' . ($instance->get('template') === null ? 'NULL' : $instance->get('template')),
			'page_baseurl = ' . ($instance->get('page_baseurl') === null ? 'NULL' : $instance->get('page_baseurl')),
			'widgetarea = ' . $instance->get('widgetarea'),
		];

		$instance->delete();

		// Reshuffle the rest of the widgets to ensure that they are ordered correctly.
		$other = WidgetInstanceModel::Find($otherCriteria, null, 'weight');
		$x = 0;
		foreach($other as $o){
			/** @var WidgetInstanceModel $o */
			$o->set('weight', ++$x);
			$o->save();
		}

		if($request->isAjax()) {
			$view->jsondata = [
				'status' => 200,
				'message' => 'Removed widget',
			];
		}
		else {
			throw new RedirectException('BACK');
		}
	}

	/**
	 * Controller view to update any instance-specific options for a given template.
	 *
	 * Usually consists of just access permissions and display template, but more options could come in the future.
	 */
	public function instance_moveup(){
		$view = $this->getView();
		$request = $this->getRequest();

		if(!\Core\user()->checkAccess('p:/core/widgets/manage')){
			return View::STATUS_ACCESSDENIED;
		}

		$instance = WidgetInstanceModel::Construct($request->getParameter(0));
		if(!$instance->exists()){
			return View::STATUS_NOTFOUND;
		}

		if(!$request->isPost()){
			return View::STATUS_BADREQUEST;
		}

		// Figure out which instance is this one -1.
		$otherCriteria = [
			'site = ' . $instance->get('site'),
			'template = ' . ($instance->get('template') === null ? 'NULL' : $instance->get('template')),
			'page_baseurl = ' . ($instance->get('page_baseurl') === null ? 'NULL' : $instance->get('page_baseurl')),
			'widgetarea = ' . $instance->get('widgetarea'),
			'weight = ' . ($instance->get('weight') - 1),
		];
		$other = WidgetInstanceModel::Find($otherCriteria, 1);

		if(!$other){
			\Core\set_message('Widget is already in the top position!', 'error');
		}
		else{
			$other->set('weight', $other->get('weight') + 1);
			$instance->set('weight', $instance->get('weight') - 1);

			$other->save();
			$instance->save();
		}

		\Core\go_back();
	}

	/**
	 * Controller view to update any instance-specific options for a given template.
	 *
	 * Usually consists of just access permissions and display template, but more options could come in the future.
	 */
	public function instance_movedown(){
		$view = $this->getView();
		$request = $this->getRequest();

		if(!\Core\user()->checkAccess('p:/core/widgets/manage')){
			return View::STATUS_ACCESSDENIED;
		}

		$instance = WidgetInstanceModel::Construct($request->getParameter(0));
		if(!$instance->exists()){
			return View::STATUS_NOTFOUND;
		}

		if(!$request->isPost()){
			return View::STATUS_BADREQUEST;
		}

		// Figure out which instance is this one -1.
		$otherCriteria = [
			'site = ' . $instance->get('site'),
			'template = ' . ($instance->get('template') === null ? 'NULL' : $instance->get('template')),
			'page_baseurl = ' . ($instance->get('page_baseurl') === null ? 'NULL' : $instance->get('page_baseurl')),
			'widgetarea = ' . $instance->get('widgetarea'),
			'weight = ' . ($instance->get('weight') + 1),
		];
		$other = WidgetInstanceModel::Find($otherCriteria, 1);

		if(!$other){
			\Core\set_message('Widget is already in the bottom position!', 'error');
		}
		else{
			$other->set('weight', $other->get('weight') - 1);
			$instance->set('weight', $instance->get('weight') + 1);

			$other->save();
			$instance->save();
		}

		\Core\go_back();
	}


	public static function _CreateUpdateHandler(Form $form){
		$baseurl = $form->getElement('baseurl')->get('value');

		$model = new WidgetModel($baseurl);
		$model->set('editurl', '/widget/update?baseurl=' . $baseurl);
		$model->set('deleteurl', '/widget/delete?baseurl=' . $baseurl);
		
		if(\Core\user()->checkAccess('p:/core/widgets/manage')){
			// Managers have additional permissions to edit advanced fields.
			$model->set('title', $form->getElementValue('title'));
			$model->set('template', $form->getElementValue('template'));
			$model->set('editpermissions', $form->getElementValue('editpermissions'));
		}
		

		$elements = $form->getElements();
		foreach($elements as $el){
			/** @var FormElement $el */
			if(strpos($el->get('name'), 'setting[') === 0){
				$name = substr($el->get('name'), 8, -1);
				$model->setSetting($name, $el->get('value'));
			}
		}
		$model->save();
		
		\Core\set_message('t:MESSAGE_SUCCESS_CORE_WIDGET_S_UPDATED', $model->get('title'));

		return 'back';
	}

	public static function _InstanceHandler(Form $form){
		$instance = $form->getModel();
	
		// Set any display settings that may be present.
		$displaySettings = $form->getElementsByName('display_setting\[.*\]');
		$displaySettingValues = [];
		foreach($displaySettings as $el){
			
			$name = substr($el->get('name'), 16, -1);
			$val  = $el->get('value');
			$displaySettingValues[$name] = $val;
		}
		$instance->set('display_settings', $displaySettingValues);
		
		$instance->save();

		return 'back';
	}
}

