<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core;

use Exception;
use PageModel;
use ReflectionClass;
use ReflectionMethod;

/**
 * Mapping utility to resolve a URI to its underlying components, including controller, method, and any parameters.
 *
 * @author charlie
 */
class BaseURLMap {

	/**
	 * @var string Original URL that was requested, primarily for reference in case something blows up
	 */
	private $_original;
	
	/**
	 * @var ReflectionClass Contains the Controller class that is resolved for this URL, if set
	 */
	private $_controller;
	
	/**
	 * @var string URL-friendly path of the Controller
	 */
	private $_controllerName;
	
	/**
	 * @var ReflectionMethod Contains the Method that is resolved for this URL, if set
	 */
	private $_method;
	
	/**
	 * @var string URL-friendly path of the Method
	 */
	private $_methodName;
	
	/**
	 * @var array Any positional arguments present in the URI
	 */
	private $_parameters = [];

	/**
	 * @var array Any named (GET) parameters present in the URI
	 */
	private $_namedParameters = [];
	
	/**
	 * @var string rewriteURL that is mapped to for this URI
	 */
	private $_rewrite;
	
	/**
	 * @var int 200 if found and 4xx otherwise
	 */
	private $_status;

	/**
	 * @var PageModel The corresponding PageModel for linked URLs, may be null if not linked or dynamic.
	 */
	private $_model;

	/**
	 * Do a thing
	 *
	 * BaseURLMap constructor.
	 * @param string $uri
	 */
	public function __construct(string $uri){

		$this->_original = $uri;
		
		if(strpos($uri, '/') !== 0){
			// All URLs should start with a slash!
			$this->_status = View::STATUS_BADREQUEST;
			return;
		}
		
		
		// Check with the PageModel if this is an aliased (or rewriteURL), of a defined baseurl.
		// We need to check if the user supplied additional parameters, these should be preserved,
		// but ignored if necessary.
		$parameters = null;
		$data = null;
		try{
			$data = PageModel::GetRoutingDataByUri($uri);
		}
		catch (Exception $e){
			// If there are GET parameters on this URL, then strip them and retry.
			$needle = strpos($uri, '?');
			if($needle !== false) {
				$parameters = substr($uri, $needle + 1);
				$uri = substr($uri, 0, $needle);

				try {
					$data = PageModel::GetRoutingDataByUri($uri);
				}
				catch (Exception $e) {
					$data = null;
				}
			}
		}

		// The page base data should be available now, this includes:
		// site
		// baseurl
		// rewriteurl
		// fuzzy
		// match
		if($data) {
			try{
				$this->_model = PageModel::Construct($data['site'], $data['baseurl']);
				$base = $this->_model->baseurl;
				$this->_rewrite = $this->_model->rewriteurl;

				// Split out the parameters from the base or rewrite;
				// these will get appended so they are passable to the calling system.
				if(strpos($uri, $data['match']) === 0) {
					$params = substr($uri, strlen($data['match']));
				}
				elseif(strpos($uri, $this->_rewrite) === 0) {
					$params = substr($uri, strlen($this->_rewrite));
				}
				else {
					$params = '';
				}

				// Tack these onto the resolved paths, they'll get processed later.
				$base .= $params;
				$this->_rewrite .= $params;
			}
			catch (Exception $ex){
				$data = null;
			}
		}

		if($data === null) {
			// None of the above checks completed
			// this is typical for dynamic-only pages or when the system isn't ready yet.
			$this->_model = null;
			$base = $this->_rewrite = $uri;
		}

		
		// Explode parts of the path ('/'), as each part indicates a different bit of the puzzle.
		$parts = explode('/', substr($base, 1));
		if(sizeof($parts) === 0){
			$this->_status = View::STATUS_BADREQUEST;
			return;
		}
		
		if(sizeof($parts) > 6){
			$this->_status = View::STATUS_URITOOLARGE;
			return;
		}
		
		// Retrieve the list of controllers from Core
		$controllers = Core::GetControllers();
		
		if(!isset($controllers[ $parts[0] ])){
			$this->_status = View::STATUS_NOTFOUND;
			return;
		}
		
		$this->_controllerName = $parts[0];
		$class = $controllers[ $parts[0] ];
		array_shift($parts);
		if(!class_exists($class)){
			// Core doesn't have a registry entry for the class.
			$this->_status = View::STATUS_NOTFOUND;
			return;
		}

		try {
			// Use reflections to inspect this class to search for the method contained therein
			$this->_controller = new ReflectionClass($class);
		}
		catch(\ReflectionException $ex) {
			// The controller's class doesn't exist.
			// This will only happen for a misconfigured registry config,
			// where a Controller is defined, but does not actually exist.
			$this->_status = View::STATUS_NOTFOUND;
			return;
		}

		$methods = [];
		foreach($this->_controller->getMethods(ReflectionMethod::IS_PUBLIC) as $m){
			/** @var $m ReflectionMethod */
			if(
				!$m->isStatic() &&
				$m->class === $class &&
				preg_match('/^[a-z][A-Za-z_0-9]*$/', $m->name)
			){
				// Only add public instanced methods!
				// private and static methods are *NOT* to be called from URI paths.
				$methods[ strtolower($m->name) ] = $m;
			}
		}
		
		// Determine the best match for this given path.
		// This is necessary because the path may have positional arguments attached.
		do{
			$attempt = implode('_', $parts);
			
			if(isset($methods[ $attempt ])){
				// Valid entry point found!
				$this->_methodName = $attempt;
				$this->_method = $methods[ $attempt ];
				break;
			}
			
			// Next.
			if(count($parts)){
				// Append the last parameter onto the array if there is one
				$this->_parameters[] = array_pop($parts);
			}
		} while(sizeof($parts) > 0);
		
		// Nothing?... Try index instead.
		if($this->_method === null && isset($methods['index'])){
			$this->_methodName = 'index';
			$this->_method = $methods['index'];
		}
		
		if($this->_method === null){
			$this->_status = View::STATUS_NOTFOUND;
			return;
		}
		
		// Parameters will contain the positional parameters from the URL in reverse order.
		// These need to get remapped back to the Request and View since we're only discovering them just now.
		$this->_parameters = array_reverse($this->_parameters);

		if($parameters !== null) {
			// Additional GET parameters were detected; grab these and add them to the list.
			parse_str($parameters, $this->_namedParameters);
		}

		// Ensure the status is good
		$this->_status = View::STATUS_OK;
	}
	
	/**
	 * Get the baseurl corresponding to this uri
	 * 
	 * If MyAppController->index() is used, '/myapp' will be returned, (the index is dropped)
	 * Unless there are parameters, then the full URL will be used to distinguish parameters from the method.
	 * 
	 * @return string
	 */
	public function getBaseURL() : string {
		
		$parts = [];
		$parts[] = $this->_controllerName;
		
		if($this->_methodName !== 'index' || count($this->_parameters)){
			$parts[] = str_replace('_', '/', $this->_methodName);
		}
		
		$parts = array_merge($parts, $this->_parameters);
		
		return '/' . implode('/', $parts);
	}

	/**
	 * Get the baseURL corresponding to this URI
	 *
	 * Will simply return the baseurl followed by ?[params] if there are named parameters present.
	 *
	 * @return string
	 */
	public function getBaseURLWithNamedParameters() : string {
		$url = $this->getBaseURL();

		if(count($this->_namedParameters) > 0) {
			$url .= '?' . http_build_query($this->_namedParameters);
		}

		return $url;
	}
	
	/**
	 * Get the Controller object for this uri
	 * 
	 * @return ReflectionClass
	 */
	public function getController() : ReflectionClass {
		return $this->_controller;
	}
	
	/**
	 * Get the default template for this page based solely on the resolved baseurl and status code.
	 * 
	 * @return string
	 */
	public function getDefaultTemplateName() : string {
		if($this->_status === View::STATUS_OK){
			return 'pages/' . $this->_controllerName . '/' . $this->_methodName;
		}
		else{
			return 'pages/error/error';
		}
	}
	
	/**
	 * Get the Method for this baseurl
	 * 
	 * @return ReflectionMethod
	 */
	public function getMethod() : ReflectionMethod {
		return $this->_method;
	}
	
	/**
	 * Get any positional parameter in this request.
	 * 
	 * These are parts of the base URL and before the '?' string.
	 * 
	 * @return array
	 */
	public function getPositionalParameters() : array {
		return $this->_parameters;
	}
	
	/**
	 * Get the rewriteurl corresponding to this uri
	 *
	 * @throws Exception If something blows up, an exception is thrown.
	 * 
	 * @return string
	 */
	public function getRewriteURL() : string {
		if($this->_rewrite === null) {
			throw new Exception($this->_original . ' does not appear to have a resolvable URL!');
		}

		$url = $this->_rewrite;

		if(count($this->_namedParameters) > 0) {
			$url .= '?' . http_build_query($this->_namedParameters);
		}

		return $url;
	}
	
	/**
	 * Get the status code of this map, usually 200, 404, or one of the other applicable View status codes
	 * 
	 * @return int
	 */
	public function getStatus() : int {
		return $this->_status;
	}

	/**
	 * Get the corresponding PageModel for this mapping, if one is set.
	 *
	 * For dynamic pages, this will be NULL!
	 *
	 * @return PageModel|null
	 */
	public function getPageModel() : ?PageModel {
		return $this->_model;
	}
}
