<?php
/**
 * General common exceptions used throughout Core
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core;
use Exception;
use Throwable;

/**
 * Any exceptions that need to trigger a Response Status code to the user agent
 */
interface ViewStatusExceptionInterface extends Throwable { }

/**
 * Any exception that is safe to be sent directly to the user; 
 * traditionally Exceptions are *not* displayed to users.
 */
interface UserPermittedExceptionInterface extends Throwable { }

/**
 * Core exceptions support automatic translation on the message field.
 */
class CoreException extends Exception {
	public function __construct(string $message = "", int $code = 0){
		if(strpos($message, 't:') === 0){
			// Support automatic translations for Core exceptions!
			$message = \t(substr($message, 2));
		}
		
		parent::__construct($message, $code, null);
	}
}

/**
 * Redirection command; expects one of the valid STATUS codes to be passed in as the second argument.
 * 
 * By default it issues a 303 redirect code (aka temporary 302).
 * 
 * It's recommended to use one of the other Exceptions instead:
 * 
 * * TemporaryFollowRedirectException   to issue a 303 (aka 302), command
 * * TemporaryResubmitRedirectException to request the user to resubmit the form on a different page
 * * PermanentFollowRedirectException   to issue a standard 301 (aka link-juice-following), command
 * * PermanentResubmitRedirectException to redirect the user to resubmit the form to a different page
 */
class RedirectException extends CoreException implements ViewStatusExceptionInterface {
	public $location;
	
	public function __construct(string $location, int $code = View::STATUS_SEEOTHER) {
		parent::__construct('Redirect Requested', $code);
		
		$this->location = \Core\resolve_link($location);
	}
}


/**
 * User does not have access to the requested page; triggers a 403 status code
 */
class AccessDeniedException extends CoreException implements ViewStatusExceptionInterface {
	public function __construct(string $message = 't:MESSAGE_EXCEPTION_ACCESS_DENIED') {
		parent::__construct($message, View::STATUS_ACCESSDENIED);
	}
}


/**
 * Something went wrong with processing the user's request, triggers a 400 status code
 */
class BadRequestException extends CoreException implements ViewStatusExceptionInterface {
	public function __construct(string $message = 't:MESSAGE_EXCEPTION_BAD_REQUEST') {
		parent::__construct($message, View::STATUS_BADREQUEST);
	}
}


/**
 * Resource not found, be it an object, controller, method, etc.  Triggers a 404 status code
 */
class NotFoundException extends CoreException implements ViewStatusExceptionInterface {
	public function __construct(string $message = 't:MESSAGE_EXCEPTION_NOT_FOUND') {
		parent::__construct($message, View::STATUS_NOTFOUND);
	}
}

/**
 * Class MethodNotAllowedException
 *
 * Request method not supported or allowed for this specific URI
 *
 * @package Core
 */
class MethodNotAllowedException extends CoreException implements ViewStatusExceptionInterface {
	public function __construct(string $message = 't:MESSAGE_EXCEPTION_METHOD_NOT_ALLOWED') {
		parent::__construct($message, View::STATUS_METHODNOTALLOWED);
	}
}

/**
 * Class MethodNotAllowedException
 *
 * Request method not supported or allowed for this specific URI
 *
 * @package Core
 */
class GoneException extends CoreException implements ViewStatusExceptionInterface {
	public function __construct(string $message = 't:MESSAGE_EXCEPTION_GONE_OR_EXPIRED') {
		parent::__construct($message, View::STATUS_GONE);
	}
}


/**
 * Link-Juice friendly redirect; the user agent should remember this and use the new URL for future calls
 * 
 * Destination is requested via a GET and any POST/PUT/DELETE data is assumed to be stored.
 * 
 * Triggers a 301 status code
 */
class PermanentFollowRedirectException extends RedirectException implements ViewStatusExceptionInterface {
	public function __construct(string $location){
		parent::__construct($location, View::STATUS_MOVEDPERMANENTLY);
	}
}


/**
 * Link-Juice friendly resubmission command; the user agent should remember this and use the new URL for future calls
 * 
 * Destination is requested via the same METHOD and any POST/PUT/DELETE data should be resubmitted
 * 
 * Triggers a 308 status code
 */
class PermanentResubmitRedirectException extends RedirectException implements ViewStatusExceptionInterface {
	public function __construct(string $location){
		parent::__construct($location, View::STATUS_PERMANENTREDIRECT);
	}
}


/**
 * Temporary redirect.
 * 
 * Destination is requested via a GET and any POST/PUT/DELETE data is assumed to be stored.
 * 
 * Triggers a 303 status code (aka 302)
 */
class TemporaryFollowRedirectException extends RedirectException implements ViewStatusExceptionInterface {
	public function __construct(string $location){
		parent::__construct($location, View::STATUS_SEEOTHER);
	}
}


/**
 * Temporary resubmission command.
 * 
 * Destination is requested via the same METHOD and any POST/PUT/DELETE data should be resubmitted
 * 
 * Triggers a 307 status code
 */
class TemporaryResubmitRedirectException extends RedirectException implements ViewStatusExceptionInterface {
	public function __construct(string $location){
		parent::__construct($location, View::STATUS_TEMPORARYREDIRECT);
	}
}

class URITooLargeException extends CoreException implements ViewStatusExceptionInterface {
	public function __construct(string $message = 't:MESSAGE_EXCEPTION_URI_TOO_LARGE') {
		parent::__construct($message, View::STATUS_URITOOLARGE);
	}
}

/**
 * Security exceptions; will trigger a 403 when sent from a Controller
 */
class SecurityException extends Exception { }

/**
 * General Model exception thrown from (somewhere in theory, nowhere in practice).
 */
class ModelException extends Exception {

}

/**
 * Validation exceptions caught by the Model for incoming data.
 *
 * Usually sent back to end users for correction on form inputs.
 */
class ModelValidationException extends ModelException {

}

/**
 * General validation exceptions that can be thrown by a Form handler,
 * not necessarily from within a Model.
 *
 * These behave similarly to ModelValidationExceptions as they are displayed to the end user,
 * but they just have a different source.
 */
class GeneralValidationException extends CoreException {

}

class UserNotActivatedException extends CoreException implements UserPermittedExceptionInterface {
	public function __construct(string $message = 't:MESSAGE_ERROR_USER_LOGIN_ACCOUNT_DEACTIVATED') {
		parent::__construct($message);
	}
}
