<?php
/**
 * 
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core;

use Core\Datamodel\DatasetWhereClause;
use Core\Date\DateTime;
use Core\Date\Timezone;
use Core\Forms\DateInput;
use Core\Forms\Form;
use Core\Templates\Template;


class FilterForm {

	/**
	 * Set to true to look for (and remember), sortkey and sortdir as well.
	 *
	 * @var bool
	 */
	public $hassort = false;

	/**
	 * Set to true to look for (and remember), the pagination values.
	 *
	 * @var bool
	 */
	public $haspagination = false;

	private $_name = null;

	private $_elements = array();

	private $_elementindexes = array();

	/**
	 * null or an array of valid sort keys, the first being the default.
	 *
	 * @var null|array
	 */
	private $_sortkeys = null;

	/**
	 * null or the sort key of the current view.
	 *
	 * @var null
	 */
	private $_sortkey = null;

	/**
	 * Sorting direction, think ascending or descending.
	 *
	 * @var string "up" or "down"
	 */
	private $_sortdir = 'down';

	/** @var array Array of limit options to render to the user agent, (gives the options of selecting different options). */
	private $_limitOptions = [25, 50, 100, 250, 500];
	
	/** @var int The limit for this filterset, only takes effect if $haspagination is set to true.  */
	private $_limit = 50;

	/**
	 * The total number of entries for this filterset.  Set automatically from applyToFactory and callable externally.
	 *
	 * @var null|int
	 */
	private $_total = null;

	/**
	 * The current page, only takes effect if $haspagination is set to true.
	 *
	 * @var int
	 */
	private $_currentpage = 1;

	/**
	 * Standard link type for a given model factory and its filter.
	 *
	 * format: key = value
	 */
	const LINK_TYPE_STANDARD = ' = ';

	/**
	 * Greater than link type for a given model factory and its filter.
	 *
	 * format: key > value
	 */
	const LINK_TYPE_GT = ' > ';

	/**
	 * Greater than or equal-to link type for a given model factory and its filter.
	 *
	 * format: key >= value
	 */
	const LINK_TYPE_GE = ' >= ';

	/**
	 * Less than link type for a given model factory and its filter.
	 *
	 * format: key < value
	 */
	const LINK_TYPE_LT = ' < ';

	/**
	 * Less than or equal-to link type for a given model factory and its filter.
	 *
	 * format: key <= value
	 */
	const LINK_TYPE_LE = ' <= ';

	/**
	 * Use the LIKE connector to find anything that starts with the value.
	 *
	 * format: key LIKE value%
	 */
	const LINK_TYPE_STARTSWITH = '_startswith_';

	/**
	 * Use the LIKE connector to find anything that contains the value.
	 *
	 * format: key LIKE %value%
	 */
	const LINK_TYPE_CONTAINS = '_contains_';
	
	/**
	 * Perform a full text search for the requested records.
	 * 
	 * @var string
	 */
	const LINK_TYPE_SEARCH = '_search_';

	/**
	 * Create a new filter form object
	 */
	public function __construct(){
		//if(!isset($_SESSION['filters'])) $_SESSION['filters'] = array();
	}
	
	//-----------------------------------------------------------------------\\
	//                                  GETTERS
	//-----------------------------------------------------------------------\\

	/**
	 * Get one element's value based on its name
	 *
	 * @param string $name The name of the element to retrieve
	 * @return mixed|null Value or null if not set.
	 */
	public function get($name){
		if(strpos($name, 'filter[') === false){
			$name = 'filter[' . $name . ']';
		}

		if(!isset($this->_elementindexes[$name])) return null;
		return $this->_elementindexes[$name]->get('value');
	}
	
	/**
	 * Get the current page that is being rendered
	 * 
	 * @return int
	 */
	public function getCurrentPage() : int {
		$maxpages = ceil($this->_total / $this->_limit);

		// The current page can't exceed past the maxpages nor can it be < 0
		return max(min($maxpages, $this->_currentpage), 1);
	}
	
	/**
	 * Get the last page number in the results
	 * 
	 * @return int
	 */
	public function getLastPage() : int {
		return (int)ceil($this->_total / $this->_limit);
	}
	
	/**
	 * Get the current view limit for number of records at a time.
	 * 
	 * @return int
	 */
	public function getLimit() : int {
		return $this->_limit;
	}

	/**
	 * Get the sort keys as an order clause, passable into the dataset system.
	 *
	 * @since 2.4.0
	 */
	public function getOrder(){
		$key = $this->getSortKey();
		$dir = $this->getSortDirection();

		if(!$key) return null;

		return $key . ' ' . ($dir == 'up' ? 'ASC' : 'DESC');
	}

	/**
	 * Get the sort direction, either up or down.
	 *
	 * @since 2.4.0
	 * @return string|null
	 */
	public function getSortDirection(){
		if(!$this->hassort) return null;
		else return $this->_sortdir;
	}

	/**
	 * Get the sort key.
	 *
	 * @return null|string
	 */
	public function getSortKey(){
		if(!$this->hassort) return null;
		elseif($this->_sortkey) return $this->_sortkey;
		elseif(is_array($this->_sortkeys) && sizeof($this->_sortkeys)) return $this->_sortkeys[0];
		else return null;
	}

	/**
	 * Get the total count for the number of records filtered.
	 *
	 * @return int|null
	 */
	public function getTotalCount(){
		return $this->_total;
	}
	
	//-----------------------------------------------------------------------\\
	//                                  CHECKS
	//-----------------------------------------------------------------------\\

	/**
	 * Return true/false on if this filter has any filters set by the user.
	 *
	 * Essentially will just check if all elements have their value set to "" or to null.
	 *
	 * @return bool
	 */
	public function checkFiltersSet() : bool {
		foreach($this->_elements as $element){
			/** @var $element FormElement */
			if($element->get('value') === ''){
				// An empty value indicates that this element is not set
				continue;
			}

			if($element->get('value') === null){
				// An empty value indicates that this element is not set
				continue;
			}

			// If this filter has a value set, then at least one filter is set!
			return true;
		}

		// No element has anything set?
		return false;
	}
	
	/**
	 * Check if there is at least one filter added to this form
	 * 
	 * @return bool
	 */
	public function checkHasFilters() : bool {
		return count($this->_elements) > 0;
	}
	
	/**
	 * Check if this is the first page in the results
	 * 
	 * @return bool
	 */
	public function checkViewingFirstPage() : bool {
		return $this->getCurrentPage() === 1;
	}
	
	/**
	 * Check if this is the last page in the results
	 * 
	 * @return bool
	 */
	public function checkViewingLastPage() : bool {
		return $this->getCurrentPage() === $this->getLastPage();
	}

	/**
	 * Set the name for this filter, required for any session saving/lookup.
	 *
	 * @param string $filtername
	 */
	public function setName($filtername){
		$this->_name = $filtername;
	}

	/**
	 * Add an element to this filter set.
	 *
	 * This is the exact same as the native Form system.
	 *
	 * @param string|FormElement $element Type of element, (or the form element itself)
	 * @param null|array         $atts [optional] An associative array of parameters for this form element
	 */
	public function addElement($element, $atts = null) {
		// Since this allows for just plain names to be submitted, translate
		// them to the form object to be rendered.

		if ($element instanceof FormElement || is_a($element, 'FormElement')) {
			// w00t, already in the right format!
			if ($atts) $element->setFromArray($atts);
		}
		elseif ($element instanceof FormGroup) {
			// w00t, already in the right format!
			if ($atts) $element->setFromArray($atts);
		}
		else {
			if (!isset(Form::$Mappings[$element])) $element = 'text'; // Default.

			$element = new Form::$Mappings[$element]($atts);
		}

		// Lastly...
		/** @var $element FormElement */

		// Make sure the name is correct and a few useful attributes are set.
		if(strpos($element->get('name'), 'filter[') === false){
			$element->set('name', 'filter[' . $element->get('name') . ']');
		}
		$element->set('class', $element->get('class') . ' filter');

		$this->_elements[] = $element;
		$this->_elementindexes[$element->get('name')] = $element;
	}

	/**
	 * Load the values from either the page request or the session data.
	 *
	 * @param Request $request
	 */
	public function load(Request $request){
		if($request->getParameter('reset')) {
			// A reset was requested, just delete all the session caches for this table
			Session::Delete('filters/' . $this->_name);
			Session::Delete('filtersort/' . $this->_name);
			Session::Delete('filterpage/' . $this->_name);
		}
		else {
			// Otherwise initially, load everything from the session.
			// ONLY do this is the *reset* key is *not* set.  This allows the session data to be reset with one quick parameter.
			$this->loadSession();
		}

		$a = array();
		$s = array();
		$p = array();

		// Check the sort keys?
		if($this->hassort){
			if($request->getParameter('sortkey')){
				$this->setSortKey($request->getParameter('sortkey'));
				$s['sortkey'] = $this->_sortkey;
			}
			if($request->getParameter('sortdir')){
				$this->setSortDirection($request->getParameter('sortdir'));
				$s['sortdir'] = $this->_sortdir;
			}
		}

		// Did the user change a filter?
		// If a filter was changed, reset back to page 1!
		if($request->getParameter('filter')){
			$filters = $request->getParameter('filter');

			foreach($filters as $f => $v){
				if(!isset($this->_elementindexes['filter[' . $f . ']'])) continue;
				/** @var $el FormElement */
				$el = $this->_elementindexes['filter[' . $f . ']'];
				$el->setValue($v);

				// Remember this for the session data.
				$a[$f] = $v;
				$this->setPage(1);
				$p['page'] = 1;
			}
		}
		// How 'bout the pagination?
		elseif($this->haspagination){
			if($request->getParameter('page')){
				$this->setPage($request->getParameter('page'));
				$p['page'] = $this->_currentpage;
				$p['limit'] = $this->_limit;
			}
			elseif($request->getParameter('limit')){
				// Reset to page 1 in the likely event that the limit changes how many pages display.
				$this->setPage(1);
				$p['page'] = 1;
				
				$this->setLimit($request->getParameter('limit'));
				$p['limit'] = $this->_limit;
			}
			// Don't change the filter sets, those have been cached and are fine as-is.
		}
		else{
			// No pagination or filters were modified... don't do anything.
		}


		if(sizeof($a)){
			Session::Set('filters/' . $this->_name, $a);
		}
		if(sizeof($s)){
			Session::Set('filtersort/' . $this->_name, $s);
		}
		if(sizeof($p)){
			Session::Set('filterpage/' . $this->_name, $p);
		}
	}

	/**
	 * Load the values from the session data.
	 *
	 * This is automatically called by the load function.
	 */
	public function loadSession(){
		if(!$this->_name){
			// Ok, no name.. no loading.
			return;
		}
		if(($filters = Session::Get('filters/' . $this->_name)) !== null){
			foreach($filters as $f => $v){
				if(!isset($this->_elementindexes['filter[' . $f . ']'])) continue;
				/** @var $el FormElement */
				$el = $this->_elementindexes['filter[' . $f . ']'];
				$el->setValue($v);
			}
		}

		if(Session::Get('filtersort/' . $this->_name) !== null){
			$this->_sortkey = Session::Get('filtersort/' . $this->_name)['sortkey'];
			$this->_sortdir = Session::Get('filtersort/' . $this->_name)['sortdir'];
		}

		$cachedPageData = Session::Get('filterpage/' . $this->_name);
		if($cachedPageData !== null && isset($cachedPageData['page'])) {
			$this->_currentpage = Session::Get('filterpage/' . $this->_name)['page'];
		}
		if($cachedPageData !== null && isset($cachedPageData['limit'])){
			$this->setLimit(Session::Get('filterpage/' . $this->_name)['limit']);
		}
	}

	/**
	 * Fetch this filter set as a string
	 *
	 * (should probably be called fetch, but whatever)
	 *
	 * @return string
	 */
	public function render(){
		return $this->_render(false);
	}

	/**
	 * Fetch this filter set as an HTML string
	 *
	 * This result set will be readonly however!
	 *
	 * @return string
	 */
	public function renderReadonly(){
		return $this->_render(true);
	}
	
	/**
	 * Alias of checkFiltersSet
	 *
	 * @return boolean
	 */
	public function hasSet(){
		return $this->checkFiltersSet();
	}

	/**
	 * Alias of checkHasFilters
	 *
	 * @return bool
	 */
	public function hasFilters(){
		return (sizeof($this->_elements) > 0);
	}

	/**
	 * Fet this filter set's pagination options as a string.
	 *
	 * @return string
	 */
	public function pagination(){
		if(!$this->haspagination) return null;

		// Give me the total number of pages given the current criteria.
		$maxpages = ceil($this->_total / $this->_limit);

		// If there are no more pages beyond the first, just return null, no pagination available!
		//if($maxpages <= 1) return null;

		// The current page can't exceed past the maxpages.
		$currentpage = min($maxpages, $this->_currentpage);
		// nor can it be < 0
		if($currentpage < 1) $currentpage = 1;

		//$currentpage = 22;

		// The number of results on either side of the current.
		$offset = 4;

		// If there are more than so many pages, only display a subset of them.
		if($maxpages > 10){
			if($currentpage > $maxpages - (($offset)*2-1)){
				$displaymin = $maxpages - (($offset)*2+1);
				$displaymax = $maxpages;
			}
			elseif($currentpage < (($offset)*2)){
				$displaymin = 1;
				$displaymax = ($offset*2+1);
			}
			else{
				$displaymin = $currentpage - ($offset);
				$displaymax = $currentpage + ($offset);
			}
		}
		else{
			$displaymin = 1;
			$displaymax = $maxpages;
		}

		// Calculate the "Displaying x - y records of z total" numbers here!

		// Total number of records, simple.
		$records_total = $this->_total;
		// Current number of records on the page, only *not* the total if the total is < limit.
		$records_current = min($this->_limit, $this->_total);
		// Starting position.... always go n+1 here because people are used to seeing "1-10" instead of "0-9"
		$records_start = ($currentpage - 1) * $this->_limit + 1;
		// And the ending position.
		$records_end = min($records_start + $records_current - 1, $this->_total);


		$tpl = Template::Factory('forms/filters-pagination.tpl');
		$tpl->assign('page_current', $currentpage);
		$tpl->assign('page_max', $maxpages);
		$tpl->assign('display_min', $displaymin);
		$tpl->assign('display_max', $displaymax);
		$tpl->assign('records_total', $records_total);
		$tpl->assign('records_current', $records_current);
		$tpl->assign('records_start', $records_start);
		$tpl->assign('records_end', $records_end);
		$tpl->assign('limit', $this->_limit);
		$tpl->assign('limit_options', $this->_limitOptions);
		return $tpl->fetch();
	}

	/**
	 * Set the sort direction for this filterset.
	 *
	 * The direction must be one of the following: "up", "down", "asc", "desc".
	 * (asc/desc are remapped to up/down automatically)
	 *
	 * @since 2.4.0
	 * @param $dir "up" or "down".  It must be up or down because fontawesome has those keys instead of "asc" and "desc" :p
	 *
	 * @return bool true/false on success or failure.
	 */
	public function setSortDirection($dir){
		$dir = strtolower($dir);
		switch($dir){
			case 'down':
			case 'up':
				$this->_sortdir = $dir;
				return true;
			case 'asc':
				$this->_sortdir = 'up';
				return true;
			case 'desc':
				$this->_sortdir = 'down';
				return true;
			default:
				return false;
		}
	}

	/**
	 * Add a single sort key onto this Filter.
	 *
	 * @since 3.4.0
	 * @param string $key
	 *
	 * @return bool
	 */
	public function addSortKey($key){
		if(!$this->hassort) return false;

		if(!is_array($this->_sortkeys)){
			$this->_sortkeys = [];
		}

		$this->_sortkeys[] = $key;
		return true;
	}

	/**
	 * Set the valid sort keys for this filterset.
	 *
	 * @since 2.4.0
	 * @param array $arr
	 *
	 * @return bool true/false on success or failure.
	 */
	public function setSortkeys($arr){
		if(!$this->hassort) return false;
		if(!is_array($arr)) return false;

		$this->_sortkeys = $arr;
		return true;
	}

	/**
	 * Set the active sort key currently.
	 * If sotkeys is populated, it MUST be one of the keys in that array!
	 *
	 * @since 2.4.0
	 * @param string $key
	 *
	 * @return bool true/false on success or failure.
	 */
	public function setSortKey($key){
		if(is_array($this->_sortkeys) && sizeof($this->_sortkeys)){
			// It's set, enforce type!
			if(in_array($key, $this->_sortkeys)){
				$this->_sortkey = $key;
				return true;
			}
			else{
				return false;
			}
		}
		else{
			// Ok, just proceed blindly.
			$this->_sortkey = $key;
			return true;
		}
	}

	/**
	 * Set the limit for pagination, will default to 50.
	 *
	 * @since 2.4.0
	 * @param int $limit
	 */
	public function setLimit(int $limit){
		$this->_limit = max(min($limit, 999), 1);
	}

	public function setPage($page){
		$this->_currentpage = (int) $page;
	}

	/**
	 * Set the total count for the number of records.
	 * This is externally available in case the factory is modified externally, which is perfectly allowed.
	 *
	 * @param int $count
	 */
	public function setTotalCount($count){
		$this->_total = (int) $count;
	}
	
	/**
	 * Manually set a filter value
	 * 
	 * @param string $filter
	 * @param mixed $value
	 */
	public function setFilterValue(string $filter, $value) {
		if(isset($this->_elementindexes['filter[' . $filter . ']'])){
			/** @var $el FormElement */
			$el = $this->_elementindexes['filter[' . $filter . ']'];
			$el->setValue($value);
		}
	}

	/**
	 * Given all the user defined filter, sort, and what not, apply those values to the ModelFactory if possible.
	 *
	 * @since 2.4.0
	 * @param ModelFactory $factory
	 */
	public function applyToFactory(ModelFactory $factory){
		if($this->hassort){
			$factory->order($this->getOrder());
		}

		if($this->haspagination){
			// Determine the starting count if the page is requested.
			if($this->_currentpage > 1){
				$startat = $this->_limit * ($this->_currentpage - 1);
				$factory->limit($this->_limit, $startat);
			}
			else{
				$factory->limit($this->_limit);
			}
		}

		foreach($this->_elements as $el){
			/** @var $el FormElement */
			$name = $el->get('name');
			$idxname = $name;

			if(strpos($name, 'filter[') === 0){
				$name = substr($name, 7, -1);
			}

			// If this element is not in the index of elements, skip to the next element.
			if(!isset($this->_elementindexes[$idxname])){
				continue;
			}

			// If this doesn't have a link attribute, just skip.
			if(!$el->get('link')){
				continue;
			}

			// No value, just skip.
			if($el->get('value') === '' || $el->get('value') === null){
				continue;
			}

			// If there is a "" option, interpret that as empty and allow "0" to be used.
			if($el->get('value') === '0'){
				if($el->get('options') && isset($el->get('options')[''])){
					// '' is set... proceed.
				}
				else{
					continue;
				}
			}

			$value = $el->get('value');

			// Was there a prefix and/or suffix requested?
			if($el->get('linkvalueprefix')){
				$value = $el->get('linkvalueprefix') . $value;
			}
			if($el->get('linkvaluesuffix')){
				$value = $value . $el->get('linkvaluesuffix');
			}

			// If this link is a date object, convert a date string to its unix timestamp representation.
			if($el instanceof DateInput || $el->get('dateformat')){
				// Default to a unix timestamp, but allow the user to override this.
				// This is useful for saving a date in the datastore as a human-readable format.
				$format = $el->get('dateformat') ? $el->get('dateformat') : 'U';
				$date = new DateTime($value);
				$value = $date->getFormatted($format, Timezone::TIMEZONE_GMT);
			}

			if($el->get('linkname')){
				$name = $el->get('linkname');
			}

			// New support for multiple link names!
			if(!is_array($name)){
				$name = [$name];
			}
			$statements = [];

			foreach($name as $n){
				switch($el->get('link')){
					case FilterForm::LINK_TYPE_STANDARD:
					case FilterForm::LINK_TYPE_GT:
					case FilterForm::LINK_TYPE_GE:
					case FilterForm::LINK_TYPE_LT:
					case FilterForm::LINK_TYPE_LE:
						$statements[] = $n . $el->get('link') . $value;
						break;
					case FilterForm::LINK_TYPE_STARTSWITH:
						$statements[] = $n . ' LIKE ' . $value . '%';
						break;
					case FilterForm::LINK_TYPE_CONTAINS:
						$statements[] = $n . ' LIKE %' . $value . '%';
						break;
					case FilterForm::LINK_TYPE_SEARCH:
						$factory->search($value);
						break;
				}
			}

			if(sizeof($statements) > 1){
				// Create a sub where clause for these.
				$subwhere = new DatasetWhereClause();
				$subwhere->setSeparator('OR');
				foreach($statements as $s){
					$subwhere->addWhere($s);
				}
				// Add this sub clause to the main where clause.
				$factory->where($subwhere);
			}
			else{
				// A single command just gets added to the main clause.
				$factory->where($statements[0]);
			}
		}

		// Might as well update the count now, it can always be updated later.
		$this->setTotalCount($factory->count());
	}
	
	public function getElements(){
		return $this->_elements;
	}

	private function _render($readonly = false){
		$filterset = $this->hasSet();

		$tpl = Template::Factory('forms/filters');
		$tpl->assign('filtersset', $filterset);
		$tpl->assign('elements', $this->_elements);
		$tpl->assign('hassort', $this->hassort);
		$tpl->assign('sortkey', $this->getSortKey());
		$tpl->assign('sortdir', $this->getSortDirection());
		$tpl->assign('readonly', $readonly);
		$tpl->assign('records_total', $this->_total);
		$tpl->assign('records_current', min($this->_limit, $this->_total));
		return $tpl->fetch();
	}
}

