<?php
/**
 * File for class ModelSchema definition in the coreplus project
 * 
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20131022.1632
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core;

use Core\Datamodel\Columns\SchemaColumn;
use Core\Datamodel\Schema;
use Core\Dict;
use Core\Model;
use ReflectionClass;

/**
 * A short teaser of what ModelSchema does.
 *
 * More lengthy description of what ModelSchema does and why it's fantastic.
 *
 * <h3>Usage Examples</h3>
 *
 *
 * @todo Write documentation for ModelSchema
 * <h4>Example 1</h4>
 * <p>Description 1</p>
 * <code>
 * // Some code for example 1
 * $a = $b;
 * </code>
 *
 *
 * <h4>Example 2</h4>
 * <p>Description 2</p>
 * <code>
 * // Some code for example 2
 * $b = $a;
 * </code>
 *
 * 
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
class ModelSchema extends Schema {
	
	private $_rawSchema = [];

	public function __construct(string $model = null){
		if($model !== null){
			$this->readModel($model);
		}
	}

	public function readModel(string $model){
		
		// RESET!
		$this->indexes     = [];
		$this->definitions = [];
		$this->order       = [];
		$this->metas       = [];
		
		$this->_rawSchema = $this->_loadSchema($model);
		
		
		foreach($this->_rawSchema as $name => $def){
			if($def['type'] == Model::ATT_TYPE_ALIAS){
				$this->aliases[$name] = $def['alias'];
			}
			elseif($def['type'] == Model::ATT_TYPE_META){
				$def['name'] = $name;
				$column = SchemaColumn::FactoryFromSchema($def);

				$this->metas[$name] = $column;
			}
			else{
				$def['name'] = $name;
				$column = SchemaColumn::FactoryFromSchema($def);

				$this->definitions[$name] = $column;
				$this->order[] = $name;
			}
		}
		

		// Construct a general object so I can use the getKeySchemas method inside.
		$ref = new ReflectionClass($model);
		$indexes = $ref->getProperty('Indexes')->getValue();
		
		foreach($indexes as $key => $dat){
			if(!is_array($dat)){
				// Models can be defined with a single element for its index.
				// This is an acceptable shorthand standard, but the lower level utilities
				// are expecting an array.
				$this->indexes[$key] = array($dat);
			}
			else{
				$this->indexes[$key] = $dat;
			}
		}
	}
	
	public function __clone() {
		foreach($this->definitions as $key => $column) {
			$this->definitions[ $key ] = clone $column;
		}

		foreach($this->metas as $key => $column) {
			$this->metas[ $key ] = clone $column;
		}
	}
	
	private function _loadSchema(string $model) : Dict {
		
		// If this Model extends another Model, then load that schema FIRST and prepend it to this object.
		// otherwise, just start with an empty array.
		// this also plays into the recursion factor.
		$parent = get_parent_class($model);
		if( !($parent === Model::class || $parent === '' || $parent === false) ) {
			$schema = $this->_loadSchema($parent);
		}
		else{
			$schema = new Dict();
		}
		
		
		$ref = new ReflectionClass($model);
		$modelSchema = $ref->getProperty('Schema')->getValue();
		
		foreach($modelSchema as $key => $dat){
			$dat['_defining_model'] = $model;
			$schema->addItem($key, $dat);
		}
		
		
		// There are a variety of dynamic columns that are defined in Core.

		if(
			$ref->getProperty('HasCreated')->getValue() &&
			!$schema->hasKey('created')
		){
			$schema->addItem('created', [ 'type' => Model::ATT_TYPE_CREATED, '_defining_model' => $model, ]);
		}
		
		if(
			$ref->getProperty('HasUpdated')->getValue() &&
			!$schema->hasKey('updated')
		){
			$schema->addItem('updated', [ 'type' => Model::ATT_TYPE_UPDATED, '_defining_model' => $model, ]);
		}
		
		if(
			$ref->getProperty('HasDeleted')->getValue() &&
			!$schema->hasKey('deleted')
		){
			$schema->addItem('deleted', [ 'type' => Model::ATT_TYPE_DELETED, '_defining_model' => $model, ]);
		}
		
		if(
			$ref->getProperty('HasSearch')->getValue() &&
			!$schema->hasKey('search_index_str')
		){
			$schema->addItem(
				'search_index_str',
				[
					'type' => Model::ATT_TYPE_SEARCH_INDEX_STR,
					'_defining_model' => $model,
					'importer' => false,
				]
			);
		}
		
		if(
			$ref->getProperty('HasSearch')->getValue() &&
			!$schema->hasKey('search_index_pri')
		){
			$schema->addItem(
				'search_index_pri',
				[
					'type' => Model::ATT_TYPE_SEARCH_INDEX_PRI,
					'_defining_model' => $model,
					'importer' => false,
				]
			);
		}
		
		if(
			$ref->getProperty('HasSearch')->getValue() &&
			!$schema->hasKey('search_index_sec')
		){
			$schema->addItem(
				'search_index_sec',
				[
					'type' => Model::ATT_TYPE_SEARCH_INDEX_SEC,
					'_defining_model' => $model,
					'importer' => false,
				]
			);
		}
		
		// Allow all supplemental models to tap into the schema too!
		if(isset(Model::$ModelSupplementals[$model])){
			foreach(Model::$ModelSupplementals[$model] as $supplemental){
				if(class_exists($supplemental)){
					$ref = new ReflectionClass($supplemental);
					if($ref->hasProperty('Schema')) {
						// Retrieve the supplemental Schema from this new component
						$s = $ref->getProperty('Schema')->getValue();

						foreach($s as $key => $dat){
							$dat['_defining_model']  = $supplemental;
							$dat['_is_supplemental'] = true;
							$schema->addItem($key, $dat);
						}
					}
				}
			}
		}
		
		return $schema;
	}
}