<?php
/**
 * @package Core
 * @since 2011.06
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright 2011-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl.html>
 * This system is licensed under the GNU LGPL, feel free to incorporate it into
 * custom applications, but keep all references of the original authors intact,
 * read the full license terms at <http://www.gnu.org/licenses/lgpl-3.0.html>,
 * and please contribute back to the community :)
 */

namespace Core;
use Core\CLI\CLI;
use Core\Filestore\File;
use Exception;

/**
 * Theme object.
 *
 * Themes consist just of the template files and corresponding assets.
 */
class Theme extends Component {
	
	// <editor-fold defaultstate="collapsed" desc="PUBLIC Load and Init Methods">

	/**
	 * @throws Exception
	 */
	private function _init() {
		if (!$this->_xmlLoader->load()) {
			throw new Exception('Invalid theme.xml file provided, unable to create Theme!');
		}

		$systemID = $this->_xmlLoader->getDOM()->doctype->systemId;
		switch($systemID){
			case 'https://corepl.us/api/6_3/theme.dtd':
				// 6.3 was released Oct 2017
				$this->_apiVersion = 63;
				break;
			case 'http://corepl.us/api/2_4/theme.dtd':
			case 'https://corepl.us/api/2_4/theme.dtd':
				// 2.4 was released around Jan 2013
				$this->_apiVersion = 24;
				break;
			case 'http://corepl.us/api/2_1/theme.dtd':
				// 2.1 was released early 2012
				$this->_apiVersion = 21;
				break;
			default:
				throw new \Exception('Unsupported theme.xml version detected!');
		}

		$this->_type = 'theme';
		$this->_name    = $this->_xmlLoader->getRootDOM()->getAttribute('name');
		$this->_version = $this->_xmlLoader->getRootDOM()->getAttribute("version");
		$this->_keyName = basename(dirname($this->_xmlFilename));
	}

	/**
	 * @throws Exception
	 */
	public function initFromFile(string $filename) {
		$this->_xmlFilename = $filename;
		$this->_xmlLoader = new XMLLoader();
		$this->_xmlLoader->setRootName('theme');
		$this->_xmlLoader->loadFromFilename($filename);
		$this->_baseDirectory = dirname($filename) . '/';

		$this->_init();
	}

	/**
	 * @throws Exception
	 */
	public function initFromString(string $xmlContents) {
		$this->_xmlLoader = new XMLLoader();
		$this->_xmlLoader->setRootName('theme');
		$this->_xmlLoader->loadFromString($xmlContents);
		$this->_baseDirectory = '';

		$this->_init();
	}
	
	/**
	 * Retrieve all the necessary information from the database regarding this Component.
	 * 
	 * Will throw an exception if the Core has not been loaded before calling!
	 * 
	 * @throws \Exception
	 */
	public function loadAllFromDB(){
		$this->_enabled = $this->_keyName === ConfigHandler::Get('/theme/selected');
	}
	
	public function registerHooks(){
		// Does nothing.
	}
	
	// </editor-fold>

	
	/**
	 * Check if this component is marked as enabled
	 * 
	 * Will always return false if called before Core has initialized the database.
	 * 
	 * @return bool
	 */
	public function isEnabled(): bool {
		return ConfigHandler::Get('/theme/selected') === $this->getKeyName();
	}
	
	
	// <editor-fold defaultstate="collapsed" desc="PUBLIC get* Methods">
	
	/**
	 * Get any "autoload" files registered in this component XML
	 * 
	 * These files are expected to be loaded as soon as the component is registered.
	 * 
	 * @return array
	 */
	public function getAutoLoads() : array {
		// Themes do not have autoload capabilities
		return [];
	}
	
	/**
	 * Get the configs defined on this component as an array of ComponentConfig objects.
	 * 
	 * @return array <<ComponentConfig>>
	 */
	public function getConfigs() : array {
		$ret = [];
		
		$node = $this->_xmlLoader->getElement('configs');
		$componentName = 'theme/' . $this->getKeyName();

		// Now, get every table under this node.
		foreach ($node->getElementsByTagName('config') as $confignode) {
			/** @var \DOMElement $confignode */
			$key = $confignode->getAttribute('key');
			
			// Themes only allow for keys starting with "/theme/"!
			// This is to encourage that all themes share a common subset of configuration options.
			// EG: if the end user sees: "Site Logo", "Business Address", "Business Phone" on one theme,
			// they would be expecting to see those same options with the same values if they change the theme,
			// (and the new theme supports those same options).
			if(strpos($key, '/theme/') === 0){
				$ret[] = new ComponentConfig($confignode, $componentName);
			}
		}
		
		return $ret;
	}
	
	/**
	 * Get the primary, (first), screenshot of this theme.
	 *
	 * @return array
	 */
	public function getScreenshot(){
		$s = $this->_xmlLoader->getElement('/screenshots/screenshot', false);

		if(!$s){
			return [
				'file' => '',
				'title' => $this->getName()
			];
		}
		else{

			$f = \Core\Filestore\Factory::File($this->getBaseDir() . $s->getAttribute('file'));

			return [
				'file' => $f,
				'title' => ($s->getAttribute('title') ? $s->getAttribute('title') : $this->getName()),
			];
		}
	}

	/**
	 * Get all screenshots in this metafile
	 *
	 * @return array
	 */
	public function getScreenshots(){
		$s = $this->_xmlLoader->getElements('/screenshots/screenshot');

		if(!$s){
			return [ ];
		}
		else{

			$f = \Core\Filestore\Factory::File($this->getBaseDir() . $s->getAttribute('file'));

			return [
				'file' => $f,
				'title' => ($s->getAttribute('title') ? $s->getAttribute('title') : $this->getName()),
			];
		}
	}
	
	/**
	 * Get all the templates registered for this theme.
	 * Each template can be a different site skin, ie: 2-column, 3-column, etc.
	 *
	 * @return array
	 */
	public function getSkins(){
		$out = [];
		$default = null;
		$admindefault = null;
		$currenttheme = false;

		// If this theme is currently selected, check the default template too.
		if($this->getKeyName() == ConfigHandler::Get('/theme/selected')){
			$default = ConfigHandler::Get('/theme/default_template');
			$admindefault = ConfigHandler::Get('/theme/default_admin_template');

			// Defaults to the main public skin.
			if(!$admindefault) $admindefault = $default;

			$currenttheme = true;
		}

		foreach($this->_xmlLoader->getElements('/skins/file') as $f){
			$basefilename = $f->getAttribute('filename');
			$filename = $this->getBaseDir() . 'skins/' . $basefilename;
			$systemSkin = $basefilename[0] === '_';

			$skin = \Core\Templates\Template::Factory($filename);

			$title = $f->getAttribute('title') ? $f->getAttribute('title') : $basefilename;

			// The return is expecting an array.
			$out[] = [
				'filename'        => $filename,
				'file'            => $basefilename,
				'title'           => $title,
				'default'         => ($default == $basefilename),
				'admindefault'    => ($admindefault == $basefilename),
				'has_stylesheets' => $skin->hasOptionalStylesheets(),
				'current_theme'   => $currenttheme,
				'is_system'       => $systemSkin,
			];
		}

		return $out;
	}

	/**
	 * Get all the email skins registered for this theme.
	 * Each template can be a different site skin, ie: 2-column, 3-column, etc.
	 *
	 * @return array
	 */
	public function getEmailSkins(){
		$out = [];
		$default = null;
		$currenttheme = false;

		// If this theme is currently selected, check the default template too.
		if($this->getKeyName() == ConfigHandler::Get('/theme/selected')){
			$default = ConfigHandler::Get('/theme/default_email_template');

			$currenttheme = true;
		}

		foreach($this->_xmlLoader->getElements('/emailskins/file') as $f){
			$basefilename = $f->getAttribute('filename');
			$filename = $this->getBaseDir() . 'emailskins/' . $basefilename;

			$skin = \Core\Templates\Template::Factory($filename);
			$title = $basefilename;

			// The return is expecting an array.
			$out[] = [
				'filename'        => $filename,
				'file'            => $basefilename,
				'title'           => $title,
				'default'         => ($default == $basefilename),
				'current_theme'   => $currenttheme,
			];
		}

		// Tack on the main default... no skin!
		$out[] = [
			'filename'        => '',
			'file'            => '',
			'title'           => '-- No Skin --',
			'default'         => ($default == ''),
			'current_theme'   => $currenttheme,
		];

		return $out;
	}
	
	// </editor-fold>

	
	// <editor-fold defaultstate="collapsed" desc="PUBLIC set* and save Methods">
	
	/**
	 * Save this theme metadata back to its XML file.
	 * Useful in packager scripts.
	 */
	public function save() {
		// Set the schema version to the newest API version.
		$this->_xmlLoader->setSchema('https://corepl.us/api/6_3/theme.dtd');
		
		// Ensure there's a required namespace on the root node.
		$this->_xmlLoader->getRootDOM()->setAttribute('xmlns:xsi', "http://www.w3.org/2001/XMLSchema-instance");

		// This needs to be the final step... write the XML doc back to the file.
		file_put_contents($this->_xmlFilename, $this->_xmlLoader->asPrettyXML());
	}

	/**
	 * Set all skin files in this theme.  Only really usable in the installer.
	 *
	 * @param $files array Array of files to set.
	 */
	public function setSkinFiles($files) {

		// This behaves slightly differently than the other ones, since this can include metadata for the skin.
		// As such, they are not simply deleted to begin with.

		// It would be nice to have them alphabetical.
		$newarray = [];
		foreach ($files as $f) {
			// Make sure that the file does not start with 'skins/'...
			if(strpos($f['file'], 'skins/') === 0) $f['file'] = substr($f['file'], 6);
			$newarray[$f['file']] = $f;
		}
		ksort($newarray);

		$used = [];
		// Instead, I'm checking each existing one.
		foreach($this->_xmlLoader->getElements('/theme/skins/file') as $el){
			$att_file = $el->getAttribute('filename');
			if(isset($newarray[$att_file])){
				$used[] = $att_file;
				$el->setAttribute('md5', $newarray[$att_file]['md5']);
			}
			else{
				// Remove it!
				$this->_xmlLoader->getElement('/theme/skins', false)->removeChild($el);
			}
		}

		// And make sure that I didn't miss any new ones.
		foreach($newarray as $f){
			if(!in_array($f['file'], $used)){
				// Make the title something generic.
				$title = substr($f['file'], 6, -4);
				$this->_xmlLoader->createElement('/theme/skins/file[@filename="' . $f['file'] . '"][@md5="' . $f['md5'] . '"][@title="' . $title . '"]');
			}
		}
	}

	/**
	 * Set all other files in this theme.  Only really usable in the installer.
	 *
	 * @param $files array Array of files to set.
	 */
	public function setOtherFiles($files) {
		// Clear out the array first.
		$this->_xmlLoader->removeElements('/theme/otherfiles/file');

		// It would be nice to have them alphabetical.
		$newarray = [];
		foreach ($files as $f) {
			$newarray[$f['file']] = $f;
		}
		ksort($newarray);

		// And recreate them all.
		foreach ($newarray as $f) {
			$this->_xmlLoader->createElement('/theme/otherfiles/file[@filename="' . $f['file'] . '"][@md5="' . $f['md5'] . '"]');
		}
	}
	
	// </editor-fold>

	/**
	 * Install this theme and its assets.
	 *
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function install(){
		
		$results = array_merge_recursive(
			$this->installAssets(),
			$this->installConfigs(),
			$this->installSkins()
		);
		
		ConfigHandler::Set('/theme/selected', $this->_keyName);
		
		return;
		
		$results = $this->_installAssets($verbosity);
		//$results = array_merge_recursive($results, $this->_installAssets($verbosity));
		var_dump($results); die();
		

		$change = $this->_parseConfigs(true, $verbosity);
		if($change !== false) $changed = array_merge($changed, $change);

		// Make sure the version is correct in the database.
		$c = new \ComponentModel('theme/' . $this->getKeyName());
		$c->set('version', $this->_version);
		$c->save();

		// Set the default skins too!
		// This is ONLY on install, (so as to not override any user-set options).
		$default = $this->_parseSkins(true, $verbose);

		// If $changes is "false" sizeof will return 1 ( it's an alias of count() ).
		// so let's compare $changes with false instead!
		if(is_array($default) && $changes !== false){
			$changes += $default;
		}

		if(is_array($changes) && $changes !== false){
			\SystemLogModel::LogInfoEvent('/updater/theme/install', 'Theme ' . $this->getName() . ' installed successfully!', implode("\n", $changes));
		}

		return $changes;
	}

	/**
	 * Reinstall this theme and its assets.
	 *
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function reinstall(){
		$results = array_merge_recursive(
			$this->installAssets(),
			$this->installConfigs()
		);
		
		// Record the results.
		$this->_logInstallResults('reinstall', $results);
		
		return $results;
	}

	/**
	 * Upgrade this theme
	 *
	 *  @return array Contains changes <<string>> and errors <<string>>
	 */
	public function upgrade(){
		$results = array_merge_recursive(
			$this->installAssets(),
			$this->installConfigs()
		);
		
		// Record the results.
		$this->_logInstallResults('upgrade', $results);
		
		return $results;
	}
	
	/**
	 * Copy in all the assets for this theme into the assets location.
	 *
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function installAssets(){
		// Will contain the report for any changes performed herein.
		$results = [
			'changes' => [],
			'errors' => [],
		];
		
		\Core\log_trace('Installing assets for Theme ' . $this->getName());
		
		
		// Array of files to process, (so the try/catch logic isn't duplicated).
		$files = [];
		$filenames = [];
		
		// Used for filename translations throughout.
		// Saved as a variable to act as a cache.
		// These variables are fully resolved with ROOT_PDIR.
		$themeBaseDir  = $this->getBaseDir();
		
		foreach($this->_xmlLoader->getElements('/assets/file') as $node){
			// Iterate over every registered theme asset and install it.
			$filename = $node->getAttribute('filename');
			$filenames[] = $filenames;

			if(file_exists(ROOT_PDIR . 'themes/custom/' . $filename)){
				// There is a file with the same relative name in themes/custom, so use that as the source.
				$f = new \Core\Filestore\Backends\FileLocal(ROOT_PDIR . 'themes/custom/' . $filename);
				$isCustom = true;
			}
			else{
				// Default is just use the actual theme file.
				$f = new \Core\Filestore\Backends\FileLocal($themeBaseDir . $filename);
				$isCustom = false;
			}

			// By using Factory as opposed to FileLocal, the destination path will get auto-resolved.
			// This is because the admin may set the path to something different than default.
			$nf = \Core\Filestore\Factory::File($filename);
			
			$files[] = [
				'name'   => $filename,
				'source' => $f,
				'dest'   => $nf,
				'custom' => $isCustom,
			];
		}
		
		// If there are custom assets not registered by any application, install them too!
		// This will allow an admin to upload additional css resources and images easily.
		$directory = new \Core\Filestore\Backends\DirectoryLocal(ROOT_PDIR . 'themes/custom/assets/');
		// For filename translations, I need just "assets/[discovered-file]" here.
		$directoryLen = strlen($directory->getPath()) - 7;
		
		$ls = $directory->ls(null, true);
		foreach($ls as $f){
			if($f instanceof \Core\Filestore\File){
				$filename = substr($f->getFilename(), $directoryLen);
				$nf = \Core\Filestore\Factory::File($filename);
				
				if(!in_array($filename, $filenames)){
					// Don't create duplicates!
					$files[] = [
						'name'   => $filename,
						'source' => $f,
						'dest'   => $nf,
						'custom' => true,
					];
				}
			}
		}
		
		
		// Now both theme assets and whatever oddity the user placed in custom/ will be in a single array
		foreach($files as $f){
			$action = $nf->exists() ? 'Replaced' : 'Installed';
			try{
				$installed = $this->_installAssetFile($f['source'], $f['dest']);
				
				if($installed){
					$results['changes'][] = $action . ' ' . ($f['custom'] ? 'CUSTOM ' : '') . $f['name'];
					\Core\log_debug($action . ' ' . ($f['custom'] ? 'CUSTOM ' : '') . $f['name']);
				}
				else{
					\Core\log_trace('Skipped ' . ($f['custom'] ? 'CUSTOM ' : '') . $f['name']);
				}
			}
			catch (\Exception $ex) {
				\Core\log_notice($ex->getMessage());
				$results['errors'][] = $ex->getMessage();
			}
		}

		
		if(sizeof($results['changes']) === 0 && sizeof($results['errors']) === 0){
			\Core\log_debug('No changes detected');
		}

		// Make sure the asset cache is purged!
		\Core\Cache::Delete('asset-resolveurl');

		return $results;
	}
	
	/**
	 * Set the default skins based on this theme
	 * 
	 * This will change the configurated skins, so be sure only to call this on install()
	 * or when you really need to.
	 *
	 * @return array Contains changes <<string>> and errors <<string>>
	 *
	 * @throws \InstallerException
	 */
	public function installSkins(){
		// Keep track of if this changed anything.
		$results = [
			'changes' => [],
			'errors'  => [],
		];

		\Core\log_trace('Setting default skins for Theme ' . $this->getName());

		// I need to get the schema definitions first.
		$node = $this->_xmlLoader->getElement('skins');

		// If requested, set those skins as the site default.
		$defaultFrontend = $node->getAttribute('default');
		$adminFrontend   = $node->getAttribute('admindefault');

		$emailNode = $this->_xmlLoader->getElement('emailskins');
		$defaultEmail = $emailNode->getAttribute('default');

		if($defaultFrontend){
			ConfigHandler::Set('/theme/default_template', $defaultFrontend);
			$results['changes'][] = 'Set default skin to ' . $defaultFrontend;
		}
		if($adminFrontend){
			ConfigHandler::Set('/theme/default_admin_template', $adminFrontend);
			$results['changes'][] = 'Set default admin skin to ' . $adminFrontend;
		}
		if($defaultEmail){
			ConfigHandler::Set('/theme/default_email_template', $defaultEmail);
			$results['changes'][] = 'Set default email skin to ' . $defaultEmail;
		}

		return $results;
	}
	
	/**
	 * Validate that a requested theme name is valid and exists on the filesystem.
	 * 
	 * @param string $theme
	 * @return bool
	 */
   public static function ValidateThemeName($theme){

	   // Static variable to use as application cache.
	   static $_cache = array();

	   // Already checked?
	   if(isset($_cache[$theme])) return $_cache[$theme];

	   // Guess not, set it to false by default.
	   $_cache[$theme] = false;

	   // Ensure that the theme is not empty, doesn't begin with a '.', is in the same directory, and exists.
	   if(!$theme) return false;
	   if($theme[0] == '.') return false;
	   if(strpos($theme, '..') !== false) return false;
	   if(!is_dir(\ROOT_PDIR . 'themes/' . $theme)) return false;

	   // And if it made it here...
	   $_cache[$theme] = true;
	   return true;
   }
   
	/**
	 * Validate that a requested theme template combination are valid and exist.
	 * 
	 * @param string $theme
	 * @param string $template 
	 */
	public static function ValidateTemplateName($theme, $template){
		if(!self::ValidateThemeName($theme)){
			return false;
		}

		$filename = \ROOT_PDIR . 'themes/' . $theme . '/skins/' . $template;

		if($template[0] == '.' || !$template || !is_readable($filename)){
			return false;
		}

		return true;
	}
	
	/**
	 * Get this application's system default theme.
	 * 
	 * This is different than the currently installed one!!!
	 * This allows developers to include a different "default" theme for their own project.
	 * 
	 * @return string
	 */
	public static function GetDefaultThemeName(){
		if(file_exists(ROOT_PDIR . '.default-theme')) {
			$theme = file_get_contents(ROOT_PDIR . '.default-theme');
		}
		else {
			// Default theme not set, we need to search for the list of themes installed and just pick one.
			$dh = opendir(ROOT_PDIR . 'themes/');
			while(($file = readdir($dh)) !== false) {
				if(is_dir(ROOT_PDIR . 'themes/' . $file) && $file !== 'custom' && $file[0] !== '.') {
					$theme = $file;
					break;
				}
			}
			closedir($dh);
		}
		
		// Strip any extra characters, like spaces, newlines, etc.
		$theme = preg_replace('/[^a-z0-9\-]/', '', $theme);
		
		return $theme;
	}
	
	/**
	 * Get the theme object of a requested theme.
	 * 
	 * @param string|null $themeName Theme name to return, null for current default
	 * @return Theme
	 */
	public static function GetTheme($themeName = null) : Theme {
		if($themeName === null){
			$themeName = ConfigHandler::Get('/theme/selected');
		}
		elseif($themeName === 'AUTO'){
			$themeName = self::GetDefaultThemeName();
		}
		
		return new Theme(ROOT_PDIR . 'themes/' . $themeName . '/theme.xml');
	}
	
	/**
	 * Get all themes installed on the site
	 * 
	 * @return Theme[]
	 */
	public static function GetAllThemes() : array {
		$ret = [];
		
		$d = opendir(ROOT_PDIR . 'themes/');
		if(!$d){
			// Unable to open directory.
			return $ret;
		}
		
		while (($file = readdir($d)) !== false) {
			if(
				is_dir(ROOT_PDIR . 'themes/' . $file) &&
				$file[0] !== '.' &&
				$file !== 'custom' &&
				file_exists(ROOT_PDIR . 'themes/' . $file . '/theme.xml')
			){
				try {
					$ret[] = new Theme(ROOT_PDIR . 'themes/' . $file . '/theme.xml');
				}
				catch(Exception $ex) {
					\Core\log_exception($ex);
				}
			}
		}
		
		closedir($d);
		
		return $ret;
	}
	
	/**
	 * Get a list of skins for the currently enabled Theme as a Dictionary.
	 * 
	 * This is useful for plugging directly into the Form system.
	 * 
	 * @return \Core\Dict
	 */
	public static function GetSkinsAsOptions() : Dict {
		$out = new Dict();
		$out->autoI18N = true;
		
		// The first option in the optionset will be -BLANK- for the default configured skin.
		$out->addItem('', 't:STRING_DEFAULT_THEME_SKIN');
		
		$currentTheme = self::GetTheme();
		$skins = $currentTheme->getSkins();

		foreach($skins as $skin){

			// Skip system skins
			if($skin['is_system']) {
				continue;
			}

			$title = $skin['title'] ? $skin['title'] : $skin['file'];

			if($skin['default'] && $skin['admindefault']) {
				$title .= ' (site default)';
			}
			elseif($skin['default']) {
				$title .= ' (default public)';
			}
			elseif($skin['admindefault']) {
				$title .= ' (default admin)';
			}

			$out->addItem(
				$skin['file'],
				$title
			);
		}
		
		return $out;
	}
	
	
	/**
	 * Perform the internal logic to install a source File to a destination File.
	 * 
	 * Returns TRUE if the file was installed, FALSE if it was skipped.
	 * 
	 * @param \Core\Filestore\File $source
	 * @param \Core\Filestore\File $dest
	 * 
	 * @throws \InstallerException
	 * @throws \Exception
	 * 
	 * @return bool
	 */
	private function _installAssetFile(\Core\Filestore\File $source, \Core\Filestore\File $dest){
		
		// If both source and destination are identical, nothing needs to be done.
		if($dest->exists() && $dest->identicalTo($source)){
			return false;
		}
		
		if(!$source->exists()){
			throw new \InstallerException('Theme asset ' . $source->getFilename() . ' does not exist, unable to install file!');
		}
		
		if(!$source->isReadable()){
			throw new \InstallerException('Theme asset ' . $source->getFilename() . ' could not be read, unable to install file!');
		}
		
		$source->copyTo($dest, true);
		return true;
	}
	
	/**
	 * Save the log report for a given install/upgrade/reinstall action.
	 * 
	 * @param string $action one of "install", "upgrade", or "reinstall".  Sets the icon and code.
	 * @param array $results Merged result from the install methods.
	 */
	private function _logInstallResults(string $action, array $results){
		// No changes than nothing to report.
		if(sizeof($results['changes']) === 0 && sizeof($results['errors']) === 0){
			return;
		}
		
		switch($action){
			case 'install':
				$icon = 'download';
				break;
			case 'reinstall':
				$icon = 'refresh';
				break;
			case 'upgrade':
				$icon = 'arrow-circle-up';
				break;
			default:
				$icon = null;
		}
		
		// Create a log object to record, (gets modified based on results)
		$log = \Core\log()
			->level(LOG_LEVEL_INFO)
			->code('/updater/theme/' . $action)
			->details(implode("\n", array_merge($results['changes'], $results['errors'])));
		
		if(sizeof($results['errors']) === 0){
			// Completely successful!
			$log->message('Theme ' . $this->getName() . ' ' . $action . ' successful!')
				->icon($icon)
				->send();
		}
		else{
			// There were errors.
			$log->message('Theme ' . $this->getName() . ' ' . $action . ' completed with errors.')
				->type('error')
				->icon('warning')
				->send();
		}
	}
}
