<?php


namespace Core\CLI;

/**
 * Command Line Table
 *
 * Renders a pretty and responsive table in the CLI, similar to markdown.
 *
 * @package Core\CLI
 */
class CLITable {
	private $_columns = [];
	private $_records = [];

	public function addNamedColumn(string $key, string $header, string $align = 'left') {
		$this->_columns[$key] = [
			'text' => $header,
			'align' => $align,
			'color' => COLOR_HEADER
		];
	}

	public function addColumn(string $header, string $align = 'left') {
		$this->_columns[] = [
			'text' => $header,
			'align' => $align,
			'color' => COLOR_HEADER
		];
	}

	public function addRecord(array $data) {
		$record = [];

		foreach($data as $idx => $value) {
			if(is_array($value)) {
				// Allow the user to modify the formatting of fields inline.
				$text = $value[0];
				$color = $value['color'] ?? null;
				$align = $value['align'] ?? null;
			}
			else {
				$text = $value;
				$color = $align = null;
			}

			if(!isset($this->_columns[$idx])) {
				trigger_error('Too many columns added!  Please ensure the number of fields in a record match the number of columns.');
				continue;
			}

			$record[$idx] = [
				'text' => $text,
				'color' => $color,
				'align' => $align
			];
		}

		$this->_records[] = $record;
	}

	public function render() {
		$widths = $this->_calculateColumnWidths();

		// Render the header
		echo COLOR_LINE . '|' . COLOR_RESET;
		foreach($this->_columns as $idx => $data) {
			if($data['text'] === null) {
				$this->_renderNull($widths[$idx]);
			}
			else {
				$this->_renderText($data['text'], $widths[$idx], $data['align'], $data['color']);
			}
		}
		echo NL;

		// Render the separator
		echo COLOR_LINE . '|' . COLOR_RESET;
		foreach($this->_columns as $idx => $data) {
			$this->_renderSeparator($widths[$idx], $data['align']);
		}
		echo NL;

		// And lastly render each record
		foreach($this->_records as $record) {
			echo COLOR_LINE . '|' . COLOR_RESET;
			foreach($this->_columns as $idx => $columnData) {
				$align = $record[$idx]['align'] ?? $columnData['align'];
				$color = $record[$idx]['color'] ?? null;

				if($record[$idx]['text'] === null) {
					$this->_renderNull($widths[$idx]);
				}
				else {
					$this->_renderText($record[$idx]['text'], $widths[$idx], $align, $color);
				}
			}
			echo NL;
		}
	}

	/**
	 * Get the total number of records set
	 *
	 * @return int
	 */
	public function getTotal() : int {
		return count($this->_records);
	}

	public function getColumns() : array {
		return $this->_columns;
	}

	public function getRecords() : array {
		return $this->_records;
	}

	private function _renderSeparator(int $width, string $align) {
		if($align == 'center') {
			$pre = ':';
			$post = ':';
		}
		elseif($align === 'right') {
			$pre = '-';
			$post = ':';
		}
		else {
			$pre = '-';
			$post = '-';
		}

		echo COLOR_LINE . $pre . str_repeat('-', $width) . $post . '|' . COLOR_RESET;
	}

	private function _renderText(string $text, int $width, string $align, string $color = null) {
		if(strlen($text) > $width) {
			// Text is too large, cut it down.
			$text = substr($text, 0, $width);
		}

		if($align === 'center') {
			$padding = ($width - strlen($text)) / 2;
			$paddingPre = 1 + floor($padding);
			$paddingPost = 1 + ceil($padding);
		}
		elseif($align === 'right') {
			$paddingPre = $width - strlen($text) + 1;
			$paddingPost = 1;
		}
		else {
			// Left / default
			$paddingPre = 1;
			$paddingPost = $width - strlen($text) + 1;
		}

		echo str_repeat(' ', $paddingPre);
		if($color) {
			echo $color;
		}
		echo $text;
		if($color) {
			echo COLOR_RESET;
		}
		echo str_repeat(' ', $paddingPost);
		echo COLOR_LINE . '|' . COLOR_RESET;
	}

	private function _renderNull(int $width) {
		$text = '';

		// Left / default
		$paddingPre = 1;
		$paddingPost = $width - strlen($text) + 1;

		echo str_repeat(' ', $paddingPre);
		echo $text;
		echo str_repeat(' ', $paddingPost);
		echo COLOR_LINE . '|' . COLOR_RESET;
	}

	private function _calculateColumnWidths() {
		$ret = [];

		foreach($this->_columns as $idx => $column) {
			$ret[ $idx ] = strlen($column['text']);
		}

		foreach($this->_records as $record) {
			foreach($record as $idx => $data) {
				$ret[ $idx ] = max($ret[ $idx ], strlen($data['text']));
			}
		}

		// Check the total width of the content and ensure it fits within the viewframe.
		// If necessary, we will trim or adjust the width of each section to make it fit.
		// The table starts after 1 character, the left border.
		$contentWidth = 1;
		foreach($ret as $val) {
			// Increase the total width by the content plus the padding and border
			$contentWidth += $val + 3;
		}

		if($contentWidth > CLI::$MaxLineWidth) {
			// Content is too wide, trim it down.
			// We could probably do some elegant algorithm to solve this, but here we'll just use brute force.
			do {

				$difference = $contentWidth - CLI::$MaxLineWidth;
				if($difference > 20) {
					$jump = 10;
				}
				elseif($difference > 4) {
					$jump = 2;
				}
				else {
					$jump = 1;
				}

				$largest = 0;
				$largestIdx = null;
				foreach($ret as $idx => $size) {
					if($size > $largest) {
						$largest = $size;
						$largestIdx = $idx;
					}
				}

				if(!$largestIdx) {
					// This shouldn't happen, but who knows.
					break;
				}
				$ret[ $largestIdx ] -= $jump;
				$contentWidth -= $jump;
			}
			while($contentWidth > CLI::$MaxLineWidth);
		}
		elseif($contentWidth < CLI::$MaxLineWidth) {
			// More common scenario, the content is smaller than the viewport.
			// Here we can easily adjust all columns to make them fit well.
			$difference = CLI::$MaxLineWidth - $contentWidth;
			$additional = ceil($difference / count($ret));
			$lastIdx = count($ret) - 1;
			$baseWidth = $contentWidth;

			foreach($ret as $idx => $size) {
				if($lastIdx !== $idx) {
					// Calculate the ratio of this column compared to the total column width
					// to determine a relative amount to increase the size of this column.
					$additional = ceil(($size / $baseWidth) * $difference);
					$ret[ $idx ] += $additional;
					$contentWidth += $additional;
				}
				else {
					// Last column needs a little more attention to ensure a proper fit.
					$ret[ $idx ] += (CLI::$MaxLineWidth - $contentWidth);
				}
			}
		}

		return $ret;
	}
}