<?php
/**
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2018  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\EmailAPI;

use Core\Filestore\File;

/**
 * The interface for email systems to implement to ensure consistent functionality.
 */
interface EmailInterface {
	
	/**
	 * Construct a new email backend to send an email.
	 */
	public function __construct();
	
	/**
	 * Get all "to" addresses for this email.
	 * 
	 * @return EmailAddress[]
	 */
	public function getTo(): array;
	
	/**
	 * Get the "from" address for this email.
	 * 
	 * @return EmailAddress
	 */
	public function getFrom(): EmailAddress;
	
	/**
	 * Get all "CC" addresses for this email.
	 * 
	 * @return EmailAddress[]
	 */
	public function getCC(): array;
	
	/**
	 * Get all "BCC" addresses for this email.
	 * 
	 * @return EmailAddress[]
	 */
	public function getBCC(): array;
	
	/**
	 * Get the "Reply-To" address for this email.
	 * 
	 * @return EmailAddress
	 */
	public function getReplyTo(): EmailAddress;
	
	/**
	 * Get the subject for this email, as set by setSubject usually.
	 * 
	 * @return string
	 */
	public function getSubject(): string;
	
	/**
	 * Get all custom headers for this email
	 * 
	 * @return array
	 */
	public function getHeaders(): array;
	
	/**
	 * Get the raw and complete EML for this email, with headers, body, and all attachments.
	 * 
	 * @return string
	 */
	public function getFullEML(): string;
	
	/**
	 * Get any attachments on this email as an array
	 * 
	 * @return EmailAttachment[]
	 */
	public function getAttachments() : array;
	
	/**
	 * Get the main body, probably as HTML
	 * 
	 * @return string
	 */
	public function getMainBody() : string;
	
	/**
	 * Set the main body on this email, optionally setting the mimetype.
	 * 
	 * This usually sets the rich/HTML version, but can be different.
	 * 
	 * @param string $body
	 * @param string $mimetype
	 * 
	 * @return EmailInterface
	 */
	public function setMainBody(string $body, string $mimetype = 'text/html') : EmailInterface;
	
	/**
	 * Set the alternate plantext-only body on this email.
	 * 
	 * @param string $body
	 * 
	 * @return EmailInterface
	 */
	public function setAltBody(string $body) : EmailInterface;
	
	/**
	 * Set the main and alt bodies of this email from rich HTML contents
	 * 
	 * @param string $body
	 * 
	 * @return EmailInterface
	 */
	//public function setBodyFromHTML(string $body) : EmailInterface;
	
	/**
	 * Set the "to" address for this email.
	 *
	 * Will clear out any other to address.
	 * 
	 * Return the parent object to allow chaining.
	 *
	 * @param string $address
	 * @param string $name
	 * 
	 * @return EmailInterface
	 */
	public function setTo(string $address, string $name = ''): EmailInterface;

	/**
	 * Convenience function for sending an email to a User account
	 *
	 * @param \UserModel $user
	 *
	 * @return EmailInterface
	 */
	public function setToUser(\UserModel $user) : EmailInterface ;

	/**
	 * Set the "from" address for this email
	 * 
	 * Return the parent object to allow chaining.
	 * 
	 * @param string $address
	 * @param string $name
	 * 
	 * @return EmailInterface
	 */
	public function setFrom(string $address, string $name = ''): EmailInterface;
	
	/**
	 * Set the "Sender" address for this email
	 * 
	 * Return the parent object to allow chaining.
	 * 
	 * @param string $address
	 * @param string $name
	 * 
	 * @return EmailInterface
	 */
	public function setSender(string $address, string $name = ''): EmailInterface;
	
	/**
	 * Set the Reply To address for this email.
	 *
	 * @param string $address
	 * @param string $name
	 * 
	 * @return EmailInterface
	 */
	public function setReplyTo(string $address, string $name = ''): EmailInterface;

	/**
	 * Set the subject of this email.
	 *
	 * @param string $subject
	 * 
	 * @return EmailInterface
	 */
	public function setSubject(string $subject): EmailInterface;
	
	/**
	 * Set a custom header on this Email object.
	 * 
	 * @param string $key
	 * @param string $value
	 * 
	 * @return EmailInterface
	 */
	public function setHeader(string $key, string $value): EmailInterface;
	
	/**
	 * Add a "to" address for this email.
	 *
	 * Can be called multiple times for sending to multiple people, will
	 * add addresses to the "to" recipient on each call.
	 *
	 * @param string $address
	 * @param string $name
	 * 
	 * @return EmailInterface
	 */
	public function addTo(string $address, string $name = ''): EmailInterface;
	
	/**
	 * Add a carbon copy email address.
	 * 
	 * Return the parent object to allow chaining.
	 * 
	 * @param string $address
	 * @param string $name
	 * 
	 * @return EmailInterface
	 */
	public function addCC(string $address, string $name = ''): EmailInterface;
	
	/**
	 * Add a blind carbon copy email address.
	 * 
	 * Return the parent object to allow chaining.
	 * 
	 * @param string $address
	 * @param string $name
	 * 
	 * @return EmailInterface
	 */
	public function addBCC(string $address, string $name = ''): EmailInterface;

	/**
	 * Add a Reply To address for this email.
	 *
	 * @param string $address
	 * @param string $name
	 * 
	 * @return EmailInterface
	 */
	public function addReplyTo(string $address, string $name = ''): EmailInterface;

	/**
	 * Add a file as an attachment
	 *
	 * @param File $file   The file to add as an attachment.
	 * @param bool $inline Set to true to request this attachment to be viewed inline.
	 *
	 * @throws \Exception
	 * 
	 * @return EmailInterface
	 */
	public function addAttachmentFromFile(File $file, $inline = false): EmailInterface;

	/**
	 * Add a file stream as an attachment
	 *
	 * @param string $data     The data to attach to this email
	 * @param string $filename Client Rendering Hint for file name
	 * @param string $mimetype Client Rendering Hint for file mimetype
	 * @param bool   $inline   Set to true to request this attachment to be viewed inline.
	 *
	 * @throws \Exception
	 * 
	 * @return EmailInterface
	 */
	public function addAttachmentFromStream(string $data, string $filename, string $mimetype, $inline = false): EmailInterface;
	
	/**
	 * Send this email!
	 *
	 * @throws \Exception
	 */
	public function send();
	
	/**
	 * Clear all recipients, (to, cc, and bcc), for this email.
	 *
	 * @throws \Exception
	 * 
	 * @return EmailInterface
	 */
	public function clearRecipients(): EmailInterface;
	
	/**
	 * Clear all attachments for this email
	 * 
	 * @return EmailInterface
	 */
	public function clearAttachments() : EmailInterface;
	
	/**
	 * Enable debug operations on this email send, if supported.
	 */
	public function enableDebug();
	
	/**
	 * Disable debug operations on this email send, if supported.
	 */
	public function disableDebug();
}
