<?php
/**
 * Description of ContentTGZ
 *
 * Provides useful extra functions that can be done with a GZipped file.
 *
 * @package
 * @since 0.1
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */


namespace Core\Filestore\Contents;

use Core\Filestore\Directory;
use Core\Filestore\Factory;
use Core\Filestore\ContentsInterface;
use Core\Filestore\File;
use Exception;

class ContentTGZ implements ContentsInterface {
	private $_file = null;

	public function __construct(File $file) {
		$this->_file = $file;
	}

	public function getContents() {
		return $this->_file->getContents();
	}

	/**
	 * Extract this archive to a requested directory.
	 *
	 * @param string|Directory $dst Destination to extract the archive to.
	 *
	 * @throws Exception
	 *
	 * @return Directory
	 */
	public function extract($destdir) : Directory {
		if($destdir instanceof Directory) {
			$d = $destdir;
		}
		else {
			// This will ensure that the destdir is properly resolved.
			$d = Factory::Directory($destdir);
		}

		if(!$d->exists()) {
			$d->mkdir();
		}

		exec('tar -xzf "' . $this->_file->getLocalFilename() . '" -C "' . $d->getPath() . '"');
		return $d;
	}

	/**
	 * Extract a single file from the tarball
	 * @throws Exception
	 */
	public function extractSingleFile(string $filename, string $destination = null) {
		$cmd = 'tar -xzf ' . escapeshellarg($this->getExtractedPackage()->getFilename()) . ' ' . escapeshellarg($filename);

		if($destination) {
			$cmd .= ' -O > ' . escapeshellarg($destination);
		}
		else {
			$cmd .= ' -O';
		}

		$outputPipes = [];
		$descriptorspec = [
			0 => array('pipe', 'r'),  // stdin
			1 => array('pipe', 'w'),  // stdout
			2 => array('pipe', 'w'),  // stderr
		];
		$process = proc_open($cmd, $descriptorspec, $outputPipes);

		$stdout = stream_get_contents($outputPipes[1]);
		fclose($outputPipes[1]);

		$stderr = stream_get_contents($outputPipes[2]);
		fclose($outputPipes[2]);

		$result = proc_close($process);

		if($result !== 0) {
			if($stderr) {
				throw new Exception($stderr);
			}
			else {
				throw new Exception('Unable to extract ' . $filename . ' from tarball due to unknown exception');
			}
		}
		else {
			return $stdout;
		}
	}

	public function listfiles() {
		$output = array();
		exec('tar -zf "' . $this->_file->getLocalFilename() . '" --list', $output);

		foreach ($output as $k => $v) {
			// Trim some characters off.
			if (strpos($v, './') === 0) $v = substr($v, 2);

			if (!$v) unset($output[$k]);
			else $output[$k] = $v;
		}

		return array_values($output);
	}
}

