<?php
/**
 * File for class PrimaryKey definition in the coreplus project
 * 
 * @package Core\GPG
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20140319.2122
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\GPG;
use Core\Filestore\File;


/**
 * A short teaser of what PrimaryKey does.
 *
 * More lengthy description of what PrimaryKey does and why it's fantastic.
 *
 * <h3>Usage Examples</h3>
 *
 *
 * @todo Write documentation for PrimaryKey
 * <h4>Example 1</h4>
 * <p>Description 1</p>
 * <code>
 * // Some code for example 1
 * $a = $b;
 * </code>
 *
 *
 * <h4>Example 2</h4>
 * <p>Description 2</p>
 * <code>
 * // Some code for example 2
 * $b = $a;
 * </code>
 *
 * 
 * @package Core\GPG
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
class PrimaryKey extends Key {

	/** @var UID[] UID keys on this primary key */
	public $uids = [];

	/** @var array Sub keys on this primary key */
	public $subkeys = [];
	
	/** @var null|array Cache of photos on the filesystem for this primary key. */
	private $_photos = null;

	/**
	 * Check and see if this key is currently valid and not expired nor revoked.
	 *
	 * Optionally, pass in an email address to check the key's specific UID and see if that subkey is valid.
	 *
	 * @param string|null $email Optional email to verify that needs to be attached to this key.
	 *
	 * @return bool
	 */
	public function isValid($email = null){
		/*
		 * o = Unknown (this key is new to the system)
		 * i = The key is invalid (e.g. due to a missing self-signature)
		 * d = The key has been disabled (deprecated - use the 'D' in field 12 instead)
		 * r = The key has been revoked
		 * e = The key has expired
		 * - = Unknown validity (i.e. no value assigned)
		 * q = Undefined validity
		 * '-' and 'q' may safely be treated as the same value for most purposes
		 * n = The key is valid
		 * m = The key is marginal valid.
		 * f = The key is fully valid
		 * u = The key is ultimately valid.  This often means that the secret key is available, but any key may be marked as ultimately valid.
		 */

		switch($this->validity){
			case 'i':
			case 'd':
			case 'D':
			case 'r':
			case 'e':
				return false;
		}

		if($email){
			// Also check for that subkey.
			$uid = $this->getUID($email);
			if(!$uid){
				// UID not attached, means it's not valid!
				return false;
			}
			if(!$uid->isValid()){
				return false;
			}
		}

		// Otherwise, all checks passed.
		return true;
	}

	/**
	 * @param $email
	 *
	 * @return UID|null
	 */
	public function getUID($email){
		foreach($this->uids as $u){
			/** @var UID $u */
			if($u->email == $email){
				return $u;
			}
		}
		return null;
	}

	/**
	 * Get the name of the first valid UID on this primary key.
	 *
	 * @return string
	 */
	public function getName() : string {
		foreach($this->uids as $u){
			/** @var UID $u */
			if($u->isValid()){
				return $u->fullname;
			}
		}
		return '';
	}

	public function getComment() : string {
		foreach($this->uids as $u){
			/** @var UID $u */
			if($u->isValid() && $u->comment){
				return $u->comment;
			}
		}
		return '';
	}
	
	/**
	 * Get the email of the first valid UID on this primary key.
	 *
	 * @return string
	 */
	public function getEmail() : string {
		foreach($this->uids as $u){
			/** @var UID $u */
			if($u->isValid()){
				return $u->email;
			}
		}
		return '';
	}

	/**
	 * Get this key as a human-friendly label
	 *
	 * @param string $format
	 * @return string
	 */
	public function getLabel(string $format = 'text/html') : string {
		$l = $this->id_short . ' - ' . $this->encryptionType . '/' . $this->encryptionBits;

		$c = $this->getComment();
		$n = $this->getName();
		$e = $this->getEmail();


		if($n) {
			$l .= ' ' . $n;
		}
		if($e) {
			if($format === 'text/html') {
				$l .= ' &lt;' . $e . '&gt;';
			}
			else {
				$l .= ' <' . $e . '>';
			}
		}

		if($c) {
			$l .= ' "' . ($format === 'text/html' ? htmlentities($c) : $c) . '"';
		}

		return $l === null ? 'Key Not Found' : $l;
	}

	/**
	 * Get an array containing all the photos for this public key
	 * 
	 * @return array
	 */
	public function getPhotos(){
		if($this->_photos === null){
			$this->_photos = [];
			// By specifying these options, the contents are extracted automatically and the filename is rendered.
			$gpg = new GPG();
			$o = $gpg->_exec('--list-keys --list-options show-photos --photo-viewer "echo %I >&2" ' . $this->fingerprint . ' 1>/dev/null ');
			
			if($o['return'] === 0){
				// In this case, stderr should contain the output.
				$files = explode("\n", $o['error']);

				foreach($files as $f){
					if($f && strpos($f, 'gpg:') !== 0){
						$this->_photos[] = \Core\Filestore\Factory::File($f);
					}
				}
			}
		}
		
		return $this->_photos;
	}

	/**
	 * Get the FIRST photo in this key
	 * 
	 * @return File|null
	 */
	public function getPhoto(){
		$p = $this->getPhotos();
		
		if(!sizeof($p)){
			return null;
		}
		else{
			return $p[0];
		}
	}

	/**
	 * Get the ASCII representation of this key
	 * 
	 * @return string
	 * 
	 * @throws \Exception
	 */
	public function getAscii() {
		die('Please extend this method in the appropriate class!');
	}

	/**
	 * Get this key in its binary format
	 */
	public function getBinary() {
		die('Please extend this method in the appropriate class!');
	}

	/**
	 * Parse a line from --with-colons output.
	 * 
	 * doc pulled form https://git.gnupg.org/cgi-bin/gitweb.cgi?p=gnupg.git;a=blob_plain;f=doc/DETAILS
	 *
	 * <pre>
	 * pub:f:1024:17:6C7EE1B8621CC013:899817715:1055898235::m:::scESC:
	 * fpr:::::::::ECAF7590EB3443B5C7CF3ACB6C7EE1B8621CC013:
	 * uid:f::::::::Werner Koch <wk@g10code.com>:
	 * uid:f::::::::Werner Koch <wk@gnupg.org>:
	 * sub:f:1536:16:06AD222CADF6A6E1:919537416:1036177416:::::e:
	 * fpr:::::::::CF8BCC4B18DE08FCD8A1615906AD222CADF6A6E1:
	 * sub:r:1536:20:5CE086B5B5A18FF4:899817788:1025961788:::::esc:
	 * fpr:::::::::AB059359A3B81F410FCFF97F5CE086B5B5A18FF4:
	 *
	 * The double --with-fingerprint prints the fingerprint for the subkeys
	 * too. --fixed-list-mode is the modern listing way printing dates in
	 * seconds since Epoch and does not merge the first userID with the pub
	 * record; gpg2 does this by default and the option is a dummy.
	 *
	 *
	 * 1. Field:  Type of record
	 * 
	 * - pub :: Public key
	 * - crt :: X.509 certificate
	 * - crs :: X.509 certificate and private key available
	 * - sub :: Subkey (secondary key)
	 * - sec :: Secret key
	 * - ssb :: Secret subkey (secondary key)
	 * - uid :: User id
	 * - uat :: User attribute (same as user id except for field 10).
	 * - sig :: Signature
	 * - rev :: Revocation signature
	 * - fpr :: Fingerprint (fingerprint is in field 10)
	 * - pkd :: Public key data [*]
	 * - grp :: Keygrip
	 * - rvk :: Revocation key
	 * - tfs :: TOFU statistics [*]
	 * - tru :: Trust database information [*]
	 * - spk :: Signature subpacket [*]
	 * - cfg :: Configuration data [*]
	 * 
	 *
	 * 2. Field:  A letter describing the calculated validity. This is a single
	 *            letter, but be prepared that additional information may follow
	 *            in some future versions. (not used for secret keys)
	 * 
	 * - o :: Unknown (this key is new to the system)
	 * - i :: The key is invalid (e.g. due to a missing self-signature)
	 * - d :: The key has been disabled
	 *    (deprecated - use the 'D' in field 12 instead)
	 * - r :: The key has been revoked
	 * - e :: The key has expired
	 * - - :: Unknown validity (i.e. no value assigned)
	 * - q :: Undefined validity.  '-' and 'q' may safely be treated as
	 * 	   the same value for most purposes
	 * - n :: The key is not valid
	 * - m :: The key is marginal valid.
	 * - f :: The key is fully valid
	 * - u :: The key is ultimately valid.  This often means that the
	 * 	   secret key is available, but any key may be marked as
	 * 	   ultimately valid.
	 * - w :: The key has a well known private part.
	 * - s :: The key has special validity.  This means that it might be
	 * 	   self-signed and expected to be used in the STEED system.
	 *
	 * If the validity information is given for a UID or UAT
	 * record, it describes the validity calculated based on this
	 * user ID.  If given for a key record it describes the best
	 * validity taken from the best rated user ID.
	 *
	 * For X.509 certificates a 'u' is used for a trusted root
	 * certificate (i.e. for the trust anchor) and an 'f' for all
	 * other valid certificates.
	 *
	 * 3. Field:  length of key in bits.
	 *
	 * 4. Field:  Algorithm:	1 = RSA
	 * 16 = Elgamal (encrypt only)
	 * 17 = DSA (sometimes called DH, sign only)
	 * 20 = Elgamal (sign and encrypt - don't use them!)
	 * (for other id's see include/cipher.h)
	 *
	 * 5. Field:  KeyID
	 *
	 * 6. Field:  Creation Date (in UTC).  For UID and UAT records, this is
	 * the self-signature date.  Note that the date is usally
	 * printed in seconds since epoch, however, we are migrating
	 * to an ISO 8601 format (e.g. "19660205T091500").  This is
	 * currently only relevant for X.509.  A simple way to detect
	 * the new format is to scan for the 'T'.
	 *
	 * 7. Field:  Key or user ID/user attribute expiration date or empty if none.
	 *
	 * 8. Field:  Used for serial number in crt records (used to be the Local-ID).
	 * For UID and UAT records, this is a hash of the user ID contents
	 * used to represent that exact user ID.  For trust signatures,
	 * this is the trust depth seperated by the trust value by a
	 * space.
	 *
	 * 9. Field:  Ownertrust (primary public keys only)
	 * This is a single letter, but be prepared that additional
	 * information may follow in some future versions.  For trust
	 * signatures with a regular expression, this is the regular
	 * expression value, quoted as in field 10.
	 *
	 * 10. Field:  User-ID.  The value is quoted like a C string to avoid
	 * control characters (the colon is quoted "\x3a").
	 * For a "pub" record this field is not used on --fixed-list-mode.
	 * A UAT record puts the attribute subpacket count here, a
	 * space, and then the total attribute subpacket size.
	 * In gpgsm the issuer name comes here
	 * An FPR record stores the fingerprint here.
	 * The fingerprint of an revocation key is stored here.
	 *
	 * 11. Field:  Signature class as per RFC-4880.  This is a 2 digit
	 * hexnumber followed by either the letter 'x' for an
	 * exportable signature or the letter 'l' for a local-only
	 * signature.  The class byte of an revocation key is also
	 * given here, 'x' and 'l' is used the same way.  IT is not
	 * used for X.509.
	 *
	 * 12. Field:  Key capabilities:
	 * 
	 * - e :: Encrypt
	 * - s :: Sign
	 * - c :: Certify
	 * - a :: Authentication
	 * - ? :: Unknown capability
	 * 
	 * A key may have any combination of them in any order.  In
	 * addition to these letters, the primary key has uppercase
	 * versions of the letters to denote the _usable_
	 * capabilities of the entire key, and a potential letter 'D'
	 * to indicate a disabled key.
	 *
	 * 13. Field:  Used in FPR records for S/MIME keys to store the
	 * fingerprint of the issuer certificate.  This is useful to
	 * build the certificate path based on certificates stored in
	 * the local keyDB; it is only filled if the issuer
	 * certificate is available. The root has been reached if
	 * this is the same string as the fingerprint. The advantage
	 * of using this value is that it is guaranteed to have been
	 * been build by the same lookup algorithm as gpgsm uses.
	 * For "uid" records this lists the preferences in the same
	 * way the gpg's --edit-key menu does.
	 * For "sig" records, this is the fingerprint of the key that
	 * issued the signature.  Note that this is only filled in if
	 * the signature verified correctly.  Note also that for
	 * various technical reasons, this fingerprint is only
	 * available if --no-sig-cache is used.
	 *
	 * 14. Field   Flag field used in the --edit menu output:
	 *
	 * 15. Field   S/N of a token
	 * 
	 * (internal protect mode 1002) or a '#' if that key is a
	 * simple stub (internal protect mode 1001)
	 * 
	 * 16. Field:  Hash algorithm
	 * 
	 * 2 = SHA-1
	 * 8 = SHA-256
	 * (for other id's see include/cipher.h)
	 * 
	 * Field 17 - Curve name
	 *            For pub, sub, sec, and ssb records this field is used for the ECC
	 *            curve name.
	 * 
	 * Field 18 - Compliance flags
	 *            Space separated list of asserted compliance modes and
	 *            screening result for this key.
	 * 
	 * Valid values are:
	 * - 8  :: The key is compliant with RFC4880bis
	 * - 23 :: The key is compliant with compliance mode "de-vs".
	 * - 6001 :: Screening hit on the ROCA vulnerability.
	 * 
	 * Field 19 - Last update
	 *            The timestamp of the last update of a key or user ID.  The update
	 *            time of a key is defined a lookup of the key via its unique
	 *            identifier (fingerprint); the field is empty if not known.  The
	 *            update time of a user ID is defined by a lookup of the key using a
	 *            trusted mapping from mail address to key.
	 * 
	 * Field 20 - Origin
	 *            The origin of the key or the user ID.  This is an integer
	 *            optionally followed by a space and an URL.  This goes along with
	 *            the previous field.  The URL is quoted in C style.
	 *
	 * All dates are displayed in the format yyyy-mm-dd unless you use the
	 * option --fixed-list-mode in which case they are displayed as seconds
	 * since Epoch.  More fields may be added later, so parsers should be
	 * prepared for this. When parsing a number the parser should stop at the
	 * first non-number character so that additional information can later be
	 * added.
	 *
	 * If field 1 has the tag "pkd", a listing looks like this:
	 * pkd:0:1024:B665B1435F4C2 .... FF26ABB:
	 * !  !   !-- the value
	 * !  !------ for information number of bits in the value
	 * !--------- index (eg. DSA goes from 0 to 3: p,q,g,y)
	 * </pre>
	 *
	 * @param string $line
	 */
	public function parseLine($line) {
		
		echo $line . "\n<br/>";
		
		$parts = explode(':', $line);
		
		echo $parts[0] . ' (' . count($parts) . ")\n<br/>";

		switch($parts[0]) {
			case 'pub':
			case 'sec':
				// Primary key, either public or secret.
				
				$s = sizeof($parts);
				
				if($s === 21){
					$this->_parseLineWithMap($parts, self::$_PubMap21);
				}
				elseif($s === 19 || $s === 16 || $s === 13){
					// Public keys on newer versions of GPG have 19 parts.
					// PUblic keys on 2.1.23 and newer have 21 parts.
					// Secret keys often have 16 parts, but the 13 length can handle it.
					$this->_parseLineWithMap($parts, self::$_PubMap13);
				}
				elseif($s === 11){
					// Public keys BEFORE being imported, eg: examine, will have only 11 parts.
					// This also includes integrated UID data.
					$this->_parseLineWithMap($parts, self::$_PubMap13);
				}
				else {
					$this->_parseLineWithMap($parts, self::$_PubMap7);
				}

				break;
			case 'fpr':
				// Fingerprint
				$this->fingerprint = $parts[9];
				break;
			case 'uid':
				// UID sub key
				$uid = new UID();
				$s = sizeof($parts);
				
				if($s === 21 || $s === 19 || $s === 11){
					// New versions of GPG have 19 parts to UIDs.
					// Older versions use 11.
					$uid->_parseKeyLine11($parts);
				}
				else {
					self::_ParseSubUIDLine5($parts, $uid);
				}
				$this->uids[] = $uid;

				break;
			case 'uat':
				// Skip user attributes.
				break;
			case 'sub':
			case 'ssb':
				$sub = new Key();
				if(sizeof($parts) == 13) {
					self::_ParseKeyLine13($parts, $sub);
				}
				else {
					self::_ParseKeyLine7($parts, $sub);
				}
				$this->subkeys[] = $sub;
				break;
			case 'sig':
				// Signatures are public signatures on a given UID.
				// This requires the last-added UID to be appended to.
				$last = current($this->uids);
				if($last){
					$last->_parseSig($parts);
				}
				
			/*
* crt = X.509 certificate
* crs = X.509 certificate and private key available
* sub = subkey (secondary key)
* ssb = secret subkey (secondary key)
* uid = user id (only field 10 is used).
* uat = user attribute (same as user id except for field 10).
* sig = signature
* rev = revocation signature
* fpr = fingerprint: (fingerprint is in field 10)
* pkd = public key data (special field format, see below)
* grp = keygrip
* rvk = revocation key
* tru = trust database information
* spk = signature subpacket
			 */
		}
	}
	
	protected function _parseKeyLine13($parts){
		/*
		 * (0-index of keys)
		 * 0.  Field:  Type of record
		 * 1.  Field:  A letter describing the calculated validity.
		 * 2.  Field:  length of key in bits.
		 * 3.  Field:  Algorithm:
		 * 4.  Field:  KeyID
		 * 5.  Field:  Creation Date (in UTC).
		 * 6.  Field:  Key or user ID/user attribute expiration date or empty if none.
		 * 7.  Field:  Used for serial number in crt records (used to be the Local-ID).
		 * 8.  Field:  Ownertrust (primary public keys only)
		 * 9.  Field:  User-ID.
		 * 10. Field:  Signature class as per RFC-4880.
		 * 11. Field:  Key capabilities
		 * 12. Field:  Used in FPR records for S/MIME keys to store the fingerprint of the issuer certificate.
		 * 13. Field:  Flag field used in the --edit menu output:
		 * 14. Field:  Used in sec/sbb to print the serial number
		 * 15. Field:  For sig records, this is the used hash algorithm:
		 */
		//$this->type = $parts[0];
		$this->validity = $parts[1];
		$this->encryptionBits = $parts[2];

		switch($parts[3]){
			case '1':
				$this->encryptionType = Key::ENCRYPTION_TYPE_RSA;
				break;
			case '16':
				$this->encryptionType = Key::ENCRYPTION_TYPE_ELGAMAL;
				break;
			case '17':
				$this->encryptionType = Key::ENCRYPTION_TYPE_DSA;
				break;
			case '20':
				$this->encryptionType = Key::ENCRYPTION_TYPE_ELGAMAL;
				break;
		}

		$this->id = $parts[4];
		$this->id_short = substr($parts[4], -8);
		$this->created = $parts[5];
		$this->expires = $parts[6];
	}

	protected function _parseKeyLine11($parts){
		var_dump($parts);
		/*
		 * (0-index of keys)
		 * 0.  Field:  Type of record
		 * 1.  Field:  A letter describing the calculated validity.
		 * 2.  Field:  length of key in bits.
		 * 3.  Field:  Algorithm:
		 * 4.  Field:  KeyID
		 * 5.  Field:  Creation Date (in UTC).
		 * 6.  Field:  Key or user ID/user attribute expiration date or empty if none.
		 * 7.  Field:  Used for serial number in crt records (used to be the Local-ID).
		 * 8.  Field:  Ownertrust (primary public keys only)
		 * 9.  Field:  User-ID.
		 * 10. Field:  Signature class as per RFC-4880.
		 * 11. Field:  Key capabilities
		 * 12. Field:  Used in FPR records for S/MIME keys to store the fingerprint of the issuer certificate.
		 * 13. Field:  Flag field used in the --edit menu output:
		 * 14. Field:  Used in sec/sbb to print the serial number
		 * 15. Field:  For sig records, this is the used hash algorithm:
		 */
		//$this->type = $parts[0];
		$this->validity = $parts[1];
		$this->encryptionBits = $parts[2];

		switch($parts[3]){
			case '1':
				$this->encryptionType = Key::ENCRYPTION_TYPE_RSA;
				break;
			case '16':
				$this->encryptionType = Key::ENCRYPTION_TYPE_ELGAMAL;
				break;
			case '17':
				$this->encryptionType = Key::ENCRYPTION_TYPE_DSA;
				break;
			case '20':
				$this->encryptionType = Key::ENCRYPTION_TYPE_ELGAMAL;
				break;
		}

		$this->id       = $parts[4];
		$this->id_short = substr($parts[4], -8);
		$this->created  = $parts[5];
		$this->expires  = $parts[6];
		$this->serial   = $parts[7];
		
		if($parts[9] && $this instanceof PublicKey) {
			$uid = new UID();
			$split = GPG::ParseAuthorString($parts[9]);
			$uid->comment = $split['comment'];
			$uid->email = $split['email'];
			$uid->fullname = $split['name'];
			$uid->validity = $this->validity;
			$uid->id_short = $this->id_short;
			$uid->id = $this->id;
			$this->uids[] = $uid;
		}
		elseif($parts[9] && $this instanceof UID) {
			$split = GPG::ParseAuthorString($parts[9]);
			$this->comment = $split['comment'];
			$this->email = $split['email'];
			$this->fullname = $split['name'];
		}
	}
	
	/**
	 * Parse a colon-delimited string for a 7-part data field, (usually from remote sources).
	 *
	 * @param     $parts
	 */
	protected function _parseKeyLine7($parts){
		/*
		 * (0-index of keys)
		 * 0.  Field:  Type of record
		 * 1.  Field:  KeyID
		 * 2.  Field:  Algorithm:
		 * 3.  Field:  length of key in bits.
		 * 4.  Field:  Creation Date (in UTC).
		 * 5.  Field:  ???
		 * 6.  Field:  ???
		 */
		//$this->type = $parts[0];
		$this->encryptionBits = $parts[3];

		switch($parts[2]){
			case '1':
				$this->encryptionType = Key::ENCRYPTION_TYPE_RSA;
				break;
			case '16':
				$this->encryptionType = Key::ENCRYPTION_TYPE_ELGAMAL;
				break;
			case '17':
				$this->encryptionType = Key::ENCRYPTION_TYPE_DSA;
				break;
			case '20':
				$this->encryptionType = Key::ENCRYPTION_TYPE_ELGAMAL;
				break;
		}

		$this->id = $parts[1];
		$this->fingerprint = $parts[1]; // On remote servers, they rarely give the fingerprint.
		$this->id_short = substr($parts[1], -8);
		$this->created = $parts[4];
		//$key->expires = $parts[6];
	}

	protected static function _ParseSubUIDLine11($parts, UID $uid){
		$uid->validity = $parts[1];
		$uid->created = $parts[5];
		$uid->expires = $parts[6];
		$uid->fingerprint = $parts[7];
		$uid->id = substr($parts[7], -16);
		$uid->id_short = substr($parts[7], -8);

		$split = GPG::ParseAuthorString($parts[9]);
		$uid->fullname = $split['name'];
		$uid->email    = $split['email'];
		$uid->comment  = $split['comment'];
	}

	protected static function _ParseSubUIDLine5($parts, UID $uid){
		/*
		 * (0-index of keys)
		 * 0.  Field:  Type of record
		 * 1.  Field:  Key name and/or email
		 * 2.  Field:  Creation Date (in UTC).
		 * 3.  Field:  ???
		 * 4.  Field:  ???
		 */
		//$uid->validity = $parts[1];
		$uid->created = $parts[2];
		//$uid->expires = $parts[6];
		//$uid->serial = $parts[7];
		$split = GPG::ParseAuthorString($parts[1]);
		$uid->fullname = $split['name'];
		$uid->email    = $split['email'];
		$uid->comment  = $split['comment'];
	}

} 