<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core\I18N;

use ReflectionClass;
use Core\Model;
use Core\ModelSupplemental;
use ConfigModel;

/**
 * Description of I18NScanner
 *
 * @author charlie
 */
class I18NScanner {
	/**
	 * Scan a directory for I18N string constants and return an array of located strings
	 * 
	 * @param string $directory Directory to scan
	 * 
	 * @return array List of results
	 */
	public static function ScanDirectory(string $directory) : array {
		
		// Here we want to find any STRING_, MESSAGE_, or FORMAT_ string in PHP, TPL, and XML files.
		$otuput = [];
		exec("find " . escapeshellarg($directory) . " -type f -regex '.*\\.\\(php\\|tpl\\|xml\\|phtml\\)' -exec egrep -R 'STRING_|MESSAGE_|FORMAT_' {} \\;", $output);

		$matches = [];
		foreach($output as $line){
			preg_match_all('/([\'"]|t:)((STRING|MESSAGE|FORMAT)_[A-Za-z0-9_ -]*)/', $line, $match);
			foreach($match[2] as $m){
				$m = trim($m);
				// If the last character of this string is NOT an underscore, then add it to the list.
				// This is to skip instances where message checks are caught, such as in this exact script!
				// As above, the exec line will trigger this catch if this check is not in place, giving false positives in the results.
				if(strrpos($m, '_') + 1 != strlen($m)){
					$matches[] = $m;

					// If this match contains _N_, then also add _0_ for 0 results and _1_ for 1 result.
					if(substr_count($m, '_N_') === 1){
						$matches[] = str_replace('_N_', '_0_', $m);
						$matches[] = str_replace('_N_', '_1_', $m);
					}
				}
			}
		}
		
		return $matches;
	}
	
	/**
	 * Scan a ConfigModel for its I18N auto-strings
	 * 
	 * @param ConfigModel $config
	 * @return array
	 */
	public static function ScanConfig(ConfigModel $config) : array {
		$matches = [];
		
		$key = I18NLoader::KeyifyString($config->get('key'));
		//$matches[] = 'STRING_CONFIG_' . $key;
		//$matches[] = 'MESSAGE_CONFIG_' . $key;

		$atts = $config->getFormAttributes();
		
		if(strpos($atts['title'], 't:') === 0){
			$matches[] = substr($atts['title'], 2);
		}
		
		if(strpos($atts['description'], 't:') === 0){
			$matches[] = substr($atts['description'], 2);
		}
		
		if(
			($atts['type'] === 'select' || $atts['type'] === 'checkboxes') &&
			isset($atts['options'])
		){
			foreach($atts['options'] as $v){
				if(strpos($v, 't:') === 0){
					$matches[] = substr($v, 2);
				}
			}
		}
		
		return $matches;
	}
	
	/**
	 * Scan a permission key for its I18N auto-strings
	 * 
	 * @param string $permission
	 * 
	 * @return array
	 */
	public static function ScanPermission( string $permission ) : array {
		$matches = [];
		
		// Parent group for this permission
		$group = substr($permission, 1, strpos($permission, '/', 1)-1);
		$matches[] = 'STRING_PERMISSION_' . strtoupper($group);
		$matches[] = 'MESSAGE_PERMISSION_' . strtoupper($group);
		
		// Keyify can handle the permission itself.
		$matches[] = 'STRING_PERMISSION_' . I18NLoader::KeyifyString($permission);
		
		return $matches;
	}
	
	public static function ScanModel( Model $model ) : array {
		$matches = [];

		$modelName = get_class($model);
		$schema = $model::GetSchema();
		
		// Schema now contains the schema for this model!
		// These should support translation strings
		foreach($schema->definitions as $column) {
			
			$i18nKey = 'MODEL_' . strtoupper($modelName . '_' . $column->field);

			if(
				$column->type === Model::ATT_TYPE_ALIAS
			){
				// Skip aliases, they don't need keys.
				continue;
			}
			
			// Skip supplementally defined attributes, they should only be present on the source component.
			if($column->model !== $modelName){
				continue;
			}

			$matches[] = 'STRING_' . $i18nKey;
			$matches[] = 'MESSAGE_' . $i18nKey;
			
			// ENUM Columns get their options mapped too!
			if($column->type === Model::ATT_TYPE_ENUM && is_array($column->options)) {
				foreach($column->options as $opt){
					$matches[] = 'VALUE_' . $i18nKey . '_' . I18NLoader::KeyifyString($opt);
				}
			}
		}
		
		// Include meta fields for the title; these can be included in listing tables.
		foreach($schema->metas as $key => $dat) {
			$i18nKey = 'MODEL_' . strtoupper($modelName . '_' . $key);
			$matches[] = 'STRING_' . $i18nKey;
		}
		
		return $matches;
	}
	
	public static function ScanSupplementalModel( ModelSupplemental $model ) : array {
		$matches = [];
		
		$modelName = get_class($model);
		// This model name needs some trimming performed to chop off the dynamic prefix and "modelsupplemental" suffix.
		$modelName = strtoupper(substr($modelName, strpos($modelName, '_')+1, -17)) . 'MODEL';
		$schema = $model::$Schema;
		
		// Schema now contains the schema for this model!
		// These should support translation strings
		foreach($schema as $column => $dat){
			$i18nKey = 'MODEL_' . $modelName . '_' . strtoupper($column);

			if(
				$dat['type'] == Model::ATT_TYPE_ALIAS
			){
				// Skip aliases, they don't need keys.
				continue;
			}

			$matches[] = 'STRING_' . $i18nKey;
			$matches[] = 'MESSAGE_' . $i18nKey;
			
			// ENUM Columns get their options mapped too!
			if($dat['type'] === Model::ATT_TYPE_ENUM && isset($dat['options'])){
				foreach($dat['options'] as $opt){
					$matches[] = 'VALUE_' . $i18nKey . '_' . I18NLoader::KeyifyString($opt);
				}
			}
		}
		
		return $matches;
	}
	
	/**
	 * Build a functional YAML file based off a string constant dictionary.
	 * 
	 * @param string $stringsFile
	 * @param string $lang
	 * @return type
	 */
	public static function BuildYAMLFile(string $stringsFile, string $lang ){
		
		$src = dirname($stringsFile) . '/';
		$strings = explode("\n", file_get_contents($stringsFile));
		
		// Is there a current version of the file?
		// If so, load the current settings to preserve data.
		if(file_exists($src . $lang . '.yml')){
			$current = yaml_parse_file($src . $lang . '.yml');
		}
		elseif(file_exists($src . 'strings.yml')){
			// Core 6.2 version of i18n; contained all strings in a single file.
			$all = yaml_parse_file($src . 'strings.yml');
			$current = [];
			foreach($all as $k => $dat){
				if(isset($dat[ $lang ])){
					$current[ $k ] = $dat[ $lang ];
				}
			}
		}
		
		$translations = [];
		foreach($strings as $string){
			$translations[ $string ] = isset($current[ $string ]) ? 
				$current[ $string ] : 
				'';
		}
		
		yaml_emit_file(
			$src . $lang . '.yml', 
			$translations, 
			YAML_UTF8_ENCODING, 
			YAML_LN_BREAK
		);
	}
}
