<?php
/**
 * 
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\ListingTable;

use Core\Datamodel\Dataset;
use Core\Datamodel\DatasetWhere;
use Core\Datamodel\DatasetWhereClause;
use Core\Date\DateTime;
use Core\Date\Timezone;
use Core\Forms\DateInput;
use Core\Forms\Form;
use Core\Forms\FormElement;
use Core\Search\Helper as SearchHelper;
use Core\Templates\Template;
use Exception;


class Filter {

	/** @var FormElement */
	private $_element;

	/** @var string */
	private $_link = '';

	/** @var string */
	private $_type;

	/** @var string */
	private $_name;

	/** @var array */
	private $_attributes = [];

	/**
	 * Standard link type for a given model factory and its filter.
	 *
	 * format: key = value
	 */
	const LINK_TYPE_STANDARD = ' = ';

	/**
	 * Greater than link type for a given model factory and its filter.
	 *
	 * format: key > value
	 */
	const LINK_TYPE_GT = ' > ';

	/**
	 * Greater than or equal-to link type for a given model factory and its filter.
	 *
	 * format: key >= value
	 */
	const LINK_TYPE_GE = ' >= ';

	/**
	 * Less than link type for a given model factory and its filter.
	 *
	 * format: key < value
	 */
	const LINK_TYPE_LT = ' < ';

	/**
	 * Less than or equal-to link type for a given model factory and its filter.
	 *
	 * format: key <= value
	 */
	const LINK_TYPE_LE = ' <= ';

	/**
	 * Use the LIKE connector to find anything that starts with the value.
	 *
	 * format: key LIKE value%
	 */
	const LINK_TYPE_STARTSWITH = '_startswith_';

	/**
	 * Use the LIKE connector to find anything that contains the value.
	 *
	 * format: key LIKE %value%
	 */
	const LINK_TYPE_CONTAINS = '_contains_';
	
	/**
	 * Perform a full text search for the requested records.
	 * 
	 * @var string
	 */
	const LINK_TYPE_SEARCH = '_search_';

	/**
	 * Create a new filter form object
	 */
	public function __construct(string $type, array $attributes) {

		if(isset($attributes['link'])) {
			$this->_link = $attributes['link'];
			unset($attributes['link']);
		}

		if(isset($attributes['name'])) {
			$this->_name = $attributes['name'];
			unset($attributes['name']);
		}

		$this->_type = $type;
		$this->_attributes = $attributes;
	}

	/**
	 * Check if this filter has a value set
	 *
	 * @return bool
	 */
	public function checkHasValue() : bool {
		if(isset($this->_attributes['value'])) {
			return $this->_attributes['value'] !== null && $this->_attributes['value'] !== '';
		}
		else {
			return false;
		}
	}

	/**
	 * Get this filter name
	 *
	 * @return string
	 */
	public function getName() : string {
		return $this->_name;
	}

	/**
	 * Get a custom attribute on this Filter, or NULL if none found
	 *
	 * @param string $attribute
	 *
	 * @return mixed|null
	 */
	public function getAttribute(string $attribute) {
		return $this->_attributes[$attribute] ?? null;
	}

	/**
	 * Get the names of the columns this index matches against
	 *
	 * @return array
	 */
	public function getLinkNames() : array {
		$names = $this->_attributes['linkname'] ?? $this->_name;

		// Ensure it's an array
		if(!is_array($names)) {
			$names = [ $names ];
		}

		return $names;
	}

	public function getLabel() : string {
		if(isset($this->_attributes['title'])) {
			$title = $this->_attributes['title'];
		}
		elseif(isset($this->_attributes['placeholder'])) {
			$title = $this->_attributes['placeholder'];
		}
		else {
			$title = $this->_name;
		}

		// Allow for inline i18n
		if(strpos($title, 't:') === 0) {
			$title = t(substr($title, 2));
		}

		return $title;
	}

	/**
	 * Get the raw value of this filter, without any transformations
	 *
	 * @return mixed
	 */
	public function getRawValue() {
		return $this->_attributes['value'];
	}

	/**
	 * Get the processed value of this filter, with any automatic processing
	 *
	 * If no value is detected, NULL is returned.
	 *
	 * @return mixed|null
	 */
	public function getProcessedValue() {
		$raw = $this->_attributes['value'] ?? null;

		if($raw === '' || $raw === null) {
			return null;
		}

		// Was there a prefix and/or suffix requested?
		$raw =
			($this->_attributes['linkvalueprefix'] ?? '') .
			$raw .
			($this->_attributes['linkvaluesuffix'] ?? '');

		// Allow automatic conversion of datetimes to other formats.
		if($this->_type === 'date') {
			$format = $this->_attributes['dateformat'] ?? 'U';
			if($format !== 'raw') {
				// Raw date strings are passed as-is with no transformation applied.
				// This allows the ISO YYYY-MM-DD to be used with no geo translation,
				// useful for entries that simply have the date recorded.
				// (Dates by themselves do NOT have timezone information!)

				try {
					$date = new DateTime($raw);
					$raw = $date->format($format, Timezone::TIMEZONE_GMT);
				}
				catch(Exception $ex) {
					// pass
				}
			}
		}

		return $raw;
	}

	/**
	 * Get the formatted value of this Filter, with any formatters applied
	 *
	 * @return mixed
	 */
	public function getFormattedValue(string $format = null) {
		$raw = $this->_attributes['value'] ?? null;

		if($raw === '' || $raw === null) {
			return null;
		}

		if(isset($this->_attributes['formatter'])) {
			// Best option, use the formatter
			return call_user_func($this->_attributes['formatter'], $raw, $format);
		}
		elseif($this->_type === 'select') {
			// Selects should use the option value when available
			if(isset($this->_attributes['options']) && isset($this->_attributes['options'][$raw])) {
				return $this->_attributes['options'][$raw];
			}
			else {
				return $raw;
			}
		}
		else {
			return $raw;
		}
	}

	/**
	 * Set the value of this filter, uses the Form system for necessary conversions.
	 *
	 * @param $value
	 */
	public function setValue($value) {
		$el = FormElement::Factory($this->_type, $this->_attributes);
		$el->setValue($value);

		// And update the cache
		$this->_attributes['value'] = $el->getValue();
	}

	public function getLinkType() : string {
		return $this->_link;
	}

	/**
	 * Render this as HTML via the FormElement system
	 *
	 * @return string
	 */
	public function render() : string {
		$el = FormElement::Factory(
			$this->_type,
			array_merge(['name' => 'filter[' . $this->_name . ']'], $this->_attributes)
		);
		return $el->render();
	}

	/**
	 * Helper method to add the results of this Filter to a raw dataset
	 *
	 * Can be used with ListingTables, manual lookups, or REST lookups
	 *
	 * @param Dataset $ds
	 * @throws \Core\Datamodel\DMIException
	 */
	public function attachToDataset(Dataset $ds) {
		$linkType = $this->getLinkType();
		if($linkType === '') {
			// Skip filters that do not have a standard link type.
			return;
		}

		// Retrieve the value with any processing modifiers necessary
		$value = $this->getProcessedValue();
		// Most filters use NULL as a non-set value, but sometimes this is actually expected and needs to be present.
		$allowNull = (isset($this->_attributes['null']) && $this->_attributes['null']);

		if($value === null && !$allowNull) {
			// Skip filters without a value set
			return;
		}

		$statements = new DatasetWhereClause();
		$statements->setSeparator('OR');
		$names = $this->getLinkNames();
		foreach($names as $n) {
			switch($linkType) {
				case Filter::LINK_TYPE_STANDARD:
					// Standard allows values to be NULL.
					if($allowNull && $value === null) {
						$statements->addWhereParts($n, DatasetWhere::OP_NULL, null);
					}
					else {
						$statements->addWhereParts($n, DatasetWhere::OP_EQ, $value);
					}
					break;
				case Filter::LINK_TYPE_GT:
					$statements->addWhereParts($n, DatasetWhere::OP_GT, $value);
					break;
				case Filter::LINK_TYPE_GE:
					$statements->addWhereParts($n, DatasetWhere::OP_GE, $value);
					break;
				case Filter::LINK_TYPE_LT:
					$statements->addWhereParts($n, DatasetWhere::OP_LT, $value);
					break;
				case Filter::LINK_TYPE_LE:
					$statements->addWhereParts($n, DatasetWhere::OP_LE, $value);
					break;
				case Filter::LINK_TYPE_STARTSWITH:
					$statements->addWhereParts($n, DatasetWhere::OP_LIKE, $value . '%');
					break;
				case Filter::LINK_TYPE_CONTAINS:
					$statements->addWhereParts($n, DatasetWhere::OP_LIKE, '%' . $value . '%');
					break;
				case Filter::LINK_TYPE_SEARCH:
					// search is the oddball out; it doesn't generate a statement.
					SearchHelper::AddSearchToDataset($value, $ds);
					break;
			}
		}

		// All statements from this Filter have been processed, add them to the dataset if necessary.
		$statementCount = count($statements->getStatements());

		if($statementCount === 1) {
			// Single query, just append the one statement
			$ds->where($statements->getStatements()[0]);
		}
		elseif($statementCount > 1) {
			// More than one query are contained, add all of them!
			$ds->where($statements);
		}
	}
}

