<?php
/**
 * File for class Table definition in the coreplus project
 *
 * @package Core\ListingTable
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20140406.2004
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\ListingTable;

use Core\Date\DateTime;
use Core\Forms\Form;
use Core\ModelFactory;
use Core\Router;
use Core\Templates\Template;
use Core\Request;
use Core\UserAgent;
use Core\View;
use Core\ViewControl;
use Core\ViewControls;
use Exception;
use Core\FilterForm;
use Iterator;
use Core\Model;
use ReflectionClass;
use Core\Core;
use Core\Utilities\Profiler\Profiler;
use Core\Utilities\Profiler\ProfilerEvent;
use Core\Session;


/**
 * A short teaser of what Table does.
 *
 * More lengthy description of what Table does and why it's fantastic.
 *
 * <h3>Usage Examples</h3>
 *
 *
 * @todo Write documentation for Table
 * <h4>Example 1</h4>
 * <p>Description 1</p>
 * <code>
 * // Some code for example 1
 * $a = $b;
 * </code>
 *
 *
 * <h4>Example 2</h4>
 * <p>Description 2</p>
 * <code>
 * // Some code for example 2
 * $b = $a;
 * </code>
 *
 *
 * @package Core\ListingTable
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
class TableOLDIGNORE implements Iterator {
	
	/** @var bool T/F if the record controls should be hover reactive */
	public $recordControlsHoverSensitive = true;
	
	/** @var bool|"auto" Set to T/F to force a UI proxy either force on or force off. */
	public $recordControlsForceProxy = 'auto';
	
	/** @var string Text of the proxy source text, no function if forceProxy is set to false */
	public $recordControlsProxyText = 't:STRING_CONTROLS';
	
	/**
	 * @var string Set to the rendering method for this table
	 * 
	 * "table-2016" for legacy 2016-circa table interface
	 * "div-2017"   for 2017-circa flat div interface
	 * "div-2018"   for 2018 full flexbox interface
	 */
	public $renderMethod = 'div-2018';
	
	/** @var bool Set to true to trigger "test" mode for this rendering.  */
	public $testmode = false;

	/** @var bool Set to false if this is a widget or sub table, will change how controls are rendered */
	public $fullpage = true;
	
	/**
	 * Set to true to allow downloading as CSV, (displays the download link)
	 * @var bool
	 */
	private $_hasCSV = false;
	
	/** @var string The session / key name of this listing table. */
	private $_name;

	/** @var ModelFactory|null The underlying factory to control this listing table. */
	private $_modelFactory;

	/** @var FilterForm The underlying FilterForm to handle pagination and filters on this listing table. */
	private $_filters;

	/** @var Form|null The underlying Form object used with quick-edit on this listing table. */
	private $_editform;

	/** @var Column[] The columns on this listing table. */
	private $_columns = [];

	/** @var bool Set to true when the filters have been applied to the Model, (set automatically). */
	private $_applied = false;

	/** @var array|null Array of the results, pulled directly from the ModelFactory's get method. */
	private $_results;

	/** @var null|ViewControls Any controls added to this table as a whole. */
	private $_controls = null;
	
	/** @var null|Model blank model required for polling a resolvable object for metadata. */
	private $_blankModel = null;

	/** @var Report[] List of reports supported on this table */
	private $_reports = [];

	/** @var array Array of  */
	private $_bulkActions = [];
	
	private $_displayOptions = true;
	
	private $_forceViewType = false;
	
	/** @var Request The user request cached from loadFromRequest. */
	protected $_request;


	public function __construct(){
		if(!$this->_name) {
			// Set the name by default to the system request's baseurl.
			// This is usually a safe default, but can be set by the caller script if desired.
			$this->setName(Router::$DefaultBaseURL->getBaseURL());
		}
		
		// Setup all default columns, (used for extended classes)
		$this->_setupColumns();
		
		// Setup all default filters, (used for extended classes)
		$this->_setupFilters();
	}


	//-----------------------------------------------------------------------\\
	//                                  GETTERS
	//-----------------------------------------------------------------------\\


	/**
	 * Get a blank Model representing this associated Model, when applicable
	 *
	 * @return Model
	 *
	 * @throws \ReflectionException
	 */
	public function getBlankModel() : Model {
		if($this->_blankModel === null){
			$n = $this->getModelFactory()->getModelName();
			$ref = new ReflectionClass($n);
			$this->_blankModel = $ref->newInstance();
		}
		
		return $this->_blankModel;
	}
	
	/**
	 * Get all columns declared on this table
	 * 
	 * @return Column[]
	 */
	public function getColumns() : array {
		return $this->_columns;
	}
	
	/**
	 * Get all VISIBLE columns declared on this table
	 * 
	 * @return Column[]
	 */
	public function getColumnsVisible() : array {
		$ret = [];
		foreach($this->_columns as $c) {
			if($c->visible) {
				$ret[] = $c;
			}
		}
		
		return $ret;
	}

	/**
	 * Get the controls for this table, if any.
	 *
	 * @return ViewControls
	 */
	public function getControls() : ViewControls {
		if($this->_controls === null){
			$this->_controls = new ViewControls();
			//$this->_controls->setProxyText('t:STRING_TABLE_ACTIONS');
			//$this->_controls->setProxyForce(true);
			
			// Include a show/hide columns link.
			// <a href="#" class="control-column-selection" title="Show / Hide Columns"><i class="icon icon-columns"></i></a>
			if($this->renderMethod === 'div-2017' || $this->renderMethod === 'table') {
				$this->_controls->addLink()
					->title('t:STRING_SHOW_HIDE_COLUMNS')
					->icon('columns')
					->classname('control-column-selection');
			}
			
			if($this->renderMethod === 'div-2018') {
				$this->_controls->addLink()
					->title('t:STRING_PRINT_PAGE')
					->icon('print')
					->classname('print-link')
					->set('_only_with_results', true);
			}
		}

		return $this->_controls;
	}

	/**
	 * Get the controls to display only if there are no results
	 *
	 * @return ViewControls
	 */
	public function getControlsForNoResults() : ViewControls {
		// Clone the built-in controls and remove any that are _only_with_results.
		$controls = $this->getControls();
		$controls = clone $controls;
		foreach($controls as $k => $c) {
			if($c['_only_with_results']) {
				unset($controls[$k]);
			}
		}
		
		return $controls;
	}

	/**
	 * Get the value of the corresponding Filter element.
	 *
	 * @param $filter string
	 * 
	 * @return mixed
	 */
	public function getFilterValue(string $filter){
		return $this->getFilters()->get($filter);
	}

	/**
	 * Get the underlying FilterForm object to control this listing table's pagination and filters.
	 *
	 * Will auto-create one if it doesn't exist.
	 *
	 * @return FilterForm
	 */
	public function getFilters() : FilterForm {
		if($this->_filters === null){
			$this->_filters = new FilterForm();
			$this->_filters->setName($this->_name);
			// Listing tables by default have pagination and sorting!
			$this->_filters->haspagination = true;
			$this->_filters->hassort = true;
		}
		
		// Ensure the filters are kept in sync with this Table, just in case the structure gets modified.
		$sortkeys = $this->getSortKeys();
		$this->_filters->hassort = (count($sortkeys) > 0);
		$this->_filters->setSortkeys($sortkeys);
		
		return $this->_filters;
	}

	/**
	 * Get any/all the listings on this table.
	 *
	 * If manual results are set, those are returned, otherwise the ModelFactory is utilized.
	 *
	 * @return array|Model
	 */
	public function getListings(){
		if(is_array($this->_results)) {
			// Manual results
			return $this->_results;
		}

		if(!$this->_applied){
			$this->getFilters()->applyToFactory($this->getModelFactory());
			$this->_applied = true;
		}

		return $this->getModelFactory()->get();
	}
	
	/**
	 * Get the "next" listing in this resultset, or NULL if at the end.
	 * 
	 * Useful for streaming large datasets without crazy memory overheads.
	 * 
	 * @return Model|null
	 */
	public function getNext() {
		if(is_array($this->_results)) {
			// Manual results already set, just pull those.
			// This requires the current to be returned, with next being called after it's queried.
			// This is because getNext() is called to pull all results, but the counter starts at the first element.

			if(($current = key($this->_results)) !== null) {
				$record = $this->_results[ $current ];
				next($this->_results);
			}
			else {
				$record = null;
			}

			return $record;
		}
		else {
			return $this->_modelFactory->getNext();
		}
	}
	
	/**
	 * Get the underlying ModelFactory for this listing table, or null if none exists.
	 *
	 * @return ModelFactory|null
	 */
	public function getModelFactory(){
		return $this->_modelFactory;
	}
	
	/**
	 * Get the name of the primary Model this data table is representing
	 * 
	 * @return string
	 */
	public function getModelName() : string {
		if(($fac = $this->getModelFactory())){
			// Usual method.
			return $fac->getModelName();
		}
		
		// If the records are manually assigned, then it may not be so easy.
		// Instead, grab the first record and return its name.
		if(sizeof($this->_results)){
			return get_class($this->_results[0]);
		}
		
		// ummm
		return '';
	}
	
	public function getName() : string {
		return $this->_name === null ? '' : $this->_name;
	}
	
	/**
	 * Get the appropriate render method for this table.
	 * 
	 * If set to auto, (default), the rendering method is guessed from the environment and UA.
	 * 
	 * @return string "div-2018", "div-2017" or "table"
	 */
	public function getRenderMethod() : string {
		return $this->renderMethod;
	}
	
	/**
	 * Get all Column keys that have sortable flags set
	 * 
	 * @return array
	 */
	public function getSortKeys() : array {
		$ret = [];
		
		foreach($this->_columns as $col) {
			if($col->sortkey) {
				$ret[] = $col->sortkey;
			}
		}
		
		return $ret;
	}

	/**
	 * Get the total count for the number of records filtered.
	 *
	 * @return int
	 */
	public function getTotalCount() : int {
		if(is_array($this->_results)) {
			return count($this->_results);
		}

		if(!$this->_applied){
			$this->getFilters()->applyToFactory($this->getModelFactory());
			$this->_applied = true;
		}

		return $this->getFilters()->getTotalCount();
	}
	
	
	
	//-----------------------------------------------------------------------\\
	//                                  CHECKS
	//-----------------------------------------------------------------------\\
	
	/**
	 * Check if this table has renderable Columns present
	 * @return bool
	 */
	public function checkHasRenderKeys() : bool {
		foreach($this->_columns as $col) {
			if($col->renderkey) {
				return true;
			}
		}
		
		return false;
	}
	
	/**
	 * Check if this filter has sortable Columns present
	 * @return bool
	 */
	public function checkHasSortKeys() : bool {
		return count($this->getSortKeys()) > 0;
	}
	
	public function checkHasDisplayOptions() : bool {
		return $this->_displayOptions;
	}

	public function checkHasBulkActions() : bool {
		return count($this->_bulkActions) > 0;
	}

	/**
	 * Check if this table has pagination, manual entries do NOT have pagination.
	 *
	 * @return bool
	 */
	public function checkHasPagination() : bool {
		// Manual entries do not have pagination
		if(is_array($this->_results)) {
			return false;
		}

		return true;
	}



	//-----------------------------------------------------------------------\\
	//                                  SETTERS
	//-----------------------------------------------------------------------\\
	
	/**
	 * Manually set the records of this listing table.
	 * By doing this, you lose ALL sorting possibilities and any dynamic nicety that the default behaviour gives.
	 * 
	 * Each record is still expected to support get as the traditional Model system does.
	 * 
	 * @param array $arrayOfRecords
	 */
	public function setRecords($arrayOfRecords){
		$this->_results = $arrayOfRecords;
		$this->_applied = true;
	}

	/**
	 * Add a new filter for this listing table.
	 *
	 * @param string $type
	 * @param array $atts
	 */
	public function addFilter($type, $atts){
		$form = $this->getFilters();
		$form->addElement($type, $atts);
	}

	/**
	 * Add a new Column onto this Listing Table.
	 *
	 * Will handle the filter sortable work automatically.
	 * 
	 * ## Usage:
	 * 
	 * ### Associative Array Example
	 * 
	 * This method can be used two ways, the first and recommended way is via an associative array.
	 * 
	 * ```
	 * $table = new Table();
	 * $table->addColumn(
	 *     [
	 *         'title' => 't:STRING_THING',
	 *         'sortkey' => 'thing',
	 *         'renderkey' => 'thing',
	 *         'visible' => true,
	 *     ]
	 * );
	 * ```
	 * 
	 * #### Associative Options
	 * 
	 * * title: (required); The title or t: string key to display
	 * * sortkey: If this column is sortable, this is the value of the ORDER BY clause
	 * * renderkey: If this Model supports automatic rendering, this is the value of the render() method
	 * * visible: Set to false to initially hide this column.  True (default), displays it.
	 * * FUTURE icon
	 * * FUTURE ??? (@todo)
	 * 
	 * ### Legacy Example
	 * 
	 * The second option is the legacy approach, which is still supported.
	 * 
	 * ```
	 * $table = new Table();
	 * $table->addColumn('t:STRING_THING', 'thing', true);
	 * ```
	 * 
	 * ### Simple Legacy Example
	 * 
	 * The shortest possible option for this command is simply pass one value as the title to display.
	 * By using this method, you will be required to render all the contents of the table.
	 * 
	 * ```
	 * $table = new Table();
	 * $table->addColumn('t:STRING_THING');
	 * ```
	 *
	 * @param array|string $dat_or_title Title or associative array containing: [title|key|renderkey|sortkey|visible]
	 * @param string|null  $sortkey      Sortkey when used in legacy mode
	 * @param boolean      $visible      T/F if this col is visible by default
	 * 
	 * @throws \Exception
	 */
	public function addColumn($dat_or_title, $sortkey = null, $visible = true){
		
		if(is_array($dat_or_title)){
			// New method supports an array as the first parameter!
			
			if(!(
				isset($dat_or_title['renderkey']) || isset($dat_or_title['key']) || isset($dat_or_title['title'])
			)){
				// At least a key or title MUST be set, (optionally both)!
				// If both keys are missing, this script will not have enough information to render the column with!
				throw new Exception('Please ensure that all listing table columns have either a title or key attribute!');
			}
			
			if(isset($dat_or_title['key'])){
				// Allow the `key` attribute to be set to set both render and sort keys.
				$dat_or_title['renderkey'] = $dat_or_title['key'];
				$dat_or_title['sortkey'] = $dat_or_title['key'];
			}
			
			// The title can be derived from the render key and the model if left empty.
			// This will pull from the native Model's i18n support structure!
			// Allow the user to override this by setting the `title` attribute though!
			$title = isset($dat_or_title['title']) ?
				$dat_or_title['title'] :
				't:STRING_MODEL_' . strtoupper($this->getModelName()) . '_' . strtoupper($dat_or_title['renderkey']);
			
			$sortkey   = isset($dat_or_title['sortkey']) ? $dat_or_title['sortkey'] : null;
			$renderkey = isset($dat_or_title['renderkey']) ? $dat_or_title['renderkey'] : null;
			$visible   = isset($dat_or_title['visible']) ? $dat_or_title['visible'] : true;
			$abbr      = isset($dat_or_title['abbr']) ? $dat_or_title['abbr'] : null;
			$group     = isset($dat_or_title['group']) ? $dat_or_title['group'] : null;
			
			// The name can come from a variety of locations, as it's generally calculated.
			// However, allow the user to override this calculation if a `name` attribute is set manually!
			if(isset($dat_or_title['name'])){
				$name = $dat_or_title['name'];
			}
			elseif($renderkey){
				$name = $renderkey;
			}
			elseif($sortkey){
				$name = $sortkey;
			}
			else{
				$name = \Core\str_to_url($title);
			}
		}
		else{
			$title     = $dat_or_title;
			$renderkey = null;
			$abbr      = null;
			$group     = null;
			$name      = \Core\str_to_url($dat_or_title);
		}
		
		$c = new Column();
		$c->title     = $title;
		$c->sortkey   = $sortkey;
		$c->visible   = $visible;
		$c->renderkey = $renderkey;
		$c->abbr      = $abbr;
		$c->group     = $group;
		$c->name      = $name;
		$this->_columns[] = $c;
	}
	
	/**
	 * Add a new Column onto this Listing Table based on the Model's schema
	 * 
	 * This requires the linked Model to contain a column with a corresponding key.
	 *
	 * @param string $key  Column key name that matches the Model
	 * 
	 * @return Column
	 */
	public function addColumnKey(string $key) : Column {
		
		$c = new Column();
		$c->title('t:STRING_MODEL_' . strtoupper($this->getModelName()) . '_' . strtoupper($key))
			->name($key)
			->key($key);
		
		$this->_columns[] = $c;
		
		return $c;
	}
	
	public function addReport(Report $report) {
		$request = Router::$DefaultRequest;

		$report->setFactory($this->getModelFactory());
		$report->setFilters($this->getFilters());
		$report->testmode = $this->testmode;
		
		$this->_reports[ $report->getName() ] = $report;
		
		if($request->getParameter('report') === $report->getName()) {
			// Ensure the filters are applied before generation!
			// This will populate the Model with the requested filters.
			$this->getListings();

			if($request->getParameter('format') === 'pdf') {
				$file = $report->generate(true);
				$file->sendToUserAgent(!$this->testmode);
			}
			else {
				$file = $report->generate(false);
				$file->sendToUserAgent();
			}

			$file->delete();
		}
	}

	/**
	 * Add a bulk action to this Table.
	 *
	 * Tables with bulk actions will enable the first column as a checkbox.
	 *
	 * @param string $title
	 * @param string $link
	 */
	public function addBulkAction(string $title, string $link) {
		$this->_bulkActions[] = [
			'title' => $title,
			'link' => $link,
		];
	}

	/**
	 * Set the model name, (and the underlying Factory object).
	 * 
	 * Please ensure to use the fully resolved class name when setting the class.
	 * 
	 * @param string $name
	 * 
	 * @throws \Exception
	 */
	public function setModelName($name){
		if(!preg_match('/Model$/', $name)){
			throw new Exception('Please set the Model name the fully resolved model class');
		}
		if(!class_exists($name)){
			throw new Exception('Class ' . $name . ' does not appear to exist!');
		}
		
		$this->_modelFactory = new ModelFactory($name);
	}

	/**
	 * Set the model factory itself.
	 *
	 * This is useful if it's a child model from another Model.
	 *
	 * @param ModelFactory $factory
	 */
	public function setModelFactory(ModelFactory $factory){
		$this->_modelFactory = $factory;
	}

	/**
	 * Set the name for this table (and the corresponding filters), required for any session saving/lookup.
	 *
	 * @param string $name
	 */
	public function setName($name){
		$name = \Core\str_to_url($name);
		
		$this->_name = $name;

		if($this->_filters !== null){
			$this->_filters->setName($this->_name);
		}
	}

	/**
	 * Set the callsmethod attribute on the edit form.
	 *
	 * @param string $method
	 */
	public function setEditFormCaller($method){
		$this->getEditForm()->set('callsmethod', $method);
	}

	/**
	 * Set the limit of results to show before pagination kicks in.
	 *
	 * @param $limit
	 */
	public function setLimit($limit){
		$this->getFilters()->setLimit($limit);
	}

	/**
	 * Set the default sort key and direction.
	 *
	 * This should be done prior to loading the results!
	 *
	 * @param string $key       The key of the column to sort by
	 * @param string $direction "DESC" or "ASC" for descending or ascending sort
	 */
	public function setDefaultSort($key, $direction = 'DESC'){
		$this->getFilters()->setSortKey($key);
		$this->getFilters()->setSortDirection($direction);
	}
	
	/**
	 * Shorthand to allow this listing table to be downloaded as a CSV.
	 * 
	 * Will set the link with a default icon and respond appropriately when N.csv is requested.
	 */
	public function setCSVDownloadable() {
		// Allow loading this data as a CSV
		$this->_hasCSV = true;
		
		$this->addControlFromArray([
			'title' => 't:STRING_DOWNLOAD_CSV',
			'link'  => '.csv',
			'icon'  => ViewControl::ICON_DOWNLOAD,
			'_only_with_results' => true,
		]);
		
		$this->_processCSVParameter();
	}
	
	public function addControlFromArray(array $data) {
		$control = new ViewControl();
		foreach($data as $k => $v){
			$control->set($k, $v);
		}

		$this->getControls()->addLink($control);
	}

	/**
	 * Add a control into the page template.
	 *
	 * Useful for embedding functions and administrative utilities inline without having to adjust the
	 * application template.
	 *
	 * @param string $title Human-friendly title to display in the mouse-over title
	 * @param string $link  Core resolvable link
	 * @param string $icon  Icon, primary item displayed to the end user
	 */
	public function addControl(string $title, string $link = null, string $icon = 'edit') {
		$control = new ViewControl();
		
		$control->title = $title;
		$control->link  = \Core\resolve_link($link);
		$control->icon  = $icon;

		$this->getControls()->addLink($control);
	}
	
	/**
	 * Manually set a filter value
	 * 
	 * @param string $filter
	 * @param mixed $value
	 */
	public function setFilterValue(string $filter, $value){
		$this->getFilters()->setFilterValue($filter, $value);
	}

	
	//-----------------------------------------------------------------------\\
	//                         REMOVERS / AND UI OVERRIDES
	//-----------------------------------------------------------------------\\
	
	
	public function removeFilters() {
		$this->_filters = new FilterForm();
		$this->_filters->setName($this->_name);
	}
	
	public function removeDisplayOptions() {
		$this->_displayOptions = false;
	}
	
	public function forceViewType(string $type) {
		switch($type) {
			case 'table':
			case 'collection':
				$this->_forceViewType = $type;
				break;
		}
	}

	//-----------------------------------------------------------------------\\
	//                                  OTHER/ACTIONABLE
	//-----------------------------------------------------------------------\\

	/**
	 * Alias of loadFromRequest()
	 *
	 * @param Request|null $request
	 */
	public function loadFiltersFromRequest($request = null){
		self::loadFromRequest($request);
	}
	
	/**
	 * Load/set the Request object to retrieve user-supplied parameters from.
	 * 
	 * By default, this will use the default system request.
	 * 
	 * @param Request $request
	 */
	public function loadFromRequest(Request $request = null) {
		$this->_setRequest($request);
		
		$this->_processFilters();
		
		$this->_processColumns();
		
		$this->_processDefaultActions();
	}

	
	/**
	 * Apply the automatic filters to the embedded Factory.
	 * 
	 * This is usually called automatically and doesn't need to be called externally
	 * unless the fully resolved Factory is required by an external source,
	 * (such as graph data or something similar).
	 */
	public function applyFilters(){
 		if(!$this->_applied){
 			$this->getFilters()->applyToFactory($this->getModelFactory());
 			$this->_applied = true;
 		}
	}

	public function render(string $section = null) : string {
		// Ensure that the filters are applied first!
		// This method is safe to call multiple times.
		$this->applyFilters();
		
		// Load the results into the internal array.
		//if($this->_results === null){
		//	$this->_results = $this->getModelFactory()->get();
		//}
		
		switch($section){
			case null:
				return $this->renderMethod === 'div-2018' ? $this->_render() : $this->_renderHead() . $this->_renderBody() . $this->_renderFoot();
			case 'head':
				return $this->_renderHead();
			case 'foot':
				return $this->_renderFoot();
			case 'filters':
				return $this->_renderFilters();
			case 'pagination':
				return $this->_renderPagination();
			case 'body':
				return $this->_renderBody();
			case 'csv':
				return $this->_renderCSV();
			default:
				return 'Unknown section to render for this listing table, [' . $section . ']';
		}
	}

	/**
	 * Send a CSV header and setup all necessary options to the View object to provide a download.
	 *
	 * All the data headers will be rendered automatically, (with the exception of the final 'controls' column).
	 *
	 * @param View  $view  Page view to manipulate
	 * @param string $title Title of the file, (will get converted to a valid URL)
	 */
	public function sendCSVHeader(View $view, $title = 'csv export'){
		$filename = \Core\str_to_url($title) . '-' . date('Y-m-d@Hi') . '.csv';

		$view->mode = View::MODE_NOOUTPUT;
		
		if($this->testmode && DEVELOPMENT_MODE){
			// Allow rendering as a test output for easy debugging for the developer.
			// For security reasons, this is only allowed when in DEVELOPMENT_MODE!
			$view->contenttype = 'text/plain';
		}
		else{
			$view->contenttype = 'text/csv';
			$view->addHeader('Content-Disposition', 'attachment; filename=' . $filename);
		}

		// Set the limits and everything as necessary.
		$this->setLimit(99999);
		$filters = $this->getFilters();
		if(!$this->_applied){
			$filters->applyToFactory($this->getModelFactory());
			$this->_results = $this->getModelFactory()->get();
			$this->_applied = true;
		}
		
		// Build the CSV header to send, (the first record).
		$header = [];
		foreach($this->_columns as $c){
			/** @var Column $c */
			$header[] = $c->getTitle();
		}
		
		// Render comment records to assist the reader of the file.
		// This is not supported by many systems, but should still render as a record for their viewing pleasure.
		$comments = [];
		
		$comments[] = ['#COMMENT', 'Generated', DateTime::Now('FDT') . ' by ' . \Core\user()->get('email')];
		foreach($filters->getElements() as $el){
			if($el->get('value')){
				if($el->get('title')) {
					$title = $el->get('title');
				}
				elseif($el->get('placeholder')) {
					$title = $el->get('placeholder');
				}
				else{
					$title = $el->get('name');
				}
				
				$comments[] = ['#COMMENT', 'Filter ' . $title, $el->get('value')];
			}
		}

		// Send the headers and start the output.
		$view->render();
		
		foreach($comments as $c){
			$this->sendCSVRecord($c);
		}
		
		$this->sendCSVRecord($header);
	}

	/**
	 * Send an indexed array to the browser as a valid CSV record.
	 *
	 * To be used in conjunction with sendCSVHeader().
	 *
	 * All scalar data in the array will be sanitized automatically.
	 *
	 * @param array $data
	 */
	public function sendCSVRecord($data){
		// CSV data is.... difficult.
		foreach($data as $k => $v){
			
			$v = trim($v);
			
			$escape = false;
			if($v == ''){
				$escape = false;
			}
			elseif(strpos($v, '"') !== false){
				// Strings with a quote in them get the quote escaped out and then wrapped in quotes.
				$escape = true;
				$v = str_replace('"', '\\"', $v);
			}
			elseif(strpos($v, ' ') !== false){
				// Strings with a space in the content get wrapped in quotes.
				$escape = true;
			}
			elseif(strpos($v, ',') !== false){
				// Strings with a comma in the content get wrapped in quotes.
				$escape = true;
			}
			elseif(strpos($v, "\n") !== false){
				// Strings with a newline in the content get wrapped in quotes.
				$escape = true;
			}
			elseif(strpos($v, "\r") !== false){
				// Strings with a newline in the content get wrapped in quotes.
				$escape = true;
			}
			
			if($escape){
				// Wrap the formatted output with quotes
				$data[$k] = '"' . $v . '"';
			}
			else{
				// Just in case the trim() did something!
				$data[$k] = $v;
			}
		}
		
		// Now...
		echo implode(',', $data) . "\r\n";
	}


	//-----------------------------------------------------------------------\\
	//                            ITERATOR ACCESS
	//-----------------------------------------------------------------------\\

	public function rewind() {
		if($this->_results !== null) {
			reset($this->_results);
		}
	}

	public function current() {
		return current($this->_results);
	}

	public function key() {
		return key($this->_results);
	}

	public function next() {
		next($this->_results);
	}

	public function valid() {
		if($this->_results === null) {
			// If there are no results, then $results will be null.
			return false;
		}
		else {
			return isset($this->_results[key($this->_results)]);
		}
	}
	
	protected function _setRequest(Request $request = null) {
		// Allow NULL to be set; this defaults to the DefaultRequest.
		$this->_request = $request === null ? Router::$DefaultRequest : $request;
	}

	//-----------------------------------------------------------------------\\
	//                              PRIVATE
	//-----------------------------------------------------------------------\\
	
	/**
	 * Checks the GET parameter `view` and saves the value if set.
	 */
	private function _processViewParameter() {
		if($this->_request->getParameter('view')) {
			// Allow the ?view=N parameter to set the view mode for this and all tables.
			switch($this->_request->getParameter('view')) {
				case 'table':
				case 'collection':
					\Core\user()->setPreference('listing-table-view-type', $this->_request->getParameter('view'));
					break;
			}
		}
	}
	
	
	/**
	 * Check the content type requested, specifically for CSVs.
	 */
	private function _processCSVParameter() {
		if(
			$this->_request !== null &&
			$this->_request->ctype === View::CTYPE_CSV &&
			$this->_hasCSV
		) {
			$this->getFilters()->setPage(1);
			$this->setLimit(9999);
			$this->render('csv');
		}
	}
	
	/**
	 * Setup all default columns for this Table
	 */
	protected function _setupColumns() {
		// Placeholder that can be extended by the calling class, by default does nothing.
	}
	
	/**
	 * Setup all default filters for this Table
	 */
	protected function _setupFilters() {
		// Placeholder that can be extended by the calling class, by default does nothing.
	}
	
	/**
	 * Process all columns to display or hide based on the user's request
	 */
	protected function _processColumns() {
		if($this->_request->getParameter('reset')) {
			// A reset was requested, just delete all the session caches for this table
			Session::Delete('tablecolumns/' . $this->_name);
			$colsDisplayed = null;
		}
		elseif($this->_request->getParameter('column')){
			$colsUser = $this->_request->getParameter('column');
			$colsDisplayed = [];
			
			foreach($this->getColumns() as $c) {
				if(in_array($c->name, $colsUser)) {
					$colsDisplayed[] = $c->name;
				}
			}
			
			Session::Set('tablecolumns/' . $this->_name, $colsDisplayed);
		}
		else {
			$colsDisplayed = Session::Get('tablecolumns/' . $this->_name);
		}
		
		
		if($colsDisplayed && is_array($colsDisplayed)) {
			// Show/Hide them!
			foreach($this->getColumns() as $c) {
				if(in_array($c->name, $colsDisplayed)) {
					$c->show();
				}
				else {
					$c->hidden();
				}
			}
		}
	}
	
	/**
	 * Process all the attached filters from the Request.
	 * 
	 * Should really be called after all fitlers are setup!
	 */
	protected function _processFilters() {
		// Forward this to the underlying Filter system so the filters can retrieve the necessary flags.
		$this->getFilters()->load($this->_request);
	}
	
	/**
	 * Process all the default actions for this Table based on what options are set
	 */
	protected function _processDefaultActions() {
		$this->_processViewParameter();
		$this->_processCSVParameter();
	}


	private function _render() : string {
		$profiler = new ProfilerEvent('Listing Table Render - ' . $this->getName());
		$profiler->start();
		
		$request      = Router::$DefaultRequest;
		$view         = Router::$DefaultView;
		$filters      = $this->getFilters();
		$sortable     = $this->checkHasSortKeys();
		$pdfAvailable = Core::IsLibraryAvailable('HTMLtoPDF');

		
		// Allow AJAX requests to pull a subset of data, namely skipping the wrapper skin.
		if($request->isAjax()) {
			$view->mastertemplate = false;
		}
		
		$tableclasses = ['listing', 'listing-table'];
		if($sortable){
			$tableclasses[] = 'listing-table-sortable';
		}
		$atts = [];
		$atts['class'] = implode(' ', $tableclasses);
		$atts['data-table-name'] = $this->_name;
		$atts['data-table-sortable'] = ($sortable ? 1 : 0);
		$attributes = '';
		foreach($atts as $k => $v){
			$attributes .= ' ' . $k . '="' . $v . '"';
		}
		
		$viewType = $this->_forceViewType ? 
			$this->_forceViewType :
			\Core\user()->getPreference('listing-table-view-type', 'table');
		$viewTypeForced = ($this->_forceViewType);
		
		$template = Template::Factory('includes/listingtable/full-2018.tpl');
		
		
		$template->assign('view_type', $viewType);
		$template->assign('view_type_forced', $viewTypeForced);
		$template->assign('filters', $filters);
		$template->assign('attributes', $attributes);
		$template->assign('edit_form', $this->_editform);
		$template->assign('table', $this);
		$template->assign('controls', $this->getControls());
		// Only assign reports if PDF generation is available.
		$template->assign('reports', $pdfAvailable ? $this->_reports : []);
		$template->assign('base_url', Router::$DefaultRequest->getBaseURL());
		$template->assign('bulk_actions', $this->_bulkActions);

		$html = $template->fetch();
		
		// Don't forget the necessary scripts!
		//\Core\view()->addScript('assets/js/core.listingtable.js', 'foot');
		
		$profiler->stop();
		return $html;
	}
	
	/**
	 * Render this table's head content, (everything above the records).
	 *
	 * @return string Full HTML Markup.
	 */
	private function _renderHead(){
		// What type of renderer should be used?
		$method = $this->getRenderMethod();
		$filters = $this->getFilters();
		$sortable = $this->checkHasSortKeys();

		if($method == 'div-2017'){
			// Render-compatible sources use the new div-based layout.
			$template = Template::Factory('includes/listingtable/div-head.tpl');
		}
		else{
			$template = Template::Factory('includes/listingtable/head.tpl');
		}
		
		
		
		
		// Set to true if there's no model factory, (ie: records were externally set/managed).
		$manualRecords = (!$this->getModelFactory());

		$tableclasses = ['listing', 'listing-table'];
		if($sortable){
			$tableclasses[] = 'listing-table-sortable';
		}
		$atts = [];
		$atts['class'] = implode(' ', $tableclasses);
		$atts['data-table-name'] = $this->_name;
		$atts['data-table-sortable'] = ($sortable ? 1 : 0);
		$attributes = '';
		foreach($atts as $k => $v){
			$attributes .= ' ' . $k . '="' . $v . '"';
		}

		if($this->_editform !== null){
			$this->addControl(
				[
					'link' => '#',
					'class' => 'control-edit-toggle',
					'icon' => 'pencil-square-o',
					'title' => 'Quick Edit',
				]
			);
		}

		$template->assign('manual_records', $manualRecords);
		$template->assign('filters', $filters);
		$template->assign('filters_rendered', $manualRecords ? '' : $this->_renderFilters());
		$template->assign('pagination_rendered', $manualRecords ? '' : $this->_renderPagination());
		$template->assign('attributes', $attributes);
		$template->assign('edit_form', $this->_editform);
		$template->assign('columns', $this->_columns);
		$template->assign('controls', $this->getControls());

		$html = $template->fetch();
		
		if($method == 'div-2017'){
			// Dynamically render each column header.
			// This needs to be dynamic because 
			
			$lastGroup = null;
			
			$html .= '<div class="listing-table-record listing-table-header">';
			
			foreach($this->_columns as $c){
				/** @var Column $c */
				// Close the previous group if this is a new one.
				// Also start the new group while we're here.
				if($c->group !== $lastGroup){
					if($lastGroup !== null){
						// Only close the previous group if there was a previous group.
						//$html .= '</div>';
					}
					
					if($c->group !== null){
						//$html .= '<div class="listing-table-group column-group-' . $c->group . '">';
					}
					
					$lastGroup = $c->group;
				}
				
				// Render the column header itself.
				$html .= $c->getDIV();
			}
			
			// If there is a group open, close that group!
			if($lastGroup !== null){
				//$html .= '</div>';
			}
			
			// Render the control links column that should be present in all tables.
			$html .= '<div class="listing-table-cell-header listing-table-cell column-controls">' .
				$this->getControls()->fetch() .
				'</div>';
			
			// Close the table-record group
			$html .= '</div>';
		}
		
		return $html;
	}
	
	/**
	 * Render a CSV in its entirety and send to the browser as a file stream.
	 */
	private function _renderCSV(){
		// Use the default View as the view to render to.
		$view = \Core\view();
		
		// Send the CSV header to the browser
		$this->sendCSVHeader($view, $this->_name);
		
		// And send each record to the browser
		$factory = $this->getModelFactory();
		while($rec = $factory->getNext()){
			$data = [];
			foreach($this->_columns as $c){
				/** @var Column $c */
				$data[] = $rec->get($c->renderkey, View::CTYPE_PLAIN);
			}
			
			$this->sendCSVRecord($data);
			$rec = null;
		}
		
		
		// The parent render() is expecting a string, so return a blank string!
		return '';
	}
	
	private function _renderBody(){
		// Each cell is primarily controlled by the contained Model for each record.
		// As such, the use of a template is not as important here.
		
		// What type of renderer should be used?
		$method = $this->getRenderMethod();
		
		$out = '';
		
		if($method == 'div-2017'){
			foreach($this as $rec){
				$out .= '<div class="listing-table-record listing-table-body ' . $rec->getClass() . '"';
				// Append any/all data tags.
				$dataTags = $rec->getDataTags();
				foreach($dataTags as $tag => $value){
					$out .= 'data-' . $tag . '="' . str_replace('"', '&quot;', $value) . '"';
				}
				$out .= '>';

				// Render each column herein
				foreach($this->_columns as $c){
					/** @var Column $c */

					// Render the column data itself.
					$out .= '<div class="listing-table-cell-ungrouped ' . $c->getClass() . '" data-title="' . $c->getTitle(true) . '">';
					$out .= $rec->get($c->renderkey, View::CTYPE_HTML);
					$out .= '</div>';
				}
				
				$lastGroup = null;

				// Render each column for use in grouping; this will skip optional columns.
				foreach($this->_columns as $c){
					/** @var Column $c */
					if(!$c->visible){
						continue;
					}
					// Close the previous group if this is a new one.
					// Also start the new group while we're here.
					if($c->group !== $lastGroup){
						if($lastGroup !== null){
							// Only close the previous group if there was a previous group.
							$out .= '</div>';
						}

						if($c->group !== null){
							$out .= '<div class="listing-table-group column-group-' . $c->group . '">';
						}

						$lastGroup = $c->group;
					}

					// Render the column data itself.
					$out .= '<div class="listing-table-cell-grouped ' . $c->getClass() . '" data-title="' . $c->getTitle(true) . '">';
					$out .= $rec->get($c->renderkey, View::CTYPE_HTML);
					$out .= '</div>';
				}

				// If there is a group open, close that group!
				if($lastGroup !== null){
					$out .= '</div>';
				}

				// Render the controls for this line item.
				$controls = ViewControls::DispatchModel($rec);
				$controls->hovercontext = $this->recordControlsHoverSensitive;
				$controls->setProxyForce($this->recordControlsForceProxy);
				$controls->setProxyText($this->recordControlsProxyText);

				$out .= '<div class="listing-table-cell column-controls">' . $controls->fetch() . '</div>';

				$out .= '</div>';
			}
		}
		else{
			foreach($this as $rec){
				$out .= '<tr>';

				// Render each column herein
				foreach($this->_columns as $c){
					/** @var Column $c */
					
					// Render the column data itself.
					$out .= '<td class="' . $c->getClass() . '">';
					$out .= $rec->get($c->renderkey, View::CTYPE_HTML);
					$out .= '</td>';
				}

				// Render the controls for this line item.
				$controls = ViewControls::DispatchModel($rec);
				$controls->hovercontext = $this->recordControlsHoverSensitive;
				$controls->setProxyForce($this->recordControlsForceProxy);
				$controls->setProxyText($this->recordControlsProxyText);

				$out .= '<td class="column-controls">' . $controls->fetch() . '</td>';

				$out .= '</tr>';
			}
		}
		
		return $out;
	}

	/**
	 * Render the filters
	 *
	 * @return string Full HTML Markup.
	 */
	private function _renderFilters(){
		$out = '';

		$f = $this->getFilters();

		if(!$f->hasFilters()){
			return '';
		}

		$out .= '<div class="screen">' . $f->render() . '</div>';
		$out .= '<div class="print">' . $f->renderReadonly() . '</div>';

		return $out;
	}

	/**
	 * Render the pagination options.
	 *
	 * @return string
	 */
	private function _renderPagination(){
		return $this->getFilters()->pagination();
	}

	/**
	 * Render this table's foot content, (everything below the records).
	 * @return string Full HTML Markup
	 */
	private function _renderFoot(){
		// What type of renderer should be used?
		$method = $this->getRenderMethod();
		
		$out = '';

		if($this->_editform !== null){
			$out .= '<tr class="edit edit-record-buttons"><td colspan="' . (sizeof($this->_columns) + 1) . '">' .
				'<a href="#" class="control-edit-toggle button">Cancel</a>' .
				'<input type="submit" value="Save Quick Edit"/>' .
				'</td></tr>';
		}

		if($method == 'div-2017'){
			$out .= '</div>';
		}
		else{
			$out .= '</table>';
		}

		if($this->_editform !== null){
			$out .= $this->_editform->render('foot');
		}

		$f = $this->getFilters();
		$out .= $f->pagination();

		if($f->getTotalCount() == 0 && !$f->hasSet()){
			return '';
		}

		// Don't forget the necessary scripts!
		\Core\view()->addScript('assets/js/core.listingtable.js', 'foot');
		\Core\view()->addScript("<script>new Core.ListingTable(\$('" . $method . "[data-table-name=\"" . $this->_name . "\"]'), '" . $this->getFilters()->getSortKey() . "', '" . $this->getFilters()->getSortDirection() . "');</script>", 'foot');

		return $out;
	}
} 
