<?php
/**
 * File for class Table definition in the coreplus project
 *
 * @package Core\ListingTable
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20140406.2004
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\ListingTable;

use Core\BadRequestException;
use Core\Datamodel\DatasetWhereClause;
use Core\Datamodel\Drivers\mysqli\mysqli_backend;
use Core\Date\DateTime;
use Core\Forms\CheckboxInput;
use Core\Forms\Form;
use Core\HookHandler;
use Core\ListingTable\Renderer\CollectionRenderer;
use Core\ListingTable\Renderer\ReportRenderer;
use Core\ListingTable\Renderer\TableRenderer;
use Core\ModelFactory;
use Core\Router;
use Core\Templates\Template;
use Core\Request;
use Core\UserAgent;
use Core\View;
use Core\ViewControl;
use Core\ViewControls;
use Exception;
use Core\FilterForm;
use Iterator;
use Core\Model;
use ReflectionClass;
use Core\Core;
use Core\Utilities\Profiler\Profiler;
use Core\Utilities\Profiler\ProfilerEvent;
use Core\Session;


/**
 * A short teaser of what Table does.
 *
 * More lengthy description of what Table does and why it's fantastic.
 *
 * <h3>Usage Examples</h3>
 *
 *
 * @todo Write documentation for Table
 * <h4>Example 1</h4>
 * <p>Description 1</p>
 * <code>
 * // Some code for example 1
 * $a = $b;
 * </code>
 *
 *
 * <h4>Example 2</h4>
 * <p>Description 2</p>
 * <code>
 * // Some code for example 2
 * $b = $a;
 * </code>
 *
 *
 * @package Core\ListingTable
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
class Table implements TableInterface {

	public $testmode = false;

	/** @var Column[] List of columns present*/
	protected $_columns = [];

	/** @var Filter[] List of filters present */
	protected $_filters = [];

	/** @var ModelFactory Underlying ModelFactory responsible for retrieving results */
	protected $_factory;

	/** @var Request */
	protected $_request;

	/** @var Action[] */
	protected $_actions = [];

	/** @var ViewControls */
	protected $_controls = null;

	/** @var Report[] */
	protected $_reports = [];

	/**
	 * @var array Array of title/link tags
	 *      + 'title' string
	 *      + 'link' string
	 */
	private $_bulkActions = [];

	protected $_defaultLimit = 50;

	protected $_forceViewMode = null;

	protected $_defaultViewMode = 'table';

	/** @var string Table name override, defaults to the Model name */
	protected $_tableName;

	/** @var string[] Class names to set on the table */
	protected $_classes = [];

	/** @var array List of scripts and their locations to render with the Table */
	protected $_scripts = [
		[ 'Core.Strings', 'library' ],
		[ 'assets/js/core.listingtable2019.js', 'foot' ],
		[ 'notify', 'library' ],
	];

	/** @var string|null Path to the baseurl to retrieve this Table */
	protected $_baseurl;

	/** @var bool Default action - Print */
	protected $_enablePrint = false;

	/** @var bool Default action - Download */
	protected $_enableDownload = true;

	/** @var bool Default action - Auto Refresh */
	protected $_enableRefresh = true;

	protected $_parameters = null;

	public function __construct() {
		// Setup all default columns, (used for extended classes)
		$this->_setupColumns();

		// Setup all default filters, (used for extended classes)
		$this->_setupFilters();

		// Setup all actions
		$this->_setupActions();

		// And the reports
		$this->_setupReports();

		// Allow scripts to extend this functionality
		$parent = get_class($this);
		$hook = str_replace('\\', '/', strtolower($parent));
		HookHandler::RegisterNewHook(
			'/core/listingtable/construct/' . $hook,
			'Constructor for ' . $parent . ', allows for manipulating the Listing Table',
			'void'
		);
		HookHandler::DispatchHook('/core/listingtable/construct/' . $hook, $this);
	}

	// <editor-fold desc="Action & Control Methods">

	/**
	 * Add a control to this table to display inline
	 *
	 * @param ViewControl $control
	 */
	public function addControl(ViewControl $control) {
		if($this->_controls === null) {
			$this->_setupControls();
		}

		$control->data('table-source', $this->getTableName());

		$this->_controls->addLink($control);
	}

	/**
	 * Add a control to this table via its title, link, and icon parameters.
	 *
	 * This provides support for Core-2018 and prior.
	 *
	 * @param string $title
	 * @param string $link
	 * @param string|null $icon
	 */
	public function addControlTLI(string $title, string $link, string $icon = null) {
		$this->addControl(ViewControl::FactoryTLI($title, $link, $icon));
	}

	/**
	 * Add a control to this page
	 *
	 * Does the same thing as $view->addControl, but
	 * useful if you need to link it to the table to preserve
	 * functionality such as 'table-only-with-results'
	 *
	 * @param ViewControl $control
	 */
	public function addControlToPage(ViewControl $control) {

		$control->data('table-source', $this->getTableName());

		Router::$DefaultView->addControl($control);
	}

	public function addAction(Action $action) {
		$this->_actions[] = $action;
	}

	public function disablePrint() {
		$this->_enablePrint = false;
	}

	public function enablePrint() {
		$this->_enablePrint = true;
	}

	public function disableDownload() {
		$this->_enableDownload = false;
	}

	public function disableRefresh() {
		$this->_enableRefresh = false;
	}

	public function getControls() : ViewControls {
		if($this->_controls === null) {
			$this->_controls = new ViewControls();

			// Setup all controls
			$this->_setupControls();
		}

		return $this->_controls;
	}

	/**
	 * Setup default controls for this Table
	 */
	protected function _setupControls() {
		$this->_controls = new ViewControls();

		$reportsEnabled = count($this->_reports) > 0;

		if($this->_enablePrint) {
			$this->addControl(
				ViewControl::Factory()
					->title('t:MESSAGE_LISTING_TABLE_PRINT_RESULTS')
					->text('t:STRING_PRINT')
					->icon('print')
					->group(ViewControl::GROUP_ADMIN)
					->classname('print-link')
					->data('table-only-with-results', true)
			);
		}

		if($this->_enableDownload) {
			$this->addControl(
				ViewControl::Factory()
					->title('t:MESSAGE_LISTING_TABLE_DOWNLOAD_RESULTS')
					->text('t:STRING_DOWNLOAD')
					->icon('download')
					->group(ViewControl::GROUP_ADMIN)
					->classname('download-link')
					->link('?action=download&table=' . $this->getTableName())
					->data('table-only-with-results', true)
			);
		}

		if($this->_enableRefresh) {
			$this->addControl(
				ViewControl::Factory()
					->title('t:MESSAGE_LISTING_TABLE_AUTO_REFRESH')
					->text('t:STRING_AUTO_REFRESH')
					->icon('sync')
					->group(ViewControl::GROUP_ADMIN)
					->classname('auto-refresh-link')
			);
		}

		if($this->_forceViewMode === null) {
			// The controller can request that a specific View mode is used.
			// If so, those control links are simply not rendered
			$this->addControl(
				ViewControl::Factory()
					->title('t:MESSAGE_LISTING_TABLE_VIEW_RESULTS_AS_GRID')
					->text('t:STRING_LISTING_TABLE_GRID')
					->icon('table')
					->link('?view=table')
					->classname('table-change-view set-view-table')
					->important(true)
					->group(ViewControl::GROUP_VIEW)
			);

			$this->addControl(
				ViewControl::Factory()
					->title('t:MESSAGE_LISTING_TABLE_VIEW_RESULTS_AS_LIST')
					->text('t:STRING_LISTING_TABLE_LIST')
					->icon('th-list')
					->link('?view=collection')
					->classname('table-change-view set-view-collection')
					->important(true)
					->group(ViewControl::GROUP_VIEW)
			);

			$control = ViewControl::Factory()
				->title('t:MESSAGE_LISTING_TABLE_VIEW_REPORTS')
				->text('t:STRING_LISTING_TABLE_REPORTS')
				->icon('chart-bar')
				->link('?view=report')
				->classname('table-change-view set-view-report')
				->important(true)
				->group(ViewControl::GROUP_VIEW);
			if(!$reportsEnabled) {
				$control->addClass('disabled')->link('#');
			}
			$this->addControl($control);
		}


		$this->addControl(
			ViewControl::Factory()
				->title('t:MESSAGE_LISTING_TABLE_OPEN_FILTERS')
				->text('t:STRING_FILTERS')
				->icon('filter')
				->classname('open-filters')
				->group(ViewControl::GROUP_MANAGE)
		);

		$this->addControl(
			ViewControl::Factory()
				->title('t:MESSAGE_LISTING_TABLE_OPEN_DISPLAY_OPTIONS')
				->text('t:STRING_LISTING_TABLE_OPTIONS')
				->icon('eye')
				->classname('open-options')
				->group(ViewControl::GROUP_MANAGE)
		);
	}

	/**
	 * Setup default actions for this Table
	 */
	protected function _setupActions() {
		// All tables support JSON loading, this provides support for retrieving the results.
		$json = new Action();
		$json->call = [$this, 'renderAsJSON'];
		$json->ctype = View::CTYPE_JSON;
		$this->addAction($json);

		$download = new Action();
		$download->matches['action'] = 'download';
		$download->call = [$this, 'renderAsDownload'];
		$this->addAction($download);
	}

	/**
	 * Process all actions provided and execute any that match the request
	 */
	protected function _processActions() {
		$matchedAction = null;
		$match = null;

		foreach($this->_actions as $action) {
			if($action->checkMatches($this->_request, $this->_parameters)) {
				if($match === null) {
					$matchedAction = $action;
					$match = $action->getMatchWeight();
				}
				elseif($action->getMatchWeight() > $match) {
					$matchedAction = $action;
					$match = $action->getMatchWeight();
				}
			}
		}

		if($matchedAction) {
			$matchedAction->execute();
		}
	}

	// </editor-fold>

	// <editor-fold desc="Column Methods">

	/**
	 * Add a new Column onto this Listing Table based on the Model's schema
	 *
	 * This requires the linked Model to contain a column with a corresponding key.
	 *
	 * @param string $key  Column key name that matches the Model
	 *
	 * @return Column
	 *
	 * @throws Exception
	 */
	public function addColumn(string $key) : Column {
		$c = new Column();
		$c->title('t:STRING_MODEL_' . strtoupper($this->getModelName()) . '_' . strtoupper($key))
			->name($key)
			->key($key);

		$this->_columns[ $key ] = $c;

		return $c;
	}

	/**
	 * Retrieve a single Column by its key name
	 *
	 * If the column does not exist, an exception will be thrown.
	 *
	 * @param string $key
	 * @return Column
	 *
	 * @throws Exception
	 */
	public function getColumn(string $key) : Column {
		if(!isset($this->_columns[ $key ] )) {
			throw new Exception('Column "' . $key . '" is not defined on table ' . $this->getModelName());
		}

		return $this->_columns[ $key ];
	}

	/**
	 * Get all columns declared on this table
	 *
	 * @return Column[]
	 */
	public function getColumns() : array {
		return $this->_columns;
	}

	/**
	 * Get all VISIBLE columns declared on this table
	 *
	 * @return Column[]
	 */
	public function getColumnsVisible() : array {
		$ret = [];
		foreach($this->_columns as $c) {
			if($c->visible) {
				$ret[] = $c;
			}
		}

		return $ret;
	}

	/**
	 * Set the visible columns on this table
	 *
	 * @param string[] $columns
	 */
	public function setVisibleColumns(array $columns) {
		$tableName = $this->getTableName();

		$cols = [];
		foreach($columns as $column) {
			try {
				// Will ensure this column actually exists.
				$this->getColumn($column);
				$cols[] = $column;
			}
			catch(Exception $ex) {
				// pass
			}
		}

		if(count($cols) === 0) {
			// No columns displayed, resets the view
			Session::Delete('tablecolumns/' . $tableName);
		}
		else {
			Session::Set('tablecolumns/' . $tableName, $cols);
		}
	}

	/**
	 * Remove a column from this Table, useful for manipulating tables on specific controllers
	 *
	 * @param string $key  Column key name that matches the Model
	 */
	public function removeColumn(string $key) {
		unset($this->_columns[ $key ]);
	}

	/**
	 * Setup all default columns for this Table
	 */
	protected function _setupColumns() {
		// Placeholder that can be extended by the calling class, by default does nothing.
	}

	// </editor-fold>


	// <editor-fold desc="Filter Methods">

	/**
	 * Add a new filter for this listing table.
	 *
	 * @param string $type
	 * @param array $attributes
	 *
	 * @return Filter
	 */
	public function addFilter(string $type, array $attributes) : Filter {
		$filter = new Filter($type, $attributes);

		$this->_filters[ $filter->getName() ] = $filter;

		return $filter;
	}

	/**
	 * Get a single filter by its key name
	 *
	 * @param string $key
	 * @return Filter
	 *
	 * @throws Exception
	 */
	public function getFilter(string $key) : Filter {
		if(!isset($this->_filters[ $key ] )) {
			throw new Exception('Filter "' . $key . '" is not defined on table ' . $this->getModelName());
		}

		return $this->_filters[ $key ];
	}

	/**
	 * Get all filters registered on this Table
	 *
	 * @return Filter[]
	 */
	public function getFilters() : array {
		return $this->_filters;
	}

	/**
	 * Get the processed value for the requested filter
	 *
	 * If the filter does not exist, NULL is returned.
	 *
	 * @param string $filter
	 *
	 * @return mixed
	 */
	public function getFilterValue(string $filter) {
		try {
			return $this->getFilter($filter)->getProcessedValue();
		}
		catch(Exception $ex) {
			return null;
		}
	}

	/**
	 * Set a group of filter values and save them in the session
	 *
	 * @param array $values
	 */
	public function setFilterValues(array $values) {
		$allValues = [];
		$tableName = $this->getTableName();

		foreach($values as $f => $v){
			try {
				$filter = $this->getFilter($f);
				$filter->setValue($v);

				if($filter->getAttribute('shared')) {
					// Support shared values, an arbituary value that can be shared between components and pages.
					$sharedKey = $filter->getAttribute('shared');
					Session::Set('sharedvalues/' . $sharedKey, $filter->getRawValue());
				}
				else {
					$allValues[ $filter->getName() ] = $filter->getRawValue();
				}
			}
			catch(Exception $ex) {
				// pass
			}
		}

		// Save these in the session
		Session::Set('tablefilters/' . $tableName, $allValues);
	}

	/**
	 * Check if there are filter values currently set from the user
	 *
	 * @return bool
	 */
	public function checkFiltersSet() : bool {
		$tableName = $this->getTableName();

		return count(Session::Get('tablefilters/' . $tableName, [])) > 0;
	}

	public function checkHasFilters() : bool {
		return is_array($this->_filters) && count($this->_filters);
	}

	/**
	 * Check if this table has the requested filter
	 *
	 * @param string $filter Name of the filter to check
	 *
	 * @return bool
	 */
	public function checkHasFilter(string $filter) : bool {
		return isset($this->_filters[ $filter ]);
	}

	/**
	 * Remove a filter from the table, useful for customizing the options.
	 *
	 * @param string $filtername
	 */
	public function removeFilter(string $filtername) {
		if(isset($this->_filters[ $filtername ])) {
			unset($this->_filters[ $filtername ]);
		}
	}

	/**
	 * Setup all default filters for this Table
	 */
	protected function _setupFilters() {
		// Placeholder that can be extended by the calling class, by default does nothing.
	}

	/**
	 * Process all the attached filters from the Request.
	 */
	protected function _processFilters() {
		$filters = $this->getFilters();
		$dataset = $this->_factory->getDataset();

		foreach($filters as $filter) {
			$filter->attachToDataset($dataset);
		}
	}

	// </editor-fold>


	// <editor-fold desc="Report Methods">

	public function addReport(Report $report) {
		// Populate the report with the necessary metadata from here
		$report->setFactory($this->getModelFactory());
		$report->setFilters($this->getFilters());
		$report->testmode = $this->testmode;

		// Add it to the list of reports
		$this->_reports[ $report->getName() ] = $report;

		if($this->_request !== null) {
			\Core\log_notice('Table->addReport called after request loaded!  Support for this will be removed in future versions.');
			$this->_processActions();
		}
	}

	/**
	 * Check if this Table has reports set
	 *
	 * @return bool
	 */
	public function checkHasReports() : bool {
		return count($this->_reports) > 0;
	}

	/**
	 * Get all the reports set on this Table
	 *
	 * @return Report[]
	 */
	public function getReports() : array {
		return $this->_reports;
	}

	/**
	 * Get the requested report by name, or an Exception if not found
	 *
	 * @param string $report
	 *
	 * @throws Exception
	 *
	 * @return Report
	 */
	public function getReport(string $report) : Report {
		if(isset($this->_reports[ $report ])) {
			return $this->_reports[ $report ];
		}
		else {
			throw new Exception('Report ' . $report . ' not registered with Table ' . $this->getTableName());
		}
	}

	/**
	 * Get the currently selected report, or an empty string if none set
	 *
	 * @return string
	 */
	public function getReportSelected() : string {
		if($this->getViewMode() === 'report') {
			// View mode is set to reports!  We can check for the report.
			$tableName = $this->getTableName();
			return Session::Get('tablereport/' . $tableName, '');
		}
		else {
			// If the report is not the current view type, just return none set
			return '';
		}
	}

	/**
	 * Set the report to be viewed as a specific report loaded in the system
	 *
	 * If the report is not available, nothing happens.
	 *
	 * If an empty string or NULL is set as the report type, the view is reverted back to the "select report" interface
	 *
	 * @param string|null $report
	 */
	public function setReport(string $report = null) {
		if($report === '' || $report === null) {
			// Empty report requested, this is allowed to blank out the selection.
			$tableName = $this->getTableName();
			Session::UnsetKey('tablereport/' . $tableName);
		}
		elseif(isset($this->_reports[ $report ])) {
			// Report located, so we can safely register it
			$tableName = $this->getTableName();
			Session::Set('tablereport/' . $tableName, $report);
		}
		// No else required, just nothing happens.
	}

	/**
	 * Set a group of report values and save them in the session
	 *
	 * @param array $values
	 */
	public function setReportOptions(array $values) {
		$allValues = [];
		$tableName = $this->getTableName();
		$reportName = $this->getReportSelected();

		if($reportName === '' || !isset($this->_reports[ $reportName ]) || count($values) === 0) {
			// If no report is set or an invalid report is set, just blank out all the values.
			$values = [];
			$reportOptions = [];
		}
		else {
			// Valid report requested, use that as a baseline of what options are allowed.
			$reportOptions = $this->_reports[ $reportName ]->getOptionsAsFormElements();
		}

		foreach($reportOptions as $f => $formElement) {
			if(isset($values[ $f ])) {
				$value = $values[ $f ];
			}
			elseif($formElement instanceof CheckboxInput) {
				// These are set to false if they're not present in the submit data,
				// because the browser just simply will not send them.
				$value = false;
			}
			else {
				// Skip this, the system will use defaults.
				$value = null;
			}

			if($value !== null) {
				// The get/set is to allow the FormElement to act on the value before saving.
				// This is because some form elements may convert incoming data to a different format
				// or perform other modifications to the data such as adding prefixes.
				$formElement->setValue($value);
				$allValues[ $f ] = $formElement->getValue();
			}
		}

		// Save these in the session
		if(count($allValues)) {
			Session::Set('tablereportoptions/' . $tableName, $allValues);
		}
		else {
			Session::UnsetKey('tablereportoptions/' . $tableName);
		}

	}

	protected function _setupReports() {
		// Placeholder that can be extended by the calling class, by default does nothing.
	}

	// </editor-fold>

	// <editor-fold desc="CSV Methods">

	/**
	 * This method is no longer used nor required; all listing tables support downloading!
	 *
	 * My thought with this is the data is presented to the user anyway, so with scrapers it can be downloaded
	 * regardless of how the site is configured.  If the user has permission to view the data, then they can download it.
	 *
	 * @deprecated 2020.12.24
	 */
	public function setCSVDownloadable() {

	}

	/**
	 * All-in-one catch for CSV and PDF downloads.
	 *
	 * Will perform the necessary action based on which view is currently active.
	 *
	 * @throws BadRequestException
	 */
	public function renderAsDownload() {
		if(!$this->_enableDownload) {
			throw new BadRequestException('Downloading of this data has been disabled!');
		}

		$mode = $this->getViewMode();
		switch($mode) {
			case 'table':
				// Tables get CSV
				$this->_generateCSV();
				break;
			case 'report':
				// Reports get PDF
				$reportName = $this->getReportSelected();
				if(!$reportName) {
					throw new BadRequestException('Please select a report to download');
				}
				$report = $this->getReport($reportName);
				$report->renderAsPDF();
				break;
			default:
				throw new BadRequestException('Requested view mode does not have a download option!');
		}
	}

	protected function _generateCSV() {
		$view  = Router::$DefaultView;


		$filename = \Core\str_to_url($this->getModelName()) . '-' . date('Ymd') . '.csv';

		if($this->testmode) {
			// Allow rendering as a test output for easy debugging for the developer.
			$view->contenttype = 'text/plain';
		}
		else {
			$view->contenttype = 'text/csv';
			$view->addHeader('Content-Disposition', 'attachment; filename=' . $filename);
		}

		// This method will stream output directly to the browser
		$view->mode = View::MODE_NOOUTPUT;
		$view->render();

		$this->_factory->limit(9999);

		// start the output
		$handle = fopen('php://output', 'w');

		// Render comment records to assist the reader of the file.
		// This is not supported by many systems, but should still render as a record for their viewing pleasure.
		$comments = [];

		$comments[] = ['#COMMENT', 'Generated', DateTime::Now('FDT') . ' by ' . \Core\user()->get('email')];
		foreach($this->_filters as $filter) {
			$value = $filter->getRawValue();

			if($value) {
				$comments[] = ['#COMMENT', 'Filter ' . $filter->getLabel(), $value];
			}
		}
		$comments[] = ['#COMMENT', ''];

		foreach($comments as $c) {
			fputcsv($handle, $c);
		}

		// Build the CSV header to send, (the first record).
		$header = [];
		foreach($this->_columns as $c){
			/** @var Column $c */
			$header[] = $c->getTitle();
		}
		fputcsv($handle, $header);


		// Retrieve all the data and push it to the browser.
		foreach($this->_factory->generator() as $model) {
			$data = [];
			foreach($this->_columns as $c) {
				/** @var Column $c */
				$data[] = $model->get($c->renderkey, View::CTYPE_CSV);
			}

			fputcsv($handle, $data);
		}

		fclose($handle);
	}

	/**
	 * Send an indexed array to the browser as a valid CSV record.
	 *
	 * To be used in conjunction with sendCSVHeader().
	 *
	 * All scalar data in the array will be sanitized automatically.
	 *
	 * @param array $data
	 */
	protected function _sendCSVRecord($data){
		// CSV data is.... difficult.
		foreach($data as $k => $v){

			$v = trim($v);

			$escape = false;
			if($v == ''){
				$escape = false;
			}
			elseif(strpos($v, '"') !== false){
				// Strings with a quote in them get the quote escaped out and then wrapped in quotes.
				$escape = true;
				$v = str_replace('"', '\\"', $v);
			}
			elseif(strpos($v, ' ') !== false){
				// Strings with a space in the content get wrapped in quotes.
				$escape = true;
			}
			elseif(strpos($v, ',') !== false){
				// Strings with a comma in the content get wrapped in quotes.
				$escape = true;
			}
			elseif(strpos($v, "\n") !== false){
				// Strings with a newline in the content get wrapped in quotes.
				$escape = true;
			}
			elseif(strpos($v, "\r") !== false){
				// Strings with a newline in the content get wrapped in quotes.
				$escape = true;
			}

			if($escape){
				// Wrap the formatted output with quotes
				$data[$k] = '"' . $v . '"';
			}
			else{
				// Just in case the trim() did something!
				$data[$k] = $v;
			}
		}

		// Now...
		echo implode(',', $data) . "\r\n";
	}

	// </editor-fold>


	// <editor-fold desc="Bulk Action Methods">

	/**
	 * Add a bulk action to this Table.
	 *
	 * Tables with bulk actions will enable the first column as a checkbox.
	 *
	 * @param string $title
	 * @param string $link
	 */
	public function addBulkAction(string $title, string $link) {
		$this->_bulkActions[] = [
			'title' => $title,
			'link' => $link,
		];
	}

	public function checkHasBulkActions() : bool {
		return count($this->_bulkActions) > 0;
	}

	// </editor-fold>

	// <editor-fold desc="Data / Model / Record Methods">

	/**
	 * Get all records *FOR THE CURRENT PAGE* on this table as a generator.
	 *
	 * If the backend supports pagination and limits, they will be applied here automatically.
	 *
	 * If you need to retrieve *ALL* records for the dataset,
	 * you will need to manually copy and manipulate the backend source
	 *
	 * @return \Generator
	 */
	public function getRecords() : \Generator {
		return $this->_factory->generator();
	}

	/**
	 * Get the name of the primary Model this data table is representing
	 *
	 * @return string
	 */
	public function getModelName() : string {
		if(!$this->_factory) {
			return 'unnamed';
		}

		return $this->_factory->getModelName();
	}

	/**
	 * Get the model factory for this table
	 *
	 * @return ModelFactory
	 *
	 * @throws Exception
	 */
	public function getModelFactory() : ModelFactory {
		if(!$this->_factory) {
			throw new Exception('No ModelFactory set for table ' . $this->getModelName());
		}

		return $this->_factory;
	}

	/**
	 * Get the sort string, formatted as "[key] [dir]"
	 *
	 * @return string
	 */
	public function getSort() : string {
		$tableName = $this->getTableName();

		$sort = Session::Get('tablesort/' . $tableName);
		if($sort === null) {
			try {
				$sort = $this->_factory->getDataset()->_order;
			}
			catch(Exception $ex) {
				$sort = null;
			}
		}

		return $sort === null || $sort === false ? '' : $sort;
	}

	/**
	 * Set the default sort key and direction.
	 *
	 * @param string $key       The key of the column to sort by
	 * @param string $direction "DESC" or "ASC" for descending or ascending sort
	 *
	 * @throws Exception
	 */
	public function setDefaultSort(string $key, string $direction = 'DESC') {
		$this->getModelFactory()->getDataset()->order($key . ' ' . $direction);
	}

	/**
	 * Set the sort key and direction and save them
	 *
	 * @param string $key       The key of the column to sort by
	 * @param string $direction "DESC" or "ASC" for descending or ascending sort
	 *
	 * @throws Exception
	 */
	public function setSort(string $key, string $direction = 'DESC') {
		// Will ensure this column exists
		$this->getColumn($key);
		$tableName = $this->getTableName();

		switch(strtolower($direction)) {
			case 'asc':
			case 'up':
				$direction = 'ASC';
				break;
			default:
				$direction = 'DESC';
		}

		$value = $key . ' ' . $direction;

		Session::Set('tablesort/' . $tableName, $value);
	}

	/**
	 * Get the display limit, ie: number of results per page
	 *
	 * @return int
	 */
	public function getLimit() : int {
		$tableName = $this->getTableName();

		return Session::Get('tablelimit/' . $tableName, $this->_defaultLimit);
	}

	/**
	 * Set the limit of results to show before pagination kicks in.
	 *
	 * @param $limit
	 */
	public function setLimit(int $limit) {
		$tableName = $this->getTableName();

		Session::Set('tablelimit/' . $tableName, min($limit, 500));
	}

	public function getPage() : int {
		$tableName = $this->getTableName();

		return Session::Get('tablepage/' . $tableName, 1);
	}

	/**
	 * Set the current Page number
	 *
	 * @param $page
	 */
	public function setPage(int $page) {
		$tableName = $this->getTableName();

		Session::Set('tablepage/' . $tableName, $page);
	}

	/**
	 * Set the model name, (and the underlying Factory object).
	 *
	 * @param string $name
	 *
	 * @throws \Exception
	 */
	public function setModelName(string $name){
		if(!preg_match('/Model$/', $name)){
			// Append 'Model' automatically
			$name .= 'Model';
		}

		if(!class_exists($name)){
			throw new Exception('Class ' . $name . ' does not appear to exist!');
		}

		$this->_factory = new ModelFactory($name);
	}

	/**
	 * Set the ModelFactory object
	 *
	 * @param ModelFactory $factory
	 */
	public function setModelFactory(ModelFactory $factory){
		$this->_factory = $factory;
	}

	/**
	 * Get the total number of records based on current filtering
	 *
	 * @return int
	 */
	public function getTotalCount() : int {
		try {
			return $this->_factory->count();
		}
		catch(Exception $ex) {
			\Core\log_exception($ex);
			return 0;
		}
	}

	// </editor-fold>


	// <editor-fold desc="General Methods">

	/**
	 * Load/set the Request object to retrieve user-supplied parameters from.
	 *
	 * By default, this will use the default system request.
	 *
	 * @param Request $request
	 */
	public function loadFromRequest(Request $request = null) {
		if($request === null) {
			// Pull this from the router instead.
			$this->_request = \Core\Router::$DefaultRequest;
		}
		else {
			$this->_request = $request;
		}

		$tableName = $this->getTableName();

		if($this->_request->getParameter('table') === $tableName) {
			// Only process this request if it was the specific table requested.
			// On most pages, there will only be one, but this is required for pages that have multiple listing tables.

			// Process all the GET parameters to retrieve/set the current request data
			$this->_parameters = $this->_request->getParameters();
			$this->_processRequestParameters();
		}
		elseif($this->_request->hasParameter($tableName)) {
			$parameters = json_decode($this->_request->getParameter($tableName), true);
			if($parameters === null) {
				// JSON decode failed, but it is a scalar value.
				// Translate this to a "quick" parameter.
				$this->_parameters = ['quick' => $this->_request->getParameter($tableName)];
				$this->_processRequestParameters();
			}
			else {
				$this->_parameters = $parameters;
				$this->_processRequestParameters();
			}
		}

		// Process all the SESSION data for stored parameters
		$this->_processSessionParameters();

		if($this->_parameters !== null) {
			// Only process this request if it was the specific table requested.
			// On most pages, there will only be one, but this is required for pages that have multiple listing tables.

			// Give the filters a chance to be processed, this will update the factory as necessary
			$this->_processFilters();

			// Check for any action that may have been requested.
			$this->_processActions();
		}
	}

	/**
	 * Load the filters and apply them to the Factory.
	 *
	 * DOES NOT INTERACT WITH THE REQUEST!
	 *
	 * This is useful if you manually build a Table for server-side operations or data retrieval.
	 */
	public function load() {
		$this->_processFilters();
	}

	/**
	 * Get the view mode for this table, taking into consideration the user preferences and forced styles
	 *
	 * @return string
	 */
	public function getViewMode() : string {
		if($this->_forceViewMode !== null) {
			return $this->_forceViewMode;
		}
		else {
			$tableName = $this->getTableName();
			return Session::Get('tableview/' . $tableName, $this->_defaultViewMode);
		}
	}

	/**
	 * Force a given view mode for this table, does not allow the user a chance to change it.
	 *
	 * @param string $mode
	 */
	public function forceViewMode(string $mode = null) {
		switch($mode) {
			case 'table':
			case 'report':
			case 'collection':
				$this->_forceViewMode = $mode;
				break;
		}
	}

	/**
	 * Add a class to this listing table
	 *
	 * @param string $class
	 */
	public function addClassname(string $class) {
		$this->_classes[] = $class;
	}

	/**
	 * Add a script to this Table that is necessary for rendering or some other JS manipulation
	 *
	 * @param string $src
	 * @param string $location
	 */
	public function addScript(string $src, string $location = 'foot') {
		$this->_scripts[] = [$src, $location];
	}

	/**
	 * Set the default view mode for this table, but allow the user a chance to change it.
	 *
	 * @param string $mode
	 */
	public function setDefaultViewMode(string $mode) {
		switch($mode) {
			case 'table':
			case 'report':
			case 'collection':
				$this->_defaultViewMode = $mode;
				break;
		}
	}

	/**
	 * Set the view mode for this table and store the preference in the user
	 *
	 * @param string $mode
	 */
	public function setViewMode(string $mode) {
		switch($mode) {
			case 'table':
			case 'collection':
			case 'report':
				// Save these in the session
				$tableName = $this->getTableName();
				Session::Set('tableview/' . $tableName, $mode);
				break;
		}
	}

	/**
	 * Set this table as embedded.
	 *
	 * Many of the controls will be disabled, and optionally a specific viewmode can be set.
	 *
	 * @param string|null $forceMode
	 */
	public function setEmbedded(string $forceMode = null) {
		if($forceMode) {
			$this->forceViewMode($forceMode);
		}

		$this->disableRefresh();
		$this->disablePrint();
		$this->disableDownload();
	}

	/**
	 * Get the table name, (just the linked Model name lowercase.
	 *
	 * @return string
	 */
	public function getTableName() : string {
		return $this->_tableName === null ? strtolower($this->getModelName()) : $this->_tableName;
	}

	/**
	 * Set the table name, useful for multiple listing tables on the same page.
	 *
	 * @param string $name
	 */
	public function setTableName(string $name) {
		$this->_tableName = $name;
	}

	/**
	 * Override the location to retrieve this Table's data
	 *
	 * @param string $url
	 */
	public function setBaseURL(string $url) {
		$this->_baseurl = $url;
	}

	/**
	 * Get the baseurl for rendering this table
	 *
	 * Usually the current page is returned, but this can be changed in the case of Widgets
	 * that may have a different source than the current page.
	 *
	 * @return string
	 */
	public function getBaseURL() : string {
		return $this->_baseurl ?: Router::$DefaultRequest->getBaseURL();
	}

	/**
	 * Render this table skeleton
	 *
	 * This will *NOT* render the data, but instead will render the interface where the data can be retrieved into.
	 *
	 * @param array $renderOptions Any options to override rendering
	 *
	 * @return string
	 *
	 * @throws Exception
	 */
	public function render(array $renderOptions = []) : string {
		$profiler = new ProfilerEvent('Listing Table Render - ' . $this->getModelName());
		$profiler->start();

		$request    = Router::$DefaultRequest;
		$view       = Router::$DefaultView;
		$filters    = $this->getFilters();
		$attributes = [];
		$controls   = $this->getControls();
		$tableName  = $this->getTableName();

		if(isset($renderOptions['class'])) {
			$tableclasses = array_merge(['listing-table', 'listing-table-' . $tableName, $renderOptions['class']], $this->_classes);
		}
		else {
			$tableclasses = array_merge(['listing-table', 'listing-table-' . $tableName], $this->_classes);
		}

		$attributes['class'] = implode(' ', $tableclasses);

		// @todo Support supplemental data tags
		$attributes['data-table-name'] = $tableName;
		$attributes['data-table-baseurl'] = $this->getBaseURL();

		$template = Template::Factory('includes/listingtable/table');


		//$template->assign('view_type', $viewType);
		//$template->assign('view_type_forced', $viewTypeForced);
		$template->assign('filters', $filters);
		$template->assign('attributes', $attributes);
		$template->assign('table', $this);
		$template->assign('controls', $controls);
		$template->assign('reports', $this->_reports);
		$template->assign('base_url', $this->getBaseURL());
		$template->assign('limit', $this->getLimit());
		$template->assign('columns', $this->getColumns());
		$template->assign('pdf_support', Core::IsLibraryAvailable('HTMLtoPDF'));
		$template->assign('bulk_actions', $this->_bulkActions);
		$template->assign('scripts', $this->_scripts);

		$html = $template->fetch();

		// Don't forget the necessary scripts!
		//\Core\view()->addScript('assets/js/core.listingtable.js', 'foot');

		$profiler->stop();
		return $html;
	}

	public function renderFilters() : string {
		$profiler = new ProfilerEvent('Listing Table Render Filters - ' . $this->getModelName());
		$profiler->start();
		$filters    = $this->getFilters();
		$template = Template::Factory('includes/listingtable/table-filters');
		$template->assign('filters', $filters);
		$html = $template->fetch();
		$profiler->stop();
		return $html;
	}

	/**
	 * Render the *DATA* of the listing table
	 *
	 * This will actually do the polling for the data in a background process after the page has loaded.
	 */
	public function renderAsJSON() {
		$profiler = new ProfilerEvent('Listing Table Render (JSON) - ' . $this->getModelName());
		$profiler->start();

		$view = Router::$DefaultView;
		$view->record = false;
		$mode = $this->getViewMode();
		$sort = $this->getSort();

		// Build the URL that the browser can reuse to view this data again
		$url = [
			//'table' => $this->getTableName(),
		];

		// Set the appropriate column sort.
		if($mode === 'table' && $sort !== '') {
			list($sortkey, $sortdir) = explode(' ', $sort);
			try {
				$col = $this->getColumn($sortkey);
				$col->currentSort = $sortdir;

				$url['sort_key'] = $sortkey;
				$url['sort_dir'] = $sortdir;
			}
			catch(Exception $ex) {
				// pass
			}
		}
		else {
			$sortkey = null;
			$sortdir = null;
		}

		switch($mode) {
			case 'collection':
				$renderer = new CollectionRenderer($this);
				$url['view'] = 'collection';
				$url['page'] = $this->getPage();
				break;
			case 'report':
				$renderer = new ReportRenderer($this);
				$url['view'] = 'report';

				if($this->getReportSelected()) {
					$report = $this->getReport($this->getReportSelected());
					$url['report_option'] = $report->getOptionValues();
					$url['report'] = $this->getReportSelected();
				}
				break;
			case 'table':
			default:
				$renderer = new TableRenderer($this);
				$mode = 'table';
				$url['view'] = 'table';

				$url['column'] = [];
				foreach($this->getColumnsVisible() as $column) {
					$url['column'][] = $column->renderkey;
				}

				$url['page'] = $this->getPage();
				$url['limit'] = $this->getLimit();
				break;
		}

		if($this->checkHasFilters()) {
			$url['filter'] = [];
			foreach($this->getFilters() as $filter) {
				if($filter->getRawValue() !== null) {
					$url['filter'][ $filter->getName() ] = $filter->getProcessedValue();
				}
			}
		}


		$view->contenttype = View::CTYPE_JSON;

		// We want to add some additional common parameters so anything listening
		// on the table knows what parameters were used for rendering.
		$view->jsondata = array_merge(
			[
				'has_filters' => $this->checkHasFilters(),
				'filters_set' => $this->checkFiltersSet(),
				'parameters' => [
					'view' => $mode,
					'sort_key' => $sortkey,
					'sort_dir' => $sortdir,
					'url' => $url,
				]
			],
			$renderer->render()
		);

		$profiler->stop();
	}

	/**
	 * Process all the standard request parameters,
	 * generally for things like resets, view changes, etc.
	 */
	private function _processRequestParameters() {
		$tableName = $this->getTableName();
		$req = $this->_request;
		$parameters = $this->_parameters;

		if(isset($parameters['reset'])) {
			// Reset the view to defaults
			Session::Delete('tablefilters/' . $tableName);
			Session::Delete('tablepage/' . $tableName);
		}

		if(isset($parameters['view'])) {
			$this->setViewMode($parameters['view']);
			// Setting / Changing the view mode will blank out any currently displayed report
			$this->setReport('');
			$this->setReportOptions([]);
		}

		if(isset($parameters['report'])) {
			$this->setReport($parameters['report']);
		}

		if(isset($parameters['report_option'])) {
			$this->setReportOptions($parameters['report_option']);
		}

		if(
			isset($parameters['sortkey']) &&
			isset($parameters['sortdir'])
		) {
			try {
				$this->setSort(
					$parameters['sortkey'],
					$parameters['sortdir']
				);
			}
			catch(Exception $ex) {
				// pass
			}
		}

		if(isset($parameters['page'])) {
			$this->setPage((int) $parameters['page']);
		}

		if(isset($parameters['limit'])) {
			// Update the limit AND page number.
			$this->setPage(1);
			$this->setLimit((int) $parameters['limit']);
		}

		if(isset($parameters['filter'])) {
			$this->setFilterValues($parameters['filter']);

			// Reset the page to 1 as well
			$this->setPage(1);
		}

		if(isset($parameters['column'])) {
			$this->setVisibleColumns($parameters['column']);
		}
	}

	/**
	 * Process all the Session data and populate the necessary destinations.
	 */
	private function _processSessionParameters() {
		$tableName = $this->getTableName();

		$filterValues = Session::Get('tablefilters/' . $tableName, []);
		$filters = $this->getFilters();
		foreach($filters as $filter) {
			$value = null;

			if($filter->getAttribute('shared')) {
				// This is a shared value, so retrieve from a different source.
				$sharedKey = $filter->getAttribute('shared');
				$value = Session::Get('sharedvalues/' . $sharedKey);
			}
			else {
				$value = $filterValues[ $filter->getName() ] ?? null;
			}

			if($value !== null) {
				$filter->setValue($value);
			}
		}


		$limit = Session::Get('tablelimit/' . $tableName, $this->_defaultLimit);
		$page = Session::Get('tablepage/' . $tableName, 1);

		if($page > 1) {
			$start = $limit * ($page - 1);
			$this->_factory->getDataset()->limit($limit, $start);
		}
		else {
			$this->_factory->getDataset()->limit($limit);
		}


		$sort = Session::Get('tablesort/' . $tableName);
		if($sort !== null) {
			try {
				$this->_factory->getDataset()->order($sort);
			}
			catch(Exception $ex) {
				// pass
			}
		}


		$columns = Session::Get('tablecolumns/' . $tableName, []);
		if(count($columns) > 0) {
			foreach($this->getColumns() as $column) {
				if(in_array($column->name, $columns)) {
					$column->show();
				}
				else {
					$column->hidden();
				}
			}
		}


		$reportName = $this->getReportSelected();
		if($reportName !== '' && isset($this->_reports[ $reportName ])) {
			$this->_reports[ $reportName ]->optionValues = Session::Get('tablereportoptions/' . $tableName, []);
		}
	}

	// </editor-fold>

	// <editor-fold desc="z_Deprecated Methods">

	/**
	 * Alias of addColumn
	 *
	 * @param string $key
	 * @return Column
	 *
	 * @deprecated 2019.06
	 *
	 * @throws Exception
	 */
	public function addColumnKey(string $key) : Column {
		return $this->addColumn($key);
	}

	/**
	 * Add a control from an associative array
	 *
	 * @deprecated 2019.06
	 *
	 * @param array $data
	 */
	public function addControlFromArray(array $data) {
		$control = new ViewControl();
		foreach($data as $k => $v){
			$control->set($k, $v);
		}

		$this->addControl($control);
	}

	/**
	 * Alias of loadFromRequest()
	 *
	 * @param Request|null $request
	 *
	 * @deprecated 2019.06
	 */
	public function loadFiltersFromRequest($request = null){
		$this->loadFromRequest($request);
	}

	/**
	 * Alias of forceViewMode
	 *
	 * @param string $type
	 *
	 * @deprecated 2019.07
	 */
	public function forceViewType($type) {
		$this->forceViewMode((string) $type);
	}

	// </editor-fold>
} 
