<?php


namespace Core\Updater;


use Core\ConfigHandler;
use Core\Date\DateTime;
use Core\Filestore\Directory;
use Core\Filestore\Factory as FileFactory;
use Core\Filestore\File;
use Exception;
use Ifsnop\Mysqldump\Mysqldump;
use ZipArchive;

class ArchiveExporter {

	private $_exportTables = [];

	private $_exportPublic = false;

	/** @var Directory */
	private $_archiveDirectory;

	private $_files;

	/** @var File */
	private $_archiveFile;

	/**
	 * Add a table to be exported in the datafile
	 *
	 * @param string $tablename
	 */
	public function exportTable(string $tablename) {
		$this->_exportTables[] = $tablename;
	}

	public function exportPublicFiles() {
		$this->_exportPublic = true;
	}

	/**
	 * Process the actual export operation with the requested assets
	 */
	public function process() : File {
		$this->_archiveDirectory = FileFactory::Directory('tmp/export-' . \Core\random_hex(8));
		$this->_archiveDirectory->mkdir();

		$this->_files = [];


		if(count($this->_exportTables)) {
			$this->_processTables();
		}

		if($this->_exportPublic) {
			$this->_processPublicFiles();
		}


		// Add all the collected files to the archive.
		$archive = new ZipArchive();
		$archivePath = $this->_archiveDirectory->getPath() . \Core\str_to_url(SITENAME) . '-archive-' . DateTime::Now() . '.zip';
		$archiveOpen = false;
		$count = 0;
		foreach($this->_files as $relpath => $fullpath) {
			++$count;

			if($count >= 1000) {
				// Reset the file locks after 1000 files to prevent file lock limits in the kernel.
				$archive->close();
				$archiveOpen = false;
				$count = 1;
			}

			if(!$archiveOpen) {
				$archive->open($archivePath, ZipArchive::CREATE);
			}

			$archive->addFile($fullpath, $relpath);
		}


		$archive->close();
		$this->_archiveFile = FileFactory::File($archivePath);

		return $this->_archiveFile;
	}

	public function __destruct() {
		if($this->_archiveDirectory !== null) {
			$this->_archiveDirectory->delete();
		}
	}

	private function _processTables() {
		// If the user opted to export database tables, export those!
		$tableDest = $this->_archiveDirectory->getPath() . 'data.sql.gz';

		$settings = ConfigHandler::LoadConfigFile('configuration');

		switch($settings['database_type']) {
			case 'mysqli':
			case 'mysql':
				$dsn = 'mysql';
				break;
			default:
				throw new Exception('Unable to export database, unsupported database type');
		}

		// mysql:host=localhost;port=3307;dbname=testdb
		$dsn .= ':host=' . $settings['database_server'] . ';port=' . $settings['database_port'] . ';dbname=' . $settings['database_name'];

		$exporter = new Mysqldump(
			$dsn,
			$settings['database_user'],
			$settings['database_pass'],
			[
				'include-tables' => $this->_exportTables,
				'compress'       => Mysqldump::GZIP,
				'add-drop-table' => true,
			]
		);

		$exporter->start($tableDest);

		$this->_files[ 'data/data.sql.gz' ] = $tableDest;
	}

	private function _processPublicFiles() {
		$dir = FileFactory::Directory('public/');
		$baselen = strlen($dir->getPath());
		foreach($dir->ls(null, true) as $file) {
			if($file instanceof File) {
				$rel = substr($file->getFilename(), $baselen);

				if(strpos($rel, 'tmp/') === 0) {
					continue;
				}
				$this->_files[ 'files/public/' . $rel ] = $file->getFilename();
			}
		}
	}
}