<?php


namespace Core\Updater;

use Core\ConfigHandler;
use Core\Date\DateTime;
use Core\Filestore\Directory;
use Core\Filestore\Factory as FileFactory;
use Core\Filestore\File;
use Exception;
use Ifsnop\Mysqldump\Mysqldump;
use ZipArchive;


class ArchiveImporter {

	/** @var ZipArchive */
	private $_archive;

	/** @var Directory */
	private $_archiveDirectory;

	public function process(File $file) {
		// Use the /private datastore here because /tmp will often get remapped due to systemd.
		$this->_archiveDirectory = FileFactory::Directory('private/tmp/import-' . \Core\random_hex(8));
		$this->_archiveDirectory->mkdir();

		$this->_archive = new ZipArchive();
		$this->_archive->open($file->getFilename());

		// Import the data based on what's enclosed.
		for( $i = 0; $i < $this->_archive->numFiles; $i++ ){
			$stat = $this->_archive->statIndex($i);

			if(strpos($stat['name'], '../') !== false) {
				// Skip files that try to extract to parent directories.
				continue;
			}

			if($stat['name'] === 'data/data.sql.gz') {
				// This contains raw SQL statements that need to run with MySQL directly
				$this->_processDatafile($stat['name']);
			}
			elseif(strpos($stat['name'], 'files/public/') === 0) {
				$this->_processPublicFile($stat['name']);
			}
		}

		$this->_archive->close();
	}

	public function __destruct() {
		if($this->_archiveDirectory !== null) {
			$this->_archiveDirectory->delete();
		}
	}

	private function _processDatafile(string $filename) {
		$path = $this->_archiveDirectory->getPath();
		$tempfile = FileFactory::File($path . $filename);

		$this->_archive->extractTo($path, $filename);

		$settings = ConfigHandler::LoadConfigFile('configuration');

		switch($settings['database_type']) {
			case 'mysqli':
			case 'mysql':
				$dsn = 'mysql';
				break;
			default:
				throw new Exception('Unable to export database, unsupported database type');
		}

		$command = <<<EOC
gunzip -c "$path$filename" | \
mysql \
 --host={$settings['database_server']} \
 --port={$settings['database_port']} \
 --user={$settings['database_user']} \
 --password="{$settings['database_pass']}" \
 --database={$settings['database_name']}
EOC;

		shell_exec($command);

		// Auto cleanup to save diskspace when possible
		$tempfile->delete();
	}

	private function _processPublicFile(string $filename) {
		$relname  = substr($filename, 13);
		$basedir  = dirname($relname);
		$destfile = FileFactory::File('public/' . $relname);
		$path     = $this->_archiveDirectory->getPath();
		$tempfile = FileFactory::File($path . $filename);


		$dir = FileFactory::Directory('public/' . $basedir);
		$dir->mkdir();

		$this->_archive->extractTo($path, $filename);
		
		$tempfile->copyTo($destfile);

		// Auto cleanup to save diskspace when possible
		$tempfile->delete();
	}
}