<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core\Utilities;

use Core\Forms\Form;
use Core\View;
use Exception;
use Core\Core;
use Core\Hook;
use Core\HookHandler;
use Core\Router;
use Core\Templates\Template;
use Core\Utilities\Profiler\DatamodelProfiler;
use Core\Filestore\FileHelper;
use Core\Utilities\Profiler\Profiler;
use function Core\is_numeric_array;

/**
 * Description of DevTools
 *
 * @author charlie
 */
class DevTools {
	private $_legend = '<div class="fieldset-title">%s<i class="icon icon-chevron-down expandable-hint"></i><i class="icon icon-chevron-up collapsible-hint"></i></div>' . "\n";
	
	private function _getHeader(string $title) : string {
		return '<fieldset class="debug-section collapsible" id="debug-section-' . strtolower(str_replace(' ', '-', $title)) . '">' .
			sprintf($this->_legend, $title);
	}
	
	private function _getFooter() : string {
		return '</fieldset>';
	}
	
	private function _formatData(array $data) : string {
		$str = '';
		$longest = max(array_map('strlen', array_keys($data)));
		
		foreach($data as $k => $v){
			$str .= str_pad($k, $longest, ' ', STR_PAD_LEFT) . ': ' . htmlentities($v) . "\n";
		}
		return '<div>' . $str . '</div>';
	}
	
	private function _getPerformanceInfo() : array {

		$request = Router::$DefaultRequest;
		$view = Router::$DefaultView;

		try{
			if(Router::$DefaultView->templatename) {
				$template = Template::ResolveFile($view->templatename);
			}
			else {
				$template = 'None';
			}

		} catch (Exception $ex) {
			$template = 'Invalid Template';
		}

		try{
			if($view->mastertemplate) {
				$skin = Template::ResolveFile($view->mastertemplate);
			}
			else {
				$skin = 'None';
			}
		} catch (Exception $ex) {
			$skin = 'Invalid Template';
		}


		$data = [
			'uri'           => $request->getBaseURL(),
			'mode'          => $view->mode,
			'type'          => $view->contenttype,
			'method'        => $request->method,
			'db_reads'      => DatamodelProfiler::GetDefaultProfiler()->readCount(),
			'db_writes'     => DatamodelProfiler::GetDefaultProfiler()->writeCount(),
			'memory'        => FileHelper::FormatSize(memory_get_peak_usage(true)),
			'time'          => Profiler::GetDefaultProfiler()->getTimeFormatted(),
			'events'        => Profiler::GetDefaultProfiler()->getRootEvent()->renderJSON(),
			'template'      => $view->templatename,
			'skin'          => $view->mastertemplate,
			'template_path' => $template,
			'skin_path'     => $skin,
		];

		return $data;
	}
	
	private function _getComponentInfo() : array {
		$data = [];
		$debugcomponents = array_merge(Core::GetComponents(), Core::GetDisabledComponents());
		// Give me sorting!
		ksort($debugcomponents);
		foreach ($debugcomponents as $v) {
			if($v->isEnabled() && $v->isReady()){
				$status = 'enabled';
			}
			elseif($v->isEnabled()){
				$status = 'error';
			}
			else{
				$status = 'disabled';
			}

			$data[] = [
				'status' => $status,
				'name' => $v->getName(),
				'version' => $v->getVersion()
			];
		}

		return $data;
	}

	private function _getFormInfo() : string {
		$data = [];

		$data[] = '<strong>Form Elements</strong>';

		$maps = Form::$Mappings;
		ksort($maps);

		foreach($maps as $key => $cls) {
			$data[] = $key . ' <i>provided by</i> ' . $cls;
		}


		$data[] = '<strong>Form Groups</strong>';

		$maps = Form::$GroupMappings;
		ksort($maps);

		foreach($maps as $key => $cls) {
			$data[] = $key . ' <i>provided by</i> ' . $cls;
		}

		return implode('<br/>', $data);
	}
	
	private function _getHookInfo() : string {
		$data = [];
		foreach(HookHandler::GetAllHooks() as $hook){
			/** @var $hook Hook */
			$data[] = $hook->name . ($hook->description ? ' <em> - ' . $hook->description . '</em>' : '');
			$data[] = '<span style="color:#999;">Return expected: ' . $hook->returnType . '</span>';
			foreach($hook->getBindings() as $b){
				$data[] = '* ' . $b['call'];
			}
			$data[] = '';
		}
		return implode("<br/>", $data);
	}
	
	private function _getLicensingInfo() : string {
		$data = [];
		$lic = Core::GetLicensedDump();
		foreach($lic as $dat){
			$licPrefix = $dat['status'] ? COLOR_SUCCESS : COLOR_ERROR;
			$data[] = $dat['component'] . ' license from ' . $dat['url'] . ' => ' . $licPrefix . $dat['message'] . COLOR_RESET;
			foreach($dat['features'] as $k => $v){
				$data[] = '* ' . $k . ': ' . $v;
			}
		}
		
		return $this->_getHeader('Licensing Information') .
			'<div>' . 
			implode('<br/>', $data) .
			'</div>' .
			$this->_getFooter();
	}
	
	private function _getIncludeInfo() : string {
		return $this->_getHeader('Included Files') .
			'<div>' . 
			'Number: ' . sizeof(get_included_files()) . '<br/>' .
			implode("<br/>", get_included_files()) .
			'</div>' .
			$this->_getFooter();
	}

	/**
	 * Attach all debug information for standard pages.
	 *
	 * @return bool
	 */
	public static function AttachDevInfo() : bool {
		if(!DEVELOPMENT_MODE){
			// This only appears when DEV_MODE is turned on.
			return true;
		}

		$view = Router::$DefaultView;

		if($view->mode === View::MODE_PAGE && $view->contenttype === View::CTYPE_HTML) {
			$tools = new self();

			$footer = <<<EOD
<a href="#" id="dev-tools-anchor" title="Open Development Tools Window">Dev Tools</a>
<div id="dev-tools-window">
	<a id="dev-tools-window-close" href="#">CLOSE</a>
	<ul id="dev-tools-window-toolbar">
		<li><a href="#" class="dev-tools-link" data-target="components">Available Components</a></li>
		<li><a href="#" class="dev-tools-link" data-target="requests">Requests and Performance</a></li>
		<li><a href="#" class="dev-tools-link" data-target="hooks">Registered Hooks</a></li>
		<li><a href="#" class="dev-tools-link" data-target="formelements">Form Elements</a></li>
	</ul>
	<div id="dev-tools-window-content">
		<div id="dev-tools-window-components"></div>
		<div id="dev-tools-window-requests"></div>
		<div id="dev-tools-window-hooks">{$tools->_getHookInfo()}</div>
		<div id="dev-tools-window-formelements">{$tools->_getFormInfo()}</div>
	</div>
</div>
EOD;
			$view->appendBodyContent($footer);
			$view->addScript('js/core.devtools.js', 'head');

			$scripts = [
				'Core.DevTools.SetComponents(' . json_encode($tools->_getComponentInfo()) . ')',
				'Core.DevTools.SetPageDetails(' . json_encode($tools->_getPerformanceInfo()) . ')',
			];
			$view->appendBodyContent('<script>' . implode("\n", $scripts) . '</script>');
		}

		/*
		$this->_getLicensingInfo() .
		$this->_getIncludeInfo() .
		 */

		return true;
	}

	/**
	 * Attach all debug info for JSON-based pages
	 *
	 * @return bool
	 */
	public static function AttachDevInfoJSON() : bool {
		if(!DEVELOPMENT_MODE){
			// This only appears when DEV_MODE is turned on.
			return true;
		}

		$view = Router::$DefaultView;
		$tools = new self();

		if($view->jsondata === null) {
			// If no JSON data is set, no automatic additions are performed.
			return true;
		}

		if(is_numeric_array($view->jsondata)) {
			// Do not attach debug information if this is a basic numeric array.
			return true;
		}

		$view->jsondata['_debug'] = [
			'page' => $tools->_getPerformanceInfo(),
		];
		return true;
	}
}
