<?php
/**
 * File for class Profiler definition in the coreplus project
 *
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20130410.1753
 * @package Core\Utilities\Profiler
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Utilities\Profiler;

use Core\Logger\LogEntry;
use Core\Session;


/**
 * Profiler gives a simple performance profiler for scripts and utilities.
 *
 * <h3>Usage</h3>
 *
 * <h4>System Profiler</h4>
 * <p>
 * Core has a system profiler running from the start of the application.
 * If FULL_DEBUG is set to true, then any event recorded there will be displayed at the end of the page execution.
 * </p>
 *
 * <code>
 * $profiler = \Core\Utilities\Profiler\Profiler::GetDefaultProfiler();
 * $profiler->record('my awesome event');
 * </code>
 *
 * <h4>Custom Profiler</h4>
 * <p>
 * To create a new profiler, (and a new timer), the following code will do that job.
 * </p>
 *
 * <code>
 * $profiler = new \Core\Utilities\Profiler\Profiler('this set');
 *
 * // Do some logic that takes some amount of time
 * // ...
 * // ...
 * $profiler->record('done with step one');
 *
 * // More stuff that takes a long time
 * // ...
 * // ...
 * $profiler->record('Finished!');
 *
 * // Display the overall time!
 * echo '&lt;h1&gt;Finished in ' . $profiler-&gt;getTimeFormatted() . '&lt;/h1&gt;';
 *
 * // Or if you want a breakdown of the events themselves...
 * echo '&lt;pre&gt;' . $profiler-&gt;getEventTimesFormatted() . '&lt;/pre&gt;';
 * </code>
 *
 * @package Core\Utilities\Profiler
 * @author Charlie Powell <charlie@evalagency.com>
 */
class DatamodelProfiler {
	private $_name;

	private $_events = [];

	private $_pointer = -1;

	private static $_DefaultProfiler;

	private $_reads = 0;

	private $_writes = 0;

	public function __construct($name){
		$this->_name = $name;

		// Do I need to register to the global "singleton" scope?
		if(self::$_DefaultProfiler === null){
			self::$_DefaultProfiler = $this;
		}
	}

	/**
	 * Get the number of reads that have been performed on this page load.
	 *
	 * @return int
	 */
	public function readCount(){
		return $this->_reads;
		//return Session::Get('datamodel_profiler_events/reads', 0);
	}

	/**
	 * Get the number of writes that have been performed on this page load.
	 *
	 * @return int
	 */
	public function writeCount(){
		return $this->_writes;
		//return Session::Get('datamodel_profiler_events/writes', 0);
	}

	/**
	 * Start recording a given query.
	 *
	 * @param string $type "read" or "write", (usually).
	 * @param string $query Human-readable version of the query string.
	 */
	public function start($type, $query){
		// Record this query!
		// This needs to include the query itself, what type it was, how long it took to execute,
		// any errors it produced, and where in the code it was called.

		if((defined('FULL_TRACE') && FULL_TRACE) || (DEVELOPMENT_MODE && $this->_pointer < 40)){
			// By skipping this in production, memory usage is cut by nearly 50% on über DB heavy pages!
			// (This occurs on pages that have more than 10k queries.
			$trace = [];
			
			$debug = debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS);
			foreach($debug as $d){
				// Skip a few obvious locations, such as this itself, the Datamodel, and the underlying DB backend.
				// The point of this trace is to see what user logic requested this!
				
				if(
					isset($d['class']) &&
					(
						$d['class'] === 'Core\\Utilities\\Profiler\\DatamodelProfiler' ||
						strpos($d['class'], 'Core\\Datamodel\\') === 0
					)
				){
					continue;
				}
				
				// Append this line onto the trace!
				$trace[] = $d;
			}
		}
		else{
			$trace = null;
		}


		$this->_pointer++;
		$this->_events[] = [
			'start'   => microtime(true),
			'type'    => $type,
			'query'   => $query,
			'trace'   => $trace,
			'memory'  => memory_get_usage(true),
			'rows'    => null,
			'error'   => null,
			'errno'   => null,
		];
	}

	public function stopSuccess($count){
		$last =& $this->_events[ $this->_pointer ];
		
		$last['time'] = microtime(true) - $last['start'];
		$last['rows'] = $count;
		
		if($last['type'] === 'read'){
			++$this->_reads;
		}
		else{
			++$this->_writes;
		}
		
		// Log this execution.
		LogEntry::Init()
			->level(LOG_LEVEL_DEBUG)
			->type('debug')
			->message($this->_getMessage($last))
			->details($this->_getDetails($last))
			->trace($last['trace'])
			->code('DMI Profiler')
			->send();

		if(
			defined('DMI_QUERY_LOG_TIMEOUT') && 
			DMI_QUERY_LOG_TIMEOUT >= 0 &&
			($last['time'] * 1000) >= DMI_QUERY_LOG_TIMEOUT
		){
			$timeFormatted = \Core\time_duration_format($time, 2);
			\Core\log_warning('[' . $timeFormatted . '] ' . $last['query']);
		}
	}

	public function stopError($code, $error){
		$last =& $this->_events[ $this->_pointer ];
		
		$last['time'] = microtime(true) - $last['start'];
		$last['errno'] = $code;
		$last['error'] = $error;
		
		if($last['type'] === 'read'){
			++$this->_reads;
		}
		else{
			++$this->_writes;
		}
		
		// Log this execution.
		LogEntry::Init()
			->level(LOG_LEVEL_INFO)
			->type('error')
			->message($this->_getMessage($last))
			->details($this->_getDetails($last))
			->trace($last['trace'])
			->code('DMIProfiler')
			->send();

		if(
			defined('DMI_QUERY_LOG_TIMEOUT') && 
			DMI_QUERY_LOG_TIMEOUT >= 0 &&
			($last['time'] * 1000) >= DMI_QUERY_LOG_TIMEOUT
		){
			$timeFormatted = \Core\time_duration_format($time, 2);
			\Core\log_warning('[' . $timeFormatted . '] ' . $last['query']);
		}
	}

	/**
	 * Get all the recorded events of this profiler as an array.
	 *
	 * @return array
	 */
	public function getEvents(){
		return $this->_events;
	}

	
	/**
	 * Get the message line based on a given event
	 * 
	 * @param array $event
	 */
	private function _getMessage(array $event) : string {
		$message = ($event['type'] === 'read' ? COLOR_BLU1 : COLOR_PRP2) . $event['type'] . COLOR_RESET;
		
		// The speed color is based on the read and the time it took.
		$timeFormatted = \Core\time_duration_format($event['time'], 2);
		$message .= ' took ';
		if(
			($event['type'] === 'read' && $event['time'] <= .001) ||
			($event['type'] === 'write' && $event['time'] <= .01)
		){
			$message .= COLOR_SUCCESS;
		}
		elseif(
			($event['type'] === 'read' && $event['time'] <= .01) ||
			($event['type'] === 'write' && $event['time'] <= .1)
		){
			$message .= COLOR_WARNING;
		}
		else{
			$message .= COLOR_ERROR;
		}
		$message .= $timeFormatted . COLOR_RESET;
		
		// Append the query!
		//$message .= ': ' . (EXEC_MODE === 'WEB' ? htmlentities($event['query'], ENT_QUOTES | ENT_HTML5) : $event['query']);
		$message .= ': ' . $event['query'];
		
		return $message;
	}
	
	/**
	 * Get the details based on a given event
	 * 
	 * @param array $event
	 */
	private function _getDetails(array $event) : string {
		$details = [];
		
		if($event['rows'] !== null){
			$details[] = 'Number of affected rows: ' . $event['rows'];
		}
		
		if($event['error']){
			$details[] = $event['error'];
		}
		
		return implode(NL, $details);
	}

	
	/**
	 * Get the first instance of the profiler
	 *
	 * @return DatamodelProfiler
	 */
	public static function GetDefaultProfiler(){
		if(self::$_DefaultProfiler === null){
			// Try to find the global one.
			global $datamodelprofiler;
			if($datamodelprofiler){
				self::$_DefaultProfiler = $datamodelprofiler;
			}
			else{
				self::$_DefaultProfiler = new self('Query Log');
			}
		}

		return self::$_DefaultProfiler;
	}
}
