<?php
/**
 * File for class Profiler definition in the coreplus project
 * 
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20130410.1753
 * @package Core\Utilities\Profiler
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Utilities\Profiler;
use Core\ConfigHandler;
use Core\i18n\I18NLoader;


/**
 * Profiler gives a simple performance profiler for scripts and utilities.
 *
 * <h3>Usage</h3>
 *
 * <h4>System Profiler</h4>
 * <p>
 * Core has a system profiler running from the start of the application.
 * If FULL_DEBUG is set to true, then any event recorded there will be displayed at the end of the page execution.
 * </p>
 *
 * <code>
 * $profiler = \Core\Utilities\Profiler\Profiler::GetDefaultProfiler();
 * $profiler->record('my awesome event');
 * </code>
 *
 * <h4>Custom Profiler</h4>
 * <p>
 * To create a new profiler, (and a new timer), the following code will do that job.
 * </p>
 *
 * <code>
 * $profiler = new \Core\Utilities\Profiler\Profiler('this set');
 *
 * // Do some logic that takes some amount of time
 * // ...
 * // ...
 * $profiler->record('done with step one');
 *
 * // More stuff that takes a long time
 * // ...
 * // ...
 * $profiler->record('Finished!');
 *
 * // Display the overall time!
 * echo '&lt;h1&gt;Finished in ' . $profiler-&gt;getTimeFormatted() . '&lt;/h1&gt;';
 *
 * // Or if you want a breakdown of the events themselves...
 * echo '&lt;pre&gt;' . $profiler-&gt;getEventTimesFormatted() . '&lt;/pre&gt;';
 * </code>
 * 
 * @package Core\Utilities\Profiler
 * @author Charlie Powell <charlie@evalagency.com>
 */
class Profiler {
	
	private $_name;
	/** @var ProfilerEvent */
	private $_rootEvent = null;
	private $_lastEvent = null;
	private $_maxEvents = 1000;


	// Counter to keep track of how many events have been added, used to skip when there are too many.
	private static $_EventCounter = 0;
	private static $_DefaultProfiler;
	// Set to false if disabled in the site configuration
	private static $_LoggingEnabled = true;
	private static $_LoggingTime = null;

	public function __construct($name){
		$this->_name = $name;

		// The first top-level event will contain all other events recorded.
		$this->_lastEvent = $this->_rootEvent = new ProfilerEvent($name, $this);

		// Do I need to register to the global "singleton" scope?
		if(self::$_DefaultProfiler === null){
			self::$_DefaultProfiler = $this;
		}

		// Is there a max number of events registered?
		if(defined('PROFILER_MAX_EVENTS')) {
			$this->_maxEvents = PROFILER_MAX_EVENTS;
		}
	}

	/**
	 * Record a standard event and its profile time from the start of the application.
	 * 
	 * @param string $event
	 */
	public function record(string $message){
		if(!$this->_checkShouldLog()) {
			return;
		}

		++self::$_EventCounter;
		$event = new ProfilerEvent($message);
		$event->commit();
	}
	
	public function registerEvent(ProfilerEvent $event) {
		if(!$this->_checkShouldLog()) {
			return;
		}

		++self::$_EventCounter;

		if($this->_lastEvent === null) {
			// Root event
			$this->_rootEvent->addEvent($event);
		}
		else {
			// Or just the latest one currently running.
			$this->_lastEvent->addEvent($event);
		}
	}
	
	/**
	 * Start a timed event that can conain other event messages contained within.
	 * 
	 * @param ProfilerEvent $event
	 */
	public function registerStartEvent(ProfilerEvent $event) {
		if(!$this->_checkShouldLog()) {
			return;
		}

		++self::$_EventCounter;

		if($this->_lastEvent === null) {
			// Root event
			$this->_rootEvent->addEvent($event);
		}
		else {
			// Or just the latest one currently running.
			$this->_lastEvent->addEvent($event);
		}
		
		$this->_lastEvent = $event;
	}
	
	/**
	 * Stop a timed event that can conain other event messages contained within.
	 *
	 * @param ProfilerEvent $event
	 */
	public function registerStopEvent(ProfilerEvent $event) {
		$this->_lastEvent = $event->parent;
	}

	/**
	 * Sometimes you just want to know how many milliseconds passed between the start of the app and now
	 * This is useful for logging utilities.
	 *
	 * @return float
	 */
	public function getTime(){
		return microtime(true) - CORE_START_MICROTIME;
	}

	/**
	 * Get the overall execution time of this profiler.
	 *
	 * This will be rounded and formatted as such:
	 * "# µs", "# ms", "# s", "# m # s", or "# h # m".
	 *
	 * @return string
	 */
	public function getTimeFormatted(){
		$time = $this->getTime();
		
		return \Core\time_duration_format($time, 4);
	}

	public function getRootEvent() : ProfilerEvent {
		return $this->_rootEvent;
	}

	/**
	 * Get the breakdown of recorded events and their time into the profiler operation.
	 * 
	 * @return string
	 */
	public function getEventTimesFormatted(){
		$out = $this->_rootEvent->render();
		
		if(EXEC_MODE === 'WEB') {
			// Tack on the simple javascript helper inline.
			$prefix = <<<EOD
<span class="profiler-event-expandall-hint"><i class="icon icon-plus-square"></i> Expand All</span><span class="profiler-event-collapseall-hint"><i class="icon icon-minus-square"></i> Collapse All</span>
<br/>
EOD;
			$out = $prefix . $out;
		}
		
		return $out;
	}

	private function _checkShouldLog() : bool {
		if(self::$_EventCounter > $this->_maxEvents) {
			// Skip if too many; this needs to be first.
			return false;
		}

		if(defined('DEVELOPMENT_MODE') && DEVELOPMENT_MODE) {
			// Development mode gets profiling.
			return true;
		}

		// Default; it's up to whether profiler logging is enabled.
		return self::$_LoggingEnabled;
	}
	
	private function _formatEvent($event){
		$in = round($event['timetotal'], 5) * 1000;
		$dcm = class_exists('I18NLoader') ? I18NLoader::GetLocaleConv('decimal_point') : '.';

		if ($in == 0){
			$time = '0000' . $dcm . '00 ms';
		}
		else{
			$parts = explode($dcm, $in);
			$whole = str_pad($parts[0], 4, 0, STR_PAD_LEFT);
			$dec   = (isset($parts[1])) ? str_pad($parts[1], 2, 0, STR_PAD_RIGHT) : '00';
			$time = $whole . $dcm . $dec . ' ms';
		}

		$mem = '[mem: ' . \Core\Filestore\format_size($event['memory']) . '] ';

		$event = $event['event'];

		return "[$time] $mem- $event";
	}
	
	
	
	/**
	 * Get the first instance of the profiler
	 *
	 * @return Profiler
	 */
	public static function GetDefaultProfiler(){
		if(self::$_DefaultProfiler === null){
			// Try to find the global one.
			global $profiler;
			if($profiler){
				self::$_DefaultProfiler = $profiler;
			}
			else{
				self::$_DefaultProfiler = new self('Root Core Profiler');
			}
		}

		return self::$_DefaultProfiler;
	}

	/**
	 * Start hook to setup the profiler with admin settings
	 */
	public static function StartHook() {
		$time = ConfigHandler::Get('/core/logs/profiler/times');
		if(!$time) {
			// Logging is disabled.
			self::$_LoggingEnabled = false;
		}
		else {
			self::$_LoggingTime = $time;
		}
	}

	/**
	 * Shutdown hook; will save the profiler log if requested.
	 */
	public static function ShutdownHook() {
		if(!self::$_LoggingEnabled) {
			return;
		}

		if(self::$_DefaultProfiler === null) {
			return;
		}

		if(self::$_LoggingTime === 'ALL') {
			// All events were requested!
			$text = self::$_DefaultProfiler->_rootEvent->renderText();
		}
		else {
			// Only log slow events
			$duration = self::$_LoggingTime / 1000;
			$slowEvents = self::$_DefaultProfiler->_rootEvent->getSlowEvents($duration);
			$text = '';
			foreach($slowEvents as $event) {
				$text .= $event->renderText();
			}
		}

		if($text) {

			// Build a simple header to log.
			$header = '[' . \Core\Date\DateTime::Now('r', \Core\Date\Timezone::TIMEZONE_DEFAULT) . '] ';
			$header .= (EXEC_MODE === 'WEB') ? REMOTE_IP : '[CLI]';
			if(isset($_SERVER['REQUEST_URI'])) {
				$header .= ' ' . $_SERVER['REQUEST_URI'];
			}
			if(isset($_SERVER['HTTP_USER_AGENT'])) {
				$header .= ' "' . $_SERVER['HTTP_USER_AGENT'] . '"';
			}
			$text = $header . "\n" . $text;

			file_put_contents(ROOT_PDIR . 'logs/profiler.log', $text, FILE_APPEND);
		}
	}
}
