<?php
/**
 * File for class Profiler definition in the coreplus project
 * 
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20130410.1753
 * @package Core\Utilities\Profiler
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Utilities\Profiler;

class ProfilerEvent {
	/**
	 * @var string An optional identifier for this log entry
	 */
	public $code = null;
	
	/**
	 * @var string An optional icon to display along with this log message
	 */
	public $icon = null;
	
	/**
	 * @var int The level of this log entry.
	 *
	 * LOG_LEVEL_CRITICAL (1)
	 * LOG_LEVEL_HIGH (2)
	 * LOG_LEVEL_INFO (3)
	 * LOG_LEVEL_DEBUG (4)
	 * LOG_LEVEL_TRACE (5)
	 */
	public $level = \LOG_LEVEL_TRACE;
	
	/**
	 * @var string The actual message to log
	 */
	public $message = null;
	
	/**
	 * @var string Type of log entry, debug, info, error, security, or some custom string
	 */
	public $type = 'info';
	
	public $parent = null;
	
	public $trace;

	public $depth = 0;

	/** @var ProfilerEvent[] */
	private $_events = [];
	private $_start;
	private $_duration;
	/** @var Profiler */
	private $_profiler;
	
	private static $_DefaultProfiler;
	
	
	public function __construct($message = null, $profiler = null) {
		$this->message = $message;
		$this->_start =  microtime(true) - CORE_START_MICROTIME;
		
		if($profiler === null) {
			if(self::$_DefaultProfiler === null) {
				self::$_DefaultProfiler = Profiler::GetDefaultProfiler();
			}
			$this->_profiler = self::$_DefaultProfiler;
		}
		else {
			$this->_profiler = $profiler;
		}
		
		if(defined('DEVELOPMENT_MODE') && DEVELOPMENT_MODE) {
			$this->trace = debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS);
		}
	}
	
	/**
	 * Commit this event, but not as a nested timed event.
	 */
	public function commit() {
		$this->_profiler->registerEvent($this);
	}
	
	/**
	 * Start and register this event as a nested-compatible event
	 */
	public function start() {
		$this->_profiler->registerStartEvent($this);
	}
	
	/**
	 * Stop and deregister an event.
	 */
	public function stop() {
		$this->_duration = $this->_getDuration();
		$this->_profiler->registerStopEvent($this);
	}
	
	/**
	 * Add a sub event to the current nested event.
	 * 
	 * @param ProfilerEvent $event
	 */
	public function addEvent(ProfilerEvent $event) {
		$this->_events[] = $event;
		$event->parent = $this;
		$event->depth = $this->depth + 1;
	}

	/**
	 * Shortcut to add a sub event to this nested Event
	 *
	 * @param string $message
	 */
	public function addEventString(string $message) {
		$event = new ProfilerEvent($message);
		$this->addEvent($event);
	}
	
	/**
	 * Render this event and any subevents into human-friendly output
	 * 
	 * @return string
	 */
	public function render() : string {
		$nested = count($this->_events) > 0;
		
		
		$start = round($this->_start * 1000, 0);
		if ($start == 0){
			$time = '0000 ms';
		}
		else{
			$time = str_pad($start, 4, 0, STR_PAD_LEFT) . ' ms';
		}
		
		
		if(EXEC_MODE === 'WEB') {
			// Web gets the fancy HTML version; CLI gets plain text.
			$out = '<div class="profiler-event-wrapper collapsed">' .
				$time . ' ' .
				$this->_getLevelString() .
				$this->_getDurationString() .
				$this->_getCodeString() . 
				($nested ? '<i class="icon icon-plus-square event-expandable-hint"></i><i class="icon icon-minus-square event-collapsible-hint"></i> ' : '') .
				'<span class="event-message" title="' . $this->_getTraceString() . '">' . $this->message . '</span>';
			
			if($nested) {
				$out .= '<ul class="profiler-events">';
				foreach($this->_events as $e) {
					$out .= '<li class="profiler-event">' . $e->render() . '</li>';
				}
				$out .= '</ul>';
			}
			
			$out .= '</div>';
		}
		else {
			$out = $time . ' ' . $this->message;
		}
		
		return $out;
	}

	/**
	 * Render this event and any subevents into log-friendly output
	 *
	 * @return string
	 */
	public function renderText() : string {
		$start = round($this->_start * 1000, 0);
		if ($start == 0){
			$time = '0000 ms';
		}
		else{
			$time = str_pad($start, 4, 0, STR_PAD_LEFT) . ' ms';
		}

		$padding = str_repeat("\t", $this->depth);

		$out = $time .
			$padding .
			' (' . $this->type . ') ' .
			$this->_getDurationString() .
			$this->_getCodeString() .
			$this->message .
			"\n";

		if(count($this->_events)) {
			foreach($this->_events as $event) {
				$out .= $event->renderText();
			}
		}

		return $out;
	}

	public function renderJSON() : array {
		$nested = count($this->_events) > 0;

		$data = [
			'start' => \Core\time_duration_format($this->_start, 0),
			'level' => $this->level,
			'type' => $this->type,
			'duration' => $this->_getDurationString(),
			'code' => $this->code,
			'message' => $this->message,
			'trace' => $this->_getTraceString(),
		];

		if($nested) {
			$data['events'] = [];
			foreach($this->_events as $event) {

				$data['events'][] = $event->renderJSON();
			}
		}
		else {
			$data['events'] = null;
		}

		return $data;
	}

	public function getSlowEvents(float $duration = 0.001) : array {
		$ret = [];

		if(count($this->_events)) {
			foreach($this->_events as $subEvent) {
				if($subEvent->_duration >= $duration) {
					$ret[] = $subEvent;
				}
			}
		}

		return $ret;
	}

	/**
	 * Get the duration of this profiler event as a string
	 *
	 * @return string
	 */
	public function getDurationString() : string {
		if($this->_duration !== null) {
			return \Core\time_duration_format($this->_duration, 0) . ' ';
		}
		elseif(count($this->_events) > 0) {
			return \Core\time_duration_format($this->_getDuration(), 0) . ' ';
		}
		else {
			return '';
		}
	}

	/**
	 * Calculate the duration of this Event
	 *
	 * @return float
	 */
	private function _getDuration() : float {
		return (microtime(true) - CORE_START_MICROTIME) - $this->_start;
	}
	
	private function _getDurationString() : string {
		if($this->_duration !== null) {
			return \Core\time_duration_format($this->_duration, 0) . ' ';
		}
		elseif(count($this->_events) > 0) {
			return \Core\time_duration_format($this->_getDuration(), 0) . ' ';
		}
		else {
			return '';
		}
	}
	
	private function _getCodeString() : string {
		if($this->code) {
			return '[' . $this->code . '] ';
		}
		else {
			return '';
		}
	}
	
	private function _getLevelString() : string {
		switch($this->level){
			case LOG_LEVEL_CRITICAL:
				return COLOR_ERROR . '(' . $this->type . ')' . COLOR_RESET . ' ';
			case LOG_LEVEL_HIGH:
				return COLOR_WARNING . '(' . $this->type . ')' . COLOR_RESET . ' ';
			case LOG_LEVEL_DEBUG:
			case LOG_LEVEL_TRACE:
				return COLOR_DEBUG . '(' . $this->type . ')' . COLOR_RESET . ' ';
			default:
				return '(' . $this->type . ') ';
		}
	}
	
	private function _getTraceString() : string {
		if(!$this->trace) {
			return '';
		}
		
		$lines = [];
		
		// Use this to trim off unnecessary text in the trace.
		$relative = dirname(__FILE__, 6);
		
		$counter = 0;
		foreach($this->trace as $t){
			$counter++;
			
			if(isset($t['class']) && isset($t['type'])){
				$call = $t['class'] . $t['type'] . $t['function'] . '()';
			}
			else{
				$call = $t['function'] . '()';
			}
			
			$line = $t['line'] ?? null;
			$file = $t['file'] ?? null;
			
			$rel = $file && strpos($file, $relative) === 0 ?
				substr($file, strlen($relative) + 1) :
				$file;
			
			$lines[] = ' #' . $counter . ': ' . $call . ($rel ? ' called from ' . $rel : '') . ($rel && $line ? ':' . $line : '');
		}
		
		return implode(NL, $lines);
	}
	
	/**
	 * Record a standard event and its profile time from the start of the application.
	 *
	 * @param string $message
	 */
	public function Record(string $message) {
		$event = new ProfilerEvent($message);
		$event->commit();
	}
}