<?php
/**
 * @package Core\Templates\Smarty
 * @since 1.9
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\Renderer\ImageRenderer;

/**
 * Render markup for an image tag in smarty
 *
 * The {img} smarty function is the recommended way to load images in templates from asset or public directories.
 * In addition to automatically resolving URLs, it can also handle server-side resizing and a few other nifty features.
 * 
 * #### "Blank" images
 * 
 * *Core 6.2.2* A "blank" image can now be requested via 'src="BLANK"'!
 * This gets replaced with a 1x1 transparent GIF image and no File magick ran.
 * Use this when you need to render an <img/> tag with no src but still be valid and with no
 * unnecessary calls back to # or about:blank or anything.
 *
 * #### Image Types &amp; Animations
 *
 * *Core 3.3.0* Image types are preserved, so if a .jpg is requested, an image/jpeg is returned.
 * This changed from the previous behaviour where all images were converted to a PNG.
 * 
 * *Core 6.2.0* SVG images are now supported natively
 *
 * Supported image types are:
 * 
 * * JPEGs (.jpg & .jpeg image/jpeg)
 * * PNGs (.png image/png)
 * * GIFs (.gif image/gif)
 * * Animated GIFs (.gif image/gif)
 * * Scalable Vector Graphics (.svg image/svg+xml)
 *
 * If an animated gif is resized, the server will attempt to preserve the animation on the resized image.
 * This is done via imagemagick, (so that library needs to be present on the server in order for this trick to work).
 *
 * #### SEO Data
 *
 * *Core 3.2.0* Alt tags are automatically added to every image that does not have the alt attribute explicitly set.
 * -This alt name is pulled from the filename of the image, with automatic capitalization and '_' => (space) converting.-
 * This alt name is first checked in the database for the user-definable title of the image.
 * If that returns nothing, the filename is used instead with automatic capitalization and formatting.
 * 
 * **Performance Reminder** If performance is required and this image is not important,
 * (such as a tiny page icon), PLEASE include the alt tag otherwise a database call is created for each image!
 *
 * #### Smarty Parameters
 *
 *  * assign
 *    * string
 *    * Set to a string to assign that variable name instead of returning the output.
 *  * dimensions
 *    * Provide both width and height in pixels, along with special instructions
 *    * Structure is "widthxheight" with no spaces between the "x" and the two integers.
 *    * Special modes available are:
 *    * Carat "`^`" at the end of the string fits the smallest dimension instead of the largest.
 *    * Exclamation mark "`!`" at the end forces size regardless of aspect ratio.
 *    * Greater than "`>`" at the end will only increase image sizes.
 *    * Less than "`<`" at the end will only decrease image sizes.
 *  * file
 *    * \Core\Filestore\File
 *    * File object passed in to display
 *    * Either "file" or "src" is required.
 *  * height
 *    * int
 *    * Maximum image height (in pixels).
 *    * If both width and height are provided, the image will be constrained to both without any distortion.
 *  * inline
 *    * int "1" or "0", (default "0")
 *    * New in Core 4.2.0
 *    * Request that the resized image be encoded as base64 and inserted inline in the markdown instead of returned as the URL.
 *  * includemeta
 *    * Set to "1" to include extra markup for the metadata
 *  * placeholder
 *    * string
 *    * placeholder image if the requested image is blank or not found.  Useful for optional fields that should still display something.
 *    * Current values: "building", "generic", "person", "person-tall", "person-wide", "photo"
 *  * src
 *    * string
 *    * Source filename to display.  This can start with "assets" for an asset, or "public" for a public file.
 *    * Either "file" or "src" is required.
 *    * Set to "BLANK" (caps matter), to render a 1x1 transparent pixel.
 *  * width
 *    * int
 *    * Maximum image width (in pixels).
 *    * If both width and height are provided, the image will be constrained to both without any distortion.
 *
 * Any other parameter is transparently sent to the resulting `<img/>` tag.
 *
 *
 * #### Example Usage
 *
 * ``` {.syntax-smarty}
 * {img src="public/gallery/photo123.png" width="123" height="123" placeholder="photo" alt="My photo 123"}
 * ```
 *
 * @param array  $params  Associative (and/or indexed) array of smarty parameters passed in from the template
 * @param Smarty $smarty  Parent Smarty template object
 *
 * @return string
 * @throws SmartyException
 */
function smarty_function_img($params, $smarty){

	global $__Core_Asset_Inline;
	
	if(isset($params['assign'])) {
		$assign = $params['assign'];
		unset($params['assign']);
	}
	else {
		$assign = null;
	}
	
	$renderer = new ImageRenderer();
	$renderer->setAttributes($params);
	
	if(isset($__Core_Asset_Inline) && $__Core_Asset_Inline) {
		$renderer->setInline();
	}
	
	$html = $renderer->fetch();

	return $assign ? $smarty->assign($assign, $html) : $html;
}