<?php
/**
 * @package Core\Templates\Smarty
 * @since 1.9
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\Router;
use Core\View;

/**
 * Manually add a widget onto a template.
 *
 * @todo Finish documentation of smarty_function_widget
 *
 * @param array  $params  Associative (and/or indexed) array of smarty parameters passed in from the template
 * @param Smarty $smarty  Parent Smarty template object
 *
 * @return mixed
 *
 * @throws SmartyException
 */
function smarty_function_widget($params, $smarty) {

	$assign  = (isset($params['assign']))? $params['assign'] : false;
	$tmpl    = $smarty->getTemplateVars('__core_template');
	$topview = Router::$DefaultView;
	$topRequest = Router::$DefaultRequest;

	// Version 2.0 uses baseurl as the defining call.
	if(isset($params['baseurl'])) {
		$api = 2.0;
		$parts = WidgetModel::SplitBaseURL($params['baseurl']);
		$baseurl = $params['baseurl'];
		$original = $params['baseurl'];
		$name = $parts['controller'];
		$method = $parts['method'];
		$parameters = $parts['parameters'];
	}
	// Version 1.0 uses name.
	elseif(isset($params['name'])) {
		$api = 1.0;
		$original = $params['name'];
		$name = $params['name'];
		// Try to look up this requested widget.
		$name .= 'Widget';
		$parameters = null;
		$method = null;
	}
	else {
		$api = 0.0;
		$name = null;
		$original = null;
		$parameters = null;
		$method = null;
	}


	if(!class_exists($name)) {
		if(DEVELOPMENT_MODE){
			return '[ERROR, Class for ' . $original . ' (' . $name . ') not found on system, widget disabled.]';
		}
		else {
			\Core\log_warning('Unable to render widget ' . $original . ' (' . $name . ') on page ' . $topRequest->getBaseURL() . ', widget not located on system');
			return '';
		}

		//throw new SmartyException('Unable to locate class [' . $name . '] for requested widget', null, null);
	}
	// @todo Add support for requiring instancing.


	// Version 1.0 API
	if($api == 1.0) {
		/** @var $w \Core\Widget */
		$w = new $name();
		$dat = $w->execute()->fetch();
	}
	// Version 2.0 API
	elseif($api == 2.0) {

		$model = WidgetModel::Construct($params['baseurl']);
		if(!$model->checkExists()) {
			if(DEVELOPMENT_MODE) {
				return '[ERROR, ' . $baseurl . ' not found on system, widget disabled.]';
			}
			else {
				\Core\log_warning('Unable to render widget ' . $baseurl . ' on page ' . $topRequest->getBaseURL() . ', widget model not located on system');
				return '';
			}
		}

		$widget = $model->getWidget();
		if($widget === null) {
			if(DEVELOPMENT_MODE) {
				return '[ERROR, ' . $baseurl . ' not found on system, widget disabled.]';
			}
			else {
				\Core\log_warning('Unable to render widget ' . $baseurl . ' on page ' . $topRequest->getBaseURL() . ', widget not located on system');
				return '';
			}
		}


		// Ensure that the widget's View knows it's linked to a parent!
		$widget->getView()->parent = $topview;

		// Populate the request with the inbound data too.
		// This allows settings to be populated from the calling template.
		unset($params['baseurl']);
		$request = $widget->getRequest();
		$request->parameters = $params;


		try {
			$return = call_user_func(array($widget, $method));
		}
		catch(Exception $ex) {
			\Core\log_exception($ex);
			return '';
		}

		$dat = null;

		if(is_int($return)) {
			throw new SmartyException("widget $name/$method returned error code $return.", null, null);
		}
		elseif($return === null) {
			// Hopefully it's setup!
			$return = $widget->getView();
		}
		// If it's just a string, return that.
		elseif(is_string($return)) {
			$dat = $return;
		}


		// If dat is still null, (it probably is still null btw), then render the template!
		if($dat === null){
			// Try to guess the templatename if it wasn't set.
			if($return->status === View::STATUS_OK && $return->contenttype == View::CTYPE_HTML && $return->templatename === null){
				$cnameshort = (strpos($name, 'Widget') == strlen($name) - 6) ? substr($name, 0, -6) : $name;
				$return->templatename = strtolower('/widgets/' . $cnameshort . '/' . $method . '.tpl');
			}

			try {
				$dat = $return->fetch();
			}
			catch(Exception $ex) {
				\Core\log_exception($ex);
				return '';
			}

		}
	}
	else{
		$dat = 'Invalid API version';
	}


	if($assign) {
		$smarty->assign($assign, $dat);
		return '';
	}
	else {
		return $dat;
	}
}
